#include "VKString.h"
#include "Sortable.h"
#include "FileManagement.h"

#include "RendererGLUT.h"
#include "DatabaseVisWindow.h"
#include "ModelDatabase.h"
#include "ShapeTriangleSoup.h"

#include "STemplate.h"
#include "VisShapeTemplate.h"
#include "AppScreenshots.h"
#include "TMatch.h"
#include "RendererOSMesa.h"
#include "TMatchAlg.h"
#include "STemplateNode.h"
#include "PartOfEdge.h"
#include "ShapeDistance.h"
#include "DistGeodesicDijkstra.h"

#include "InitWordNet.h"

#include "CreateTemplate.h"
#include "AutoSegment.h"


STemplate * AutoSegment(ModelDatabase * db, const VKString & modelname, bool readManifold, int nseg)
{
	Model3D * m = db->GetModel(modelname);
	assert(m!=NULL);
	
	GeomShape * shape = ReadModelGeometry(db, modelname, "main_Even1000.pnts", readManifold);
	assert(shape!=NULL);
	ShapePointCloud * pnts = shape->SafeCastToPointCloud();
	assert(pnts!=NULL);
	std::cout<<"Initializing template from geometry: "<<modelname.c_str()<<" with "<<nseg<<" segments"<<std::endl;
	// create initial template
	// NOTE: ASSUMING first nseg points are iteratively farthest points
	std::vector<std::vector<int> > initClustering;
	CreateInitialSegmentation(pnts, nseg, initClustering);	// initialize segmentation
	STemplate * st = NULL;
	for (int i=0; i<10; i++)
	{
		st = CreateInitialBoxes(pnts, nseg, initClustering);
		
		// do optimization: alternate MRF and Labeling w/o deformation penalty
		std::vector<int> emptySubset;
		TMatch * match = st->CreateSeedConfig(0, pnts, emptySubset, R3identity_affine);
		TMatchAlgParams params;
		
		TMatchEnergy * energy = new TMatchEnergy(match);
	//	energy->s_fEdgeEnergyWeight = 10;

		int deformAndRelabel = TMatchEnergy::ET_Pnts | TMatchEnergy::ET_NodeFit2Pnts | TMatchEnergy::ET_PntEdges;
		//deformAndRelabel |= TMatchEnergy::ET_NodeXForm | TMatchEnergy::ET_NodeShape ;
		deformAndRelabel |= TMatchEnergy::ET_PntsDistance | TMatchEnergy::ET_DistsModelSpace;
		energy->SetTerms(deformAndRelabel);
		TMatchAlg::OptimizeConfigured(st, energy, params);
			
		// update template from the match, return template
		st->SetDeformAsRestPose(energy->GetMatch());

		// copy labels
		for (int j=0; j<(int)initClustering.size(); j++)
			initClustering[j].clear();
		
		// re-initialize
		for (int j=0; j<match->NShapePrimitives(); j++)
		{
			int nid = match->LocalNodeID(j);
			assert(nid < (int)initClustering.size());
			initClustering[nid].push_back(j);
		}
	}
	assert(st!=NULL);
	return st;
}



STemplate * CreateInitialBoxes(ShapePointCloud * pnts, int nseg,
							   const std::vector<std::vector<int> > & initClustering)
{	
	STemplate * st = new STemplate(NULL);
	st->UseMeshInParts(-1);	// boxes
	st->m_CreatedFromModels.push_back(pnts->GetModel()->ModelName());
	
	// create all nodes
	// nodes
	STemplateNode * root = new STemplateNode("AutoSeg_Root", (int)st->m_AllNodes.size(), st);
	st->m_AllNodes.push_back(root);
	
	const std::vector<int> subsetFaces;
	for (int sID=0; sID<nseg; sID++)
	{
		// get (or create) node for a segment
		VKString label = VKString("AutoSeg_")+VKString::number(nseg);
		// create node
		STemplateNode * tNode = new STemplateNode(label, (int)st->m_AllNodes.size(), st);
		st->m_AllNodes.push_back(tNode);
		// add geometries / transformations
		tNode->InitShapeAndXForm(0, NULL, 1, 0, pnts, initClustering[sID], -1, subsetFaces);
		R3Vector v = tNode->GetMeanScale(0);
	}
	
	// add parents
	for (int i=1; i<(int)st->m_AllNodes.size(); i++)
	{
		STemplateNode * tNode = st->m_AllNodes[i];
		st->m_AllNodes[0]->AddPartOfChild(tNode, st, 0, false);
		assert (tNode->PartOfParent()!=NULL);
		tNode->PartOfParent()->UpdateXForm(0, NULL, tNode, 1, 0);
		tNode->PartOfParent()->UpdateNParts(0, 1, true);	// include part
	}

	// add null node
	STemplateNode * nullNode = new STemplateNode("NULLNode", (int)st->m_AllNodes.size(), st);
	nullNode->InitNullGeometry();
	st->m_AllNodes.push_back(nullNode);
	st->m_AllNodes[0]->AddPartOfChild(nullNode, st, 0, true);

	return st;
}

void CreateInitialSegmentation(ShapePointCloud * pnts, int nseg,
							   std::vector<std::vector<int> > & initClustering)	// initialize segmentation
{
	for (int i=0; i<nseg; i++)
		initClustering.push_back(std::vector<int>());
	
	bool manifold = pnts->GetMesh()->SafeCastToManifold()!=NULL;
	ShapePointCloud * iterFarthest = pnts;
	DistGeodesicDijkstra * geoDist = NULL;
	if (manifold)	// compute geodesic iterative farthest
	{
		ShapeMesh * mesh = (ShapeMesh*)pnts->GetMesh();
		geoDist = new DistGeodesicDijkstra(mesh, "geofar");
		mesh->AddDistanceMetric(ShapeDistance::m_sDefaultGeodesic, geoDist);

		iterFarthest = ShapePointCloud::CreateIterFar("geofar", mesh->GeometryName(), (Model3D*)pnts->GetModel(), nseg,
													  ShapeDistance::m_sDefaultGeodesic, pnts);
	}
	
	//double distThresh = .5 * pnts->AverageRadius(pnts->Centroid());	// vk: const
	double distThresh = -1;
//	if (!manifold)
//		distThresh = ;
	// assign based on geodesic distances (is it important?)
//	for (int i=0; i<nseg; i++)
//	{
//		std:;cout<<"FAR["<<i<<"]: "<<iterFarthest->PointPosition(i).X()<<" ";
//		std::cout<<iterFarthest->PointPosition(i).Y()<<" "<<iterFarthest->PointPosition(i).Z()<<std::endl;;
//	}
	for (int p=0; p<pnts->NumPoints(); p++)
	{
		int bestSegment = -1;
		double minDist = -1;
		double minEuclDist = -1;
		R3Point pos = pnts->PointPosition(p);
		for (int i=0; i<nseg; i++)
		{
			double euclDist = R3Distance(iterFarthest->PointPosition(i), pos);;
			double dist = euclDist;
//			if (geoDist!=NULL)
//				dist = geoDist->NormalizedDistance(iterFarthest->GetPoint(i), pnts->GetPoint(p));
			if (dist < minDist || bestSegment==-1)
			{
				minEuclDist = euclDist;
				minDist = dist;
				bestSegment = i;
			}
		}
		assert(bestSegment>=0 && bestSegment <(int)initClustering.size());
		if (minEuclDist < distThresh || distThresh==-1)
			initClustering[bestSegment].push_back(p);
	}
}




