#include "MatchHeaders.h"

/// Functions defined in MatchTemplate

void ReadModelNames(VKStringList & store, const VKString & filename)
{
	std::ifstream textStream(filename.c_str());
	assert(textStream.is_open());
	while (!textStream.eof())
	{
		VKString mName;
		mName.readToken(textStream);
		if (!textStream.eof())
			store.push_back(mName);
	}
	textStream.close();
}

STemplate * SingleTemplate(ModelDatabase * db, const VKString & mName, const VKString & geomName,
						   const VKString & labelsName, bool trustLabels, double useMeshThresh)
{
	GeomShape * shape = NULL;
	ShapeLabels * labels = NULL;
	ShapeLabels * meshLabels = NULL;
	if (geomName.endsWith(".pnts"))
	{
		GeomShape * mesh = ReadModelGeometry(db, mName, Model3D::m_sMainGeomName, false, labelsName);
		ShapeValue * vals = mesh->GetShapeVal(labelsName);
		assert(vals!=NULL && vals->SafeCastToLabels()!=NULL);
		GeomShape * pnts = ReadModelGeometry(db, mName, geomName, false);
		meshLabels = vals->SafeCastToLabels();
		labels = ShapeLabels::CopyLabels(vals->SafeCastToLabels(), pnts);
	}
	else
	{
		shape = ReadModelGeometry(db, mName, geomName, false, labelsName);
		assert(shape!=NULL && shape->GetShapeVal(labelsName)!=NULL);
		labels = shape->GetShapeVal(labelsName)->SafeCastToLabels();
		assert(labels!=NULL);
		meshLabels = labels;
	}
	return STemplate::CreateTemplate(labels, trustLabels, false, useMeshThresh, meshLabels);
}

GeomShape * ReadModelGeometry(ModelDatabase * db, const VKString & mName, const VKString & geomName,
							  bool readManifold, const VKString & geomValue)
{
	assert(db!=NULL);
	Model3D * m = db->GetModel(mName);
	if (m==NULL)
		std::cout<<"[ERROR] Cannot find model: "<<mName.c_str()<<std::endl;
	assert(m!=NULL);
	VKString geomFile = db->GetModelGeometryFilename(m->ModelName(), geomName);
	VKString mDir = db->GetDBDirname()+"/Models/";
	
	if (m->GetGeomShape(geomName)==NULL)
	{
		if (geomName.endsWith(".pnts"))
		{
			assert(geomName.startsWith(Model3D::m_sMainGeomName));
			ReadModelGeometry(db, mName, Model3D::m_sMainGeomName, readManifold);
			m->ReadGeometry(GeomShape::SHAPE_TYPE_POINT_CLOUD, geomName, geomFile, Model3D::m_sMainGeomName);
		}
		else
		{
			VKString geomFile = db->GetModelGeometryFilename(m->ModelName(), geomName)+".polysoup.off";
			GeomShape::GeomShapeTypes mt = GeomShape::SHAPE_TYPE_TRIANGLE_SOUP;
			if (readManifold)
				mt = GeomShape::SHAPE_TYPE_MANIFOLD;
			m->ReadGeometry(mt, geomName, geomFile);
		}
	}
	GeomShape * shape = m->GetGeomShape(geomName);
	assert(shape!=NULL);
	if (geomValue!="none")
	{
		VKString shapeValFile = mDir+db->ShapeValueFilename(m->ModelName(), geomName, geomValue);
		if (shape->GetShapeVal(geomValue)==NULL)
			m->ReadGeomValue("ShapeLabels", shape, geomValue, shapeValFile);
	}
	return shape;
}

void WriteAlignedModel(const TMatch * match, const VKString & filename)
{
	FileManagement::CreateDirsOnPathToFile(filename.c_str());
	const R3Affine & xForm = match->GlobalXForm();
	const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
	const ShapeMesh * mesh = pnts->GetMesh();
	assert(mesh!=NULL);
	
	ShapeTriangleSoup * xformed = new ShapeTriangleSoup();
	
	for (int i=0; i<mesh->NVertices(); i++)
	{
		R3Point p = mesh->VertexPosition(i);
		xForm.Apply(p);
		xformed->AddVertex(p[0], p[1], p[2]);
	}
	
	for (int i=0; i<mesh->NFaces(); i++)
		xformed->AddFace(mesh->VertexOnFace(i, 0), mesh->VertexOnFace(i, 1), mesh->VertexOnFace(i, 2));
	
	xformed->WriteOff(filename, ShapeTriangleSoup::OFF_ASCII);
	
}

void WriteTemplateMetaData(const STemplate * tmplt, const VKString & filename)
{
	std::ofstream textStream(filename.c_str());
	assert(textStream.is_open());
	
	textStream<<"Parts\n";
	for (int i=0; i<tmplt->NNodes(); i++)
	{
		VKString nodeName = tmplt->GetNode(i)->NodeName();
		VKString nodeType = tmplt->GetNode(i)->GetPart(0)->GetBoxTypeStr();
		double nodeErr = tmplt->GetNode(i)->GetPart(0)->GetFittingErr();
		textStream<<nodeName.c_str()<<" "<<nodeType.c_str()<<" "<<nodeErr<<"\n";
	}
	
	textStream.close();
}
void WriteLabelsExternal(ModelDatabase * db, const TMatch * match, const VKString & dirName, const VKString & equivLabels)
{
    const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
	assert(pnts!=NULL);
	const ShapeMesh * mesh = match->GetMatchedGeomAsPnts()->GetMesh();
	assert(mesh!=NULL);

    VKString modelFile = dirName + "/"+pnts->GetModel()->ModelName()+".off";
	//WriteAlignedModel(match, modelFile);
	const ShapeManifoldMesh * manifold = mesh->SafeCastToManifold();
	ShapeTriangleSoup * triSoup = (ShapeTriangleSoup*)mesh->SafeCastToTriangleSoup();
	if (manifold!=NULL)
		((R3Mesh*)manifold->GetMesh())->WriteFile(modelFile.c_str());
	else if (triSoup!=NULL)
		triSoup->WriteOff(modelFile, ShapeTriangleSoup::OFF_ASCII);
	else
		assert(false);
	// write ground truth
	VKString perFaceSegments = dirName + "/"+pnts->GetModel()->ModelName()+".truth.seg";
	std::vector<VKStringList> equivalenceList;
	std::map<VKString, int> label2ListID;
	
	// get ground truth (mesh labels)
	
	VKString gtLabelsName = ShapeValue::m_sDefaultSceneGraphLabels;
	VKString mName = mesh->GetModel()->ModelName();
	const GeomShape * gtShape = ReadModelGeometry(db, mName, Model3D::m_sMainGeomName, true, gtLabelsName);
	const ShapeMesh * gtMesh = (gtShape!=NULL ? gtShape->SafeCastToMesh() : NULL);
	const ShapeValue * gtVals = (gtMesh!=NULL ? gtMesh->GetShapeVal(gtLabelsName) : NULL);
	const ShapeLabels * gtLabels = (gtVals!=NULL ? gtVals->SafeCastToLabels() : NULL);
	assert(gtLabels!=NULL);

	if (equivLabels!="none")
		ReadEquivalentLabels(equivLabels, equivalenceList, label2ListID);
	
	std::map<VKString, int> uniqueName2UniqueID;
	VKStringList uniqueNames;
	for (int i=1; i<gtLabels->NNodes(); i++)
	{
		VKString label = gtLabels->NodeLabel(i);
		if (uniqueName2UniqueID.find(label) == uniqueName2UniqueID.end())
		{
			uniqueName2UniqueID[label] = (int)uniqueNames.count();
			uniqueNames.push_back(label);
		}
	}

	std::ofstream gtStream(perFaceSegments.c_str());
	assert(gtStream.is_open());
	gtStream<<"Labels: "<<uniqueNames.count()<<"\n";
	for (int i=0; i<(int)uniqueNames.count(); i++)
		gtStream<<uniqueNames[i].c_str()<<"\n";
	gtStream<<"NFaces "<<mesh->NFaces()<<"\n";
	for (int i=0; i<mesh->NFaces(); i++)
	{
		ShapePoint pntOnFace(mesh, i, 1./3., 1./3.);
		int gtSegID = (int)gtLabels->Value(pntOnFace, 0);
		VKString gtLabel = MainName(gtLabels->NodeLabel(gtSegID), equivalenceList, label2ListID);
		gtStream<<uniqueName2UniqueID[gtLabel]<<"\n";
	}
	gtStream.close();

	
    VKString labelsFile = dirName+"/"+pnts->GetModel()->ModelName()+".pnt.labels";
	std::ofstream textStream(labelsFile.c_str());
	assert(textStream.is_open());
    textStream<<"NNodes "<<match->NTemplateNodes()<<"\n";
    for (int i=0; i<match->NTemplateNodes(); i++)
    {
        const TemplateNodeDeform & deform = match->Deformation(i);
		VKString nodeName = match->GetSTemplate()->GetNode(deform.NodeID())->NodeName();
        textStream<<MainName(nodeName, equivalenceList, label2ListID).c_str()<<" ";
        textStream<<deform.NodeID()<<" "<<deform.NParts()<<" "<<deform.PartID()<<"\n";
    }
    
    // node names
    // n points
    textStream<<"NPoints "<<pnts->NumPoints()<<"\n";
	int fid;
	double b1, b2;
	const ShapeMesh * testMesh;
    for (int i=0; i<pnts->NumPoints(); i++)
    {
        ShapePoint pnt = pnts->GetPoint(i);
		R3Point pos = pnt.GetPosition();
        int nodeID = match->LocalNodeID(i);
		//VKString label = MainName(gtLabels->NodeLabel(gtSegID), equivalenceList, label2ListID);
		pnt.GetBarycentricCoords(&fid, &b1, &b2, &testMesh);
		textStream<<fid<<" "<<b1<<" "<<b2<<" "<<(1. - b1 - b2)<<" ";
        textStream<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<" "<<nodeID<<"\n";
    }
	
    textStream.close();
}

STemplate * InitTemplate(const VKString & topologyFile, const VKString & trainGeomName,
						 const VKString & trainLabelsName, ModelDatabase * db)
{
	VKString flatSegName = "MinDepthSeg";
	// read topology file
	std::ifstream topologyFStream(topologyFile.c_str());
	assert(topologyFStream.is_open());
	VKString rootStr;
	rootStr.readLine(topologyFStream);
	std::map<VKString, int> children;
	std::map<VKString, VKString> equivalence;
	while (!topologyFStream.eof())
	{
		VKString line;
		line.readLine(topologyFStream);
		if (line=="")
			break;
		
		line = line.replace(" - ", "");
		VKStringList lines = line.split(", ");
		if (lines.count()>0)
		{
			for (int i=0; i<lines.count(); i++)
			{
				//std::cout<<"Keyword: |"<<lStr.c_str()<<"|"<<std::endl;
				children[lines[i]] = (int)children.size();
				equivalence[lines[i]] = lines[0];
			}
		}
	}
	
	// find appropriate segmentations (at level of labels)
	for (int i=0; i<db->NumModels(); i++)
	{
		// read segmentations
		Model3D * m = db->GetModel(i);
		assert(m!=NULL);
		VKString mName = m->ModelName();
		GeomShape * g = ReadModelGeometry(db, mName, trainGeomName, false);
		assert(g!=NULL);
		GeomShape * mesh = ReadModelGeometry(db, mName, Model3D::m_sMainGeomName, false, trainLabelsName);
		ShapeValue * vals = mesh->GetShapeVal(trainLabelsName);
		assert(vals!=NULL && vals->SafeCastToLabels()!=NULL);
		ShapePointCloud * pnts = g->SafeCastToPointCloud();
		assert(pnts!=NULL);
		ShapeLabels * labels = ShapeLabels::CopyLabels(vals->SafeCastToLabels(), pnts);
		assert(labels!=NULL);
		
		// find appropriate depth of segmentation
		// for every label - find depth
		int minDepth = -1;
		for (int l=0; l<labels->NNodes(); l++)
		{
			const VKString & nl = labels->NodeLabel(l);
			int depth = labels->NodeDepth(l);
			if ((minDepth < 0 || minDepth > depth) && equivalence.find(nl)!=equivalence.end())
				minDepth = depth;
		}
		
		// expand to minimal depth: create flat segmentation
		std::vector<SceneGraphLabel*> allLabels;
		ShapeLabels * newLabels = new ShapeLabels(pnts, flatSegName);
		SceneGraphLabel * root = new SceneGraphLabel();
		root->m_sLabel = rootStr;
		root->m_iID = (int)allLabels.size();
		root->m_pParent = NULL;
		allLabels.push_back(root);
		
		std::vector<int> pnts2segs;
		for (int p=0; p<pnts->NumPoints(); p++)
			pnts2segs.push_back(0);
		
		for (int l=0; l<labels->NNodes(); l++)
		{
			if (labels->NodeDepth(l) == minDepth || labels->NodeNChildren(l) == 0)
			{
				SceneGraphLabel * newNode = new SceneGraphLabel();
				std::map<VKString, VKString>::iterator iter = equivalence.find(labels->NodeLabel(l));
				if (iter==equivalence.end())
					newNode->m_sLabel = labels->NodeLabel(l);
				else
					newNode->m_sLabel = iter->second;
				newNode->m_iID = (int)allLabels.size();
				newNode->m_pParent = root;
				root->m_apChildren.push_back(newNode);
				allLabels.push_back(newNode);
				
				std::vector<int> allPntIDs;
				labels->NodeAllPoints(l, allPntIDs);
				for (int p=0; p<(int)allPntIDs.size(); p++)
				{
					int pid = allPntIDs[p];
					assert(pid >= 0 && pid < (int)pnts2segs.size());
					pnts2segs[pid] = newNode->m_iID;
				}
			}
		}
		
		newLabels->SetAllLabels(allLabels);
		newLabels->SetPntsSegments(pnts2segs);
		pnts->AddShapeVal(newLabels->GetShapeValueName(), newLabels);
	}
	
	// find pairs that share at least one label
	//	VKGraph graph(db->NumModels();
	//	for (int i=0; i<db->NumModels(); i++)
	//	for (int j=i+1; j<db->NumModels(); j++)
	//	{
	//	}
	
	// do consistent segmentation?
	std::cout<<"NOT FINISHED IMPLEMENTATION YET: InitTemplate"<<std::endl;
	exit(0);
}


STemplate * ReadTemplateFromDir(const VKString & tmpltDir)
{
	STemplate * st = new STemplate(NULL);
	PartBox rootEllipsoid(st->GetAllParts());
	R3Triad xyz(R3Vector(1, 0, 0), R3Vector(0, 1, 0), R3Vector(0, 0, 1));
	rootEllipsoid.InitBox(R3Point(0,0,0), xyz, R3Vector(1, 1, 1), 0);
	
	st->InitAndAddNewNode(-1, "Root", rootEllipsoid, 0);
	
	// go over all text files
	VKStringList allfiles = FileManagement::ListAllFiles(tmpltDir.c_str());
	for (int i=0; i<allfiles.count(); i++)
	{
		if (allfiles[i].endsWith(".txt"))
		{
			std::cout<<"Reading "<<allfiles[i].c_str()<<std::endl;
			std::ifstream textStream((tmpltDir+"/"+allfiles[i]).c_str());
			assert(textStream.is_open());
			VKString label;
			label.readLine(textStream);
			label = label.replace(" ", "").replace("\n", "");
			while (!textStream.eof())
			{
				int nparts, nmodels;
				textStream>>nparts>>nmodels;
				
				for (int pid = 0; pid < nparts; pid++)
				{
					double xmin, ymin, zmin, xmax, ymax, zmax;
					textStream>>xmin>>ymin>>zmin>>xmax>>ymax>>zmax;
					R3Point pmin(xmin, ymin, zmin);
					R3Point pmax(xmax, ymax, zmax);
					PartBox currPart(st->GetAllParts());
					currPart.InitBox((pmin+pmax)/2, xyz, (pmax-pmin)/2, 0);
					if ((label=="seat" && nparts==1) || (label == "leg" && nparts==4)
						|| (label=="back" && nparts==1))
						st->InitAndAddNewNode(0, label, currPart, 0);
				}
			}
		}
	}
	return st;
}
