#include "MRFOptimization.h"

/////  MRFProblem /////

MRFProblem::MRFProblem(int nNodes, int nLabels)
{
	m_iNNodes = nNodes;
	m_iNLabels = nLabels;
}

MRFProblem::~MRFProblem()
{
	
}

MRFProblemArray * MRFProblem::SafeCastToArray()
{
	return NULL;
}

int MRFProblem::LabelCost(int labelID) const
{
	return 0;
}

int MRFProblem::NNodes()
{
	return m_iNNodes;
}

int MRFProblem::NLabels()
{
	return m_iNLabels;
}


/////  MRFProblemArray /////
MRFProblemArray::MRFProblemArray(int nNodes, int nLabels, VKGraph * g, int * dataCosts, int * labelsCosts)
: MRFProblem(nNodes, nLabels)
{
	m_RescaleDouble2Int = 1.;
	m_pGraph = g;
	
	if (dataCosts==NULL)
	{
		m_aDataCost = new int[nNodes * nLabels];
		for (int n=0; n<NNodes(); n++)
		for (int l=0; l<NLabels(); l++)
			m_aDataCost[n*NLabels()+l] =  (int)NodeCost(n, l);
	}
	else 
		m_aDataCost = dataCosts;
	
	if (labelsCosts==NULL)
	{
		m_aLabelsCost = new int[nLabels * nLabels];
		for (int l1=0; l1<NLabels(); l1++)
		for (int l2=0; l2<NLabels(); l2++)
			m_aLabelsCost[l1*NLabels()+l2] = (int)LabelCost(l1, l2);
	}
	else 
		m_aLabelsCost = labelsCosts;
}

MRFProblemArray::MRFProblemArray(int nNodes, int nLabels, const VKGraph * g, 
								 const double * dataCosts, const double * labelsCosts, double d2int)
: MRFProblem(nNodes, nLabels)
{
	m_RescaleDouble2Int = d2int;
	m_pGraph = g;
	m_aDataCost = new int[nNodes * nLabels];	
	m_aLabelsCost = new int[nLabels * nLabels];
	
	for (int i=0; i<nNodes*nLabels; i++)
		m_aDataCost[i] = (int)(dataCosts[i] * m_RescaleDouble2Int);
	for (int i=0; i<nLabels*nLabels; i++)
		m_aLabelsCost[i] = (int)labelsCosts[i];
	
//	for (int i=0; i<g->NumEdges(); i++)
//		g->Edge(i)->weight *= d2int;
}

void MRFProblemArray::SetNewDataCosts(const double * newCosts)
{
	for (int i=0; i<m_iNNodes*m_iNLabels; i++)
		m_aDataCost[i] = (int)(newCosts[i] * m_RescaleDouble2Int);
}


MRFProblemArray::~MRFProblemArray()
{
	delete [] m_aDataCost;
	delete [] m_aLabelsCost;
	//delete m_pGraph;
}

MRFProblemArray * MRFProblemArray::SafeCastToArray()
{
	return this;
}

int MRFProblemArray::EdgeCost(int nodeID1, int nodeID2, int label1ID, int label2ID) const
{
	assert(false);
}

int MRFProblemArray::NodeCost(int nodeID, int labelID) const
{
	assert(false);	// needs to be implemented in a child if not provided
}
	
int MRFProblemArray::NNeighbors(int nodeID) const
{
	assert(m_pGraph!=NULL);
	return ((VKGraph*)m_pGraph)->Node(nodeID)->NEdges();
}

int MRFProblemArray::Neighbor(int nodeID, int nID, int * weight) const
{
	assert(m_pGraph!=NULL);
	VKGraphNode * n = ((VKGraph*)m_pGraph)->Node(nodeID);
	*weight = (int)(n->Edge(nID)->weight * m_RescaleDouble2Int);
	return n->Neighbor(nID)->id;
}

int MRFProblemArray::LabelCost(int label1ID, int label2ID) const
{
	assert(false);	// needs to be implemented if not specified
}

const int * MRFProblemArray::DataCosts() const
{
	return m_aDataCost;
}

const int * MRFProblemArray::LabelsCosts() const
{
	return m_aLabelsCost;
}



/////  MRFOptimizer /////
MRFOptimizer::MRFOptimizer(MRFProblem * problem)
{
	try 
	{
		m_pProblem = problem;
		assert(m_pProblem!=NULL && m_pProblem->NLabels()>0 && m_pProblem->NNodes());
		//std::cout<<"graph: "<<m_pProblem->NNodes()<<" x "<<m_pProblem->NLabels()<<std::endl;
		m_pGeneralGraphOptim = new GCoptimizationGeneralGraph(m_pProblem->NNodes(), m_pProblem->NLabels());
		UpdateDataCosts();
		MRFProblemArray * probArr = m_pProblem->SafeCastToArray();
		if (probArr!=NULL)
		{
			m_pGeneralGraphOptim->setSmoothCost((int*)probArr->LabelsCosts());
			m_pGeneralGraphOptim->setLabelCost(0);
			for (int n1=0; n1<m_pProblem->NNodes(); n1++)
			for (int nn=0; nn<m_pProblem->NNeighbors(n1); nn++)
			{
				int w;
				int n2 = m_pProblem->Neighbor(n1, nn, &w);
				assert(n2 < m_pProblem->NNodes());
				//std::cout<<"SetNhd: "<<n1<<" "<<n2<<" "<<w<<" / "<<m_pProblem->NNodes()<<std::endl;
				m_pGeneralGraphOptim->setNeighbors(n1, n2, (int)w);
			}
		}
		else
			assert(false);	
	}
	catch (GCException exception)
	{
		exception.Report();
		assert(false);
	}
	catch (...) 
	{
		std::cout<<"[ERROR] MRF Initialization failed with unknown exception"<<std::endl;
		assert(false);
	}
}

MRFOptimizer::~MRFOptimizer()
{
	delete m_pGeneralGraphOptim;
}

void MRFOptimizer::Optimize(int nIterations)
{
	// run expansion for 2 iterations. For swap use gc->swap(num_iterations);
	try 
	{
		m_pGeneralGraphOptim->expansion(nIterations);
	}
	catch (GCException exception)
	{
		exception.Report();
		assert(false);
	}
	catch (...) 
	{
		std::cout<<"[ERROR] MRF Optimization failed with unknown exception"<<std::endl;
		assert(false);
	}
}

int MRFOptimizer::Energy(int * data, int * smooth, int * label) const
{
	if (data!=NULL)
		*data = m_pGeneralGraphOptim->giveDataEnergy();
	if (smooth!=NULL)
		*smooth = m_pGeneralGraphOptim->giveSmoothEnergy();
	if (label!=NULL)
		*label = m_pGeneralGraphOptim->giveLabelEnergy();
	return m_pGeneralGraphOptim->compute_energy();
}

void MRFOptimizer::UpdateDataCosts()
{
	MRFProblemArray * probArr = m_pProblem->SafeCastToArray();
	
	if (probArr!=NULL)
		m_pGeneralGraphOptim->setDataCost((int*)probArr->DataCosts());
	else
		assert(false);
}

void MRFOptimizer::SetLabel(int nodeID, int labelID)
{
	m_pGeneralGraphOptim->setLabel(nodeID, labelID);
}

int MRFOptimizer::Label(int nodeID) const
{
	return m_pGeneralGraphOptim->whatLabel(nodeID);
}






