#ifndef __Clustering_H
#define __Clustering_H

#include "LinAlgMatrixReal.h"

class ClusteringAlgorithm;

struct ClusteredNode
{
	int id;
	float error;
};

/**
 * Describes a clustering problem.
 * i.e. stores nodes + features...
 */ 
class ClusteringProblem
{
	public:
		
		friend class ClusteringAlgorithm;

		ClusteringProblem(int nNodes, int nFeatures);
		ClusteringProblem(LinAlgMatrixReal * features);
		virtual ~ClusteringProblem();
		virtual void SetFeatures(int nodeID, const double * features);
		virtual void SetFeatures(int nodeID, const float * features);
		
	// problem
		virtual int NNodes();
		virtual int NFeatures();
	
	// solution
		virtual int NClusters();
		virtual int ClusterSize(int clusterID);
		virtual const std::vector<ClusteredNode> & GetCluster(int clusterID);
		virtual void SetNodeErr(int clusterID, int nodeID, double err);
		virtual std::vector<int> SortClusters();
		virtual void ClearSolution();
		
		virtual void WriteSolution(const char * filename);
		virtual void ReadSolution(const char * filename);
	
		bool m_bVerbose;
	
	protected:
		std::vector<std::vector<ClusteredNode> > m_Clusters;
		std::vector<double> m_ClusterTightness;
		LinAlgMatrixReal * m_pFeatures;
};

/**
 * Clustering algorithm interface
 */
class ClusteringAlgorithm
{
	public:
		virtual void Cluster(ClusteringProblem & problem) = 0;
	
	protected:
		virtual LinAlgMatrixReal * Features(ClusteringProblem & problem);
		virtual std::vector<std::vector<ClusteredNode> > & Clusters(ClusteringProblem & problem);
		virtual std::vector<double> & ClusterTightness(ClusteringProblem & problem);
};

/**
 * Collapse nearest: iteratively join nearest clusters 
 * until certain number of clusters is reached or threshold exceeds
 * (note that the threshold can be estimated from the data)
 */ 
class ClusterNearest : public ClusteringAlgorithm
{
	public:	
		ClusterNearest();
		ClusterNearest(double threshold);
		ClusterNearest(int nClusters);
		virtual ~ClusterNearest();
		virtual void Cluster(ClusteringProblem & problem);
	
	protected:		
		double m_fMaxDistance;
		int m_iMaxClusters;		
};

/**
 * simple k-means clustering
 */ 
class KMeansClustering : public ClusteringAlgorithm
{
	public:
		KMeansClustering(int nClusters, int maxIterations = 1000, 
						 const LinAlgMatrixReal * seedCentroids = NULL);
		virtual ~KMeansClustering();
		virtual void Cluster(ClusteringProblem & problem);
	
	protected:
		int m_iMaxIterations;
		int m_iNClusters;
		LinAlgMatrixReal * m_pCentroids;
		int * m_CentroidNorms;
};

// todo: mean shift, collapse nearest

#endif

