#ifdef INCLUDE_FLANN
	// see http://people.cs.ubc.ca/~mariusm/index.php/FLANN/FLANN
	#include "flann/flann.hpp"
#endif


#include "LinAlgMatrixSparseReal.h"
#include "LinAlgMatrixReal.h"
#include "LinAlgVectorReal.h"


#ifndef __DiffusioMap_H
#define __DiffusioMap_H

/** 
 * Only keeps diffusion distances + search space in 'spectral embedding space'.
 * NOTE: only FLANN version is currently supported (for nearest-neighbor search)
 *			If necessary - FLANN can also support readius search.
 */ 
class DiffusionMap
{	
	public:
		enum DiffusedSearchType
		{
			SEARCH_EXHAUSTIVE,
			SEARCH_USE_FLANN
		};
	
	// initializing the diffusion map
		DiffusionMap(int N, int nEigs, double diffTime = 1.,
					 const std::vector<int> & numPointsInGroup = std::vector<int>(),
					 DiffusedSearchType srch = SEARCH_USE_FLANN, 
					 SparseMtxDataTypes mtxLib = SPARSE_MTX_GENERAL);
		~DiffusionMap();
		void FreeData();
	
	// fill the correspondences and embed
		void AddCorrValue(int i, int j, double val);
		void ClearAllCorrValues();
		void Embed(bool addIdentities=true, bool symmetrize=true, 
				   bool rowNormalize=true, bool normalizeEigenvectors=true, 
				   double * timeInSec = NULL, const char * prefix = NULL);

		// Queries: search for neighbors 
		//		(note: relatively slow - cache output for real-time apps!)
		double ComputeExactDistance(int i, int j) const;
		void ClosestPntsKNN(int pntID, int numNN,
							int ** closestIDs, double ** closestDists, int * nPoints);
		void ClosestPntsInGroupKNN(int pntID, int groupID, int numNN,
								   int ** closestIDs, double ** closestDists, int * nPoints);

	// Map parameters (some can be changed interactively)
		int NEntries() const;
		int NEigs() const;
	
		double DiffTime() const;
		void SetTime(double t);
		void SetSearchType(DiffusedSearchType srchTp);
	
		void ClearNNCache();
	
	// Get raw data (just in case)
		void Eigs(LinAlgMatrixReal ** eigVecs, LinAlgVectorReal ** eigVals);
		void Corrs(LinAlgMatrixSparseReal ** corrs);
	
	// reading / writing the data: 
		void WriteBinaryData(FILE * file) const;
		void ReadBinaryData(FILE * file) const;
		
	protected:	
	// to save: input params
		int m_iN;
		int m_iNEigs;
		double m_fDiffTime;
		SparseMtxDataTypes m_MtxLib;
		DiffusedSearchType m_SearchType;	
		
	// to save: correspondence values
		LinAlgMatrixSparseReal * m_pCorrVals;
		
	// to save: eigen-decomposition (embedded values)
		LinAlgMatrixReal * m_pEigenVectors;
		LinAlgVectorReal * m_pEigenValues;
	
	// to save: grouping (e.g. points per model)	
		int m_iNGroups;
		int * m_aPointToGroup;
		int * m_aGroupToOffset;
		int * m_aGroupToNumPoints;

	// temporary: nearest-neighbor search data and functions
		void FillInputOutputPtrs();
#ifdef INCLUDE_FLANN
		flann::Index<flann::L2<double> > * m_pEmbeddedSpaceANN;
		flann::Matrix<double> * m_pANNDataset;
		flann::Matrix<int> * m_pANNOutIndices;
		flann::Matrix<double> * m_pANNOutDistances;
		flann::Matrix<double> * m_pANNQuery;
		void BuildFLANNIndex();
	
	// per group (e.g. per model) flann
		flann::Index<flann::L2<double> > ** m_pEmbeddedSpaceANNPerGroup;
		flann::Matrix<double> ** m_pANNDatasetPerGroup;
		void BuildFLANNForGroup(int groupID);
		void FillInputOutputPtrsFLANN();
#endif
	// temporary: NN library data
		double * m_aQueryFeatures;
		int * m_aClosestPntIDs;
		double * m_aClosestDistances;
};

#endif





