
#include "LinAlg.h"
#include "svdlib.h"
#include <math.h>
#include "LinAlgVectorReal.h"
#include "LinAlgMatrixReal.h"

class LinAlgMatrixComplex;
class LinAlgVectorReal;

#ifndef __LINALG_MATRIX_SPARSE_REAL_H
#define __LINALG_MATRIX_SPARSE_REAL_H

// Implementations:
//		LinAlgMatrixSparseRealCXSparse
//		LinAlgMatrixSparseLibSVBC
//		LinAlgMatrixSparseRealARPACK
//		LinAlgMatrixSparseRealGeneral

//#define WITH_CXSPARSE 
//#define WITH_ARPACK

enum SparseMtxDataTypes
{
	SPARSE_MTX_CXSPARSE,
	SPARSE_MTX_LIBSVDC,
	SPARSE_MTX_ARPACK,
	SPARSE_MTX_GENERAL		// mostly uses matlab
};

class LinAlgMatrixSparseReal
{
	public:
		friend class LinAlgVectorReal;
		friend class LinAlgMatrixComplex;	
		friend class LinAlgMatrixReal;

		LinAlgMatrixSparseReal(int m, int n, SparseMtxDataTypes tp);
		virtual ~LinAlgMatrixSparseReal();
		
		virtual void AddVal(int i, int j, double val) = 0;
		virtual SparseMtxDataTypes GetMatrixDataType() const;
		virtual int Rows() const;
		virtual int Cols() const;

	protected: 
		SparseMtxDataTypes m_SparseMtxType;
		int m, n;
};

#ifdef WITH_CXSPARSE
/**
 * Supports: factorization, solving system of linear equations
 */
class LinAlgMatrixSparseRealCXSparse : public LinAlgMatrixSparseReal
{
	public:
		LinAlgMatrixSparseRealCXSparse(int m, int n, int maxNonzero = -1);
		virtual ~LinAlgMatrixSparseRealCXSparse();
	
		virtual void AddVal(int i, int j, double val);
			// stores factorization result in the matrix
		virtual void Factorize(LinAlg::MatrixFactorizationType factType);	
		virtual void SolveFactorized(const LinAlgVectorReal & b, LinAlgVectorReal &x) const;
			// note: matrix content might be destroyed
		virtual void Solve(const LinAlgVectorReal & b, LinAlgVectorReal &x);

	
	protected:
		LinAlg::MatrixFactorizationType m_PreFactorized;
		cs_di * m_pData;
		cs_di * m_pPreFactorizedData;
		css * m_pPreFactorizedS;
		csn * m_pPreFactorizedN;
		double * m_aPreAllocatedX;	// solution to A x = b
		double * m_aPreFactorizedWorkspace;
};
#endif

/**
 * Supports: symmetric eigendecomposition
 */ 
class LinAlgMatrixSparseRealLibSVDC : public LinAlgMatrixSparseReal
{
	public:
		LinAlgMatrixSparseRealLibSVDC(int m, int n, int maxNonzero);
		virtual ~LinAlgMatrixSparseRealLibSVDC();
		virtual void AddVal(int i, int j, double val);
			// NOTE: only for symmetric matrices
		virtual void EigsSymmetric(LinAlgVectorReal & eigenValues, 
								   LinAlgMatrixReal & eigenVectors, 
								   int dimensions);

	protected:
		SMat m_pSVDLibData;
		int m_iCurrSVDLibDataCol;
		int m_iCurrSVDLibDataRow;
		int m_iCurrentSVDLibDataN;
};

/**
 * Supports: runs MATLAB (as an external script)
 */
class LinAlgMatrixSparseRealGeneral : public LinAlgMatrixSparseReal
{
	public:
		LinAlgMatrixSparseRealGeneral(int m, int n);
		LinAlgMatrixSparseRealGeneral(const LinAlgMatrixSparseRealGeneral & M);
		virtual ~LinAlgMatrixSparseRealGeneral();
	
		virtual double GetVal(int i, int j);
		virtual void AddVal(int i, int j, double val);
		// A = (A + A') / 2.
		virtual void Symmetrize();
		virtual void RowNormalize(LinAlgVectorReal * saveRows = NULL, double norm = 1);
		virtual void ColNormalize(LinAlgVectorReal * saveCols = NULL, double norm = 1);

		virtual void Eigs(LinAlgVectorReal & eigenValues, 
						  LinAlgMatrixReal & eigenVectors, 
						  int dimensions, 
						  const char * uniquePrefix = "none",
						  double * timeInSec = NULL);
		
		virtual void WriteMatrixBin(const char * filename);
		virtual void ReadMatrixBin(const char * filename);
		
		virtual void WriteMatrixBin(FILE * file);
		virtual void ReadMatrixBin(FILE * file);
	
	protected:
		std::map<int, double> * m_ArrayOfColsMapRows;
};

//#ifdef WITH_ARPACK
///**
// * Supports: general eigendecomposition
// */
//class LinAlgMatrixSparseRealARPACK : public LinAlgMatrixSparseReal
//{
//	public:
//	
//};
//#endif


// TODO: conversions
//virtual LinAlgMatrixSparseRealGeneral * ConvertMatrixToGeneral();



#endif
