/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ShapeCorrs.h"
#include "CorrValue.h"
#include "ShapePointCloud.h"
VKString ShapeCorrs::m_sDefaultCorrValName = "default";

VKString ShapeCorrs::CoreCorrsTypeToStr(CoreCorrsType tp)
{
	if (tp==CORECORRS_CorrsCollection)
		return "CORECORRS_CorrsCollection";
	else if (tp==CORECORRS_CorrsCollectionMap)
		return "CORECORRS_CorrsCollectionMap";
	else if (tp==CORECORRS_CorrsPair)
		return "CORECORRS_CorrsPair";
	else if (tp==CORECORRS_CorrsPairMap)
		return "CORECORRS_CorrsPairMap";
	else if (tp==CORECORRS_CorrsOther)
		return "CORECORRS_CorrsOther";
	else
		assert(false);	
}

ShapeCorrs::CoreCorrsType ShapeCorrs::StrToCoreCorrsType(const VKString & str)
{
	if (str=="CORECORRS_CorrsCollection")
		return CORECORRS_CorrsCollection;
	else if (str=="CORECORRS_CorrsCollectionMap")
		return CORECORRS_CorrsCollectionMap;
	else if (str=="CORECORRS_CorrsPair")
		return CORECORRS_CorrsPair;
	else if (str=="CORECORRS_CorrsPairMap")
		return CORECORRS_CorrsPairMap;
	else if (str=="CORECORRS_CorrsOther")
		return CORECORRS_CorrsOther;
	else
		assert(false);
}


ShapeCorrs::ShapeCorrs(const VKString & corrsName, CoreCorrsType corrsType, DBCorrs * corrs)
{
	m_sCorrName = corrsName;
	m_CorrsTp = corrsType;
	m_pDBCorrs = corrs;
}

ShapeCorrs::~ShapeCorrs()
{
	
}

VKString ShapeCorrs::GetCorrsClassName() const
{
	return "ShapeCorrs";
}

const VKString & ShapeCorrs::GetCorrsName() const
{
	return m_sCorrName;
}

ShapeCorrs::CoreCorrsType ShapeCorrs::GetCorrsType()
{
	return m_CorrsTp;
}

bool ShapeCorrs::IsShapeSupportedDomain(const GeomShape * shape) const
{
	std::vector<const GeomShape *> domain;
	Domain(domain);
	for (int i=0; i<(int)domain.size(); i++)
		if (domain[i]==shape)
			return true;
	return false;
}

bool ShapeCorrs::IsShapeSupportedRange(const GeomShape * shape) const
{
	std::vector<const GeomShape *> range;
	Range(range);	
	for (int i=0; i<(int)range.size(); i++)
		if (range[i]==shape)
			return true;
	return false;
}

bool ShapeCorrs::PointInDomain(const ShapePoint & pnt) const
{
	return IsShapeSupportedDomain(pnt.GetPntsSet());
}

int ShapeCorrs::PairwiseCorrs(const ShapePoint & pnt,
							   GeomShape * toShape,
							   std::vector<ShapePoint> & corrs, 
							   const VKString & corrValName,
							   std::vector<double> * corrValue)
{
	std::vector<ShapePoint> allCorrs;
	std::vector<double> allCorrVals;
	std::vector<double> * allCorrValsPtr = NULL;
	if (corrValue!=NULL)
		allCorrValsPtr = &allCorrVals;
	
	int nCorrs = GetCorrs(pnt, allCorrs, corrValName, allCorrValsPtr);
	assert(nCorrs==(int)allCorrs.size());
	if (corrValue!=NULL)
		assert(nCorrs==(int)allCorrVals.size());
	
	for (int i=0; i<nCorrs; i++)
	{
		if (allCorrs[i].SameMeshGeometry(toShape))
		{
			corrs.push_back(allCorrs[i]);
			if (corrValue!=NULL)
				corrValue->push_back(allCorrVals[i]);
		}
	}
	return (int)corrs.size();
}

ShapePoint ShapeCorrs::BestCorr(const ShapePoint & pnt,
								GeomShape * toShape,
								const VKString & corrValName,
								double * corrValue)
{
	std::vector<ShapePoint> allCorrs;
	std::vector<double> allCorrVals;
	
	int nCorrs = GetCorrs(pnt, allCorrs, corrValName, &allCorrVals);
	assert(nCorrs==(int)allCorrs.size());
	assert(nCorrs==(int)allCorrVals.size());
	
	int bestCorrID = -1;
	double bestCorrValue = -1;
	for (int i=0; i<nCorrs; i++)
	{
		if (allCorrs[i].SameMeshGeometry(toShape)
			 && allCorrVals[i] > bestCorrValue)
		{
			bestCorrID = i;
			bestCorrValue = allCorrVals[i];
		}
	}
	
	if (corrValue!=NULL)
		*corrValue = bestCorrValue;
	
	if (bestCorrID < 0)
		return ShapePoint();
	
	return allCorrs[bestCorrID];
}

void ShapeCorrs::WriteShapeCorrs(const VKString & filename)
{
	FILE * file = fopen(filename.c_str(), "wb");
	WriteShapeCorrs(file);
	fclose(file);
}

void ShapeCorrs::ReadShapeCorrs(const VKString & filename)
{
	FILE * file = fopen(filename.c_str(), "rb");
	ReadShapeCorrs(file);
	fclose(file);	
}

void ShapeCorrs::WriteShapeCorrs(FILE * file)
{
	std::cout<<"[WARNING] WriteShapeCorrs not implemented in the child"<<std::endl;
}

void ShapeCorrs::ReadShapeCorrs(FILE * file)
{
	std::cout<<"[WARNING] ReadShapeCorrs not implemented in the child"<<std::endl;
}

void ShapeCorrs::AddCorrValue(const VKString & corrValName, 
							  CorrValue * corrVal)
{
	assert(m_NameToCorrValID.find(corrValName)==m_NameToCorrValID.end());
	m_NameToCorrValID[corrValName] = NExtraCorrValues();
	m_CorrVals.push_back(corrVal);
}

int ShapeCorrs::NExtraCorrValues() const 
{
	return (int)m_CorrVals.size();
}

const VKString & ShapeCorrs::CorrValueName(int corrValID) const 
{
	return m_CorrVals[corrValID]->CorrValueName();
}

bool ShapeCorrs::CorrValueSupported(const VKString & corrValName) const 
{
	if (corrValName==m_sDefaultCorrValName)
		return IsDeafultSupported();
	return (m_NameToCorrValID.find(corrValName)!=m_NameToCorrValID.end());
}

bool ShapeCorrs::IsDeafultSupported() const 
{
	return false;
}

double ShapeCorrs::DefaultIntegratedValue(const VKString & ) const
{
	return -1;
}

double ShapeCorrs::ExtraCorrValue(const ShapePoint & p1, 
							 const ShapePoint & p2,
							 const VKString & corrValName)
{
	CorrValue * cv = GetExtraCorrValue(corrValName);
	return cv->Value(p1, p2);
}

void ShapeCorrs::ExtraCorrValues(const ShapePoint & p1, 
								 std::vector<ShapePoint> & p2,
								 const VKString & corrValName, 
								 std::vector<double> & vals)
{
	CorrValue * cv = GetExtraCorrValue(corrValName);
	for (int i=0; i<(int)p2.size(); i++)
 		vals.push_back(cv->Value(p1, p2[i]));
}

CorrValue * ShapeCorrs::GetExtraCorrValue(const VKString & corrValName)
{
	assert(corrValName!=m_sDefaultCorrValName);
	std::map<VKString, int>::iterator iter = m_NameToCorrValID.find(corrValName);
	assert(iter!=m_NameToCorrValID.end());
	return m_CorrVals[iter->second];
}


