/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include <vector>
#include <map>
#include <iostream>
#include <assert.h>

#include "ParamParser.h"

#include "Model3D.h"
#include "ShapeManifoldMesh.h"
#include "ShapeTriangleSoup.h"
#include "FileManagement.h"
#include "ModelDatabase.h"
#include "ShapePointCloud.h"
#include "DistEuclidean.h"
#include "DistGeodesicDijkstra.h"
#include "Model3DParser.h"
#include "ShapeImg.h"
#include "ShapeDescriptor.h"
#include "ShapeValue.h"
#include "ShapeMaterials.h"
#include "ShapeLabels.h"


VKStringList Model3D::m_sEmptyStrList;
const VKString Model3D::m_sMainGeomName = "main";
const VKString Model3D::m_sThumbGeomName = "thumb";

std::map<VKString, Model3DParser*>  Model3D::m_TypeToModelParser;

void Model3D::AddParser(const VKString & pName, Model3DParser * parser)
{
	assert(m_TypeToModelParser.find(pName)==m_TypeToModelParser.end());
	m_TypeToModelParser[pName] = parser;
}

Model3D * Model3D::CreateFromFile(const VKString & filename, int readingFlags, int creatingFlags,
								  const VKString & modelName, int modelID)
{
	Model3D * model = new Model3D(modelName, modelID);
	
	if (filename.endsWith(".model3d"))
		model->ReadModelDBFormat(filename.FilePath(), filename, readingFlags);
	else if (filename.endsWith(".pnts"))
	{
		// TODO: point cloud (without mesh)
		assert(false);
	}
	else if (filename.endsWith(".seg"))		// psb segmentation...
		model->ReadModelPSBSegmentedFormat(filename.FilePath(), filename);
	else if (filename.endsWith(".psb"))
		model->ReadModelPSBFormat(filename.FilePath(), filename);
	else 
	{
		bool couldRead = model->ReadUnrecognizedFormat(filename.FilePath(), filename);
		if (!couldRead) // should be a mesh
		{
			VKString geomName = m_sMainGeomName;
			
			GeomShape * geometry = NULL;
			if (readingFlags & (ModelDatabase::READ_MESH_AS_SOUP))
				geometry = new ShapeTriangleSoup(filename, geomName, model);
			else
				geometry = new ShapeManifoldMesh(GeomShape::SHAPE_TYPE_MANIFOLD,
												 filename, geomName, model);
			model->AddGeomShape(geomName, geometry);		
		}
	}
		
	// create stuff
	if (creatingFlags!=0)
	{
		// add vertices
		int nGeom = model->NGeometries();
		for (int i=0; i<nGeom; i++)
		{
			ShapePointCloud * vts = NULL; 
			GeomShape * geom = model->GetGeomShape(i);
			GeomShape::GeomShapeTypes tp = geom->GetShapeType();
			
			// only for meshes
			if (tp==GeomShape::SHAPE_TYPE_MANIFOLD 
//				|| tp==GeomShape::SHAPE_TYPE_MID_EDGE_FLATTENED
				|| tp==GeomShape::SHAPE_TYPE_TRIANGLE_SOUP)
			{
				// all vertices
				if (creatingFlags & ModelDatabase::CREATE_PNTS_ALLVTX)
				{
					VKString pName = geom->GeometryName()+"_"+ShapePointCloud::m_sDefaultAllVtx;
					vts = ShapePointCloud::CreateAllVtxSet(pName, geom->GeometryName(), model);
					model->AddGeomShape(pName, vts);
				}
				
				// all faces
				if (creatingFlags & ModelDatabase::CREATE_PNTS_ALLFACES)
				{
					VKString pName = geom->GeometryName()+"_"+ShapePointCloud::m_sDefaultAllFace;
					vts = ShapePointCloud::CreateAllFacesSet(pName, geom->GeometryName(), model);
					model->AddGeomShape(pName, vts);
				}
			}
		}
		
		// add distances to 'main' geometry
		GeomShape * mainShape = model->GetGeomShape(m_sMainGeomName);
		// euclidean distances
		if (creatingFlags & ModelDatabase::CREATE_DIST_EUCLIDEAN)
		{
			assert(mainShape!=NULL);
			DistEuclidean * eucl = new DistEuclidean(mainShape, ShapeDistance::m_sDefaultEuclidean);
			mainShape->AddDistanceMetric(ShapeDistance::m_sDefaultEuclidean, eucl);
		}
		// geodesic distances
		GeomShape::GeomShapeTypes tp = mainShape->GetShapeType();
		if (tp==GeomShape::SHAPE_TYPE_MANIFOLD )
		{
			if (creatingFlags & ModelDatabase::CREATE_DIST_GEODESIC)
			{
				VKString geodName = ShapeDistance::m_sDefaultGeodesic;
				assert(mainShape!=NULL);
				DistGeodesicDijkstra * geod = new DistGeodesicDijkstra(mainShape, geodName);
				mainShape->AddDistanceMetric(geodName, geod);
			}
		}
	}
	
	return model;
}

Model3D::Model3D(const VKString & modelName,
				 int modelID)
{
	m_sInfoFile = "none";
	m_pMetaData = NULL;
	m_ModelID = modelID;
	m_ModelName = modelName;
}

Model3D::~Model3D()
{
	for (int i=0; i<(int)m_GeomShapes.size(); i++)
		delete m_GeomShapes[i];
}

void Model3D::SetMetaData(const VKString & infoFile, ParamParser * metaData)
{
	m_sInfoFile = infoFile;
	m_pMetaData = metaData;
}

VKStringList & Model3D::GetMetaData(const VKString & field, bool * exists)
{
	*exists = false;
	if (m_pMetaData==NULL)
		return m_sEmptyStrList;
	
	if (m_pMetaData->NModules()==1)
	{
		VKString moduleName = m_pMetaData->ModuleName(0);
		return m_pMetaData->GetStrValues(moduleName, field, *exists);		
	}
	return m_sEmptyStrList;
}

void Model3D::WriteMetaData()
{
	assert(m_pMetaData!=NULL);
	assert(m_sInfoFile!="none");
	m_pMetaData->WriteParams(m_sInfoFile);
}

void Model3D::CreateMetaField(const VKString & field)
{
	assert(m_pMetaData!=NULL);
	assert(m_pMetaData->NModules()==1);
	m_pMetaData->SetParam(m_pMetaData->ModuleName(0), field); 
}

int Model3D::NGeometries()
{
	return (int)m_GeomShapes.size();
}

GeomShape * Model3D::GetGeomShape(int shapeID)
{
	assert(shapeID >= 0 && shapeID < (int)m_GeomShapes.size());
	return m_GeomShapes[shapeID];
}

GeomShape * Model3D::GetGeomShape(const VKString & shapeName)
{
	std::map<VKString, int>::iterator iter = m_GeomShapeToID.find(shapeName);
	if (iter==m_GeomShapeToID.end())
	{
//		std::cout<<"[WARNING] Cannot find geometry: "<<shapeName.c_str();
//		std::cout<<" in model "<<m_ModelName.c_str()<<std::endl;
		return NULL;
	}
	assert(iter!=m_GeomShapeToID.end());
	return m_GeomShapes[iter->second];
}

const GeomShape * Model3D::GetGeomShape(int shapeID) const
{
	return ((Model3D*)this)->GetGeomShape(shapeID);
}

const GeomShape * Model3D::GetGeomShape(const VKString & shapeName) const
{
	return ((Model3D*)this)->GetGeomShape(shapeName);
}

bool Model3D::GeometryExists(const VKString & shapeName) const
{
	std::map<VKString, int>::iterator iter;
	iter = ((std::map<VKString, int>&)m_GeomShapeToID).find(shapeName);
	return (iter!=m_GeomShapeToID.end());
}

void Model3D::AddGeomShape(const VKString & shapeName, GeomShape * shape)
{
	assert(m_GeomShapeToID.find(shapeName)==m_GeomShapeToID.end());
	m_GeomShapeToID[shapeName] = (int)m_GeomShapes.size();
	m_GeomShapes.push_back(shape);
}

void Model3D::RemoveGeomShape(const VKString & shapeName)
{
	std::map<VKString, int>::iterator iter = m_GeomShapeToID.find(shapeName);
	int locID = iter->second;
	if (iter==m_GeomShapeToID.end())
		std::cout<<"[ERROR] Cannot remove shape from model: "<<shapeName.c_str()<<std::endl;;
	assert(iter!=m_GeomShapeToID.end());
	assert(locID < (int)m_GeomShapes.size());
	m_GeomShapes.erase(m_GeomShapes.begin() + locID);
	m_GeomShapeToID.erase(iter);
	for (std::map<VKString, int>::iterator i = m_GeomShapeToID.begin(); 
		 i != m_GeomShapeToID.end(); i++)
	{
		if (i->second > locID)
			i->second--;
		else
			assert(i->second<locID);
	}
}


void Model3D::SetModelName(const VKString & modelName)
{
	m_ModelName = modelName;
}

GeomShape * Model3D::ReadGeometry(GeomShape::GeomShapeTypes tp, const VKString & geomName, 
								  const VKString & geomFile, const VKString & pntsCloudMesh)
{
	GeomShape * shape = CreateGeometry(tp, geomName, pntsCloudMesh);
	assert(shape!=NULL);
	shape->ReadGeometryFile(geomFile);
	AddGeomShape(geomName, shape);
	return shape;
}

GeomShape * Model3D::CreateGeometry(GeomShape::GeomShapeTypes tp, 
									const VKString & geomName, 
									const VKString & pntCloudMeshName)
{
	GeomShape * shape = NULL;
	if (tp == GeomShape::SHAPE_TYPE_POINT_CLOUD)
	{
		shape = new ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD,
									geomName, pntCloudMeshName, this);
	}
	else if (tp==GeomShape::SHAPE_TYPE_POINT_CLOUD_ALLVTX)
	{
		assert(pntCloudMeshName!="none");
		shape = ShapePointCloud::CreateAllVtxSet(geomName, pntCloudMeshName, this);
	}
	else if (tp==GeomShape::SHAPE_TYPE_POINT_CLOUD_ALLFACE)
	{
		assert(pntCloudMeshName!="none");
		shape = ShapePointCloud::CreateAllFacesSet(geomName, pntCloudMeshName, this);
	}
	else if (tp == GeomShape::SHAPE_TYPE_MANIFOLD)
		shape = new ShapeManifoldMesh(GeomShape::SHAPE_TYPE_MANIFOLD,
									  "none", geomName, this);
//	else if (tp == GeomShape::SHAPE_TYPE_MID_EDGE_FLATTENED)
//		shape = new ShapeManifoldMesh(GeomShape::SHAPE_TYPE_MID_EDGE_FLATTENED,
//									  "none", geomName, this);
	else if (tp == GeomShape::SHAPE_TYPE_IMAGE)
		shape = new ShapeImg(geomName, this);
	else if (tp == GeomShape::SHAPE_TYPE_DESCRIPTOR)
		shape = new ShapeDescriptor(geomName, this);
	else if (tp == GeomShape::SHAPE_TYPE_TRIANGLE_SOUP)
		shape = new ShapeTriangleSoup("none", geomName, this);
	else 
		assert(false);
	return shape;
}

ShapeValue * Model3D::ReadGeomValue(const VKString & className, GeomShape * shape, 
									const VKString & valueName, const VKString & shapeValFile)
{
	ShapeValue * val = CreateGeomValue(className, shape, valueName);
	VKString gName = shape->GeometryName();
	if (!FileManagement::FileExists(shapeValFile.c_str()))
		std::cout<<"[ERROR] Cannot find file: "<<shapeValFile.c_str()<<std::endl;
		
	assert(FileManagement::FileExists(shapeValFile.c_str()));
	if (className=="ShapeLabels" && shapeValFile.endsWith(".seg"))
		val->SafeCastToLabels()->ReadFromSegFile(shapeValFile);
	else if (className=="ShapeLabels" && shapeValFile.endsWith(".ply"))
		val->SafeCastToLabels()->ReadFromPlyFile(shapeValFile);
	else
		val->ReadShapeValues(shapeValFile);
	shape->AddShapeVal(valueName, val);
	return val;
}

ShapeValue * Model3D::CreateGeomValue(const VKString & className, 
									  GeomShape * shape, const VKString & valueName)
{
	if (className=="ShapeValue")
		return new ShapeValue(shape, valueName);
	else if (className=="ShapeMaterials")
		return new ShapeMaterials(shape, valueName);
	else if (className=="ShapeLabels")
		return new ShapeLabels(shape, valueName);
	else
	{
		std::cout<<"[ERROR] Cannot create geometry: "<<className.c_str()<<std::endl;
		assert(false);
	}
	
}

void Model3D::WriteModelDBFormat(const VKString & modelDir,
								 const VKString & modelFile,
								 int writingFlags, const VKString & writeArg)
{
	VKString geomDir = ModelDatabase::GetModelGeometryDir();
	if (writingFlags & ModelDatabase::WRITE_MODEL3D_FILES)
	{
		FileManagement::CreateDirsOnPathToFile(modelFile.c_str());
		
		std::ofstream textStream(modelFile.c_str());
		assert(textStream.is_open());
		textStream<<m_ModelName.c_str()<<":\n";

		// save geometry names to Model3D file
		textStream<<"\tNGeom\t"<<(int)m_GeomShapes.size()<<"\n";
		for (int i=0; i<(int)m_GeomShapes.size(); i++)
		{
			GeomShape * geom = m_GeomShapes[i];
			VKString geomName = geom->GeometryName();
			assert(geomName!="none");
			textStream<<"\tGeom"<<i<<"Name\t"<<geomName.c_str()<<"\n";
			GeomShape::GeomShapeTypes tp = geom->GetShapeType();
			textStream<<"\tGeom"<<i<<"Type\t"<<GeomShape::GetGeomShapeTypeStr(tp).c_str()<<"\n";
			VKString relFile = geomDir + "/"+geom->GeometryFile();
			textStream<<"\tGeom"<<i<<"File\t"<<relFile.c_str()<<"\n";
			if (tp==GeomShape::SHAPE_TYPE_POINT_CLOUD
				|| tp==GeomShape::SHAPE_TYPE_POINT_CLOUD_ALLVTX
				|| tp==GeomShape::SHAPE_TYPE_POINT_CLOUD_ALLFACE)
			{
				ShapePointCloud * pnts = geom->SafeCastToPointCloud();
				assert(pnts!=NULL);
				VKString meshName = pnts->GetMesh()->GeometryName();
				textStream<<"\tGeomPnts"<<i<<"MeshName\t"<<meshName.c_str()<<"\n";
			}
			
			// save model values
			textStream<<"\tGeom"<<i<<"NVals\t"<<geom->NShapeVals()<<"\n";
			for (int j=0; j<(int)geom->NShapeVals(); j++)
			{
				ShapeValue * geomVals = geom->GetShapeVal(j);
				assert(geomVals!=NULL);
				textStream<<"\tGeom"<<i<<"Val"<<j<<"Name\t";
				textStream<<geomVals->GetShapeValueName().c_str()<<"\n";
				textStream<<"\tGeom"<<i<<"Val"<<j<<"Class\t";
				textStream<<geomVals->GetClassName().c_str()<<"\n";
			}
		}
	
		textStream.close();
	}
	
//	std::cout<<"Saving geom files: "<<m_GeomShapes.size()<<std::endl;
	// save geometry files
	if ((writingFlags & ModelDatabase::WRITE_MODEL3D_GEOMS)!=0
		|| (writingFlags & ModelDatabase::WRITE_ARG_SINGLE_GEOM)!=0)
	{
		for (int i=0; i<m_GeomShapes.size(); i++)
		{
			GeomShape * geom = m_GeomShapes[i];

			if ((writingFlags & ModelDatabase::WRITE_MODEL3D_GEOMS)!=0
				|| writeArg==geom->GeometryName())
			{
				VKString geomFile = modelDir+"/" + geomDir+ "/" + geom->GeometryFile();
				FileManagement::CreateDirsOnPathToFile(geomFile.c_str());
				geom->WriteGeometryFile(geomFile);
				
				for (int j=0; j<(int)geom->NShapeVals(); j++)
				{
					ShapeValue * geomVals = geom->GetShapeVal(j);
					assert(geomVals!=NULL);
					VKString gName = geom->GeometryName();
					VKString gvName = geomVals->GetShapeValueName();
					VKString gValsFile;
					gValsFile = ModelDatabase::ShapeValueFilename(modelDir, gName, gvName);
					FileManagement::CreateDirsOnPathToFile(gValsFile.c_str());
					geomVals->WriteShapeValues(gValsFile);
				}
			}
		}
	}
}

bool Model3D::ReadUnrecognizedFormat(const VKString & directory, const VKString & filename)
{
	std::map<VKString, Model3DParser*>::iterator iter;
	bool couldRead = false;
	for (iter = m_TypeToModelParser.begin(); iter != m_TypeToModelParser.end(); iter++)
	{
		if (filename.endsWith(iter->first))
		{
			iter->second->LoadModelFile(this, filename, m_sMainGeomName, ModelDatabase::EVERYTHING);
			couldRead = true;
		}
	}
	return couldRead;
}

void Model3D::ReadModelPSBSegmentedFormat(const VKString & directory, const VKString & filename)
{
	assert(filename.endsWith(".seg"));
	// check *ply, *seg files
	VKString plyFile = filename;
	plyFile.replace(".seg", ".ply");
	VKString segFile = filename;
	if (!FileManagement::FileExists(plyFile.c_str()))
		std::cout<<"[ERROR] Cannot find: "<<plyFile.c_str()<<std::endl;
	assert(FileManagement::FileExists(plyFile.c_str()));
	assert(FileManagement::FileExists(segFile.c_str()));
	
	// parse *seg file
	std::ifstream segStream(segFile.c_str());
	assert(segStream.is_open());
	VKStringList segNames;
	while (!segStream.eof())
	{
		VKString segName;
		segName.readToken(segStream);
		if (segStream.eof())
			break;
		int sID;
		segStream>>sID;
		assert(sID==segNames.count());
		int lID = segName.lastIndexOf("-");
		segNames.push_back(segName.mid(0, lID));
		//std::cout<<"[debuginfo] segname: "<<segNames[sID].c_str()<<std::endl;
	}
	segStream.close();
	
	// parse *ply from PSB (NOT able to parse general PLY!!!)
	std::ifstream textStream(plyFile.c_str());
	assert(textStream.is_open());
	VKString tempStr;
	int nVert, nFace;
	tempStr.readLine(textStream);		assert(tempStr=="ply");
	tempStr.readLine(textStream);		assert(tempStr=="format ascii 1.0");
	tempStr.readToken(textStream);		assert(tempStr=="element");
	tempStr.readToken(textStream);		assert(tempStr=="vertex");
	textStream>>nVert;				
	tempStr.readLine(textStream);	// read end of line
	tempStr.readLine(textStream);		assert(tempStr=="property float x");
	tempStr.readLine(textStream);		assert(tempStr=="property float y");
	tempStr.readLine(textStream);		assert(tempStr=="property float z");
	tempStr.readToken(textStream);		assert(tempStr=="element");
	tempStr.readToken(textStream);		assert(tempStr=="face");
	textStream>>nFace;
	tempStr.readLine(textStream);	// read end of line
	tempStr.readLine(textStream);		assert(tempStr=="property list uchar int vertex_indices");
	tempStr.readLine(textStream);		assert(tempStr=="property int segment_id");
	tempStr.readLine(textStream);		assert(tempStr=="end_header");	
	
	ShapeTriangleSoup * mesh = new ShapeTriangleSoup("none", m_sMainGeomName, this);

		// read vertices
	double x, y, z;
	for (int v=0; v<nVert; v++)
	{
		textStream>>x>>y>>z;
		mesh->AddVertex(x, y, z);		
	}
	
	ShapeLabels * labels = new ShapeLabels(mesh, ShapeValue::m_sDefaultSceneGraphLabels);
	int nv;
	int * readVals = new int[nFace * 4];
	
	std::vector<std::vector<int> > groupedFaces;
	std::vector<SceneGraphLabel*> labelsTree;
	SceneGraphLabel * root = new SceneGraphLabel();
	root->m_sLabel = "PSB_root";
	root->m_iID = 0;
	root->m_pParent = NULL;
	labelsTree.push_back(root);
	
	std::vector<SceneGraphGeometry*> labelsGeoms;
		// read faces (with last integer == segment ID)
	for (int f=0; f<nFace; f++)
	{
		textStream>>nv;
		assert(nv==3);
		textStream>>readVals[f*4+0]>>readVals[f*4+1]>>readVals[f*4+2]>>readVals[f*4+3];
		assert(!textStream.eof());
		while (readVals[f*4+3] >= (int)groupedFaces.size())
			groupedFaces.push_back(std::vector<int>());
		groupedFaces[readVals[f*4+3]].push_back(f);
		
//		if (f==0)
//			std::cout<<"fv=["<<readVals[f*4+0]<<" "<<readVals[f*4+1]<<" "<<readVals[f*4+2]<<"]"<<std::endl;

	}
	textStream.close();
	
//	std::cout<<"Geometry: "<<filename.right(8).c_str()<<std::endl;
	std::vector<int> segIDs;
	for (int s=0; s<(int)groupedFaces.size(); s++)
	{
		SceneGraphLabel * newNode = new SceneGraphLabel();
		assert(s < segNames.count());
		newNode->m_sLabel = segNames[s];
		newNode->m_iID = (int)labelsTree.size();
		root->m_apChildren.push_back(newNode);
		newNode->m_pParent = root;
		int gID = (int)labelsGeoms.size();
		newNode->m_aiGeometryID.push_back(gID);
		if (groupedFaces.size()>0)
			newNode->m_iFirstFaceInGeometry.push_back(mesh->NFaces());
		else
			newNode->m_iFirstFaceInGeometry.push_back(-1);
		newNode->m_iNFacesInGeometry.push_back((int)groupedFaces[s].size());
		
		SceneGraphGeometry * newGeometry = new SceneGraphGeometry();
		newGeometry->m_iID = gID;
		newGeometry->m_sGeometryName = segNames[s];
		
		labelsTree.push_back(newNode);
		labelsGeoms.push_back(newGeometry);
		
		for (int fID=0; fID < (int)groupedFaces[s].size(); fID++)
		{
			segIDs.push_back(newNode->m_iID);
			int f = groupedFaces[s][fID];
			mesh->AddFace(readVals[f*4+0], readVals[f*4+1], readVals[f*4+2]);
		}
	}
	
	labels->SetFaceSegments(segIDs);
	labels->SetAllLabels(labelsTree);
	labels->SetAllGeometries(labelsGeoms);	
	
	mesh->AddShapeVal(labels->GetShapeValueName(), labels);
	AddGeomShape(mesh->GeometryName(), mesh);
	delete [] readVals;
}

void Model3D::ReadModelPSBFormat(const VKString & directory, const VKString & filename)
{
	ParamParser params(filename);
	bool valid;
	VKString geomType = params.GetStrValue("ModelPSB", "GeometryType", valid);
	
	GeomShape::GeomShapeTypes tp = GeomShape::SHAPE_TYPE_MANIFOLD;
	if (valid)
		tp = GeomShape::GetGeomShapeType(geomType);

	VKString geometryFile = params.GetStrValue("ModelPSB", "Geometry", valid);
	assert(valid);
	VKString relGeomFile = directory + "/" + geometryFile;
//	VKString fName = relGeomFile.FileName();
	GeomShape * shape = CreateGeometry(tp, m_sMainGeomName);
	shape->ReadGeometryFile(relGeomFile);
	AddGeomShape(m_sMainGeomName, shape);
	
	VKStringList gtPnts = params.GetStrValues("ModelPSB", "CorrPntSets", valid);
	if (valid)
	{
		VKStringList gtGroups = params.GetStrValues("ModelPSB", "CorrGroups");
		assert(gtPnts.count()==gtGroups.count());
		
		// for all feature points with ground truth
		for (int i=0; i<gtPnts.count(); i++)
		{
			ShapePointCloud * pnts = NULL;
			VKString pntsName = m_sMainGeomName+"_gtpts_"+gtGroups[i];
			VKString extFile = directory+"/"+gtPnts[i];
			if (gtPnts[i]=="AllVertices")
				pnts = ShapePointCloud::CreateAllVtxSet(pntsName, m_sMainGeomName, this);
			else
				pnts = ShapePointCloud::CreateFromExternalVtxList(pntsName, m_sMainGeomName, 
																  this, extFile);

			AddGeomShape(pntsName, pnts);
		}
		
		// TODO: load camera?
		
		VKStringList gtSymm = params.GetStrValues("ModelPSB", "Symm", valid);	
		// TODO: add symmetric map
	}
}

void Model3D::ReadModelDBFormat(const VKString & directory, 
								const VKString & filename, int readingFlags)
{
	if (readingFlags==0)
		return;
	
	std::ifstream textStream(filename.c_str());
	// main file should be there, geometries are optional, just print error
	if (!textStream.is_open())
		std::cout<<"[ERROR] Could not read: "<<filename.c_str()<<" (dir="<<directory.c_str()<<")"<<std::endl;
	assert(textStream.is_open());
	
	VKString tempStr;
	tempStr.readToken(textStream);
	assert(tempStr==(m_ModelName+":"));
	
	tempStr.readToken(textStream);
	assert(tempStr=="NGeom");
	int nShapes = 0;
	textStream>>nShapes;

	std::vector<int> geometries;
	VKStringList fileNames; 
	
	// parse file: read shapes
	for (int i=0; i<nShapes; i++)
	{
		tempStr.readToken(textStream);
		assert(tempStr==VKString("Geom")+VKString::number(i)+"Name");
		VKString geomName;
		geomName.readToken(textStream);
		tempStr.readToken(textStream);
		assert(tempStr==VKString("Geom")+VKString::number(i)+"Type");
		VKString geomType;
		geomType.readToken(textStream);
		GeomShape::GeomShapeTypes tp;
		tp = GeomShape::GetGeomShapeType(geomType);

		tempStr.readToken(textStream);
		assert(tempStr==VKString("Geom")+VKString::number(i)+"File");
		VKString relFile;
		relFile.readToken(textStream);
		fileNames.push_back(directory+"/"+relFile);
		
		geometries.push_back(NGeometries());
		VKString pntCloudMeshName = "none";
		if (tp==GeomShape::SHAPE_TYPE_POINT_CLOUD
			|| tp==GeomShape::SHAPE_TYPE_POINT_CLOUD_ALLVTX
			|| tp==GeomShape::SHAPE_TYPE_POINT_CLOUD_ALLFACE)
		{
			tempStr.readToken(textStream);
			assert(tempStr==VKString("GeomPnts")+VKString::number(i)+"MeshName");
			pntCloudMeshName.readToken(textStream);
		}
		GeomShape * shape = CreateGeometry(tp, geomName, pntCloudMeshName);
		assert(shape!=NULL);
		AddGeomShape(geomName, shape);
		
		tempStr.readToken(textStream);
		assert(tempStr==VKString("Geom")+VKString::number(i)+"NVals");
		VKString nValsStr;
		nValsStr.readToken(textStream);
		bool ok;
		int nVals = nValsStr.toInt(&ok);
		assert(ok);
		
		for (int j=0; j<nVals; j++)
		{
			VKString valName;
			VKString valClass;
			VKString prefix=VKString("Geom")+VKString::number(i)+"Val"+VKString::number(j);
			
			tempStr.readToken(textStream);
			assert(tempStr==prefix+"Name");
			valName.readToken(textStream);

			tempStr.readToken(textStream);
			assert(tempStr==prefix+"Class");
			valClass.readToken(textStream);

			ShapeValue * values = CreateGeomValue(valClass, shape, valName);
			shape->AddShapeVal(valName, values);
		}
	}
	textStream.close();
	
	// read geometries
	for (int i=0; i<(int)geometries.size(); i++)
	{
		GeomShape * shape = GetGeomShape(geometries[i]);
		shape->ReadGeometryFile(fileNames[i]);
		for (int j=0; j<shape->NShapeVals(); j++)
		{
			ShapeValue * val = shape->GetShapeVal(j);
			VKString gName = shape->GeometryName();
			VKString gvName = val->GetShapeValueName();
			VKString gValsFile;
			gValsFile = ModelDatabase::ShapeValueFilename(directory, gName, gvName);
			assert(FileManagement::FileExists(gValsFile.c_str()));
			val->ReadShapeValues(gValsFile);
		}
	}

}

void Model3D::RearrangeAxes(char newOrder[3], int newSigns[3])
{
	for (int i=0; i<(int)m_GeomShapes.size(); i++)
		m_GeomShapes[i]->RearrangeAxes(newOrder, newSigns);
}

const VKString & Model3D::ModelName() const 
{
	return m_ModelName;
}

const VKString & Model3D::ModelFile() const 
{
	assert(false);
}

int Model3D::ModelID() const
{
	return m_ModelID;
}
void Model3D::UpdateID(int newID)
{
	m_ModelID = newID;
}

