/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "Model3D.h"
#include "ModelDatabase.h"
#include "FileManagement.h"
#include "ShapePointCloud.h"
#include "TimeProfiler.h"

void ModelDatabase::GetModelFiles(const VKString & inputDir, const InputDirFormat dirFormat,
								  const VKStringList & listModelsToLoad, 
								  VKStringList & modelNames, VKStringList & modelFiles)
{
	TimeProfiler timer;
	if (!FileManagement::DirExists(inputDir.c_str()))
		std::cout<<"[ERROR] Cannot find directory: "<<inputDir.c_str()<<std::endl;
	assert(FileManagement::DirExists(inputDir.c_str()));
	
	if (listModelsToLoad.count()>500)	
		std::cout<<"Directory listing ("<<listModelsToLoad.count()<<" dirs)"<<std::endl;
	for (int i=0; i<listModelsToLoad.count(); i++)
	{
		if (listModelsToLoad.count()>500 && i%500==0)
			timer.WriteProgress("listmodels", i, listModelsToLoad.count());
		VKString mFile;
		VKString mName;
		
		if (dirFormat==DBINPUTDIR_MyFormat) // parse Models
		{	
			mFile = GetModelFilename(inputDir, listModelsToLoad[i]);
			mName = listModelsToLoad[i];
		}
		else if (dirFormat==DBINPUTDIR_MeshFiles)
		{
			mFile = inputDir + "/" + listModelsToLoad[i];
			mName = mFile.FileName();
		}
		else if (dirFormat==DBINPUTDIR_ColladaDirs)
		{
			VKString mDir = inputDir + "/" + listModelsToLoad[i] + "/models/";
			VKStringList modelFiles = FileManagement::ListAllFiles(mDir.c_str());
			if (modelFiles.count()!=1)
			{
				std::cout<<"[ERROR] Cannot interpret "<<modelFiles.count();
				std::cout<<" files in: "<<mDir.c_str()<<std::endl;
				for (int fID=0; fID < modelFiles.count(); fID++)
					std::cout<<"   "<<modelFiles[fID].c_str()<<std::endl;
			}
			assert(modelFiles.count()==1);
			mFile = mDir + modelFiles[0];
			mName = listModelsToLoad[i];
		}
		else
			assert(false);
		modelFiles.push_back(mFile);
		modelNames.push_back(mName);
	}
	if (listModelsToLoad.count()>500)
		std::cout<<std::endl;
}

VKStringList ModelDatabase::GetAllModelsInDir(const VKString & inputDir, 
											  const InputDirFormat dirFormat)
{
	VKStringList toLoad;
	if (!FileManagement::DirExists(inputDir.c_str()))
		std::cout<<"[ERROR] directory does not exist: "<<inputDir.c_str()<<std::endl;
	assert(FileManagement::DirExists(inputDir.c_str()));
	
	if (dirFormat==DBINPUTDIR_MyFormat)
	{
		// parse Models, Applications, CorrCollections dirs
		assert(FileManagement::DirExists((inputDir+"/Models").c_str()));

		VKString mListFile = inputDir+"/../ModelList.txt";
		// check if there is text file
		if (FileManagement::FileExists(mListFile.c_str()))
		{
			std::ifstream mListStream(mListFile.c_str());
			assert(mListStream.is_open());
			while (!mListStream.eof())
			{
				VKString currLine;
				currLine.readLine(mListStream);
				if (currLine!="")
					toLoad.push_back(currLine);
			}
			mListStream.close();
		}
		else			// if not list all files...
			toLoad = FileManagement::ListAllSubdirs((inputDir+"/Models").c_str());
	}
	else if (dirFormat==DBINPUTDIR_MeshFiles)
	{
		// list of all geometry files
		toLoad = FileManagement::ListAllFiles(inputDir.c_str());
		for (int i=0; i<toLoad.count(); i++)
		{
			if (toLoad[i].startsWith("."))
			{
				toLoad.remove(i);
				i--;
			}
		}
		
		if (toLoad.count()==0)
		{
			std::cout<<"[WARNING] Could not find mesh files in "<<inputDir.c_str()<<std::endl;
			if (FileManagement::DirExists((inputDir+"/Models").c_str()))
				std::cout<<"\tTry reading DB as MyFormat"<<std::endl;
			else if (FileManagement::ListAllSubdirs(inputDir.c_str()).count()>0)
				std::cout<<"\tTry reading DB as ColladaDirs"<<std::endl;
		}		
	}
	else if (dirFormat==DBINPUTDIR_ColladaDirs)
	{
		// list of dirs/models/*dae
		assert(FileManagement::DirExists(inputDir.c_str()));
		toLoad = FileManagement::ListAllSubdirs(inputDir.c_str());
	}
	else
		assert(false);
	return toLoad;
}

ModelDatabase * ModelDatabase::CreateSingleModelDB(Model3D * m, 
												   const VKString & dbDir,
												   const VKString & dbName)
{
	ModelDatabase * db = new ModelDatabase(dbDir, dbName);
	db->AddModel(m);
	return db;
}

ModelDatabase * ModelDatabase::CreatePairModelsDB(Model3D * m1, Model3D * m2, 
												  const VKString & dbDir,
												  const VKString & dbName)
{
	ModelDatabase * db = new ModelDatabase(dbDir, dbName);
	db->AddModel(m1);
	db->AddModel(m2);
	return db;
}

/**
 * Loads all models in the database
 */
ModelDatabase * ModelDatabase::CreateDBFromDir(const VKString & inputDir,
											   InputDirFormat dirFormat,
											   VKStringList modelsToLoad,											   
											   int readingFlags, int createFlags,
											   const VKString & dbDir, const VKString & dbName)
{
	if (modelsToLoad.count()==0)
		modelsToLoad = GetAllModelsInDir(inputDir, dirFormat);
	else if (modelsToLoad.count()==1 && modelsToLoad[0] == "LOAD_NO_MODELS")
		modelsToLoad.clear();
	ModelDatabase * db = new ModelDatabase(dbDir, dbName);
	db->ReadEverythingInADir(inputDir, dirFormat, modelsToLoad, readingFlags, createFlags);
	return db;
}

void ModelDatabase::ReadEverythingInADir(const VKString & inputDir, InputDirFormat dirFormat,
										 VKStringList & modelsToLoad,
										 int readingFlags, int createFlags)
{
	VKStringList modelNames;
	VKStringList modelFiles;
	GetModelFiles(inputDir, dirFormat, modelsToLoad, modelNames, modelFiles);
	assert(modelNames.count()==modelFiles.count());
	TimeProfiler timer;
	if (modelNames.count()>5)
		std::cout<<"Reading DB"<<std::endl;
	for (int i=0; i<modelNames.count(); i++)
	{
		if (modelNames.count()>5)
			timer.WriteProgress("readingdb", i, modelNames.count());
		Model3D * m = Model3D::CreateFromFile(modelFiles[i], readingFlags, createFlags,
											  modelNames[i], NumModels()); 
		AddModel(m);
	}
	if (modelNames.count()>5)
		std::cout<<std::endl;
}

ModelDatabase::ModelDatabase(const VKString & dbDir,
							 const VKString & dbName)
{
	m_sDBName = dbName;
	m_sDBDir = dbDir;

}

ModelDatabase::~ModelDatabase()
{
	for (int i=0; i<(int)m_Models.size(); i++)
		delete m_Models[i];
}

void ModelDatabase::WriteDB(int writingFlags,
							const VKString & writeArg)
{
	if (m_sDBDir=="none")
		std::cout<<"[WARNING] Cannot write: DB Directory not defined"<<std::endl;
	
	// write geometries (meshes + point sets)
	for (int i=0; i<NumModels(); i++)
	{
		Model3D * m = GetModel(i);
		VKString mName = m->ModelName();
		m->WriteModelDBFormat(GetModelDirname(mName), 
							  GetModelFilename(mName), writingFlags, writeArg);
	}
		
	// TODO: write collections
}

void ModelDatabase::ReadDB()
{
	if (m_sDBDir=="none")
		std::cout<<"[WARNING] Cannot read: DB Directory not defined"<<std::endl;	

	VKStringList allModels = GetAllModelsInDir(m_sDBDir, DBINPUTDIR_MyFormat);
	ReadEverythingInADir(m_sDBDir, DBINPUTDIR_MyFormat, allModels, EVERYTHING, NOTHING);
}

int ModelDatabase::NumModels()
{
	return (int)m_Models.size(); 
}

void ModelDatabase::AddModel(Model3D * model)
{
	VKString mName = model->ModelName();
	if (m_MapModelNameToModelID.find(mName)!=m_MapModelNameToModelID.end())
	{
		std::cout<<"[ERROR] Name duplicate: "<<mName.c_str()<<std::endl;
		assert(false);
		VKString mNameNew = mName+"_v1";
		int c = 1;
		while (m_MapModelNameToModelID.find(mNameNew)!=m_MapModelNameToModelID.end())
			mNameNew = mName + "_v"+VKString::number(++c);
		mName = mNameNew;
	}
	m_MapModelNameToModelID[mName] = (int)m_Models.size();
	m_Models.push_back(model);
}

void ModelDatabase::RemoveModel(int modelID)
{
	assert(modelID < (int)m_Models.size() && modelID>=0);
	Model3D * model = m_Models[modelID];
	m_MapModelNameToModelID.erase(model->ModelName());
	m_Models.erase(m_Models.begin() + modelID);
}

Model3D * ModelDatabase::GetModel(const VKString & modelName)
{
	std::map<VKString, int>::iterator iter;
	iter = m_MapModelNameToModelID.find(modelName);
	if (iter==m_MapModelNameToModelID.end())
		return NULL;
	return GetModel(iter->second);
}

Model3D * ModelDatabase::GetModel(int modelID)
{
	if (!(modelID < (int)m_Models.size() && modelID>=0))
		std::cout<<"[ERROR] Cannot get: "<<modelID<<" / "<<m_Models.size()<<std::endl;
	assert(modelID < (int)m_Models.size() && modelID>=0);
	return m_Models[modelID];
}

VKString ModelDatabase::GetDBDescriptorFilename()
{
	if (m_sDBDir=="none")
		return "none";
	return GetDBDirname() + "/database.txt";
}

VKString ModelDatabase::GetDBDirname()
{
	if (m_sDBDir=="none")
		return "none";
	return m_sDBDir;
}

VKString ModelDatabase::GetModelDirname(const VKString & modelName)
{
	if (m_sDBDir=="none")
		return "none";
	return GetModelDirname(GetDBDirname(), modelName);
}

VKString ModelDatabase::GetModelFilename(const VKString & modelName)
{
	return GetModelDirname(modelName) + "/"+modelName+".model3d";
}

VKString ModelDatabase::GetMetaDataFilename(const VKString & modelName, const VKString & metaData)
{
	return GetDBDirname()+"../meta/"+modelName+"/"+metaData;
}

VKString ModelDatabase::GetModelDirname(const VKString & dbDir,
										const VKString & modelName)
{
	return dbDir+"/Models/"+modelName;
}

VKString ModelDatabase::GetModelFilename(const VKString & dbDir,
										 const VKString & modelName)
{
	return dbDir + "/Models/" + modelName + "/" + modelName + ".model3d";
}


VKString ModelDatabase::GetModelGeometryFilename(const VKString & modelName, 
												 const VKString & geometryFile)
{
	if (m_sDBDir=="none")
		return "none";

	return GetModelDirname(modelName)+"/"+GetModelGeometryDir() + "/"+geometryFile;
}

VKString ModelDatabase::GetModelGeometryDir()
{
	return VKString("Geometry");
}

VKString ModelDatabase::ShapeValueFilename(const VKString & modelDir, 
										   const VKString & geomName,
										   const VKString & geomValueName)
{
	VKString retStr = modelDir + "/" + GetModelGeometryDir();
	retStr += VKString("/")+geomName+"Vals/"+geomValueName+".vals";
	return retStr;
}

//void ModelDatabase::WriteDB()
//{
//	
//	FILE * file = fopen(filename.c_str(), "wb");
//	assert(file!=NULL);
//}
//
//void ModelDatabase::ReadDB(const VKString & filename,
//						   int & nModels, int & maximalUniqueID,
//						   VKStringList * modelNames = NULL,
//						   std::vector<int> * modelIDs = NULL)
//{
//	
//}


