/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ShapeMesh.h"
#include "VKGraph.h"
#include "ShapeLabels.h"
#include "TimeProfiler.h"

ShapeMesh::ShapeMesh(GeomShape::GeomShapeTypes shapeTp, 
					 const VKString & geomName,
					 Model3D * model)
: GeomShape(shapeTp, geomName, model)
{
	m_pSearchTreeMesh = NULL;
	m_pSearchTree = NULL;
}

ShapeMesh::~ShapeMesh()
{
	if (m_pSearchTreeMesh!=NULL)
		delete m_pSearchTreeMesh;
	if (m_pSearchTree!=NULL)
		delete m_pSearchTree;
}

ShapeMesh * ShapeMesh::SafeCastToMesh()
{
	//std::cout<<"Safe cast to mesh"<<std::endl;
	return this;
}

void ShapeMesh::RemoveFaces(int f1, int f2)
{
	for (int f=f1; f<=f2; f++)
		RemoveFace(f);
}

R3Point ShapeMesh::Centroid() const
{
	// Return centroid of mesh
	double area = 0;
	R3Point centroid(0,0,0);
	for (int i = 0; i < NFaces(); i++) 
	{
		double face_area = FaceArea(i);
		centroid += face_area * FaceCentroid(i);
		area += face_area;
	}
	
	// Return centroid
	if (area == 0) return centroid;
	else return centroid / area;
}

double ShapeMesh::AverageRadius(const R3Point & centroid) const
{	
	// Compute average distance between a position on the surface and a center point
	RNArea area = 0;
	RNLength distance = 0;
	for (int i = 0; i < NFaces(); i++) 
	{
		RNArea face_area = FaceArea(i);
		distance += face_area * R3Distance(FaceCentroid(i), centroid);
		area += face_area;
	}
	
	// Return weighted average
	if (area == 0) return 0;
	else return distance / area;
}

R3Triad ShapeMesh::PCAFromFaces(const std::vector<int> & faces, double * variances, double * center) const
{
	// find centroid
	R3Point centroid(0,0,0);
	double A = 0;
	for (int i=0; i<(int)faces.size(); i++)
	{
		double a = FaceArea(faces[i]);
		centroid += FaceCentroid(faces[i])*a;
		A += a;
	}
	centroid /= A;
	
	RNScalar m[9] = { 0 };
	// face-based covariance matrix
	for (int i=0; i<(int)faces.size(); i++)
	{
		double a = FaceArea(faces[i]);
		R3Point p = VertexPosition(VertexOnFace(faces[i], 0));
		R3Point q = VertexPosition(VertexOnFace(faces[i], 1));
		R3Point r = VertexPosition(VertexOnFace(faces[i], 2));
		R3Point fc = FaceCentroid(faces[i]);
		
		m[0] += a/12.*(9.*fc.X()*fc.X()+p.X()*p.X()+q.X()*q.X()+r.X()*r.X());
		m[4] += a/12.*(9.*fc.Y()*fc.Y()+p.Y()*p.Y()+q.Y()*q.Y()+r.Y()*r.Y());
		m[8] += a/12.*(9.*fc.Z()*fc.Z()+p.Z()*p.Z()+q.Z()*q.Z()+r.Z()*r.Z());
		m[1] += a/12.*(9.*fc.X()*fc.Y()+p.X()*p.Y()+q.X()*q.Y()+r.X()*r.Y());
		m[3] += a/12.*(9.*fc.X()*fc.Y()+p.X()*p.Y()+q.X()*q.Y()+r.X()*r.Y());
		m[2] += a/12.*(9.*fc.X()*fc.Z()+p.X()*p.Z()+q.X()*q.Z()+r.X()*r.Z());
		m[6] += a/12.*(9.*fc.X()*fc.Z()+p.X()*p.Z()+q.X()*q.Z()+r.X()*r.Z());
		m[5] += a/12.*(9.*fc.Y()*fc.Z()+p.Y()*p.Z()+q.Y()*q.Z()+r.Y()*r.Z());
		m[7] += a/12.*(9.*fc.Y()*fc.Z()+p.Y()*p.Z()+q.Y()*q.Z()+r.Y()*r.Z());
	}
	
	// normalize covariance matrix (subtract mean)
	m[0] = m[0] / A - centroid.X()*centroid.X();
	m[4] = m[4] / A - centroid.Y()*centroid.Y();
	m[8] = m[8] / A - centroid.Z()*centroid.Z();
	m[1] = m[1] / A - centroid.X()*centroid.Y();
	m[3] = m[3] / A - centroid.X()*centroid.Y();
	m[2] = m[2] / A - centroid.X()*centroid.Z();
	m[6] = m[6] / A - centroid.X()*centroid.Z();
	m[5] = m[5] / A - centroid.Y()*centroid.Z();
	m[7] = m[7] / A - centroid.Y()*centroid.Z();
	
	// Compute eigenvalues and eigenvectors
	RNScalar U[9];
	RNScalar W[3];
	RNScalar Vt[9];
	RNSvdDecompose(3, 3, m, U, W, Vt);  // m == U . DiagonalMatrix(W) . Vt

	// Copy principle axes into more convenient form
	// W has eigenvalues (greatest to smallest) and Vt has eigenvectors (normalized)
	R3Vector axes[3];
	for (int i = 0; i < 3; i++) 
	for (int j = 0; j < 3; j++) 
		axes[i][j] = Vt[3*i+j];
	
	// Flip axes so that "heavier" on positive side for first two dimensions
	int positive_count[3] = { 0, 0, 0 };
	int negative_count[3] = { 0, 0, 0 };
	for (int i = 0; i < (int)faces.size(); i++) 
	{
		R3Vector vertex_vector = FaceCentroid(faces[i]) - centroid;
		for (int j = 0; j < 2; j++)
		{
			RNScalar dot = axes[j].Dot(vertex_vector);
			if (dot > 0.0) positive_count[j]++;
			else negative_count[j]++;
		}
	}
	for (int j =0; j < 2; j++) 
		if (positive_count[j] < negative_count[j]) 
			axes[j].Flip();
	
	// Set third axis to form orthonormal triad with other two
	axes[2] = axes[0] % axes[1];
	
	// Just checking
	assert(RNIsEqual(axes[0].Length(), 1.0, RN_BIG_EPSILON));
	assert(RNIsEqual(axes[1].Length(), 1.0, RN_BIG_EPSILON));
	assert(RNIsEqual(axes[2].Length(), 1.0, RN_BIG_EPSILON));
	assert(RNIsZero(axes[0].Dot(axes[1]), RN_BIG_EPSILON));
	assert(RNIsZero(axes[1].Dot(axes[2]), RN_BIG_EPSILON));
	assert(RNIsZero(axes[0].Dot(axes[2]), RN_BIG_EPSILON));
	
	// Return variances (eigenvalues)
	if (variances!=NULL) 
	{
		variances[0] = W[0];
		variances[1] = W[1];
		variances[2] = W[2];
	}
	
	if (center!=NULL)
	{
		center[0] = centroid.X();
		center[1] = centroid.Y();
		center[2] = centroid.Z();
	}
	// Return triad
	return R3Triad(axes[0], axes[1], axes[2]);
}

R3Triad ShapeMesh::PCAFromFaces(double * variances) const
{
	assert(false);	// todo (as above)
}


R3Box ShapeMesh::BBox() const
{
	assert(NVertices()>0);
	R3Point p = VertexPosition(0);
	R3Box bbox(p.X(), p.Y(), p.Z(), p.X(), p.Y(), p.Z());
	for (int i=0; i<NVertices(); i++)
	{
		R3Point p = VertexPosition(i);
		bbox.Union(p);
	}
	return bbox;
}

void ShapeMesh::AdjacentFaces(int fID, std::vector<int> & fids) const
{
	int v1[3];
	for (int i=0; i<3; i++)
		v1[i] = VertexOnFace(fID, i);
	
	std::map<int, int> face2count;
	
	for (int i=0; i<3; i++)
	{
		int vid = VertexOnFace(fID, i);
		int nf = NAdjacentFaces(vid);
		for (int f=0; f<nf; f++)
		{
			int fID2 = FaceOnVertex(vid, f);
			std::map<int, int>::iterator iter = face2count.find(fID2);
			if (iter == face2count.end())
				face2count[fID2] = 1;
			else
				iter->second++;
		}
	}
	
	for (std::map<int, int>::iterator iter = face2count.begin(); iter != face2count.end(); iter++)
	{
		if (iter->second==2)
			fids.push_back(iter->first);
	}
}

void ShapeMesh::EdgeBetweenFaces(int f1, int f2, int & v1, int & v2) const
{
	bool first = false;
	bool second = false;
	for (int i=0; i<3; i++)
	for (int j=0; j<3; j++)
	{
		if (VertexOnFace(f1, i)==VertexOnFace(f2, j))
		{
			if (!first)
			{
				first = true;
				v1 = VertexOnFace(f1, i);
			}
			else if (!second)
			{
				second = true;
				v2 = VertexOnFace(f1, i);
			}
			else
				assert(false);
		}
	}
	assert(second && v1 != v2);
}

double ShapeMesh::Area() const
{
	double area = 0;
	for (int f=0; f<NFaces(); f++)
		area += FaceArea(f);
	return area;
}

R3Point ShapeMesh::FaceCentroid(int fID) const
{
	return FacePoint(fID, 1./3., 1./3.);
}

int ShapeMesh::VertexValence(int vID) const
{
	return (int)AdjacentVertices(vID).size();
}

ShapePoint ShapeMesh::GetShapePointOnMesh(const R3Point & pnt) const
{
	ShapeMesh * ncThis = (ShapeMesh*)this;

	if (m_pSearchTree==NULL)
		ncThis->m_pSearchTree = CreateSearchTree(&(ncThis->m_pSearchTreeMesh), 
												 ncThis->m_SearchTreeToTriangleMap,
												 ncThis->m_SearchTreeToVertexMap);	
	R3MeshIntersection closest;
	m_pSearchTree->FindClosest(pnt, closest);
	assert(closest.type!=R3_MESH_NULL_TYPE);
	
	R3Mesh * mesh = m_pSearchTreeMesh;
	assert(mesh!=NULL);
	int tID = mesh->FaceID(closest.face);
	R3Point bary = mesh->FaceBarycentric(closest.face, closest.point);
	
	if (m_SearchTreeToTriangleMap.size()==0 && m_SearchTreeToVertexMap.size()==0)
	{
		return ShapePoint(ncThis, tID, bary.X(), bary.Y());
	}
	else	// triangle ids and vertex ids are not the same
	{
		int vID1 = mesh->VertexID(mesh->VertexOnFace(closest.face, 0));
		int vID2 = mesh->VertexID(mesh->VertexOnFace(closest.face, 1));
		int vID3 = mesh->VertexID(mesh->VertexOnFace(closest.face, 2));
		
		assert(tID < (int)m_SearchTreeToTriangleMap.size());
		assert(vID1 < (int)m_SearchTreeToVertexMap.size());
		assert(vID2 < (int)m_SearchTreeToVertexMap.size());
		assert(vID3 < (int)m_SearchTreeToVertexMap.size());
		
		int newTriID = m_SearchTreeToTriangleMap[tID];
		int newVID1 = m_SearchTreeToVertexMap[vID1];
		int newVID2 = m_SearchTreeToVertexMap[vID2];
		int newVID3 = m_SearchTreeToVertexMap[vID3];
		
		int newVIDOnTriangle1 = VertexOnFace(newTriID, newVID1);
		int newVIDOnTriangle2 = VertexOnFace(newTriID, newVID2);
		int newVIDOnTriangle3 = VertexOnFace(newTriID, newVID3);
		
		double b1 = -1;
		double b2 = -1;
		
		if (newVIDOnTriangle1 == newVID1)
			b1 = bary.X();
		else if (newVIDOnTriangle1 == newVID2)
			b1 = bary.Y();
		else if (newVIDOnTriangle1 == newVID3)
			b1 = bary.Z();
		
		if (newVIDOnTriangle2 == newVID1)
			b2 = bary.X();
		else if (newVIDOnTriangle2 == newVID2)
			b2 = bary.Y();
		else if (newVIDOnTriangle2 == newVID3)
			b2 = bary.Z();
		//std::cout<<"Bary: "<<b1<<" "<<b2<<" "<<(1-b1-b2)<<std::endl;
		assert(b1>=0 && b2>=0);
		return ShapePoint(ncThis, newTriID, b1, b2);
	}
}

R3Point ShapeMesh::GetPointPositionOnMesh(const R3Point & pnt) const
{
	if (m_pSearchTree==NULL)
	{
		ShapeMesh * ncThis = (ShapeMesh*)this;
		ncThis->m_pSearchTree = CreateSearchTree(&(ncThis->m_pSearchTreeMesh), 
												 ncThis->m_SearchTreeToTriangleMap,
												 ncThis->m_SearchTreeToVertexMap);	
	}

	assert(m_pSearchTree!=NULL);
	R3MeshIntersection closest;
	m_pSearchTree->FindClosest(pnt, closest);
	return closest.point;
}

const ShapeLabels * ShapeMesh::GetConnectedComponents() const
{
//	TimeProfiler tempTimer;
//	tempTimer.startedProcess("GetConnectedComponents");
	// find connected components
	std::vector<std::vector<int> > cc2vtx;
	std::vector<int> vtx2cc;
	VKGraph graph(NVertices());

	//std::cout<<"Creating V: "<<NVertices()<<" F: "<<NFaces()<<std::endl;
	for (int i=0; i<NVertices(); i++)
	{
		std::vector<int> neighbors = AdjacentVertices(i);
		for (int n=0; n<(int)neighbors.size(); n++)
		{
			if (i < neighbors[n])
				graph.AddEdge(i, neighbors[n], 1.);
		}
	}
	
//	tempTimer.startedProcess("FindConnectedComponents");
	graph.FindConnectedComponents(&cc2vtx, &vtx2cc);
//	tempTimer.finishedProcess("FindConnectedComponents", true);
	
	
	// sort CC by size
	std::vector<Sortable> sortedBySize;
	for (int i=0; i<(int)cc2vtx.size(); i++)
		sortedBySize.push_back(Sortable(-(double)cc2vtx[i].size(), NULL, i));
	std::sort(sortedBySize.begin(), sortedBySize.end());
	
	// create label
	ShapeLabels * newSeg = new ShapeLabels((ShapeMesh*)this, ShapeValue::s_sDefaultConnectedComponents);
	std::vector<int> perFaceSegs;
	std::vector<SceneGraphLabel*> nodes;
	std::vector<SceneGraphGeometry*> geoms;
	
	SceneGraphLabel * root = new SceneGraphLabel();
	root->m_sLabel = "CC_root";
	root->m_iID = 0;
	root->m_pParent = NULL;
	nodes.push_back(root);
	
	for (int f=0; f<NFaces(); f++)
		perFaceSegs.push_back(-1);
	
//	tempTimer.startedProcess("sortedBySize");
	for (int i=0; i<(int)sortedBySize.size(); i++)
	{
		// create node for each cluster		
		std::vector<int> & cluster = cc2vtx[sortedBySize[i].id];
		SceneGraphLabel * newNode = new SceneGraphLabel();
		newNode->m_sLabel = VKString::number(i)+"_"+VKString::number((int)cluster.size());
		newNode->m_iID = i+1;
		newNode->m_pParent = root;
		nodes.push_back(newNode);
		root->m_apChildren.push_back(newNode);
		
		// create clusters
		for (int c=0; c<(int)cluster.size(); c++)
		{
			for (int flid=0; flid<NAdjacentFaces(cluster[c]); flid++)
			{
				int f = FaceOnVertex(cluster[c], flid);
				perFaceSegs[f] = i+1;
			}
		}
	}
//	tempTimer.finishedProcess("sortedBySize", true);
	newSeg->SetFaceSegments(perFaceSegs);
	newSeg->SetAllLabels(nodes);
	newSeg->SetAllGeometries(geoms);
	((ShapeMesh*)this)->AddShapeVal(newSeg->GetShapeValueName(), newSeg);
//	tempTimer.finishedProcess("GetConnectedComponents", true); 
	return newSeg;
}


