/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ShapePoint.h"
#include "ShapeMesh.h"
#include "ShapePointCloud.h"
#include "Model3D.h"

/// POINT DATA ////
PointData::PointData(const R3Point & p )
{
	m_Point = p;
}

PointDataAtVertex * PointData::AtVtxData()
{
	return NULL;
}

PointDataOnFace * PointData::OnFaceData()
{
	return NULL;
}

PointDataMeshless * PointData::MeshlessData()
{
	return NULL;
}

/// POINT DATA: Meshless ////
PointDataMeshless::PointDataMeshless(const R3Point & p)
: PointData(p)
{
}

const ShapeMesh * PointDataMeshless::GetMeshGeometry() const
{
	return NULL;
}

R3Point PointDataMeshless::GetPosition() const
{
	return m_Point;
}

R3Vector PointDataMeshless::GetNormal() const
{
	return R3Vector(0,0,0);	//todo: store normals
}

PointDataMeshless * PointDataMeshless::MeshlessData()
{
	return this;
}

PointData * PointDataMeshless::Copy() const
{
	return new PointDataMeshless(m_Point);
}


/// POINT DATA: At Vertex ////
PointDataAtVertex::PointDataAtVertex(int vID, const ShapeMesh * mesh, const R3Point &p )
: PointData(mesh!=NULL ? mesh->VertexPosition(vID) : p)
{
	m_pMesh = mesh;
	m_iVertID = vID;
}

R3Point PointDataAtVertex::GetPosition() const
{
	return m_Point;
}

R3Vector PointDataAtVertex::GetNormal() const
{
	assert(m_pMesh!=NULL);
	return m_pMesh->VertexNormal(m_iVertID);
}


PointDataAtVertex * PointDataAtVertex::AtVtxData()
{
	return this;
}

const ShapeMesh * PointDataAtVertex::GetMeshGeometry() const
{
	return m_pMesh;
}

PointData * PointDataAtVertex::Copy() const
{
	return new PointDataAtVertex(m_iVertID, m_pMesh, m_Point);
}

/// POINT DATA: On Face ////
PointDataOnFace::PointDataOnFace(int tID, double b1, double b2, const ShapeMesh * mesh)
: PointData(mesh!=NULL ? mesh->FacePoint(tID, b1, b2) : R3Point(0,0,0))
{
	m_iTriID = tID;
	m_Bary[0] = b1;
	m_Bary[1] = b2;
	
	m_pMesh = mesh;
	assert(m_pMesh!=NULL);
}

PointDataOnFace::PointDataOnFace(int tID, double b1, double b2, 
								 const ShapeMesh * mesh, const R3Point & point)
: PointData(point)	// point has to be defined: reading, writing
{
	m_iTriID = tID;
	m_Bary[0] = b1;
	m_Bary[1] = b2;
	
	m_pMesh = mesh;
}

R3Point PointDataOnFace::GetPosition() const
{
	return m_Point;
}

R3Vector PointDataOnFace::GetNormal() const
{
	return m_pMesh->FaceNormal(m_iTriID);
}

PointDataOnFace * PointDataOnFace::OnFaceData()
{
	return this;
}

PointData * PointDataOnFace::Copy() const
{
	PointData * copy = new PointDataOnFace(m_iTriID, m_Bary[0], m_Bary[1], m_pMesh, m_Point);
	return copy;
}

const ShapeMesh * PointDataOnFace::GetMeshGeometry() const
{
	return m_pMesh;
}


//// ShapePoint //////
ShapePoint::ShapePoint()
{
	m_pPntData = NULL;
	m_pPntsSet = NULL;
	m_iIDInSet = -1;
}
bool ShapePoint::Invalid() const
{
	return (m_pPntData==NULL);
}

ShapePoint::ShapePoint(const ShapePoint & shapePnt)
{
	if (shapePnt.m_pPntData==NULL)
		m_pPntData = NULL;
	else
		m_pPntData = shapePnt.m_pPntData->Copy();

	m_pPntsSet = shapePnt.m_pPntsSet;	
	m_iIDInSet = shapePnt.m_iIDInSet;
}

ShapePoint & ShapePoint::operator=(const ShapePoint & shapePnt)
{
	if (m_pPntData!=NULL)
		delete m_pPntData;
	
	if (shapePnt.m_pPntData==NULL)
		m_pPntData = NULL;
	else
		m_pPntData = shapePnt.m_pPntData->Copy();

	m_pPntsSet = shapePnt.m_pPntsSet;
	m_iIDInSet = shapePnt.m_iIDInSet;
	return *this;
}

ShapePoint::ShapePoint(const R3Point & p, const ShapePointCloud * pnts, int pID)
{
	m_iIDInSet = pID;
	m_pPntsSet = pnts;
	m_pPntData = new PointDataMeshless(p);
}

ShapePoint::ShapePoint(const ShapeMesh * mesh, int vID, 
					   const R3Point & pos, const ShapePointCloud * pnts, int pID)
{
	m_iIDInSet = pID;
	m_pPntsSet = pnts;
	m_pPntData = new PointDataAtVertex(vID, mesh, pos);
}

ShapePoint::ShapePoint(const ShapeMesh * mesh, int tID, double b1, double b2, 
					   const ShapePointCloud * pnts, int pID)
{
	m_pPntsSet = pnts;
	m_iIDInSet = pID;
	m_pPntData = new PointDataOnFace(tID, b1, b2, mesh);
}

ShapePoint::ShapePoint(const ShapeMesh * mesh, int tID, double b1, double b2, 
					   const R3Point & pos, const ShapePointCloud * pnts, int pID)
{
	m_pPntsSet = pnts;
	m_iIDInSet = pID;
	m_pPntData = new PointDataOnFace(tID, b1, b2, mesh, pos);
}

ShapePoint::~ShapePoint()
{
	delete m_pPntData;
}

ShapePoint::ShapePointType ShapePoint::GetPntType() const
{

	if (m_pPntData->AtVtxData()!=NULL)
		return SHAPE_POINT_AT_VTX;
	else if (m_pPntData->OnFaceData()!=NULL)
		return SHAPE_POINT_ON_FACE;
	else if (m_pPntData->MeshlessData()!=NULL)
		return SHAPE_POINT_MESHLESS;
	
	assert(false);
	return SHAPE_POINT_AT_VTX;
}

int ShapePoint::VertexID(bool * atVtx) const
{
	PointDataAtVertex * vtxData = m_pPntData->AtVtxData();
	if (atVtx == NULL)
	{
		assert(vtxData!=NULL);
		return vtxData->m_iVertID;
	}
	else
	{
		if (vtxData!=NULL)
		{
			*atVtx = true;
			return vtxData->m_iVertID;
		}
		else
			*atVtx = false;
	}
	return -1;
}

void ShapePoint::GetBarycentricCoords(int * fID, double * b1, double * b2,
									  const ShapeMesh ** mesh) const
{
	PointDataAtVertex * vtxData = m_pPntData->AtVtxData();
	PointDataOnFace * faceData = m_pPntData->OnFaceData();
	if (vtxData!=NULL)
	{
		assert(vtxData->m_pMesh!=NULL);	// mesh wasn't loaded
		*mesh = vtxData->m_pMesh;
		int vid = vtxData->m_iVertID;
		if (vtxData->m_pMesh->NAdjacentFaces(vid)==0)
		{
			vid = -1;
			double minDist = -1;
			for (int v=0; v<vtxData->m_pMesh->NVertices(); v++)
			{
				double dist = R3Distance(GetPosition(), vtxData->m_pMesh->VertexPosition(v));
				if ((vid == -1 || minDist >= dist) && vtxData->m_pMesh->NAdjacentFaces(v)>0)
				{
					minDist = dist;
					vid = v;
				}
			}
		}
		
		*fID = vtxData->m_pMesh->FaceOnVertex(vid);
		assert(*fID>=0);
		
		*b1 = 0;
		*b2 = 0;
		if (vtxData->m_pMesh->VertexOnFace(*fID, 0)==vid)
			*b1 = 1.;
		else if (vtxData->m_pMesh->VertexOnFace(*fID, 1)==vid)
			*b2 = 1.;
		else
			assert(vtxData->m_pMesh->VertexOnFace(*fID, 2)==vid);
	}
	else if (faceData!=NULL)
	{
		assert(faceData->m_pMesh!=NULL);; // mesh wasn't loaded
		*mesh = faceData->m_pMesh;
		*fID = faceData->m_iTriID;
		*b1 = faceData->m_Bary[0];
		*b2 = faceData->m_Bary[1];
	}
	else
		assert(false);
}

int ShapePoint::PointID() const
{
	return m_iIDInSet;
}

void ShapePoint::SetPointID(int pID)
{
	m_iIDInSet = pID;
}

const ShapeMesh * ShapePoint::GetMeshGeometry() const
{
	return m_pPntData->GetMeshGeometry();
}

bool ShapePoint::SameMeshGeometry(const ShapeMesh * shape ) const
{
	return (GetMeshGeometry()==shape);
}

bool ShapePoint::SameMeshGeometry(const GeomShape * shape ) const
{
	const ShapeMesh * mesh = ((GeomShape*)shape)->SafeCastToMesh();
	
	if (mesh==NULL)
	{
		const ShapePointCloud * pnts = ((GeomShape*)shape)->SafeCastToPointCloud();
		if (pnts==NULL)
			return false;
		else
			return SameMeshGeometry(pnts->GetMesh());
	}
	else
		return SameMeshGeometry(mesh);
	
}

const Model3D * ShapePoint::GetModel() const
{
	const ShapePointCloud * pnts = GetPntsSet();
	if (pnts!=NULL)
		return pnts->GetModel();
	
	const ShapeMesh * mesh = GetMeshGeometry();
	if(mesh!=NULL)
		return mesh->GetModel();
	assert(mesh!=NULL || pnts!=NULL);
	return NULL;
}

const ShapePointCloud * ShapePoint::GetPntsSet() const
{
	return m_pPntsSet;
}

void ShapePoint::SetPntsSet(const ShapePointCloud * pntsSet)
{
	m_pPntsSet = pntsSet;
}

bool ShapePoint::SamePntsSet(const ShapePointCloud * pntsSet) const
{
	return (GetPntsSet()==pntsSet);
}

R3Point ShapePoint::GetPosition() const
{
	assert(m_pPntData!=NULL);
	return m_pPntData->GetPosition();
}

R3Vector ShapePoint::GetNormal() const
{
	assert(m_pPntData!=NULL);
	return m_pPntData->GetNormal();
}

bool ShapePoint::operator != (const ShapePoint & p) const
{
	return !((*this)==p);
}

bool ShapePoint::operator == (const ShapePoint & p) const
{
	bool sameInSet = (m_pPntsSet==p.m_pPntsSet) && (m_iIDInSet==p.m_iIDInSet);
	// todo: can also check if on same mesh and same barycentric coords...
	return sameInSet;
}


