/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "Sortable.h"
#include "VkFunctions.h"
#include "ShapeDistance.h"

#include "ShapePointCloud.h"
#include "ShapeManifoldMesh.h"
#include "Model3D.h"
#include "DistGeodesicDijkstra.h"
#include "TimeProfiler.h"

PointInSet::PointInSet()
{
}

R3Point PointInSet::GetPosition() const
{
	return pnts->PointPosition(id);
}

ShapePoint PointInSet::GetModelPoint() const
{
	return pnts->GetPoint(id);
}

R3Point global_PointPosition(const PointInSet * pnt, void * )
{
	return pnt->GetPosition();
}

R2Point global_PointPosition2D(const PointInSet * pnt, void * )
{
	R3Point pos = pnt->GetPosition();
	assert(pos.Z()==0);
	return R2Point(pos.X(), pos.Y());
}

////

VKString ShapePointCloud::m_sDefaultAllVtx = "AllVertices";
VKString ShapePointCloud::m_sDefaultAllFace = "AllFaces";

VKString ShapePointCloud::ShapeMeshFile(const VKString & pntSetFile)
{
	VKString fName = pntSetFile.FileName();
	VKString fName2 = fName.FileName();
	return fName2;
}

PointCloudAllVertices * ShapePointCloud::CreateAllVtxSet(const VKString & pntSetName,
														 const VKString & meshGeomName,
														 Model3D * model)
{
	return new PointCloudAllVertices(pntSetName, meshGeomName, model);
}

PointCloudAllFaces * ShapePointCloud::CreateAllFacesSet(const VKString & pntSetName, 
														const VKString & meshGeomName,
														Model3D * model)
{
	return new PointCloudAllFaces(pntSetName, meshGeomName, model);
}

ShapePointCloud * ShapePointCloud::CreateFromExternalVtxList(const VKString & pntSetName, 
															 const VKString & meshGeomName,
															 Model3D * model,															  
															 const VKString & externalFile)
{
	ShapePointCloud * pnts = new ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD,
												 pntSetName, meshGeomName, model);
	std::ifstream textStream(externalFile.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] Cannot read "<<externalFile.c_str()<<std::endl;
	assert(textStream.is_open());
	int vID;
	double x, y, z;
	while (!textStream.eof())
	{
		textStream>>vID>>x>>y>>z;
		if (textStream.eof())
			break;
		int pID = pnts->NumPoints();
		pnts->AddPoint(ShapePoint(pnts->m_pGeometry, vID, R3Point(x,y,z), pnts, pID));
	}
	textStream.close();
	return pnts;
}

ShapePointCloud * ShapePointCloud::CreateRandom(const VKString & pntSetName,
												const VKString & meshGeomName,
												Model3D * model, int nPnts)
{
	ShapePointCloud * pnts = new ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD,
												 pntSetName, meshGeomName, model);

	GeomShape * shape = model->GetGeomShape(meshGeomName);
	assert(shape!=NULL);
	ShapeMesh * mesh = shape->SafeCastToMesh();
	assert(mesh!=NULL);

	int nFaces = mesh->NFaces();
	// compute face areas
	double * faceProb = new double[nFaces];
	double totalArea = 0;
	for (int i=0; i<nFaces; i++)
	{
		faceProb[i] = mesh->FaceArea(i);
		totalArea += faceProb[i];
	}

	// order faces randomly
	std::vector<int> faceOrder;
	for (int i=0; i<nFaces; i++)
		faceOrder.push_back(i);
	
	for (int i=0; i<nFaces; i++)
	{
		int r = rand() % nFaces;
		int fID = faceOrder[r];
		faceOrder[r] = faceOrder[i];
		faceOrder[i] = fID;
	}
	
	// keep adding faces
	double cumProb = 0;
	int nAdded = 0;
	for (int i=0; i<(int)faceOrder.size(); i++)
	{
		int fID = faceOrder[i];
		cumProb += (faceProb[fID] / totalArea);
		int nExpected = (int)ceil(cumProb * (double)nPnts);
//		std::cout<<i<<": "<<cumProb<<" -> "<<nExpected<<std::endl;
		while (nAdded < nExpected && nAdded < nPnts)
		{
			RNScalar r1 = sqrt(RNRandomScalar());
			RNScalar r2 = RNRandomScalar();
			RNScalar t0 = (1.0 - r1);
			RNScalar t1 = r1 * (1.0 - r2);
			pnts->AddPoint(ShapePoint(mesh, fID, t0, t1, pnts, pnts->NumPoints()));
			nAdded++;
		}
		if (nAdded==nPnts)
			break;
	}
//	std::cout<<"nadded = "<<nAdded<<" / "<<nPnts<<std::endl;
	if (nAdded != nPnts)
	{
		std::cout<<"[ERROR] Could not add points: "<<model->ModelName().c_str()<<std::endl;
		std::cout<<"          Added: "<<nAdded<<" vs "<<nPnts<<std::endl;
		std::cout<<"          Mesh faces: "<<mesh->NFaces()<<std::endl;
	}
	assert(nAdded == nPnts);
	delete [] faceProb;
	return pnts;
}


ShapePointCloud * ShapePointCloud::CreateIterFar(const VKString & pntSetName, const VKString & meshGeomName,
												 Model3D * model, int nPnts, const VKString & distanceMetric,
												 ShapePointCloud * selectFromSet, int firstPnt, bool verbose)
{
	assert(nPnts>0);
	GeomShape * shape = NULL;
	if (model!=NULL)
		shape = model->GetGeomShape(meshGeomName);
	else
		shape = selectFromSet;
	assert(shape!=NULL);
	ShapeDistance * dist = shape->GetDistanceMetric(distanceMetric);
	assert(dist!=NULL);
		
	// distribute n * 10 'evenly sampled points' (by triangle area)
	ShapePointCloud * randPnts = selectFromSet;
	if (verbose)
		std::cout<<"Creating Iter Farthest: "<<nPnts<<"\n\tFinding Rand Pnts "<<std::endl;
	if (randPnts==NULL)
		randPnts = CreateRandom(pntSetName, meshGeomName, model, 20 * nPnts);	
	
//	// add all vertices
//	for (int i=0; i<mesh->NVertices(); i++)
//		 randPnts->AddPoint(ShapePoint(pnts->m_pGeometry, i, pnts->NumPoints()));
	
	// iteratively remove closest until have n
	bool deleteFirst = true;
	ShapePointCloud * pnts = new ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD,
												 pntSetName, meshGeomName, model);
	assert(firstPnt>=0 && firstPnt < randPnts->NumPoints());
	pnts->AddPoint(randPnts->GetPoint(firstPnt));
		
	std::vector<Sortable> minDistances;
	for (int i=0; i<randPnts->NumPoints(); i++)
		minDistances.push_back(Sortable(RN_INFINITY, NULL, i));
	minDistances.erase(minDistances.begin());

	TimeProfiler timer;
	if (verbose && nPnts>200)
		std::cout<<"\tIter Farthest: "<<nPnts<<std::endl;

	while (pnts->NumPoints() < nPnts)
	{
		if (verbose && pnts->NumPoints()%200==199)
			timer.WriteProgress("IterFar", pnts->NumPoints(), nPnts);
		const ShapePoint & p1 = pnts->GetPoint(pnts->NumPoints()-1);
			
		for (int i=0; i<(int)minDistances.size(); i++)
		{
			int pID = minDistances[i].id;
			const ShapePoint & p2 = randPnts->GetPoint(pID);

			minDistances[i].value = vkMin(minDistances[i].value, 
										  dist->NormalizedDistance(p1, p2));
		}
		std::sort(minDistances.begin(), minDistances.end());
				
		int pNew = minDistances[(int)minDistances.size()-1].id;
		if (deleteFirst)
		{
			pnts->ReplacePoint(0, randPnts->GetPoint(pNew));
			for (int i=0; i<(int)minDistances.size(); i++)
				minDistances[i].value = RN_INFINITY;
			deleteFirst = false;
		}
		else
			pnts->AddPoint(randPnts->GetPoint(pNew));
		minDistances.erase(minDistances.begin()+(minDistances.size()-1));
	}
	if (verbose && nPnts>200)
		std::cout<<std::endl;
	
	if (selectFromSet==NULL)
		delete randPnts;
	
	assert(pnts->NumPoints() == nPnts);	
	return pnts;
}


ShapePointCloud * ShapePointCloud::CreateAGDExtrema(const VKString & pntSetName,
													const VKString & meshGeomName,
													Model3D * model, int nPnts,
													bool verbose,
													int nSampPnts, double minDist, double smoothSigma, 
													bool takeMinima, bool takeMaxima)
{
	assert(takeMinima||takeMaxima);
	// approxiamte number of points (might not find the exact number)
	if (nPnts<=0)
		nPnts = 5; // tweakable
	
	// minimal distance between points
	if (minDist==-1)
		minDist = sqrt(1./((double)nPnts * 3.14));	// tweakable
	
	// do gaussian smoothing of 'feature' values
	double sigma2;
	if (smoothSigma==-1)
		sigma2 = pow(1. / nPnts, 2.); // tweakable
	else
		sigma2 = pow(smoothSigma, 2.);
	
	double smoothRadius = 2. * sqrt(sigma2);

	if (nSampPnts==-1)
		nSampPnts = 20*nPnts; // tweakable
		
	// get mesh ptr
	GeomShape * shape = model->GetGeomShape(meshGeomName);
	assert(shape!=NULL);
	ShapeMesh * mesh = shape->SafeCastToMesh();
	assert(mesh!=NULL);
	int nVert = mesh->NVertices();
	
	// get geodesic distance metric
	ShapeDistance * distPtr = NULL;
	DistGeodesicDijkstra * geoDist = NULL;
	bool deleteGeodesic = false;
	distPtr = mesh->GetDistanceMetric(ShapeDistance::m_sDefaultGeodesic);
	if (distPtr==NULL)
	{
		geoDist = new DistGeodesicDijkstra(shape, "Geodesic");
		deleteGeodesic = true;
	}
	else
		geoDist = distPtr->SafeCastToGeodesic();
	assert(geoDist!=NULL);

	// find iterative farthest vertices
	if (verbose)
		std::cout<<"Extrema AGD (V="<<nVert<<"):\n\tEven pnts: "<<nSampPnts<<std::endl;

	TimeProfiler timer;
	double * distsToSet = new double[nVert];
	
	int maxDistVtx = 0;
	geoDist->PrecomputeRow(0);
	int nVals = -1;
	const double * distsToVtx = geoDist->GetPrecomputedRow(0, &nVals);
	assert(nVals==nVert);
	for (int i=0; i<nVert; i++)
	{
		distsToSet[i] = RN_INFINITY;
		if (maxDistVtx==0 || distsToVtx[maxDistVtx] < distsToVtx[i])
			maxDistVtx = i;
	}	
	
	std::vector<int> sampleAGDVtxs;
	assert(maxDistVtx>=0);
	sampleAGDVtxs.push_back(maxDistVtx);
	distsToSet[maxDistVtx] = 0;
	while ((int)sampleAGDVtxs.size() < nSampPnts)
	{
		if (verbose)
			timer.WriteProgress("IterFarSamples", (int)sampleAGDVtxs.size(), nSampPnts);
		int lastAdded = sampleAGDVtxs[(int)sampleAGDVtxs.size()-1];
		geoDist->PrecomputeRow(lastAdded);
		distsToVtx = geoDist->GetPrecomputedRow(lastAdded, &nVals);
		assert(nVals==nVert);
		
		maxDistVtx = -1;
		double maxDist = -RN_INFINITY;
		for (int i=0; i<nVert; i++)
		{
			distsToSet[i] = vkMin(distsToSet[i], distsToVtx[i]);
			if (distsToSet[i] > maxDist && distsToSet[i]>0)
			{
				maxDist = distsToSet[i];
				maxDistVtx = i;
			}
		}
		assert(maxDistVtx>=0);
		sampleAGDVtxs.push_back(maxDistVtx);
		distsToSet[maxDistVtx] = 0;
		//minDistances.erase(minDistances.begin()+(minDistances.size()-1));
	}
	if (verbose)
		std::cout<<"\n\tCompute AGD:"<<std::endl;
	
	delete [] distsToSet;
	assert((int)sampleAGDVtxs.size()==nSampPnts);
	
	// compute average geodesic distances
	double * aveDists = new double[nSampPnts];
	for (int sID=0; sID<nSampPnts; sID++)
	{
		if (verbose)
			timer.WriteProgress("AGDComp", sID, nSampPnts);

		double sumDist = 0;
		geoDist->PrecomputeRow(sampleAGDVtxs[sID]);
		int nDists = 0;
		const double * dists = geoDist->GetPrecomputedRow(sampleAGDVtxs[sID], &nDists);
		for (int dID=0; dID<nDists; dID++)
			sumDist += geoDist->NormalizedDistance(dists[dID]);
		aveDists[sID] = (sumDist / nVert);
	}
	
	if (verbose)
		std::cout<<"\n\tSmooth AGD:"<<std::endl;

	// smooth average geodesic distances
	double distNorm = 1. / geoDist->NormalizedDistance(1.);
	double * smoothDists = new double[nSampPnts];
	ShapePointCloud * sampleAGDPnts = new ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD,
														  pntSetName+"Removeme", meshGeomName, model);
	for (int i=0; i<nSampPnts; i++)
		sampleAGDPnts->AddPoint(ShapePoint(mesh, sampleAGDVtxs[i]));

	const R3Kdtree<const PointInSet *> * kdTree = sampleAGDPnts->GetKdTree();
	
	for (int sID=0; sID<nSampPnts; sID++)
	{
		if (verbose)
			timer.WriteProgress("AGDSmooth", sID, nSampPnts);

		double sumW = 0;
		double sumV = 0;
		RNArray<const PointInSet *> neighbors;
		ShapePoint s = sampleAGDPnts->GetPoint(sID);
		kdTree->FindAll(sampleAGDPnts->PointPosition(sID), 0, smoothRadius * distNorm, neighbors);
		for (int nID=0; nID < neighbors.NEntries(); nID++)
		{
			int pID = neighbors[nID]->GetModelPoint().PointID();
			assert(pID < sampleAGDPnts->NumPoints());
			double d = geoDist->NormalizedDistance(s, neighbors[nID]->GetModelPoint());
			if (d < smoothRadius)
			{
				double w = exp(-d*d / sigma2);
				sumV += w * aveDists[pID];
				sumW += w;
			}
		}
		smoothDists[sID] = sumV / sumW;
	}
	
	// find extrema (use euclidean to find NN - then find exact distance)
	double * minDistsToSet = new double[sampleAGDPnts->NumPoints()];
	for (int sID=0; sID<sampleAGDPnts->NumPoints(); sID++)
		minDistsToSet[sID] = RN_INFINITY;
		
	ShapePointCloud * agdPnts = new ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD,
													pntSetName, meshGeomName, model);
	if (verbose)
		std::cout<<"\n\tCheck Extrema:"<<std::endl;

	for (int sID=0; sID<sampleAGDPnts->NumPoints(); sID++)
	{		
		if (verbose)
			timer.WriteProgress("AGDExtremaCheck", sID, nSampPnts);
		
		// check if too close
		if (minDistsToSet[sID] < minDist)	
			continue;
		
		// find nearest neighbors
		RNArray<const PointInSet *> neighbors;
		ShapePoint s = sampleAGDPnts->GetPoint(sID);
		R3Point sP = sampleAGDPnts->PointPosition(sID);
		kdTree->FindAll(sP, 0, minDist*distNorm, neighbors);
		
		// if not - should have larger neighborhood... shouldn't get there
		assert (neighbors.NEntries()>1);
		
		// is local extrema? type of extrema?
		bool minima = takeMinima;
		bool maxima = takeMaxima;
		double myvalue = smoothDists[sID];
		for (int nID=0; nID < neighbors.NEntries() && (minima || maxima); nID++)
		{
			int pID = neighbors[nID]->GetModelPoint().PointID();
			assert(pID < sampleAGDPnts->NumPoints());
			double d = geoDist->NormalizedDistance(s, neighbors[nID]->GetModelPoint());
			if (d < minDist && pID!=sID)
			{
				if (smoothDists[pID] <= myvalue)
					minima = false;
				if (smoothDists[pID] >= myvalue)
					maxima = false;
			}
		}
		
		// add point, update distances to set
		if (minima || maxima)
		{
			agdPnts->AddPoint(s);
			for (int i=0; i<sampleAGDPnts->NumPoints(); i++)
			{
				double d = geoDist->NormalizedDistance(s, sampleAGDPnts->GetPoint(i));
				minDistsToSet[i] = vkMin(minDistsToSet[i], d);
			}
		}
	}
	if (verbose)
		std::cout<<"\n\tFound "<<agdPnts->NumPoints()<<" extrema of agd"<<std::endl;
	
//	if (deleteGeodesic)
//		delete geoDist;

	delete sampleAGDPnts;
	delete [] aveDists;
	delete [] smoothDists;
	delete [] minDistsToSet;
	return agdPnts;	
}

ShapePointCloud * ShapePointCloud::CreateCurvatureExtrema(const VKString & pntSetName,
														  const VKString & meshGeomName,
														  Model3D * model, int nPnts)
{
	assert(false);
	return NULL;
}

ShapePointCloud::ShapePointCloud(GeomShape::GeomShapeTypes tp,
								 const VKString & pntSetName,
								 const VKString & meshGeomName,
								 const Model3D * model)
: GeomShape(tp, pntSetName, model)
{
	m_pGeometry = NULL;
	if (meshGeomName!="none")
	{
		assert(model!=NULL);
		const GeomShape * shape = model->GetGeomShape(meshGeomName);
		if (shape==NULL)
			std::cout<<"[ERROR] Mesh was not detected (for pnt cloud): "<<meshGeomName.c_str()<<std::endl;
		assert(shape!=NULL);
		m_pGeometry = shape->SafeCastToMesh();
		assert(m_pGeometry!=NULL);
	}
	
	m_pPntsSearch2D = NULL;
	m_pPntsSearch = NULL;
	m_PointsInSet = NULL;
	m_pPCAAxes = NULL;
	m_pUpVecForAxes = NULL;
}

ShapePointCloud::ShapePointCloud(const ShapeMesh * meshGeom, const VKString & pntSetName)
: GeomShape(GeomShape::SHAPE_TYPE_POINT_CLOUD, pntSetName, meshGeom!=NULL ? meshGeom->GetModel() : NULL)
{
	m_pGeometry = meshGeom;
	m_pPntsSearch2D = NULL;
	m_pPntsSearch = NULL;
	m_PointsInSet = NULL;
	m_pPCAAxes = NULL;
	m_pUpVecForAxes = NULL;
}

ShapePointCloud::~ShapePointCloud()
{
	if (m_pPntsSearch == NULL)
		delete m_pPntsSearch;
	if (m_PointsInSet == NULL)
		delete m_PointsInSet;
	if (m_pUpVecForAxes == NULL)
		delete m_pUpVecForAxes;
	if (m_pPCAAxes == NULL)
		delete m_pPCAAxes;
}

void ShapePointCloud::WriteGeometryFile(const VKString & geomFile)
{
	// exception: set of all vertices, set of all faces
	if (geomFile.endsWith(m_sDefaultAllVtx)
		|| geomFile.endsWith(m_sDefaultAllFace))
		return;
	
	// write all data
	FILE * file = fopen(geomFile.c_str(), "wb");
	assert(file!=NULL);
	WritePointSet(file, m_PointSet);
	fclose(file);
}

void ShapePointCloud::ReadGeometryFile(const VKString & geomFile)
{
	if (geomFile.endsWith(m_sDefaultAllVtx)
		|| geomFile.endsWith(m_sDefaultAllFace))
		return;
	
	FILE * file = fopen(geomFile.c_str(), "rb");
	if (file==NULL)
		std::cout<<"[ERROR] Cannot read "<<geomFile.c_str()<<std::endl;
	assert(file!=NULL);
	std::vector<ShapePoint> pnts;
	ReadPointSet(m_pGeometry, this, file, pnts);
	fclose(file);
	for (int i=0; i<(int)pnts.size(); i++)
		AddPoint(pnts[i]);
}

void ShapePointCloud::WritePointSet(FILE * file, const std::vector<ShapePoint> & pntsSet)
{
	// general point sets: 'at vertex' or 'on face' points
	int nPnts = (int)pntsSet.size();
	int * vertexIDs = NULL;
	int * triIDs = NULL;
	double * baryIDs = NULL;
	double * pnts = NULL;
	ShapePoint::ShapePointType * pntTypes = new ShapePoint::ShapePointType[nPnts];
	int nVtxPnts = 0;
	int nTriPnts = 0;
	int nMeshless = 0;
	
	for (int i=0; i<(int)pntsSet.size(); i++)
	{
		pntTypes[i] = pntsSet[i].GetPntType();
		if (pntTypes[i]==ShapePoint::SHAPE_POINT_AT_VTX)
			nVtxPnts++;
		else if (pntTypes[i]==ShapePoint::SHAPE_POINT_ON_FACE)
			nTriPnts++;
		else if (pntTypes[i]==ShapePoint::SHAPE_POINT_MESHLESS)
			nMeshless++;
		else
			assert(false);
	}
	
	if (nVtxPnts>0)
		vertexIDs = new int[nVtxPnts];
	
	if (nTriPnts>0)
	{
		triIDs = new int[nTriPnts];
		baryIDs = new double[2 * nTriPnts];
	}

	if (pntsSet.size()>0)
		pnts = new double[((int)pntsSet.size())*3];
	
	int cVtx = 0;
	int cTri = 0;
	int cMshless = 0;
	for (int i=0; i<(int)pntsSet.size(); i++)
	{
		R3Point p = pntsSet[i].m_pPntData->m_Point;
		if (pntTypes[i]==ShapePoint::SHAPE_POINT_AT_VTX)
		{
			PointDataAtVertex * vtxData = pntsSet[i].m_pPntData->AtVtxData();
			assert(vtxData!=NULL);
			vertexIDs[cVtx] = vtxData->m_iVertID;
			cVtx++;
		}
		else if (pntTypes[i]==ShapePoint::SHAPE_POINT_ON_FACE)
		{
			PointDataOnFace * faceData = pntsSet[i].m_pPntData->OnFaceData();
			assert(faceData!=NULL);
			triIDs[cTri] = faceData->m_iTriID;
			baryIDs[2*cTri+0] = faceData->m_Bary[0];
			baryIDs[2*cTri+1] = faceData->m_Bary[1];
			cTri++;
		}
		else if (pntTypes[i]==ShapePoint::SHAPE_POINT_MESHLESS)
		{
			PointDataMeshless * mshlssData = pntsSet[i].m_pPntData->MeshlessData();
			assert(mshlssData!=NULL);
			cMshless++;
		}
		else
			assert(false);
		
		pnts[3*i+0] = p.X();
		pnts[3*i+1] = p.Y();
		pnts[3*i+2] = p.Z();
	}
	
	assert(nVtxPnts+nTriPnts+nMeshless==nPnts);
	fwrite(&nVtxPnts, sizeof(int), 1, file);
	fwrite(&nTriPnts, sizeof(int), 1, file);
	fwrite(&nMeshless, sizeof(int), 1, file);
	fwrite(pntTypes, sizeof(ShapePoint::ShapePointType), nPnts, file);
	if (nVtxPnts>0)
		fwrite(vertexIDs, sizeof(int), nVtxPnts, file);
	if (nTriPnts>0)
	{
		fwrite(triIDs, sizeof(int), nTriPnts, file);
		fwrite(baryIDs, sizeof(double), 2*nTriPnts, file);
	}

	if (pntsSet.size()>0)
		fwrite(pnts, sizeof(double), 3*(int)pntsSet.size(), file);
	
	// cleanup arrays
	delete [] pntTypes;
	if (vertexIDs!=NULL)
		delete [] vertexIDs;
	if (triIDs!=NULL)
	{
		assert(baryIDs!=NULL);
		delete [] triIDs;
		delete [] baryIDs;
	}
	if (pnts!=NULL)
		delete [] pnts;
}

void ShapePointCloud::ReadPointSet(const ShapeMesh * mesh, const ShapePointCloud * partOfSet,
								   FILE * file, std::vector<ShapePoint> & pntsSet)
{
	// read: 'at vertex' or 'on face' points
	int nVtxPnts = 0;
	int nTriPnts = 0;
	int nMeshless = 0;
	
	fread(&nVtxPnts, sizeof(int), 1, file);
	fread(&nTriPnts, sizeof(int), 1, file);
	fread(&nMeshless, sizeof(int), 1, file);
	
	int nPnts = nVtxPnts + nTriPnts + nMeshless;
	ShapePoint::ShapePointType * pntTypes = new ShapePoint::ShapePointType[nPnts];	
	int * vertexIDs = NULL;
	int * triIDs = NULL;
	double * baryIDs = NULL;
	double * pnts = NULL;
	
	if (nVtxPnts>0)
		vertexIDs = new int[nVtxPnts];
	if (nTriPnts>0)
	{
		triIDs = new int[nTriPnts];
		baryIDs = new double[2 * nTriPnts];
	}

	if (nPnts>0)
		pnts = new double[3*nPnts];
	
	fread(pntTypes, sizeof(ShapePoint::ShapePointType), nPnts, file);
	
	if (nVtxPnts>0)
		fread(vertexIDs, sizeof(int), nVtxPnts, file);
	
	if (nTriPnts>0)
	{
		fread(triIDs, sizeof(int), nTriPnts, file);
		fread(baryIDs, sizeof(double), 2*nTriPnts, file);
	}
	
	if (nPnts>0)
		fread(pnts, sizeof(double), 3*nPnts, file);
	
	int cVtx = 0;
	int cTri = 0;
	int cMshless = 0;
	// create points
	for (int i=0; i<nPnts; i++)
	{
		R3Point p(pnts[3*i+0], pnts[3*i+1], pnts[3*i+2]);
		
		if (pntTypes[i]==ShapePoint::SHAPE_POINT_AT_VTX)
		{
			pntsSet.push_back(ShapePoint(mesh, vertexIDs[cVtx], p, partOfSet, i));
			cVtx++;
		}
		else if (pntTypes[i]==ShapePoint::SHAPE_POINT_ON_FACE)
		{
			double b1 = baryIDs[cTri*2+0];
			double b2 = baryIDs[cTri*2+1];
			pntsSet.push_back(ShapePoint(mesh, triIDs[cTri], b1, b2, p, partOfSet,  i));
			cTri++;
		}
		else if (pntTypes[i]==ShapePoint::SHAPE_POINT_MESHLESS)
		{
			pntsSet.push_back(ShapePoint(p, partOfSet, i));
			cMshless++;			
		}
		else
			assert(false);
	}
	
	// cleanup arrays
	delete [] pntTypes;
	if (vertexIDs!=NULL)
		delete [] vertexIDs;
	if (triIDs!=NULL)
	{
		assert(baryIDs!=NULL);
		delete [] triIDs;
		delete [] baryIDs;
	}
	if (pnts!=NULL)
		delete [] pnts;
}

const ShapeMesh * ShapePointCloud::GetMesh() const
{
	return m_pGeometry;
}

ShapePoint ShapePointCloud::GetPoint(int pntID) const
{
	if (!(pntID>=0 && pntID < (int)m_PointSet.size()))
	{
		std::cout<<"[ERROR] Invalid point id: "<<pntID<<" / "<<m_PointSet.size()<<std::endl;
		assert(false);
	}
	return m_PointSet[pntID];
}

R3Point ShapePointCloud::PointPosition(int pntID) const
{
	assert(pntID>=0 && pntID < (int)m_PointSet.size());
	return m_PointSet[pntID].GetPosition();
}

R3Vector ShapePointCloud::PointNormal(int pntID) const
{
	return GetPoint(pntID).GetNormal();
}

int ShapePointCloud::AddPoint(const ShapePoint & pnt)
{
	m_PointSet.push_back(pnt);
	int lastID = (int)m_PointSet.size()-1;
	m_PointSet[lastID].SetPointID(lastID);
	m_PointSet[lastID].SetPntsSet(this);
	return lastID;
}

void ShapePointCloud::ReplacePoint(int pntID, const ShapePoint & newPnt)
{
	assert(pntID>=0 && pntID < (int)m_PointSet.size());
	m_PointSet[pntID] = newPnt;
	m_PointSet[pntID].SetPointID(pntID);
	m_PointSet[pntID].SetPntsSet(this);
}

void ShapePointCloud::RemovePoint(int pntID)
{
	assert(pntID>=0 && pntID < (int)m_PointSet.size());
	m_PointSet.erase(m_PointSet.begin()+pntID);
	for (int i=pntID; i<(int)m_PointSet.size(); i++)
		m_PointSet[i].SetPointID(i);
}

int ShapePointCloud::NumPoints() const
{
	return (int)m_PointSet.size();
}

ShapePointCloud * ShapePointCloud::SafeCastToPointCloud()
{
	return this;
}

R3Point ShapePointCloud::Centroid() const
{
	R3Point centroid(0,0,0);
	for (int i=0; i<NumPoints(); i++)
		centroid += GetPoint(i).GetPosition();
	return (centroid / NumPoints());
}

double ShapePointCloud::AverageRadius(const R3Point & centroid) const
{
	double dist = 0;
	for (int i=0; i<NumPoints(); i++)
		dist += R3Distance(centroid, GetPoint(i).GetPosition());
	return (dist / NumPoints());
}

R3Triad ShapePointCloud::PCAAxes(R3Vector * upVector) const
{
	bool recompute = (m_pPCAAxes==NULL);
	if (!recompute && upVector!=NULL)
	{
		recompute = (m_pUpVecForAxes==NULL);
			// just for debugging. make sure that it works 
			// instead of assertion - just recompute.
		assert((*upVector) == (*m_pUpVecForAxes));
	}
	else if (!recompute && upVector==NULL)
	{
		recompute = (m_pUpVecForAxes!=NULL);
		if (m_pUpVecForAxes!=NULL)
			delete m_pUpVecForAxes;
	}
	
	if (recompute)
	{
		ShapePointCloud * pnts = (ShapePointCloud*)this;
		if (upVector==NULL)
			pnts->m_pPCAAxes = new R3Triad(ComputePCAAxesNoUp());
		else 
		{
			pnts->m_pUpVecForAxes = new R3Vector(*upVector);
			pnts->m_pPCAAxes = new R3Triad(ComputePCAAxesUpVec(*upVector));
		}
	}
	
	assert(m_pPCAAxes!=NULL);
	return *m_pPCAAxes;
}

R3Triad ShapePointCloud::PCAFromPoints(const std::vector<int> & pntIDs, double * vars, double * ctr) const
{
	// Get centroid
	R3Point centroid(0,0,0);
	
	// Compute covariance matrix
	double sumW = 0;
	RNScalar m[9] = { 0 };
	
	for (int i = 0; i < (int)pntIDs.size(); i++) 
		centroid += PointPosition(pntIDs[i]);
	centroid /= (int)pntIDs.size();
	
	for (int i = 0; i < (int)pntIDs.size(); i++) 
	{
		R3Point p = PointPosition(pntIDs[i]);
		RNScalar x = p.X() - centroid[0];
		RNScalar y = p.Y() - centroid[1];
		RNScalar z = p.Z() - centroid[2];
		double w = 1.;
		m[0] += w * x*x;
		m[4] += w * y*y;
		m[8] += w * z*z;
		m[1] += w * x*y;
		m[3] += w * x*y;
		m[2] += w * x*z;
		m[6] += w * x*z;
		m[5] += w * y*z;
		m[7] += w * y*z;
		sumW += w;
	}
	
	// Normalize covariance matrix
	assert(sumW>0);
	for (int i = 0; i < 9; i++) m[i] /= sumW;
	
	// Compute eigenvalues and eigenvectors
	RNScalar U[9];
	RNScalar W[3];
	RNScalar Vt[9];
	RNSvdDecompose(3, 3, m, U, W, Vt);  // m == U . DiagonalMatrix(W) . Vt
	
	// Copy principle axes into more convenient form
	// W has eigenvalues (greatest to smallest) and Vt has eigenvectors (normalized)
	R3Vector axes[3];
	for (int i = 0; i < 3; i++) 
		for (int j = 0; j < 3; j++) 
			axes[i][j] = Vt[3*i+j];
	
	// Flip axes so that "heavier" on positive side for first two dimensions
	int positive_count[3] = { 0, 0, 0 };
	int negative_count[3] = { 0, 0, 0 };
	for (int i = 0; i < NumPoints(); i++) 
	{
		R3Vector vertex_vector = PointPosition(i) - centroid;
		for (int j = 0; j < 2; j++)
		{
			RNScalar dot = axes[j].Dot(vertex_vector);
			if (dot > 0.0) positive_count[j]++;
			else negative_count[j]++;
		}
	}
	for (int j =0; j < 2; j++) 
		if (positive_count[j] < negative_count[j]) 
			axes[j].Flip();
	
	// Set third axis to form orthonormal triad with other two
	axes[2] = axes[0] % axes[1];
	
	// Just checking
	assert(RNIsEqual(axes[0].Length(), 1.0, RN_BIG_EPSILON));
	assert(RNIsEqual(axes[1].Length(), 1.0, RN_BIG_EPSILON));
	assert(RNIsEqual(axes[2].Length(), 1.0, RN_BIG_EPSILON));
	assert(RNIsZero(axes[0].Dot(axes[1]), RN_BIG_EPSILON));
	assert(RNIsZero(axes[1].Dot(axes[2]), RN_BIG_EPSILON));
	assert(RNIsZero(axes[0].Dot(axes[2]), RN_BIG_EPSILON));
	
	// Return variances (eigenvalues)
	if (vars!=NULL) 
	{
		vars[0] = W[0];
		vars[1] = W[1];
		vars[2] = W[2];
	}
	
	if (ctr!=NULL)
	{
		ctr[0] = centroid.X();
		ctr[1] = centroid.Y();
		ctr[2] = centroid.Z();
	}
	
	// Return triad
	return R3Triad(axes[0], axes[1], axes[2]);
}

R3Triad ShapePointCloud::ComputePCAAxesNoUp(double * variances) const
{
	// Get centroid
	R3Point centroid = Centroid();
	
	// Compute covariance matrix
	double sumW = 0;
	RNScalar m[9] = { 0 };
	
	for (int i = 0; i < NumPoints(); i++) 
	{
		R3Point p = PointPosition(i);
		RNScalar x = p.X() - centroid[0];
		RNScalar y = p.Y() - centroid[1];
		RNScalar z = p.Z() - centroid[2];
		double w = 1.;
		m[0] += w * x*x;
		m[4] += w * y*y;
		m[8] += w * z*z;
		m[1] += w * x*y;
		m[3] += w * x*y;
		m[2] += w * x*z;
		m[6] += w * x*z;
		m[5] += w * y*z;
		m[7] += w * y*z;
		sumW += w;
	}
	
	// Normalize covariance matrix
	assert(sumW>0);
	for (int i = 0; i < 9; i++) m[i] /= sumW;
	
	// Compute eigenvalues and eigenvectors
	RNScalar U[9];
	RNScalar W[3];
	RNScalar Vt[9];
	RNSvdDecompose(3, 3, m, U, W, Vt);  // m == U . DiagonalMatrix(W) . Vt
	
	// Copy principle axes into more convenient form
	// W has eigenvalues (greatest to smallest) and Vt has eigenvectors (normalized)
	R3Vector axes[3];
	for (int i = 0; i < 3; i++) 
	for (int j = 0; j < 3; j++) 
		axes[i][j] = Vt[3*i+j];
	
	// Flip axes so that "heavier" on positive side for first two dimensions
	int positive_count[3] = { 0, 0, 0 };
	int negative_count[3] = { 0, 0, 0 };
	for (int i = 0; i < NumPoints(); i++) 
	{
		R3Vector vertex_vector = PointPosition(i) - centroid;
		for (int j = 0; j < 2; j++)
		{
			RNScalar dot = axes[j].Dot(vertex_vector);
			if (dot > 0.0) positive_count[j]++;
			else negative_count[j]++;
		}
	}
	for (int j =0; j < 2; j++) 
		if (positive_count[j] < negative_count[j]) 
			axes[j].Flip();
	
	// Set third axis to form orthonormal triad with other two
	axes[2] = axes[0] % axes[1];
	
	// Just checking
	assert(RNIsEqual(axes[0].Length(), 1.0, RN_BIG_EPSILON));
	assert(RNIsEqual(axes[1].Length(), 1.0, RN_BIG_EPSILON));
	assert(RNIsEqual(axes[2].Length(), 1.0, RN_BIG_EPSILON));
	assert(RNIsZero(axes[0].Dot(axes[1]), RN_BIG_EPSILON));
	assert(RNIsZero(axes[1].Dot(axes[2]), RN_BIG_EPSILON));
	assert(RNIsZero(axes[0].Dot(axes[2]), RN_BIG_EPSILON));
	
	// Return variances (eigenvalues)
	if (variances!=NULL) 
	{
		variances[0] = W[0];
		variances[1] = W[1];
		variances[2] = W[2];
	}
	
	// Return triad
	return R3Triad(axes[0], axes[1], axes[2]);
}

R3Triad ShapePointCloud::ComputePCAAxesUpVec(const R3Vector & upVec, double * variances) const
{
//	upVec.Normalize();
	
	// just to add stability, pick x, y, or z least parallel to upVec
	int notUp = 0;
	if (vkAbs(upVec.X()) <= vkAbs(upVec.Y()))
	{
		if (vkAbs(upVec.X()) < vkAbs(upVec.Z()))
			notUp = 0;
		else
			notUp = 2;
	}
	else
	{
		if (vkAbs(upVec.Y()) < vkAbs(upVec.Z()))
			notUp = 1;
		else
			notUp = 2;		
	}
			
	// find (arbitrary) 2D plane orthogonal to the up vector
	R3Vector axis1a = upVec;
	R3Vector axis1b = R3xyz_triad[notUp];
	R3Vector axis1c = axis1a % axis1b;
	axis1b = axis1a % axis1c;
	R3Triad xyzWithUp(axis1a, axis1b, axis1c);
	xyzWithUp.Normalize();
	const R4Matrix & xyzToUp = xyzWithUp.Matrix();
	
	// Get centroid
	R3Point centroid = Centroid();
	
	// Compute covariance matrix
	double sumW = 0;
	RNScalar m[4] = { 0 };
	double varZ = 0;
	for (int i = 0; i < NumPoints(); i++) 
	{
		R3Point p = PointPosition(i);
		double w = 1.;
		
		R3Point xyzVec(p[0] - centroid[0], 
					   p[1] - centroid[1],
					   p[2] - centroid[2]);
		R3Point xyzWithUpVec = xyzToUp * xyzVec;
		
		varZ += w * pow(xyzWithUpVec[0], 2.);
		
		double x = xyzWithUpVec[1];
		double y = xyzWithUpVec[2];
		m[0] += w * x*x;
		m[1] += w * x*y;
		m[2] += w * x*y;	
		m[3] += w * y*y;
		sumW += w;
	}
	
	// Normalize covariance matrix
	assert(sumW>0);
	for (int i = 0; i < 4; i++) m[i] /= sumW;
	varZ /= sumW;
	
	// Compute eigenvalues and eigenvectors
	RNScalar U[4];
	RNScalar W[2];
	RNScalar Vt[4];
	RNSvdDecompose(2, 2, m, U, W, Vt);  // m == U . DiagonalMatrix(W) . Vt
	
	// Copy principle axes into more convenient form
	// W has eigenvalues (greatest to smallest) and Vt has eigenvectors (normalized)
	if (variances!=NULL)
	{
		variances[0] = varZ;
		variances[1] = W[0];
		variances[2] = W[1];
	}
	R3Vector axes[3];
	axes[0] = upVec;
	
	axes[1] = Vt[0] * xyzWithUp[1]  + Vt[1] * xyzWithUp[2];
	axes[2] = Vt[2] * xyzWithUp[1]  + Vt[3] * xyzWithUp[2];
	
	// Flip axes so that "heavier" on positive side for first two dimensions
	int positive_count[2] = { 0, 0 };
	int negative_count[2] = { 0, 0 };
	for (int i = 0; i < NumPoints(); i++) 
	{
		R3Vector vertex_vector = PointPosition(i) - centroid;
		for (int j = 1; j < 2; j++) 
		{
			RNScalar dot = axes[j].Dot(vertex_vector);
			if (dot > 0.0) positive_count[j]++;
			else negative_count[j]++;
		}
	}
	
	for (int j =1; j < 2; j++) 
		if (positive_count[j] < negative_count[j]) 
			axes[j].Flip();
	
	// Set third axis to form orthonormal triad with other two
	axes[2] = axes[0] % axes[1];
	axes[0].Normalize();
	axes[1].Normalize();
	axes[2].Normalize();
	
	
	// Just checking
	assert(RNIsEqual(axes[0].Length(), 1.0, RN_BIG_EPSILON));
	assert(RNIsEqual(axes[1].Length(), 1.0, RN_BIG_EPSILON));
	assert(RNIsEqual(axes[2].Length(), 1.0, RN_BIG_EPSILON));
	assert(RNIsZero(axes[0].Dot(axes[1]), RN_BIG_EPSILON));
	assert(RNIsZero(axes[1].Dot(axes[2]), RN_BIG_EPSILON));
	assert(RNIsZero(axes[0].Dot(axes[2]), RN_BIG_EPSILON));
		
	// Return triad
	return R3Triad(axes[0], axes[1], axes[2]);
}

R3Affine ShapePointCloud::PCANormTransform(R3Vector * upVec,
										   bool translate, bool rotate, bool scale) const
{
	// Initialize transformation
	R3Affine affine(R3identity_affine);
	
	// Compute center of mass
	R3Point centroid = Centroid();
	
	// Translate center of mass back to original (if not translating)
	if (!translate) {
		affine.Translate(centroid.Vector());
	}
	
	// Scale by inverse of radius
	if (scale) 
	{
		RNScalar radius = AverageRadius(centroid);
		if (RNIsPositive(radius)) affine.Scale(1.0 / radius);
	}
	
	// Rotate to align principal axes with XYZ
	if (rotate) {
		RNScalar variances[3];
		R3Triad triad = PCAAxes(upVec);
		if (!rotate) affine.Transform(R3Affine(triad.InverseMatrix()));
		affine.Transform(R3Affine(triad.InverseMatrix()));
	}
	
	// Translate center of mass to origin
	affine.Translate(-(centroid.Vector()));
	
	// Return PCA normalization transformation
	return affine;
}

R3Kdtree<const PointInSet *> * ShapePointCloud::CreateSubsetKdTree(const std::vector<int> & subset) const
{
	if (m_pPntsSearch==NULL)
		GetKdTree();
	assert(m_PointsInSet!=NULL);
	
	RNArray<const PointInSet *> array;	
	for (int i = 0; i < (int)subset.size(); i++) 
		array.Insert(&(m_PointsInSet[subset[i]]));
	
	return new R3Kdtree<const PointInSet *>(array, &global_PointPosition, NULL);
}

const R3Kdtree<const PointInSet *> * ShapePointCloud::GetKdTree() const
{
	if (m_pPntsSearch==NULL)
	{
		ShapePointCloud * ncptr = (ShapePointCloud *)this;
		if (m_PointsInSet==NULL)
		{		
			ncptr->m_PointsInSet = new PointInSet[NumPoints()];	
			for (int i = 0; i < NumPoints(); i++) 
			{
				ncptr->m_PointsInSet[i].id = i;
				ncptr->m_PointsInSet[i].pnts = ncptr;
			}
		}	
		RNArray<const PointInSet *> array;	
		for (int i = 0; i < NumPoints(); i++) 
			array.Insert(&(m_PointsInSet[i]));

		ncptr->m_pPntsSearch = new R3Kdtree<const PointInSet *>(array, &global_PointPosition, NULL);
	}
	return m_pPntsSearch;
}

const R2Kdtree<const PointInSet *> * ShapePointCloud::GetKdTree2D() const
{
	if (m_pPntsSearch2D==NULL)
	{
		ShapePointCloud * ncptr = (ShapePointCloud *)this;
		if (m_PointsInSet==NULL)
		{		
			ncptr->m_PointsInSet = new PointInSet[NumPoints()];	
			for (int i = 0; i < NumPoints(); i++) 
			{
				ncptr->m_PointsInSet[i].id = i;
				ncptr->m_PointsInSet[i].pnts = ncptr;
			}
		}
		
		RNArray<const PointInSet *> array;	
		for (int i = 0; i < NumPoints(); i++) 
			array.Insert(&(m_PointsInSet[i]));
		
		ncptr->m_pPntsSearch2D = new R2Kdtree<const PointInSet *>(array, &global_PointPosition2D, NULL);
	}
	return m_pPntsSearch2D;
}

void ShapePointCloud::RearrangeAxes(char newOrder[3], int newSigns[3])
{
	for (int i=0; i<(int)m_PointSet.size(); i++)
	{
		R3Point oldPnt = m_PointSet[i].m_pPntData->m_Point;
		R3Point newPnt;
		for (int d=0; d<3; d++)
		{
			int dim = ((int)newOrder[d]-(int)'x');
			assert(dim>=0 && dim<3);
			newPnt[d] = newSigns[d] * oldPnt[dim];
		}
		m_PointSet[i].m_pPntData->m_Point = newPnt;
	}
}


///// ALL VERTICES ///////
PointCloudAllVertices::PointCloudAllVertices(const VKString & pntSetName,
											 const VKString & meshGeomName,
											 Model3D * model)
: ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD_ALLVTX,
				  pntSetName, meshGeomName, model)

{	
}

ShapePoint PointCloudAllVertices::GetPoint(int pntID) const
{
	return ShapePoint(m_pGeometry, pntID, R3Point(0,0,0), this, pntID);
}

R3Point PointCloudAllVertices::PointPosition(int pntID) const
{
	return GetMesh()->VertexPosition(pntID);
}

int PointCloudAllVertices::AddPoint(const ShapePoint & pnt)
{
	assert(false);
}

void PointCloudAllVertices::ReplacePoint(int id, const ShapePoint & newPnt)
{
	assert(false);
}

void PointCloudAllVertices::RemovePoint(int id)
{
	assert(false);
}

int PointCloudAllVertices::NumPoints() const
{
	return GetMesh()->NVertices();
}

void PointCloudAllVertices::WriteGeometryFile(const VKString & )
{
	return;
}

void PointCloudAllVertices::ReadGeometryFile(const VKString & )
{
	return;
}

void PointCloudAllVertices::RearrangeAxes(char newOrder[3], int newSigns[3])
{
	// assume mesh was permuted
}


///// ALL FACES ///////
PointCloudAllFaces::PointCloudAllFaces(const VKString & pntSetName,
									   const VKString & meshGeomName,
									   Model3D * model)
: ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD_ALLFACE,
				  pntSetName, meshGeomName, model)
{
}

ShapePoint PointCloudAllFaces::GetPoint(int pntID) const
{
	return ShapePoint(m_pGeometry, pntID, 1./3., 1./3., this, pntID);
}

R3Point PointCloudAllFaces::PointPosition(int pntID) const
{
	return GetMesh()->FaceCentroid(pntID);
}

int PointCloudAllFaces::AddPoint(const ShapePoint & pnt)
{
	assert(false);
}

void PointCloudAllFaces::ReplacePoint(int id, const ShapePoint & newPnt)
{
	assert(false);
}

void PointCloudAllFaces::RemovePoint(int id)
{
	assert(false);
}

int PointCloudAllFaces::NumPoints() const
{
	return GetMesh()->NFaces();
}

void PointCloudAllFaces::WriteGeometryFile(const VKString & )
{
	return;
}

void PointCloudAllFaces::ReadGeometryFile(const VKString & )
{
	return;
}

void PointCloudAllFaces::RearrangeAxes(char newOrder[3], int newSigns[3])
{
	// assume mesh was permuted
}


