/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ShapeTriangleSoup.h"
#include "ShapeValue.h"
#include "Model3D.h"

ShapeTriangleSoup::ShapeTriangleSoup(const VKString & geomFile,
									 const VKString & geomName,
									 Model3D * model)
: ShapeMesh(GeomShape::SHAPE_TYPE_TRIANGLE_SOUP, geomName, model)
{
	if (geomFile!="none")
		ReadGeometryFile(geomFile);
}

ShapeTriangleSoup::~ShapeTriangleSoup()
{
}

ShapeTriangleSoup * ShapeTriangleSoup::SafeCastToTriangleSoup()
{
	return this;
}

void ShapeTriangleSoup::WriteGeometryFile(const VKString & geomFile)
{
	if (geomFile.endsWith(".off"))
		WriteOff(geomFile);
	else
	{
		std::cout<<"[WARNING] Unknown format: "<<geomFile.c_str();
		std::cout<<". Did not write mesh."<<std::endl;
	}
}

void ShapeTriangleSoup::ReadGeometryFile(const VKString & geomFile)
{
	//std::cout<<"Reading triangle soup"<<std::endl;
	if (geomFile.endsWith(".off"))
		ReadOff(geomFile);
	else
	{
		std::cout<<"[WARNING] Attempting to read "<<geomFile.c_str();
		std::cout<<" using R3Mesh."<<std::endl;
		R3Mesh * mesh = new R3Mesh();
		int couldRead = mesh->ReadFile(geomFile.c_str());
		if (couldRead>0)
		{
			for (int i=0; i<mesh->NVertices(); i++)
			{
				R3Point p = mesh->VertexPosition(mesh->Vertex(i));
				AddVertex(p.X(), p.Y(), p.Z());
			}
			
			for (int i=0; i<mesh->NFaces(); i++)
			{
				R3MeshFace * f = mesh->Face(i);
				int v0 = mesh->VertexID(mesh->VertexOnFace(f, 0));
				int v1 = mesh->VertexID(mesh->VertexOnFace(f, 1));
				int v2 = mesh->VertexID(mesh->VertexOnFace(f, 2));
				AddFace(v0, v1, v2);
			}
		}
		delete mesh;
	}
//	else
//	{
//		std::cout<<"[WARNING] Unknown format: "<<geomFile.c_str();
//		std::cout<<". Did not read mesh."<<std::endl;
//	}
}

void ShapeTriangleSoup::WriteOff(const VKString & filename, OffFormat format)
{
	switch(format)
	{
		case OFF_ASCII:
			WriteOffASCII(filename);
			break;
		case OFF_BIN_SINGLE:
		case OFF_BIN_DOUBLE:
			WriteOffBinary(filename, format==OFF_BIN_SINGLE);
			break;
		default:
			std::cout<<"[ERROR] Cannot write format "<<format<<std::endl;
	}
}

void ShapeTriangleSoup::ReadOff(const VKString & filename, OffFormat format)
{
	if (format == OFF_AUTO)
	{
		FILE *fp = fopen(filename.c_str(), "rb");
		if (fp==NULL)
			std::cout<<"[ERROR] Cannot find: "<<filename.c_str()<<std::endl;
		assert(fp!=NULL);
		char header[9];
		header[8] = 0;
		fread(header, sizeof(char), 8, fp);
		if (strcmp(header, "OFF1    ")==0)
			format = OFF_BIN_SINGLE;
		else if (strcmp(header, "OFF2    ")==0)
			format = OFF_BIN_DOUBLE;
		else if (strncmp(header, "OFF", 3)==0)
			format = OFF_ASCII;
		else
			std::cout<<"[ERROR] Unknown off header: "<<header<<std::endl;
		fclose(fp);
	}
	
	switch(format)
	{
		case OFF_ASCII:
			ReadOffASCII(filename);
			break;
		case OFF_BIN_SINGLE:
			ReadOffBinary(filename, format == OFF_BIN_SINGLE);
			break;
		case OFF_BIN_DOUBLE:
			ReadOffBinary(filename, format == OFF_BIN_SINGLE);
			break;
		default:
			std::cout<<"Unknown OFF format: "<<format<<std::endl;
			assert(false);
	}
}

void ShapeTriangleSoup::ReadOffASCII(const VKString & filename)
{
	// Open file
	FILE *fp;
	if (!(fp = fopen(filename.c_str(), "r"))) 
	{
		std::cout<<"[ERROR] Unable to open (read) "<<filename.c_str()<<std::endl;
		assert(false);
	}
	
	// Read file
	int nverts = 0;
	int nfaces = 0;
	int nedges = 0;
	int line_count = 0;
	int vertex_count = 0;
	int face_count = 0;
	char buffer[1024];
	char header[64];
	while (fgets(buffer, 1023, fp)) {
		// Increment line counter
		line_count++;
		
		// Skip white space
		char *bufferp = buffer;
		while (isspace(*bufferp)) bufferp++;
		
		// Skip blank lines and comments
		if (*bufferp == '#') continue;
		if (*bufferp == '\0') continue;
		
		// Check section
		if (nverts == 0) {
			// Read header keyword
			if (strstr(bufferp, "OFF")) {
				// Check if counts are on first line
				int tmp;
				if (sscanf(bufferp, "%s%d%d%d", header, &tmp, &nfaces, &nedges) == 4) {
					nverts = tmp;
				}
			}
			else {
				// Read counts from second line
				if ((sscanf(bufferp, "%d%d%d", &nverts, &nfaces, &nedges) != 3) || (nverts == 0)) {
					std::cout<<"[ERROR] Unable to read header "<<line_count;
					std::cout<<" file="<<filename.c_str()<<std::endl;
					fclose(fp);
					assert(false);
				}
			}
		}
		else if (vertex_count < nverts) {
			// Read vertex coordinates
			double x, y, z;
			if (sscanf(bufferp, "%lf%lf%lf", &x, &y, &z) != 3) {
				std::cout<<"[ERROR] Unable to read vtx coords "<<line_count;
				std::cout<<" file="<<filename.c_str()<<std::endl;
				fclose(fp);
				assert(false);
			}
			
			// Create vertex
			AddVertex(x, y, z);
			
			// Increment counter
			vertex_count++;
		}
		else if (face_count < nfaces) {
			// Read number of vertices in face 
			int face_nverts = 0;
			bufferp = strtok(bufferp, " \t");
			if (bufferp) face_nverts = atoi(bufferp);
			else {
				std::cout<<"[ERROR] Unable to read face "<<line_count;
				std::cout<<" file="<<filename.c_str()<<std::endl;
				fclose(fp);
				assert(false);
			}
			
			// Read vertex indices for face
			bufferp = strtok(NULL, " \t");
			assert(bufferp);
			int v1 = atoi(bufferp);
			bufferp = strtok(NULL, " \t");
			assert(bufferp);
			int v2 = atoi(bufferp);
			bufferp = strtok(NULL, " \t");
			assert(bufferp);
			int v3 = atoi(bufferp);
			AddFace(v1, v2, v3);
			// Increment counter
			face_count++;
		}
		else {
			// Should never get here
			std::cout<<"[ERROR] extr text line "<<line_count;
			std::cout<<" file="<<filename.c_str()<<std::endl;
			assert(false);
		}
	}
	
	// Close file
	fclose(fp);
}

void ShapeTriangleSoup::ReadOffBinary(const VKString & filename, bool single)
{
	FILE * fp = fopen(filename.c_str(), "rb");
	assert(fp!=NULL);
	char header[9];
	header[8] = 0;
	fread(header, sizeof(char), 8, fp);
	assert(strcmp(header, "OFF1    ")==0 && single
		   || strcmp(header, "OFF2    ")==0 && !single);
	int nvertices, nfaces, nedges;
	fread(&nvertices, sizeof(int), 1, fp);
	fread(&nfaces, sizeof(int), 1, fp);
	fread(&nedges, sizeof(int), 1, fp);
	
	if (single)
	{
		float * vtxData = new float[3*nvertices];
		fread(vtxData, sizeof(float), 3*nvertices, fp);
		for (int i=0; i<nvertices; i++)
			AddVertex((double)vtxData[3*i+0], (double)vtxData[3*i+1], (double)vtxData[3*i+2]);
		delete [] vtxData;
	}
	else
	{
		double * vtxData = new double[3*nvertices];
		fread(vtxData, sizeof(double), 3*nvertices, fp);
		for (int i=0; i<nvertices; i++)
			AddVertex(vtxData[3*i+0], vtxData[3*i+1], vtxData[3*i+2]);
		delete [] vtxData;
	}
	
	int nIds = 0;
	int vtxIds[3];
	for (int i=0; i<nfaces; i++)
	{	
		fread(&nIds, sizeof(int), 1, fp);
		assert(nIds==3);	// only triangles for now
		fread(vtxIds, sizeof(int), 3, fp);
		AddFace(vtxIds[0], vtxIds[1], vtxIds[2]);
	}
	
	fclose(fp);
}

void ShapeTriangleSoup::WriteOffASCII(const VKString & filename)
{
	// Open file
	FILE *fp;
	if (!(fp = fopen(filename.c_str(), "w"))) 
	{
		std::cout<<"[ERROR] Unable to open "<<filename.c_str()<<std::endl;
		assert(false);
	}
	
	// Write header
	fprintf(fp, "OFF\n");
	fprintf(fp, "%d %d %d\n", NVertices(), NFaces(), 0);
	
	// Write vertices
	for (int i = 0; i < NVertices(); i++) {
		const R3Point& p = VertexPosition(i);
		fprintf(fp, "%g %g %g\n", p.X(), p.Y(), p.Z());
	}
	
	// Write Faces
	for (int i = 0; i < NFaces(); i++) 
	{
		int v1 = VertexOnFace(i, 0);
		int v2 = VertexOnFace(i, 1);
		int v3 = VertexOnFace(i, 2);
		fprintf(fp, "3 %d %d %d\n", v1, v2, v3);
	}
	
	// Close file
	fclose(fp);
}

void ShapeTriangleSoup::WriteObj(const VKString & filename)
{
	std::ofstream textStream(filename.c_str());
	assert(textStream.is_open());
	textStream<<"# Studio\n";
	textStream<<"g Polygonal_Model_1\n";
	textStream<<"# Number of geometric vertices: "<<NVertices()<<"\n";
	for (int i=0; i<NVertices(); i++)
	{
		R3Point p = VertexPosition(i);
		textStream<<"v "<<p.X()<<" "<<p.Y()<<" "<<p.Z()<<"\n";
	}
	textStream<<"# Number of texture vertices: 0\n";
	textStream<<"g Polygonal_Model_1 Triangles_0\n";
	textStream<<"# Number of triangles: "<<NFaces()<<"\n";
	for (int f = 0; f<NFaces(); f++)
	{
		int v1 = VertexOnFace(f, 0);
		int v2 = VertexOnFace(f, 1);
		int v3 = VertexOnFace(f, 2);
		textStream<<"f "<<v1<<" "<<v2<<" "<<v3<<"\n";
	}
	textStream<<"# end of file\n"<<std::endl;
	textStream.close();
}

void ShapeTriangleSoup::WriteOffBinary(const VKString & filename, bool single)
{
	FILE * fp = fopen(filename.c_str(), "wb");
	assert(fp!=NULL);
	VKString header=VKString("OFF")+(single ? "1" : "2") + "    ";
	fwrite(header.c_str(), sizeof(char), 8, fp);
	int nvertices = NVertices();
	int nfaces = NFaces();
	int nedges = 0;
	fwrite(&nvertices, sizeof(int), 1, fp);
	fwrite(&nfaces, sizeof(int), 1, fp);
	fwrite(&nedges, sizeof(int), 1, fp);
	
	if (single)
	{
		float * vtxData = new float[3*nvertices];
		for (int i=0; i<nvertices; i++)
		{
			R3Point vp = VertexPosition(i);
			vtxData[3*i+0] = (float)vp.X();
			vtxData[3*i+1] = (float)vp.Y();
			vtxData[3*i+2] = (float)vp.Z();
		}
		fwrite(vtxData, sizeof(float), 3*nvertices, fp);
		delete [] vtxData;
	}
	else
	{
		double * vtxData = new double[3*nvertices];
		for (int i=0; i<nvertices; i++)
		{
			R3Point vp = VertexPosition(i);
			vtxData[3*i+0] = vp.X();
			vtxData[3*i+1] = vp.Y();
			vtxData[3*i+2] = vp.Z();
		}
		fwrite(vtxData, sizeof(double), 3*nvertices, fp);
		delete [] vtxData;	
	}
	
	int nIds = 3;
	int vtxIds[3];
	for (int i=0; i<nfaces; i++)
	{	
		for (int v=0; v<3; v++)
			vtxIds[v] = VertexOnFace(i, v);
		
		fwrite(&nIds, sizeof(int), 1, fp);
		fwrite(vtxIds, sizeof(int), 3, fp);
	}
	
	fclose(fp);
}

void ShapeTriangleSoup::ReadPly(const VKString & filename, const VKString & readValsName)
{
	// parse *ply from PSB (NOT able to parse general PLY!!!)
	std::ifstream textStream(filename.c_str());
	if (!textStream.is_open())
		std::cout<<"[ERROR] Invalid ply file: "<<filename.c_str()<<std::endl;
	assert(textStream.is_open());
	VKString tempStr;
	int nVert, nFace;
	tempStr.readLine(textStream);		assert(tempStr=="ply");
	tempStr.readLine(textStream);		assert(tempStr=="format ascii 1.0");
	tempStr.readToken(textStream);		assert(tempStr=="element");
	tempStr.readToken(textStream);		assert(tempStr=="vertex");
	textStream>>nVert;
	tempStr.readLine(textStream);	// read end of line
	tempStr.readLine(textStream);		assert(tempStr=="property float x");
	tempStr.readLine(textStream);		assert(tempStr=="property float y");
	tempStr.readLine(textStream);		assert(tempStr=="property float z");
	tempStr.readToken(textStream);		assert(tempStr=="element");
	tempStr.readToken(textStream);		assert(tempStr=="face");
	textStream>>nFace;
	tempStr.readLine(textStream);	// read end of line
	tempStr.readLine(textStream);		assert(tempStr=="property list uchar int vertex_indices");
	
	VKStringList valNames;
	while (tempStr!="end_header")
	{
		tempStr.readLine(textStream);
		if (tempStr=="end_header")
			break;
		assert(tempStr.startsWith("property int"));
		tempStr = tempStr.replace("property int ", "");
		valNames.push_back(tempStr);
	}
	
	//ShapeTriangleSoup * mesh = new ShapeTriangleSoup("none", m_sMainGeomName, this);
	
	// read vertices
	double x, y, z;
	for (int v=0; v<nVert; v++)
	{
		textStream>>x>>y>>z;
		AddVertex(x, y, z);
	}
	
	ShapeLabels * labels = new ShapeLabels(this, readValsName);
	int nv;
	int nentries = 3 + valNames.count();
	int * readVals = new int[nFace * nentries];
	
	std::vector<std::vector<int> > groupedFaces;
	std::vector<SceneGraphLabel*> labelsTree;
	SceneGraphLabel * root = new SceneGraphLabel();
	root->m_sLabel = "PSB_root";
	root->m_iID = 0;
	root->m_pParent = NULL;
	labelsTree.push_back(root);
	
	std::vector<SceneGraphGeometry*> labelsGeoms;
	// read faces (with last integer == segment ID)
	for (int f=0; f<nFace; f++)
	{
		textStream>>nv;
		assert(nv==3);
		for (int i=0; i<nentries; i++)
			textStream>>readVals[f*nentries+i];
		assert(!textStream.eof());
		while (readVals[f*nentries+3] >= (int)groupedFaces.size())
			groupedFaces.push_back(std::vector<int>());
		groupedFaces[readVals[f*nentries+3]].push_back(f);		
	}
	textStream.close();
	
	std::vector<int> segIDs;
	for (int s=0; s<(int)groupedFaces.size(); s++)
	{
		SceneGraphLabel * newNode = new SceneGraphLabel();
		newNode->m_sLabel = VKString("Seg_")+VKString::number(s);
		newNode->m_iID = (int)labelsTree.size();
		root->m_apChildren.push_back(newNode);
		newNode->m_pParent = root;
		int gID = (int)labelsGeoms.size();
		newNode->m_aiGeometryID.push_back(gID);
		if (groupedFaces.size()>0)
			newNode->m_iFirstFaceInGeometry.push_back(NFaces());
		else
			newNode->m_iFirstFaceInGeometry.push_back(-1);
		newNode->m_iNFacesInGeometry.push_back((int)groupedFaces[s].size());
		
		SceneGraphGeometry * newGeometry = new SceneGraphGeometry();
		newGeometry->m_iID = gID;
		newGeometry->m_sGeometryName = newNode->m_sLabel;
		
		labelsTree.push_back(newNode);
		labelsGeoms.push_back(newGeometry);
		
		for (int fID=0; fID < (int)groupedFaces[s].size(); fID++)
		{
			segIDs.push_back(newNode->m_iID);
			int f = groupedFaces[s][fID];
			AddFace(readVals[f*nentries+0], readVals[f*nentries+1], readVals[f*nentries+2]);
		}
	}
	
	labels->SetFaceSegments(segIDs);
	labels->SetAllLabels(labelsTree);
	labels->SetAllGeometries(labelsGeoms);
	
	AddShapeVal(labels->GetShapeValueName(), labels);
	delete [] readVals;
}

R3Mesh * ShapeTriangleSoup::CreateR3Mesh() const
{
	R3Mesh * mesh = new R3Mesh();
	
	std::vector<R3MeshVertex *> vtx;
	for (int i=0; i<NVertices(); i++)
		vtx.push_back(mesh->CreateVertex(VertexPosition(i)));
	
	for (int i=0; i<NFaces(); i++)
	{
		R3MeshVertex * v1 = vtx[VertexOnFace(i, 0)];
		R3MeshVertex * v2 = vtx[VertexOnFace(i, 1)];
		R3MeshVertex * v3 = vtx[VertexOnFace(i, 2)];
		R3MeshFace * face = mesh->CreateFace(v1, v2, v3);
		
		if (face==NULL) face = mesh->CreateFace(v1, v3, v2);
		if (face==NULL) face = mesh->CreateFace(v2, v1, v3);
		if (face==NULL) face = mesh->CreateFace(v2, v3, v1);
		if (face==NULL) face = mesh->CreateFace(v3, v1, v2);
		if (face==NULL) face = mesh->CreateFace(v3, v2, v1);
		if (face==NULL)
		{
			v1 = mesh->CreateVertex(mesh->VertexPosition(v1));
			v2 = mesh->CreateVertex(mesh->VertexPosition(v2));
			v3 = mesh->CreateVertex(mesh->VertexPosition(v3));
			face = mesh->CreateFace(v1, v2, v3);
			assert(face!=NULL);
		}
	}
	return mesh;
}

int ShapeTriangleSoup::AddVertex(double x, double y, double z,
								 double nx, double ny, double nz)
{
	int vID = (int)m_Vertices.size();
	m_Vertices.push_back(MeshElementVertex());
	m_Vertices[vID].p[0] = x;
	m_Vertices[vID].p[1] = y;
	m_Vertices[vID].p[2] = z;
	m_Vertices[vID].vID = vID;
	if (nx != 0 || ny != 0 || nz != 0)
	{
		m_Vertices[vID].validNorm = true;
		m_Vertices[vID].normal = R3Vector(nx, ny, nz);
	}
	else
		m_Vertices[vID].validNorm = false;
	return vID;
}

int ShapeTriangleSoup::AddFace(int v1, int v2, int v3)
{
	int fID = (int)m_Triangles.size();
	m_Triangles.push_back(MeshElementTriangle());
	m_Triangles[fID].v[0] = v1;
	m_Triangles[fID].v[1] = v2;
	m_Triangles[fID].v[2] = v3;
	m_Triangles[fID].tID = fID;
	
	m_Vertices[v1].faces.push_back(fID);
	m_Vertices[v2].faces.push_back(fID);
	m_Vertices[v3].faces.push_back(fID);
	return fID;
}

void ShapeTriangleSoup::RemoveFaces(int f1, int f2)
{
	for (int i=0; i<NShapeVals(); i++)
		GetShapeVal(i)->RemoveFaces(f1, f2);	// removes isolated vertices as well..

	std::vector<int> v2visit;
	for (int f=f1; f<=f2; f++)
	for (int vid=0; vid<3; vid++)
		v2visit.push_back(m_Triangles[f].v[vid]);
		
	// erase face, re-index faces after f
	m_Triangles.erase(m_Triangles.begin() + f1, m_Triangles.begin() + (f2+1));
	for (int i=f1; i<(int)m_Triangles.size(); i++)
		m_Triangles[i].tID = i;

	int ndelete = f2-f1+1;
	// re-index vertices refering to those faces
	for (int i=0; i < (int)v2visit.size(); i++)
	{
		int vid = v2visit[i];
		for (int j=0; j<(int)m_Vertices[vid].faces.size(); j++)
		{
			int & currF = m_Vertices[vid].faces[j];
			if (currF>=f1 && currF <= f2)	// erase
			{
				m_Vertices[vid].faces.erase(m_Vertices[vid].faces.begin()+j);
				j--;
			}
			else if (m_Vertices[vid].faces[j] > f2) // re-index
				m_Vertices[vid].faces[j]-=ndelete;
		}
	}
	
	// remove vertices with no faces
	for (int i=0; i<(int)v2visit.size(); i++)
		if (m_Vertices[v2visit[i]].faces.size()==0)
			RemoveVertex(v2visit[i]);

}

void ShapeTriangleSoup::RemoveFace(int f)
{
	// update labels
	for (int i=0; i<NShapeVals(); i++)
		GetShapeVal(i)->RemoveFace(f);	// removes isolated vertices as well..

	// store triangles
	int v[3];
	for (int i=0; i<3; i++)
		v[i] = m_Triangles[f].v[i];
	
	// erase face, re-index faces after f
	m_Triangles.erase(m_Triangles.begin() + f);
	for (int i=f; i<(int)m_Triangles.size(); i++)
		m_Triangles[i].tID = i;
	
	// re-index vertices refering to those faces
	int deleted = 0;
	for (int i=0; i < (int)m_Vertices.size(); i++)
	{
		for (int j=0; j<(int)m_Vertices[i].faces.size(); j++)
		{
			if (m_Vertices[i].faces[j] == f)	// erase
			{
				m_Vertices[i].faces.erase(m_Vertices[i].faces.begin()+j);
				j--;
				deleted++;
			}
			else if (m_Vertices[i].faces[j] > f) // re-index
				m_Vertices[i].faces[j]--;
		}
	}
	assert(deleted==3);

	// remove vertices with no faces
	for (int i=0; i<3; i++)
		if (m_Vertices[v[i]].faces.size()==0)
			RemoveVertex(v[i]);
}

void ShapeTriangleSoup::RemoveVertex(int v)
{
	assert(m_Vertices[v].faces.size()==0);
	m_Vertices.erase(m_Vertices.begin() + v);
	for (int f=0; f<(int)m_Triangles.size(); f++)
	for (int lv=0; lv <3; lv++)
	{
		if (m_Triangles[f].v[lv] > v)	// update index
			m_Triangles[f].v[lv]--;
		else	// make sure v it not referenced anywhere...
			assert(m_Triangles[f].v[lv]!=v);	
	}
}

// vertices
int ShapeTriangleSoup::NVertices() const
{
	return (int)m_Vertices.size();
}

R3Point ShapeTriangleSoup::VertexPosition(int vID) const
{
	const double * p = m_Vertices[vID].p;
	return R3Point(p[0], p[1], p[2]); 
}

R3Vector ShapeTriangleSoup::VertexNormal(int vID) const
{
	// THIS IS NOT THE RIGHT WAY TO INTERPOLATE NORMALS!
	if (m_Vertices[vID].validNorm)
		return m_Vertices[vID].normal;
	
	((MeshElementVertex&)m_Vertices[vID]).validNorm = true;
//	((MeshElementVertex&)m_Vertices[vID]).normal = R3zero_vector;
//	for (int f=0; f < (int)m_Vertices[vID].faces.size(); f++)
//	{
//		int fID = m_Vertices[vID].faces[f];
//		((MeshElementVertex&)m_Vertices[vID]).normal += FaceArea(fID) * FaceNormal(fID);
//	}
	assert((int)m_Vertices[vID].faces.size()>0);
	((MeshElementVertex&)m_Vertices[vID]).normal = FaceNormal(m_Vertices[vID].faces[0]);
	((MeshElementVertex&)m_Vertices[vID]).normal.Normalize();
	
	return m_Vertices[vID].normal;
}

int ShapeTriangleSoup::FaceOnVertex(int vID) const
{
	return FaceOnVertex(vID, 0);
}

int ShapeTriangleSoup::NAdjacentFaces(int vID) const
{
	assert(vID < NVertices());
	return (int)m_Vertices[vID].faces.size();
}

int ShapeTriangleSoup::FaceOnVertex(int vID, int nFace) const
{
	bool validFace = vID < NVertices() && nFace < (int)m_Vertices[vID].faces.size();
	if (!validFace)
	{
		std::cout<<"[ERROR] Requesting invalid face on vertex: v="<<vID<<" / "<<NVertices();
		std::cout<<" f="<<nFace<<" / "<<m_Vertices[vID].faces.size()<<std::endl;
	}
	assert(validFace);
	return m_Vertices[vID].faces[nFace];
}

std::vector<int> ShapeTriangleSoup::AdjacentVertices(int vID) const
{
	std::map<int, bool> uniqueVtx;
	assert(vID < NVertices());
	
	for (int i=0; i<(int)m_Vertices[vID].faces.size(); i++)
	{
		int fID = m_Vertices[vID].faces[i];
		uniqueVtx[VertexOnFace(fID, 0)] = true;
		uniqueVtx[VertexOnFace(fID, 1)] = true;
		uniqueVtx[VertexOnFace(fID, 2)] = true;
	}
	
	std::vector<int> neighbors;
	for (std::map<int, bool>::iterator iter = uniqueVtx.begin();
		 iter != uniqueVtx.end(); iter++)
		neighbors.push_back(iter->first);
	return neighbors;
}

// edges
int ShapeTriangleSoup::NEdges() const
{
	int nEdges = 0;
	for (int vID=0; vID < NVertices(); vID++)
		nEdges += VertexValence(vID);
	return nEdges / 2;
}

// faces
int ShapeTriangleSoup::NFaces() const
{
	return (int)m_Triangles.size();
}

int ShapeTriangleSoup::VertexOnFace(int fID, int vID) const
{
	return m_Triangles[fID].v[vID];
}

R3Point ShapeTriangleSoup::FacePoint(int fID, double b1, double b2) const
{
	R3Point point = R3zero_point;
	int v1 = m_Triangles[fID].v[0];
	int v2 = m_Triangles[fID].v[1];
	int v3 = m_Triangles[fID].v[2];
	point += b1 * VertexPosition(v1);
	point += b2 * VertexPosition(v2);
	point += (1.-b1-b2) * VertexPosition(v3);
	return point;
}

R3Point ShapeTriangleSoup::FaceBarycentric(int fID, const R3Point & point) const
{
	// From http://www.devmaster.net/wiki/Ray-triangle_intersection
	int v1 = m_Triangles[fID].v[0];
	int v2 = m_Triangles[fID].v[1];
	int v3 = m_Triangles[fID].v[2];
	R3Point p0 = VertexPosition(v1);
	R3Point p1 = VertexPosition(v2);
	R3Point p2 = VertexPosition(v3);
	R3Vector b = p1 - p0;
	R3Vector c = p2 - p0;
	R3Vector p = point - p0;
	RNDimension dim = FaceNormal(fID).MaxDimension();
	RNDimension dim0 = (dim+1)%3;
	RNDimension dim1 = (dim+2)%3;
	RNScalar denom = b[dim1]*c[dim0] - b[dim0]*c[dim1];
	if (denom == 0) return R3zero_point;
	RNScalar b1 = (p[dim1]*c[dim0] - p[dim0]*c[dim1]) / denom;
	RNScalar b2 = (p[dim1]*b[dim0] - p[dim0]*b[dim1]) / -denom;
	RNScalar b0 = 1 - b1 - b2;
	return R3Point(b0, b1, b2);

}

R3Vector ShapeTriangleSoup::FaceNormal(int fID) const
{
	R3Point p0 = VertexPosition(m_Triangles[fID].v[0]);
	R3Point p1 = VertexPosition(m_Triangles[fID].v[1]);
	R3Point p2 = VertexPosition(m_Triangles[fID].v[2]);

	R3Vector v1 = p1-p0;
	R3Vector v2 = p2-p0;
	R3Vector v3 = v1 % v2;
	v3.Normalize();
	return v3;
}

double ShapeTriangleSoup::FaceArea(int fID) const
{
	R3Point p0 = VertexPosition(m_Triangles[fID].v[0]);
	R3Point p1 = VertexPosition(m_Triangles[fID].v[1]);
	R3Point p2 = VertexPosition(m_Triangles[fID].v[2]);
	
	R3Vector v1 = p1-p0;
	R3Vector v2 = p2-p0;
	R3Vector v3 = v1 % v2;
	return 0.5 * v3.Length();
}

void ShapeTriangleSoup::RearrangeAxes(char newOrder[3], int newSigns[3])
{
//	std::cout<<"name="<<GetModel()->ModelName().c_str()<<"::"<<GeometryName().c_str()<<std::endl;
//	std::cout<<"re-arranging "<<newOrder[0]<<" "<<newOrder[1]<<" "<<newOrder[2]<<" "<<std::endl;
//	for (int d=0; d<3; d++)
//	{
//		int dim = ((int)newOrder[d]-(int)'x');
//		assert(dim>=0 && dim<3);
//		std::cout<<"newpnt["<<d<<"] = old["<<dim<<"]"<<std::endl;
//	}

	for (int i=0; i<(int)m_Vertices.size(); i++)
	{
		double * oldPnt = m_Vertices[i].p;
		double newPnt[3];
		for (int d=0; d<3; d++)
		{
			int dim = ((int)newOrder[d]-(int)'x');
			assert(dim>=0 && dim<3);
			newPnt[d] = newSigns[d] * oldPnt[dim];
		}
		for (int d=0; d<3; d++)
			m_Vertices[i].p[d] = newPnt[d];
	}
}

R3MeshSearchTree * ShapeTriangleSoup::CreateSearchTree(R3Mesh ** meshPtr,
													   std::vector<int> & triangleMap,
													   std::vector<int> & vertexMap) const 
{
	R3Mesh * mesh = new R3Mesh();
	*meshPtr = mesh;
	
	for (int tID = 0; tID < NFaces(); tID++)
	{
		int v1 = VertexOnFace(tID, 0);
		int v2 = VertexOnFace(tID, 1);
		int v3 = VertexOnFace(tID, 2);
		
		R3Point v1Pos = VertexPosition(v1);
		R3Point v2Pos = VertexPosition(v2);
		R3Point v3Pos = VertexPosition(v3);
		
		R3Vector normal = FaceNormal(tID);
		
		R3MeshVertex * v1Ptr = mesh->CreateVertex(v1Pos, normal);
		R3MeshVertex * v2Ptr = mesh->CreateVertex(v2Pos, normal);
		R3MeshVertex * v3Ptr = mesh->CreateVertex(v3Pos, normal);
		
		R3MeshFace * facePtr = mesh->CreateFace(v1Ptr, v2Ptr, v3Ptr);
		
		int v1ID = mesh->VertexID(v1Ptr);
		int v2ID = mesh->VertexID(v2Ptr);
		int v3ID = mesh->VertexID(v3Ptr);
		
		int faceID = mesh->FaceID(facePtr);
		
		assert(v1ID==(int)vertexMap.size());
		vertexMap.push_back(v1);
		assert(v2ID==(int)vertexMap.size());
		vertexMap.push_back(v2);
		assert(v3ID==(int)vertexMap.size());
		vertexMap.push_back(v3);
		
		assert(faceID==(int)triangleMap.size());
		triangleMap.push_back(tID);
	}
	
	for (int i=0; i<(int)vertexMap.size(); i++)
		assert(vertexMap[i]>=0);
	
	return new R3MeshSearchTree(mesh);
}

