/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __CoAlignedParts_H
#define __CoAlignedParts_H

#include "Model3D.h"
#include "ShapeMesh.h"
#include "ShapeLabels.h"
#include "ShapePointCloud.h"
#include "ShapeLocalPCA.h"

#include "PartBox.h"
#include "TemplateNodeDeform.h"

class CoAlignedParts;

/**
 * Part - can be a node matched to point cloud, or an agglomeration of points
 */
class MatchedPart
{
	public:
		friend class CoAlignedParts;
	
		MatchedPart(int matchID, int deformID);
		int m_iMatchID;
		int m_iDeformID;	// can be -1 if taken from multiple
		
		std::vector<int> m_PntsSubset; // must be defined if m_iDeformID==-1, otherwise can be empty (all points)
		std::vector<double> m_Confidences;
		std::map<int, int> m_ToSubsetMap;
};

/**
 * Several parts
 */
class PartCluster
{
	public:
		friend class CoAlignedParts;
	
		PartCluster(int clusterID, int seedID, int nodeID, int partID, int nParts);
	
		int m_iClusterID;
		int m_iSeedID;	// -1 for mixed
		int m_iNodeID;	// -1 for mixed
		int m_iPartID;	// -1 for mixed
		int m_iNParts;	// -1 for mixed
		
		std::vector<MatchedPart> m_AllParts;

		double m_AveConfidence;
		double m_MinConfidence;
		double m_MaxConfidence;
};

/**
 * A clustering of parts
 */
class ClusterPartPair
{
	public:
		int m_iClusterID;
		int m_iPartID;
};

class PartClustering
{
	public:
		
		friend class CoAlignedParts;
		PartClustering(int clusteringID, int seedID, int nodeID, int partID, int nParts);
		void GetMatch(int matchID, int nodeID, int * clusterID, int * partInClusterID, bool & valid) const;
		
		int m_iClusteringID;
		int m_iSeedID;	// -1 for mixed
		int m_iNodeID;	// -1 for mixed
		int m_iPartID;	// -1 for mixed
		int m_iNParts;	// -1 for mixed

		std::vector<PartCluster> m_AllClusters;
	
		bool m_bPrecomputedMatch2Cluster;
		std::vector<std::map<int, ClusterPartPair> > m_Match2ClusterAndPartID;
};

/**
 * model in cluster
 */
class ModelInCluster
{
	public:
		TMatchEnergy * m_pModelMatch;
		int m_iModelID;
		double m_fOldEnergy;
		double m_fFitToSeedEnergy;
		int m_iNLabelsChanged;
	
		bool m_bLowEnergy;
		bool m_bLabelsSame;
		bool m_bEnergyImproved;
};

/**
 * cluster of models
 */
class ClusterOfModels
{
	public:
		TMatch * m_pSeedDeform;
		int m_iCurrentSeedID;
		VKString m_sClusterName;
		std::vector<ModelInCluster> m_ClusterOfModels;
};


/**
 * This class supports various operations on a collection of models co-aligned to template.
 */
class CoAlignedParts
{
	public:
		enum ConfidenceCalcFlags
		{
			CCF_Dist2Tmplt 		= 1<<0,
			CCF_Dist2All 		= 1<<1
		};
		static void PrintLog(const VKString & str, std::ofstream * fileStream);
	
		CoAlignedParts(STemplate * st, std::vector<TMatchEnergy * > & matches);
		virtual ~CoAlignedParts();
	
		virtual int NClusterings() const;
		virtual int GetClusteringID(int seedID, int nodeID, int partID, int nParts) const;
		virtual const PartClustering & GetClustering(int clusteringID) const;
		
		virtual int NClusters(int clusteringID) const;
		virtual const PartCluster & GetCluster(int clusteringID, int clusterID) const;
		
		virtual int NPartsInCluster(int clusteringID, int clusterID) const;
		virtual const MatchedPart & GetPart(int clusteringID, int clusterID, int partID) const;

		// compute per-point value indicating whether point fits in a cluster
		virtual void ComputeConfidences(int clusteringID, int clusterID, int CCFflags);
		virtual double GetPartConfidenceValue(int clusteringID, int clusterID, int partID,
											  int matchID, bool & valid) const;
		virtual double GetConfidenceValue(int clusteringID, int clusterID, int partID,
										  int matchID, int pointID, bool & valid) const;
		virtual void GetConfidenceValBounds(int clusteringID, int clusterID, double * minValue, double * maxValue) const;
		virtual void GetClusterForPoint(int clusteringID, int matchID, int pointID,
										int * clusterID, int * partInClusterID, bool & valid) const;
		virtual void GetClusterWithPart(int clusteringID, int matchID, int nodeID,
									   int * clusterID, int * partInClusterID, bool & valid) const;
	
		// cluster based on current matching (i.e. by parts)
		virtual int GetClusteringFromMatchesID(int configD);
		virtual int CreateClustersFromMatches(int configID);
		
		// cluster
		virtual void ClusterStyles(int configID, int nodeID, int partID, int nParts);
		virtual void ClusterNullNodes(int configID);
	
		// updater
		virtual VKString UpdateTemplateFromAllMatches(const VKString & updateLog, int totalShapes);
		virtual void AddMatchesToBestSeeds(const std::vector<int> & matches, double distThreshold,
										   std::vector<ClusterOfModels> & seeds, std::vector<int> & notIncluded,
										   std::vector<int> & notIncludedNewConfigs, std::ofstream * logStream);
		virtual void AgglomerativeClustering(const std::vector<int> & elements, double terminationThreshold,
											 std::vector<std::vector<int> > & outClusters, std::ofstream * logStream);
		
//		virtual void ClusterSeedDeforms(double energyThresh,
//										std::vector<ClusterOfModels> & outSeeds, std::ofstream * logStream);
//		virtual void FitToDeforms(double energyThresh, int energyTerms,
//								  std::vector<ClusterOfModels> & inOutClusters, std::ofstream * logStream);
//		virtual void ExtractFinalClusters(const std::vector<ClusterOfModels> & inClusters,
//										  double energyThresh, int nMinModelsToAddNewSeed,
//										  std::vector<ClusterOfModels> & outClusters, std::ofstream * logStream);
		virtual void UpdateConfig(ClusterOfModels & cluster, double energyUpdThresh, std::ofstream * logStream);
//		virtual bool UpdateTemplateFromAllMatches_Nov12(const VKString & updateLog);
	
	protected:
		STemplate * m_pST;
		std::vector<TMatchEnergy*> & m_Matches;

	// 'style' clusters (of parts).
		std::vector<PartClustering> m_AllClusterings;
	
	// special clusterings...
		std::map<int, int> m_Seed2FromMatchesClustering;
};

#endif



