/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "PartLocalGeometry.h"
#include "VkFunctions.h"
#include "TimeProfiler.h"

//////////// LOCAL SHAPE DESCRIPTOR ////////////
VKString PartLocalGeometry::s_sDefaultLocPCAName = "DefaultLocPCA";

int PartLocalGeometry::GetNhdSizeConstant()
{
	return -1;
}

double PartLocalGeometry::GetNhdDist()
{
	return .3;
}


PartLocalGeometry::PartLocalGeometry(const ShapePointCloud * pnts, int offset, int npoints, bool freePnts)
{
	m_PCAName = s_sDefaultLocPCAName;
	m_pPnts = pnts;
	m_iOffset = offset;
	m_iNPoints = npoints;
	m_pLocalPCA = NULL;
	m_aTempFeatureStorage = new double[NLocalFeatures()];
	m_bFreePnts = freePnts;
}

PartLocalGeometry::~PartLocalGeometry()
{
	delete [] m_aTempFeatureStorage;
	if (m_bFreePnts)
	{
		delete m_pPnts;
		delete m_pLocalPCA;
	}
}

R3Triad PartLocalGeometry::LocalOrient(int gpntid, double * eigs, const R3Affine & xform) const
{
//	std::cout<<"local orient: "<<gpntid<<" / "<<m_iNPoints<<" / "<<m_pPnts->NumPoints()<<std::endl;
	assert(gpntid>=0 && gpntid < m_iNPoints);
	const R3Triad & axes = GetLocalPCAVals()->GetPCA(m_pPnts->GetPoint(m_iOffset + gpntid), eigs);
	
	R3Vector a1 = axes[0];
	R3Vector a2 = axes[1];
	R3Vector a3 = axes[2];
	xform.ApplyInverseTranspose(a1);
	xform.ApplyInverseTranspose(a2);
	xform.ApplyInverseTranspose(a3);
	
	R3Triad pntOrient(a1,a2,a3);
	pntOrient.Normalize();
	return pntOrient;

}

const double * PartLocalGeometry::LocalOrientAsFeatures(int pntID, const R3Affine & xform) const
{
	double pi = 3.14159265359;
	double eigs[3];
	const R3Triad & lo = LocalOrient(pntID, eigs, xform);
	if (eigs[0] < 0)
	{
		//std::cout<<"Invalid: "<<eigs[0]<<std::endl;
		for (int i=0; i<NLocalFeatures(); i++)
			m_aTempFeatureStorage[i] = -1;
		return m_aTempFeatureStorage;
	}

	int i0, i1, i2;
	if (eigs[0] > eigs[1] && eigs[0] > eigs[2])			// eig 0 - highest
	{
		i0 = 0;
		if (eigs[2] > eigs[1]) { i1 = 2;    i2 = 1; }	// eig0 >= eig2 >= eig1
		else				   { i1 = 1;    i2 = 2; }	// eig0 >= eig1 >= eig2
	}
	else if (eigs[1] > eigs[2])							// eig 1 - highest
	{
		i0 = 1;
		if (eigs[0] >= eigs[2]) { i1 = 0; 	i2 = 2; }	// eig1 >= eig0 >= eig2
		else 					{ i1 = 2; 	i2 = 0; }	// eig1 >= eig2 >= eig0
	}
	else												// eig 2 - highest
	{
		i0 = 2;
		if (eigs[0] >= eigs[1]) { i1 = 0; 	i2 = 1; }	// eig2 >= eig0 >= eig1
		else 					{ i1 = 1; 	i2 = 0; }	// eig2 >= eig1 >= eig0
	}
	assert(eigs[i0] >= eigs[i1] && eigs[i1] >= eigs[i2]);
	
	// eigs
	m_aTempFeatureStorage[0] = eigs[i1] / eigs[i0];		//
	m_aTempFeatureStorage[1] = eigs[i2] / eigs[i0];		//
	
	// normal
	m_aTempFeatureStorage[2] = acos(vkAbs(lo[i2].Z()))/pi; 		//
	m_aTempFeatureStorage[3] = acos(vkAbs(lo[i2].X()))/pi; 		//
	
	// principal direction
	m_aTempFeatureStorage[4] = acos(vkAbs(lo[i0].Z()))/pi;		//
	m_aTempFeatureStorage[5] = acos(vkAbs(lo[i0].X()))/pi; 		//
	return m_aTempFeatureStorage;
}

int PartLocalGeometry::NPnts() const
{
	//return m_pPnts->NumPoints();
	return m_iNPoints;
}

int PartLocalGeometry::NLocalFeatures()
{
	// theta - anlge w.r.t. 'up' vector (z-axis)
	// phi - angle w.r.t. a vector orthogonal to 'up' (x-axis)
	return 6;	// [ eig1 / eig0, eig2 / eig1, n_theta, n_phi, d_theta, d_phi ]
	//return 2;
}

const ShapeLocalPCA * PartLocalGeometry::GetLocalPCAVals() const
{
	if (m_pLocalPCA!=NULL)
		return m_pLocalPCA;
	
	// get pca vals
	ShapePointCloud * ncPnts = (ShapePointCloud*)m_pPnts;
	assert(ncPnts!=NULL);
	ShapeLocalPCA * pcaVals = NULL;
	ShapeValue * vals = ncPnts->GetShapeVal(m_PCAName);
	if (vals==NULL)	// default null name
	{
		pcaVals = new ShapeLocalPCA(ncPnts, m_PCAName, GetNhdDist(), GetNhdSizeConstant(), true);
		pcaVals->ComputePntsPCA();
		ncPnts->AddShapeVal(m_PCAName, pcaVals);
	}
	else
	{
		assert(vals!=NULL && vals->GetClassName()=="ShapeLocalPCA");
		pcaVals = (ShapeLocalPCA*)vals;
	}
	assert(pcaVals!=NULL);
	((PartLocalGeometry*)this)->m_pLocalPCA = pcaVals;
	assert(m_pLocalPCA!=NULL);
	return m_pLocalPCA;
}


//////////// MAPPING OF NODE IDS ///////////
TrainingNodeIDMap::TrainingNodeIDMap(const std::vector<int> & nodeIDs, int nMaxOrigNodes)
{
	std::map<int, int> id2classID;
	int currClaID = 0;
	int nOrigNodes = 0;
	for (int i=0; i<(int)nodeIDs.size(); i++)
	{
		if (id2classID.find(nodeIDs[i]) == id2classID.end())
			id2classID[nodeIDs[i]] = currClaID++;
		nOrigNodes = vkMax(nodeIDs[i], nOrigNodes);
	}
	
	m_iMaxOrigNodes = nOrigNodes+1;
	m_iNClasses = currClaID;
	m_aiOrigNode2Classifier = new int[m_iMaxOrigNodes];
	m_aiClassifier2OrigNode = new int[currClaID];
	for (int i=0; i<m_iMaxOrigNodes; i++)
		m_aiOrigNode2Classifier[i] = -1;
	
	for (std::map<int, int>::iterator iter = id2classID.begin(); iter != id2classID.end(); iter++)
	{
		assert(iter->first < nOrigNodes+1);
		assert(iter->second < currClaID);
		m_aiOrigNode2Classifier[iter->first] = iter->second;
		m_aiClassifier2OrigNode[iter->second] = iter->first;
	}
}

TrainingNodeIDMap::~TrainingNodeIDMap()
{
	delete [] m_aiClassifier2OrigNode;
	delete [] m_aiOrigNode2Classifier;
}

TrainingNodeIDMap * TrainingNodeIDMap::CreateCopy() const
{
	TrainingNodeIDMap * tmap = new TrainingNodeIDMap();
	tmap->m_aiClassifier2OrigNode = new int[m_iNClasses];
	tmap->m_aiOrigNode2Classifier = new int[m_iMaxOrigNodes];
	for (int i=0; i<m_iNClasses; i++)
		tmap->m_aiClassifier2OrigNode[i] = m_aiClassifier2OrigNode[i];
	for (int i=0; i<m_iMaxOrigNodes; i++)
		tmap->m_aiOrigNode2Classifier[i] = m_aiOrigNode2Classifier[i];
	tmap->m_iNClasses = m_iNClasses;
	tmap->m_iMaxOrigNodes = m_iMaxOrigNodes;

	return tmap;
}

int TrainingNodeIDMap::Classifier2OrigNodeID(int claID) const
{
	assert(claID < m_iNClasses);
	return m_aiClassifier2OrigNode[claID];
}

int TrainingNodeIDMap::OrigNodeID2Classifier(int nodeID) const
{
	if (nodeID >= m_iMaxOrigNodes)
		std::cout<<"[ERROR] Invalid orig node: "<<nodeID<<" / "<<m_iMaxOrigNodes<<std::endl;
	assert(nodeID < m_iMaxOrigNodes);
	return m_aiOrigNode2Classifier[nodeID];
}

int TrainingNodeIDMap::NumUniqueClasses() const
{
	return m_iNClasses;
}

int TrainingNodeIDMap::MaxOrigNodeIDs() const
{
	return m_iMaxOrigNodes;
}

TrainingNodeIDMap::TrainingNodeIDMap()
{
	m_iNClasses = -1;
	m_iMaxOrigNodes = -1;
	m_aiClassifier2OrigNode = NULL;
	m_aiOrigNode2Classifier = NULL;
}

void TrainingNodeIDMap::WriteNodeIDMap(FILE * file) const
{
	fwrite(&m_iNClasses, sizeof(int), 1, file);
	fwrite(&m_iMaxOrigNodes, sizeof(int), 1, file);
	fwrite(m_aiClassifier2OrigNode, sizeof(int), m_iNClasses, file);
	fwrite(m_aiOrigNode2Classifier, sizeof(int), m_iMaxOrigNodes, file);
}

TrainingNodeIDMap * TrainingNodeIDMap::ReadNodeIDMap(FILE * file)
{
	TrainingNodeIDMap * t = new TrainingNodeIDMap();
	fread(&(t->m_iNClasses), sizeof(int), 1, file);
	fread(&(t->m_iMaxOrigNodes), sizeof(int), 1, file);
	t->m_aiClassifier2OrigNode = new int[t->m_iNClasses];
	t->m_aiOrigNode2Classifier = new int[t->m_iMaxOrigNodes];
	fread(t->m_aiClassifier2OrigNode, sizeof(int), t->m_iNClasses, file);
	fread(t->m_aiOrigNode2Classifier, sizeof(int), t->m_iMaxOrigNodes, file);
	return t;
}

int TrainingNodeIDMap::Update(const std::vector<int> & nodeIDs)
{
	int nAdded = 0;
	bool requiresUpdate = false;
	std::vector<int> orignode2classifier;
	std::vector<int> classifier2orignode;

	for (int j=0; j<m_iMaxOrigNodes; j++)
		orignode2classifier.push_back(m_aiOrigNode2Classifier[j]);
	for (int j=0; j<m_iNClasses; j++)
		classifier2orignode.push_back(m_aiClassifier2OrigNode[j]);

	for (int i=0; i<(int)nodeIDs.size(); i++)
	{
		int nid = nodeIDs[i];
		while (nid >= (int)orignode2classifier.size())
			orignode2classifier.push_back(-1);

		if (orignode2classifier[nid]==-1)
		{
			requiresUpdate = true;
			orignode2classifier[nid] = (int)classifier2orignode.size();
			classifier2orignode.push_back(nid);
			nAdded++;
		}
	}

	delete [] m_aiOrigNode2Classifier;
	delete [] m_aiClassifier2OrigNode;
	
	m_iMaxOrigNodes = (int)orignode2classifier.size();
	m_iNClasses = (int)classifier2orignode.size();
	m_aiOrigNode2Classifier = new int[m_iMaxOrigNodes];
	m_aiClassifier2OrigNode = new int[m_iNClasses];
	
	for (int j=0; j<m_iMaxOrigNodes; j++)
		m_aiOrigNode2Classifier[j] = orignode2classifier[j];
	
	for (int j=0; j<m_iNClasses; j++)
		m_aiClassifier2OrigNode[j] = classifier2orignode[j];
	
	return nAdded;
}



//////////// TRAINING DATA ////////////
LocalGeometryTrainingSet::LocalGeometryTrainingSet(const std::vector<const PartLocalGeometry *> & shapes,
												   const std::vector<int> & nodeID, int nMaxOrigNodes)
: m_Shapes(shapes), m_NodeIDs(nodeID)
{
	m_pTrainingIDMapping = new TrainingNodeIDMap(nodeID, nMaxOrigNodes);
	
	m_iNFeatures = -1;
	assert(shapes.size()==nodeID.size());
	for (int i=0; i<(int)shapes.size(); i++)
	{
		if (m_iNFeatures==-1)
			m_iNFeatures = shapes[i]->NLocalFeatures();
		else
			assert(m_iNFeatures==shapes[i]->NLocalFeatures());
	}
	
	m_iNExamples = 0;
	for (int i=0; i<(int)m_Shapes.size(); i++)
		m_iNExamples += m_Shapes[i]->NPnts();
		
	m_aiExampleClasses = new int[m_iNExamples];
	m_afExampleFeatures = new double[m_iNFeatures * m_iNExamples];
	int example = 0;
	std::ofstream testDebug("test.txt");
	for (int i=0; i<(int)m_Shapes.size(); i++)
	{
		int classID = m_pTrainingIDMapping->OrigNodeID2Classifier(m_NodeIDs[i]);
		for (int p = 0; p<(int)m_Shapes[i]->NPnts(); p++)
		{
			const double * feat = m_Shapes[i]->LocalOrientAsFeatures(p, R3identity_affine);
			testDebug<<classID<<"\n";
			for (int f=0; f<m_iNFeatures; f++)
			{
				m_afExampleFeatures[example*m_iNFeatures + f] = feat[f];
				testDebug<<feat[f]<<" ";
			}
			testDebug<<"\n";
			// deleteme:
			R3Point pos = m_Shapes[i]->m_pPnts->PointPosition(p);
			testDebug<<pos[0]<<" "<<pos[1]<<" "<<pos[2]<<"\n";
			
			m_aiExampleClasses[example] = classID;
			example++;
		}
	}
	testDebug.close();
	assert(example == m_iNExamples);
}

LocalGeometryTrainingSet::LocalGeometryTrainingSet(const LocalGeometryTrainingSet & training)
: m_Shapes(training.m_Shapes), m_NodeIDs(training.m_NodeIDs)
{
	assert(false);	// potentially unsfe...
}

LocalGeometryTrainingSet::~LocalGeometryTrainingSet()
{
	delete [] m_aiExampleClasses;
	delete [] m_afExampleFeatures;
}

long LocalGeometryTrainingSet::numExamples() const
{
	return m_iNExamples;
}

long LocalGeometryTrainingSet::numFeatures() const
{
	return m_iNFeatures;
}


void LocalGeometryTrainingSet::getFeature(long feature_index, std::vector<double> & values) const
{
	values.clear();
	for (int i=0; i<m_iNExamples; i++)
		values.push_back(m_afExampleFeatures[i * m_iNFeatures + (int)feature_index]);
}

void LocalGeometryTrainingSet::getClasses(std::vector<long> & classes) const
{
	classes.clear();
	for (int i=0; i<m_iNExamples; i++)
		classes.push_back(m_aiExampleClasses[i]);
}

TrainingNodeIDMap * LocalGeometryTrainingSet::GetNodeIDMapping() const
{
	return (TrainingNodeIDMap*)m_pTrainingIDMapping;
}


//////////// CLASSIFIER /////////////
LocalGeometryClassifier::LocalGeometryClassifier(const LocalGeometryTrainingSet & trainingData)
{
	m_ClassifierType = CLASSIFY_GMM;	// maybe needs further debugging.
	//m_ClassifierType = CLASSIFY_JB;
	
	m_pTrainingIDMapping = trainingData.GetNodeIDMapping();

	// train classifier
	m_iNFeatures = trainingData.numFeatures();
	m_iNNodes = m_pTrainingIDMapping->NumUniqueClasses();
	m_afTempProbStorageClassifier = new double[m_pTrainingIDMapping->NumUniqueClasses()];
	m_afTempProbStorageReturn = new double[m_pTrainingIDMapping->MaxOrigNodeIDs()];

//	m_pJB = NULL;
	
	if (m_ClassifierType==CLASSIFY_JB)
		TrainJointBoost(trainingData);
	else if (m_ClassifierType==CLASSIFY_GMM)
		TrainGMM(trainingData);
}

LocalGeometryClassifier::LocalGeometryClassifier()
{
	m_ClassifierType = CLASSIFY_NULL;
	m_pTrainingIDMapping = 0;
	
	m_iNFeatures = -1;
	m_iNNodes = -1;
	m_afTempProbStorageClassifier = NULL;
	m_afTempProbStorageReturn = NULL;
	
//	m_pJB = NULL;
}

LocalGeometryClassifier::~LocalGeometryClassifier()
{
	delete [] m_afTempProbStorageClassifier;
	delete [] m_afTempProbStorageReturn;
}

LocalGeometryClassifier * LocalGeometryClassifier::CreateCopy()
{
	LocalGeometryClassifier * rc = new LocalGeometryClassifier();
	rc->m_ClassifierType = m_ClassifierType;
	
	rc->m_pTrainingIDMapping = m_pTrainingIDMapping->CreateCopy();
	rc->m_afTempProbStorageClassifier = new double[rc->m_pTrainingIDMapping->NumUniqueClasses()];
	rc->m_afTempProbStorageReturn = new double[rc->m_pTrainingIDMapping->MaxOrigNodeIDs()];
	
	rc->m_iNFeatures = m_iNFeatures;
	rc->m_iNNodes = m_iNNodes;
	
	for (int i=0; i<(int)m_GMM.size(); i++)
		rc->m_GMM.push_back(m_GMM[i]->CreateCopy()->SafeCastToGaussian());
//	assert(m_pJB==NULL);
	return rc;
}

void LocalGeometryClassifier::WriteClassifier(FILE * file) const
{
	assert(m_ClassifierType==CLASSIFY_GMM);	// for now GMM only
	fwrite(&m_ClassifierType, sizeof(ClassifierType), 1, file);
	fwrite(&m_iNFeatures, sizeof(int), 1, file);
	fwrite(&m_iNNodes, sizeof(int), 1, file);
	int ngmm = (int)m_GMM.size();
	fwrite(&ngmm, sizeof(int), 1, file);
	m_pTrainingIDMapping->WriteNodeIDMap(file);
	for (int i=0; i<ngmm; i++)
		m_GMM[i]->WriteToFile(file);
}

LocalGeometryClassifier * LocalGeometryClassifier::ReadClassifier(FILE * file)
{
	LocalGeometryClassifier * rc = new LocalGeometryClassifier();
	fread(&(rc->m_ClassifierType), sizeof(ClassifierType), 1, file);
	assert(rc->m_ClassifierType==CLASSIFY_GMM);	// for now GMM only
	
	fread(&rc->m_iNFeatures, sizeof(int), 1, file);
	fread(&rc->m_iNNodes, sizeof(int), 1, file);

	int ngmm = -1;
	fread(&ngmm, sizeof(int), 1, file);
	rc->m_pTrainingIDMapping = TrainingNodeIDMap::ReadNodeIDMap(file);
	rc->m_afTempProbStorageClassifier = new double[rc->m_pTrainingIDMapping->NumUniqueClasses()];
	rc->m_afTempProbStorageReturn = new double[rc->m_pTrainingIDMapping->MaxOrigNodeIDs()];
	assert(rc->m_iNNodes==rc->m_pTrainingIDMapping->NumUniqueClasses());
//	int nMaxOrigNodes = rc->m_pTrainingIDMapping->MaxOrigNodeIDs();
//	if (rc->m_iNNodes!=nMaxOrigNodes)
//		std::cout<<"[ERROR] Numbfer of nodes does not match"<<rc->m_iNNodes<<" vs "<<nMaxOrigNodes<<std::endl;
//	assert(rc->m_iNNodes==nMaxOrigNodes);
	
	for (int i=0; i<ngmm; i++)
		rc->m_GMM.push_back(ContProbDistribution::CreateFromFile(file)->SafeCastToGaussian());
	return rc;
}

void LocalGeometryClassifier::UpdateClassifier(const LocalGeometryTrainingSet & trainingData)
{
	assert(m_ClassifierType==CLASSIFY_GMM);	// No easy way to update joint boost
	// NOTE: map might change since new items might be added
	int nNewClasses = m_pTrainingIDMapping->Update(trainingData.m_NodeIDs);
	m_iNNodes = m_pTrainingIDMapping->NumUniqueClasses();
	assert((int)m_GMM.size() + nNewClasses == m_pTrainingIDMapping->NumUniqueClasses());
	for (int i=0; i<nNewClasses; i++)
		m_GMM.push_back(new ProbGaussian(m_iNFeatures));

	delete [] m_afTempProbStorageClassifier;
	delete [] m_afTempProbStorageReturn;
	m_afTempProbStorageClassifier = new double[m_pTrainingIDMapping->NumUniqueClasses()];
	m_afTempProbStorageReturn = new double[m_pTrainingIDMapping->MaxOrigNodeIDs()];
	
	TrainGMM(trainingData);
}

const double * LocalGeometryClassifier::LogProbs(int NFeat, const double * features) const
{
	assert(NFeat == m_iNFeatures);
	
//	std::cout<<"Getting log prob - 1"<<std::endl;
	if (m_ClassifierType==CLASSIFY_JB)
		LogProbsJointBoost(NFeat, features);
	else if (m_ClassifierType==CLASSIFY_GMM)
		LogProbsGMM(NFeat, features);
	else
		assert(false);

//	std::cout<<"Getting log prob - 2. Storage: "<<m_afTempProbStorageReturn<<" & "<<m_afTempProbStorageClassifier<<std::endl;
	
	for (int i=0; i<m_pTrainingIDMapping->MaxOrigNodeIDs(); i++)
	{
//		std::cout<<" i = "<<i<<" / "<<m_pTrainingIDMapping->MaxOrigNodeIDs()<<std::endl;
		int claID = m_pTrainingIDMapping->OrigNodeID2Classifier(i);
		if (claID==-1)
			m_afTempProbStorageReturn[i] = -1;
		else
			m_afTempProbStorageReturn[i] = m_afTempProbStorageClassifier[claID];
	}
//	std::cout<<"Getting log prob - 3"<<std::endl;
	
	return m_afTempProbStorageReturn;
}

void LocalGeometryClassifier::TrainGMM(const LocalGeometryTrainingSet & trainingData)
{
	assert(m_GMM.size()==0 || (int)m_GMM.size()==m_pTrainingIDMapping->NumUniqueClasses());
	if (m_GMM.size()==0)
	{
		for (int i=0; i<m_pTrainingIDMapping->NumUniqueClasses(); i++)
			m_GMM.push_back(new ProbGaussian(m_iNFeatures));
	}
	for (int i=0; i<(int)trainingData.m_Shapes.size(); i++)
	{
//		std::cout<<"i: "<<i<<" / ("<<trainingData.m_Shapes.size()<<" "<<trainingData.m_NodeIDs.size()<<")"<<std::endl;
//		std::cout<<"  nodeid = "<<trainingData.m_NodeIDs[i]<<std::endl;

		int claID = m_pTrainingIDMapping->OrigNodeID2Classifier(trainingData.m_NodeIDs[i]);
		assert(claID>=0);
		if (claID < 0 || claID>=(int)m_GMM.size())
			std::cout<<"[ERROR] Invalid cluster id: "<<claID<<" / "<<m_GMM.size()<<std::endl;
		assert(claID>=0 && claID < (int)m_GMM.size());
		
		const PartLocalGeometry * pg = trainingData.m_Shapes[i];
		
		for (int e = 0; e < pg->NPnts(); e++)
		{
			const double * lo = pg->LocalOrientAsFeatures(e, R3identity_affine);
			if (lo[0]>=0)
				m_GMM[claID]->UpdateDistribution(lo, true);
		}
	}
	
//	for (int claID = 0; claID < (int)m_GMM.size(); claID++)
//	{
//		const double * meanGMM = m_GMM[claID]->GetMostProbable();
//		std::cout<<claID<<": [ ";
//		for (int f = 0; f < m_iNFeatures; f++)
//			std::cout<<meanGMM[f]<<" ";
//		std::cout<<"]"<<std::endl;
//	}
}

void LocalGeometryClassifier::LogProbsGMM(int NFeat, const double * features) const
{
	// should fill: m_afTempProbStorageClassifier
	assert((int)m_GMM.size()==m_pTrainingIDMapping->NumUniqueClasses());
	for (int i=0; i<(int)m_pTrainingIDMapping->NumUniqueClasses(); i++)
		m_afTempProbStorageClassifier[i] = -m_GMM[i]->LogProb(features);
}

const ProbGaussian * LocalGeometryClassifier::GetGaussian(int localNodeID) const
{
	int claID = m_pTrainingIDMapping->OrigNodeID2Classifier(localNodeID);
	assert(claID>=0 && claID < (int)m_GMM.size());
	return m_GMM[claID];
}

void LocalGeometryClassifier::TrainJointBoost(const LocalGeometryTrainingSet & trainingData)
{
	assert(false);
//	Thea::JointBoost::Options opt = Thea::JointBoost::Options::defaults();
//	opt.setMinBoostingRounds(2 * m_iNNodes)		// 4 * m_iNNodes
//	.setMaxBoostingRounds(4 * m_iNNodes)		// 20 * m_iNNodes
//	.setMinFractionalErrorReduction(0.0000001)
//	.setFeatureSamplingFraction(0.25)
//	.setMaxThresholdsFraction(0.003)
//	.setForceGreedy(true)
//	.setVerbose(false);
//	
//	m_pJB = new Thea::JointBoost(m_iNNodes, m_iNFeatures, opt);
//	m_pJB->train(trainingData);
}

void LocalGeometryClassifier::LogProbsJointBoost(int NFeat, const double * features) const
{
	assert(false);
	// Should fill: m_afTempProbStorageClassifier
//	m_pJB->predict(features, m_afTempProbStorageClassifier);
	for (int i=0; i<(int)m_pTrainingIDMapping->NumUniqueClasses(); i++)
		m_afTempProbStorageClassifier[i] = -log(vkMin(m_afTempProbStorageClassifier[i], 1.));
}


