/***************************************************************************
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "VkFunctions.h"
#include "PartOfEdge.h"
#include "PartTransforms.h"
#include "TMatch.h"
#include "TMatchEnergy.h"


PartTransforms::PartTransforms(int xformflags)
{
	m_iXFormFlags = xformflags;
	
	int nFields;
	int offset = Offset(PARTXFORM_ScaleIsotropic, m_iXFormFlags, &nFields, &m_iNTotalFields);
	assert(m_iNTotalFields>0);
	
	m_aTransform = new double[m_iNTotalFields];
}

PartTransforms::~PartTransforms()
{	
	for (int i=0; i<(int)m_apXForms.size(); i++)
		delete m_apXForms[i];
	delete [] m_aTransform;
}

PartTransforms * PartTransforms::CreateCopy() const
{
	PartTransforms * newXform = new PartTransforms(m_iXFormFlags);
	
	for (int i=0; i<(int)m_apXForms.size(); i++)
		newXform->m_apXForms.push_back(m_apXForms[i]->CreateCopy());
	
	return newXform;
}

int PartTransforms::NSamples() const
{
	int nSamples = -1;
	for (int i=0; i<(int)m_apXForms.size(); i++)
	{
		if (i==0)
			nSamples = m_apXForms[i]->NSamples();
		else
			assert(nSamples==m_apXForms[i]->NSamples());
	}
	assert(nSamples>=0);
	return nSamples;
}

int PartTransforms::XFormFlags() const
{
	return m_iXFormFlags;
}

void PartTransforms::UpdateTransformStats(int partID, const PartBox & parent, const PartBox & child)
{
	while(partID >= m_apXForms.size())
		m_apXForms.push_back(new ProbGaussian(m_iNTotalFields));

    const R3Triad & op = parent.UndeformedOrientation();
    const R3Triad & oc = child.UndeformedOrientation();
    
    
	ComputeAligningTransform(parent.UndeformedCentroid(), op[0], op[1], op[2], 
                             child.UndeformedCentroid(), oc[0], oc[1], oc[2], m_iXFormFlags, m_aTransform);

	const double * xform = m_apXForms[partID]->GetMostProbable();
	m_apXForms[partID]->UpdateDistribution(m_aTransform, true);
	xform = m_apXForms[partID]->GetMostProbable();
}

void PartTransforms::UpdateTransformStats(const TemplateNodeDeform & deform, bool include)
{
	while(deform.PartID() >= m_apXForms.size())
		m_apXForms.push_back(new ProbGaussian(m_iNTotalFields));

	//assert(deform.PartID() >= 0 && deform.PartID() < (int)m_apXForms.size());
	
	int xformType, xformDims;
	const double * xform = deform.GetTransform(xformDims, xformType);
	assert(xformType==m_iXFormFlags);
	m_apXForms[deform.PartID()]->UpdateDistribution(xform, include);
}

void PartTransforms::Reset()
{
	for (int i=0; i<(int)m_apXForms.size(); i++)
		m_apXForms[i]->Reset();
}

void PartTransforms::Normalize(const R3Point & centroid, double aveRadius)
{
	assert(m_iXFormFlags == PARTXFORM_Translation);
	for (int i=0; i<(int)m_apXForms.size(); i++)
	for (int d=0; d<3; d++)
		m_apXForms[i]->Rescale(d, -centroid[d], 1./aveRadius);

}


ContProbDistribution * PartTransforms::GetPartTransform(int partID)
{
	assert(partID>=0 && partID < (int)m_apXForms.size());
	return m_apXForms[partID];
}

const ContProbDistribution * PartTransforms::GetPartTransform(int partID) const
{
	return ((PartTransforms*)this)->GetPartTransform(partID);
}




///// STATIC HELPER FUNCTIONS /////

std::vector<PartTransforms::TransformTypes> PartTransforms::m_Types;
std::vector<int> PartTransforms::m_TypeFields;

R3Affine PartTransforms::CreateAffineTransform(const double * transform, int typeFlags)
{
	R3Affine xform = R3identity_affine;
	int offset = Offset(PARTXFORM_ScaleIsotropic, typeFlags);
	if (offset!=-1)
		xform.Scale(transform[offset]);
	
	offset = Offset(PARTXFORM_ScaleAnisotropic, typeFlags);
	if (offset!=-1)
		xform.Scale(R3Vector(transform[offset+0], transform[offset+1], transform[offset+2]));
	
	offset = Offset(PARTXFORM_RotationEuler, typeFlags);
	if (offset!=-1)
		xform.Rotate(R3Vector(transform[offset+0], transform[offset+1], transform[offset+2]));
	
	offset = Offset(PARTXFORM_RotationQuaternion, typeFlags);
	assert(offset==-1);
	
	offset = Offset(PARTXFORM_Translation, typeFlags);
	assert(offset==0);
	
	if (offset!=-1)
		xform.Translate(R3Vector(transform[offset+0], transform[offset+1], transform[offset+2]));
	
	return xform;
}

int PartTransforms::Offset(TransformTypes type, int typeFlags, int * nFields, int * nTotalFields)
{
	int nFlds = 0;
	int nTotFlds = 0;
	int offset = -1;
	
	if (m_Types.size()==0)
	{
		m_Types.push_back(PARTXFORM_ScaleIsotropic);
		m_TypeFields.push_back(1);
		m_Types.push_back(PARTXFORM_ScaleAnisotropic);
		m_TypeFields.push_back(3);
		m_Types.push_back(PARTXFORM_Translation);
		m_TypeFields.push_back(3);
		m_Types.push_back(PARTXFORM_RotationEuler);
		m_TypeFields.push_back(3);
		m_Types.push_back(PARTXFORM_RotationQuaternion);
		m_TypeFields.push_back(3);
	}
	
	for (int i=0; i<(int)m_Types.size(); i++)
	{
		if (typeFlags & m_Types[i])
		{
			if (type == m_Types[i])
			{
				offset = nTotFlds;
				nFlds = m_TypeFields[i];
			}
			nTotFlds += m_TypeFields[i];
		}
	}
	
	if (nFields!=NULL)
		*nFields = nFlds;
	if (nTotalFields!=NULL)
		*nTotalFields = nTotFlds;
	
	return offset;
}

void PartTransforms::ComputeAligningTransform(const R3Point & pcenter, const R3Vector & paxesX,
											  const R3Vector & paxesY, const R3Vector & paxesZ, 
											  const R3Point & ccenter, const R3Vector & caxesX,
											  const R3Vector & caxesY, const R3Vector & caxesZ, 
											  int typeFlags, double * transform)
{
	int offset = -1;
	// relative translation (shift)
	offset = Offset(PARTXFORM_Translation, typeFlags);
	if (offset>=0)
	{
		transform[offset+0] = ccenter.X()-pcenter.X();
		transform[offset+1] = ccenter.Y()-pcenter.Y();
		transform[offset+2] = ccenter.Z()-pcenter.Z();
	}
	
	// todo: relative scale (based on average radius of bounding box)
	offset = Offset(PARTXFORM_ScaleIsotropic, typeFlags);
	assert(offset==-1);
	
	// todo: relative scale of bounding boxes (anisotropic)
	offset = Offset(PARTXFORM_ScaleAnisotropic, typeFlags);
	assert(offset==-1);
	
	// todo: rotation using Euler angles...
	offset = Offset(PARTXFORM_RotationEuler, typeFlags);
	assert(offset==-1);
	
	// todo: rotation using quaternions...
	offset = Offset(PARTXFORM_RotationQuaternion, typeFlags);
	assert(offset==-1);
}

void PartTransforms::WriteTransforms(FILE * file) const
{
	fwrite(&(m_iXFormFlags), sizeof(int), 1, file);
	
	int nXForms = (int)m_apXForms.size();	// might not be the same as NParts (due to NullNode)
	fwrite(&nXForms, sizeof(int), 1, file);
	for (int pInst = 0; pInst < nXForms; pInst++)
		m_apXForms[pInst]->WriteToFile(file);
}

PartTransforms * PartTransforms::ReadTransforms(FILE * file, STemplate::STemplateVersions version)
{
	int flags;
	fread(&(flags), sizeof(int), 1, file);
	PartTransforms * xforms = new PartTransforms(flags);
	
	int nXForms;
	fread(&nXForms, sizeof(int), 1, file);
	for (int pInst = 0; pInst < nXForms; pInst++)
		xforms->m_apXForms.push_back(ContProbDistribution::CreateFromFile(file));
	
	return xforms;
}



