/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "VkFunctions.h"
#include "PartsCollection.h"
#include "TMatchEnergy.h"
#include "ShapeTriangleSoup.h"
#include "ShapeManifoldMesh.h"

//////// PartGeometry //////////
PartGeometry::PartGeometry()
{
	m_sPartLabel = "Unknown";
	m_pNormalizedPart = NULL;
	m_pNormalizedPartMesh = NULL;
	m_iNPoints = 0;
}

PartGeometry::PartGeometry(const ShapePointCloud * pnts, const std::vector<int> & subset, const VKString & partLabel,
						   const R3Point & centroid, const R3Triad & orientation, const R3Vector & scale,
						   const std::vector<int> & facesSubsetInput)
{
	std::vector<int> facesSubset = facesSubsetInput;
//	m_pLocalPCA = NULL;
	m_iNPoints = pnts->NumPoints();
	
	m_sPartLabel = partLabel;
	
	R3Point c =  pnts->Centroid();
	double aveRad = pnts->AverageRadius(c);
	
	const ShapeMesh * mesh = pnts->GetMesh();
	m_pNormalizedPartMesh = NULL;
	if (mesh!=NULL)
	{
		if (mesh->SafeCastToTriangleSoup()!=NULL)
			m_pNormalizedPartMesh = new ShapeTriangleSoup();
		else if (mesh->SafeCastToManifold()!=NULL)
			m_pNormalizedPartMesh = new ShapeManifoldMesh();
		else
			assert(false);
	}
	m_pNormalizedPart = new ShapePointCloud(m_pNormalizedPartMesh, partLabel); 
	
	// if mesh exists - copy mesh first (remember triangles), then points (same triangles)
	if (m_pNormalizedPartMesh!=NULL)
	{
		assert(facesSubset.size()>0);
		// find new faces (and map to old faces)
		std::vector<int> orig2newTriangleMap;
		std::vector<int> orig2newVerticesMap;
		for (int i=0; i<mesh->NFaces(); i++)
			orig2newTriangleMap.push_back(-1);
		for (int i=0; i<mesh->NVertices(); i++)
			orig2newVerticesMap.push_back(-1);
		
		// add triangles not included
		for (int i=0; i<(int)subset.size(); i++)
		{
			int fID;
			double b1, b2;
			const ShapeMesh * pntMesh=NULL;
			pnts->GetPoint(subset[i]).GetBarycentricCoords(&fID, &b1, &b2, &pntMesh);
			facesSubset.push_back(fID);
		}
			
		// add new triangles and faces
		for (int i=0; i<(int)facesSubset.size(); i++)
		{
			int nvIDs[3];
			if (orig2newTriangleMap[facesSubset[i]]!=-1)
				continue;	// avoid duplicates
			orig2newTriangleMap[facesSubset[i]] = m_pNormalizedPartMesh->NFaces();
			for (int v=0; v<3; v++)
			{
				int vTmp = mesh->VertexOnFace(facesSubset[i], v);
				if (orig2newVerticesMap[vTmp] == -1)
				{
					R3Point vPos = mesh->VertexPosition(vTmp);
					for (int d=0; d<3; d++)
						vPos[d] = (vPos[d] - c[d]) / aveRad - centroid[d];
					orig2newVerticesMap[vTmp] = m_pNormalizedPartMesh->AddVertex(vPos.X(), vPos.Y(), vPos.Z());
				}
				nvIDs[v] = orig2newVerticesMap[vTmp];
			}
			
			m_pNormalizedPartMesh->AddFace(nvIDs[0], nvIDs[1], nvIDs[2]);
		}
		
		// create points
		for (int i=0; i<(int)subset.size(); i++)
		{
			int fID;
			double b1, b2;
			const ShapeMesh * pntMesh=NULL;
			pnts->GetPoint(subset[i]).GetBarycentricCoords(&fID, &b1, &b2, &pntMesh);
			assert(pntMesh==mesh);
			assert(fID < (int)orig2newTriangleMap.size() && fID>=0);
			int newFID = orig2newTriangleMap[fID];
			assert(newFID>=0);
			m_pNormalizedPart->AddPoint(ShapePoint(m_pNormalizedPartMesh, newFID, b1, b2, m_pNormalizedPart, i));
		}
	}
	else
	{
		// if mesh does not exist - just copy points
		for (int i=0; i<(int)subset.size(); i++)
		{
			R3Point pos = pnts->PointPosition(subset[i]);
			for (int d=0; d<3; d++)
				pos[d] = (pos[d] - c[d]) / aveRad - centroid[d];
			m_pNormalizedPart->AddPoint(ShapePoint(pos));
		}
	}
	
	m_Centroid = centroid;
	m_Orientation = orientation;
	m_Scale = scale;
}

PartGeometry::~PartGeometry()
{
	if (m_pNormalizedPart!=NULL)
		delete m_pNormalizedPart;
}

PartGeometry * PartGeometry::CreateCopy() const
{
	PartGeometry * pg = new PartGeometry();
	pg->m_Centroid = m_Centroid;
	pg->m_Orientation = m_Orientation;
	pg->m_Scale = m_Scale;
	
	pg->m_sPartLabel = m_sPartLabel;
	pg->m_iNPoints = m_iNPoints;
	pg->m_pNormalizedPartMesh = m_pNormalizedPartMesh;
	pg->m_pNormalizedPart = m_pNormalizedPart;
	return pg;	
}

const ShapePointCloud * PartGeometry::GetNormalizedPart() const
{
	return m_pNormalizedPart;
}

const ShapeMesh * PartGeometry::GetNormalizedMesh() const
{
	return m_pNormalizedPartMesh;
}


R3Point PartGeometry::PointPosition(int pntID) const
{
	return m_pNormalizedPart->PointPosition(pntID);
}

double PartGeometry::Area() const
{
	return (double)m_pNormalizedPart->NumPoints() / (double)m_iNPoints;
}

int PartGeometry::NearestPointID(const R3Point & partSpacePos) const
{
	const R3Kdtree<const PointInSet *> * kdTree = m_pNormalizedPart->GetKdTree();
	const PointInSet * closest = kdTree->FindClosest(partSpacePos);
	assert(closest!=NULL);
	return closest->id;

}

void PartGeometry::WritePart(FILE * file) const
{
	// write label
	VKString::WriteString(m_sPartLabel, file);
	
	// write mesh if it exists
	bool meshExists = (m_pNormalizedPartMesh!=NULL);
	fwrite(&meshExists, sizeof(bool), 1, file);
	if (meshExists)	// mesh + point barycentric coords
	{
		GeomShape::GeomShapeTypes type = m_pNormalizedPartMesh->GetShapeType();
		int nVertices = m_pNormalizedPartMesh->NVertices();
		int nFaces = m_pNormalizedPartMesh->NFaces();

		fwrite(&type, sizeof(GeomShape::GeomShapeTypes), 1, file);
		fwrite(&nVertices, sizeof(int), 1, file);
		fwrite(&nFaces, sizeof(int), 1, file);
		for (int i=0; i<nVertices; i++)
		{
			R3Point vp = m_pNormalizedPartMesh->VertexPosition(i);
			double xyz[3];
			xyz[0] = vp.X();
			xyz[1] = vp.Y();
			xyz[2] = vp.Z();
			fwrite(xyz, sizeof(double), 3, file);
		}
		for (int i=0; i<nFaces; i++)
		{
			int vids[3];
			vids[0] = m_pNormalizedPartMesh->VertexOnFace(i, 0);
			vids[1] = m_pNormalizedPartMesh->VertexOnFace(i, 1);
			vids[2] = m_pNormalizedPartMesh->VertexOnFace(i, 2);
			fwrite(vids, sizeof(int), 3, file);
		}
	}

	// write points
	std::vector<ShapePoint> pnts;
	for (int i=0; i<m_pNormalizedPart->NumPoints(); i++)
		pnts.push_back(m_pNormalizedPart->GetPoint(i));
	
	ShapePointCloud::WritePointSet(file, pnts);
	
	double c[3], o[9], s[3];
	for (int i=0; i<3; i++)
	{
		c[i] = m_Centroid[i];
		s[i] = m_Scale[i];
		for (int j=0; j<3; j++)
			o[i*3 + j] = m_Orientation[i][j];
	}
	fwrite(c, sizeof(double), 3, file);
	fwrite(s, sizeof(double), 3, file);
	fwrite(o, sizeof(double), 9, file);
}

PartGeometry * PartGeometry::ReadPart(FILE * file)
{
	PartGeometry * partGeom = new PartGeometry();
	
	// read label
	partGeom->m_sPartLabel = VKString::ReadString(file);
	
	// write mesh if it exists
	bool meshExists = false;
	fread(&meshExists, sizeof(bool), 1, file);
	if (meshExists)	// read mesh
	{
		GeomShape::GeomShapeTypes type;
		int nVertices = 0;
		int nFaces = 0;
		fread(&type, sizeof(GeomShape::GeomShapeTypes), 1, file);
		fread(&nVertices, sizeof(int), 1, file);
		fread(&nFaces, sizeof(int), 1, file);
		
		partGeom->m_pNormalizedPartMesh = NULL;
		if (type==GeomShape::SHAPE_TYPE_MANIFOLD)
			partGeom->m_pNormalizedPartMesh = new ShapeManifoldMesh();
		else if (type==GeomShape::SHAPE_TYPE_TRIANGLE_SOUP)
			partGeom->m_pNormalizedPartMesh = new ShapeTriangleSoup();
		else
			std::cout<<"[ERROR] Invalid geometry type: "<<type<<std::endl;
		assert(partGeom->m_pNormalizedPartMesh!=NULL);
		for (int i=0; i<nVertices; i++)
		{
			double xyz[3];
			fread(xyz, sizeof(double), 3, file);
			partGeom->m_pNormalizedPartMesh->AddVertex(xyz[0], xyz[1], xyz[2]);
		}
		for (int i=0; i<nFaces; i++)
		{
			int vids[3];
			fread(vids, sizeof(int), 3, file);
			partGeom->m_pNormalizedPartMesh->AddFace(vids[0], vids[1], vids[2]);
		}
	}
	
	// read points 
	std::vector<ShapePoint> pnts;
	partGeom->m_pNormalizedPart = new ShapePointCloud(partGeom->m_pNormalizedPartMesh, partGeom->m_sPartLabel);
	ShapePointCloud::ReadPointSet(partGeom->m_pNormalizedPartMesh, partGeom->m_pNormalizedPart, file, pnts);
	for (int i=0; i<(int)pnts.size(); i++)
		partGeom->m_pNormalizedPart->AddPoint(pnts[i]);
	
	double c[3], o[9], s[3];
	fread(c, sizeof(double), 3, file);
	fread(s, sizeof(double), 3, file);
	fread(o, sizeof(double), 9, file);
	
	partGeom->m_Centroid = R3Point(c[0], c[1], c[2]);
	partGeom->m_Scale = R3Vector(s[0], s[1], s[2]);
	partGeom->m_Orientation = R3Triad(R3Vector(o[0], o[1], o[2]), R3Vector(o[3], o[4], o[5]), R3Vector(o[6], o[7], o[8]));

	return partGeom;
}

const R3Point & PartGeometry::Centroid() const
{
	return m_Centroid;
}

const R3Triad & PartGeometry::Orientation() const
{
	return m_Orientation;
}

const R3Vector & PartGeometry::Scale() const
{
	return m_Scale;
}

//R3Triad PartGeometry::GetLocalPCA(const ShapePoint & pnt, double tPntEigs[3] ) const
//{
//	assert(false);
//	int knn = PartLocalGeometry::GetNhdSizeConstant();
//	double rad = PartLocalGeometry::GetNhdDist();	// vk: const - same as in energy
//	 
//	if (m_pLocalPCA==NULL && m_pNormalizedPart->NumPoints()>knn)
//	{
//		PartGeometry * ncThis = (PartGeometry*)this;
//		ncThis->m_pLocalPCA = new ShapeLocalPCA(m_pNormalizedPart, "locPCA", rad, knn, false);	
//		ncThis->m_pLocalPCA->ComputePntsPCA();
//	}
//	
//	if (m_pLocalPCA!=NULL)
//		return m_pLocalPCA->GetPCA(pnt, tPntEigs);
//	else
//	{
//		tPntEigs[0] = -1;
//		tPntEigs[1] = -1;
//		tPntEigs[2] = -1;
//		return m_Orientation;
//	}
//}


//////// PartsCollection //////////
PartsCollection::PartsCollection()
{
}

PartsCollection::~PartsCollection()
{
	for (int i=0; i<(int)m_AllParts.size(); i++)
		delete m_AllParts[i];
}

int PartsCollection::NParts() const
{
	return (int)m_AllParts.size();
}

int PartsCollection::AddPart(PartGeometry * part) 
{
	m_AllParts.push_back(part);
	return (int)m_AllParts.size()-1;
}

const PartGeometry * PartsCollection::GetPart(int partID) const
{
	assert(partID >= 0 && partID < (int)m_AllParts.size());
	return m_AllParts[partID];
}

void PartsCollection::WriteParts(FILE * file) const
{
	int nParts = (int)m_AllParts.size();
	fwrite(&nParts, sizeof(int), 1, file);
	for (int i=0; i<nParts; i++)
		m_AllParts[i]->WritePart(file);
}

PartsCollection * PartsCollection::ReadParts(FILE * file)
{
	PartsCollection * parts = new PartsCollection();
	int nParts;
	fread(&nParts, sizeof(int), 1, file);
	for (int i=0; i<nParts; i++)
		parts->m_AllParts.push_back(PartGeometry::ReadPart(file));
	return parts;
}


