/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "ProbDistribution.h"
#include "VkFunctions.h"
#include "LinAlg.h"

//////// ProbDistribution /////////
ContProbDistribution::ContProbDistribution(int nDims, int nSamples)
{
	m_iNDims = nDims;
	m_iNSamples = nSamples;
	m_aTempRVHolder = new double[m_iNDims];
}

ContProbDistribution::~ContProbDistribution()
{
	delete [] m_aTempRVHolder;
}

ContProbDistribution * ContProbDistribution::CreateFromFile(FILE * file)
{
	VKString className = VKString::ReadString(file);
	int nDims = 0;
	fread(&nDims, sizeof(int), 1, file);
	
	ContProbDistribution * newProb = NULL;
	if (className=="ProbGaussian")
		newProb = new ProbGaussian(nDims);
	assert(newProb!=NULL);
	newProb->ReadFromFile(file);
	return newProb;
}

ProbGaussian * ContProbDistribution::SafeCastToGaussian()
{
	return NULL;
}

const ProbGaussian * ContProbDistribution::SafeCastToGaussian() const
{
	return ((ContProbDistribution*)this)->SafeCastToGaussian();
}

int ContProbDistribution::NDims() const
{
	return m_iNDims;
}

int ContProbDistribution::NSamples() const
{
	return m_iNSamples;
}


//////// DistrGaussian /////////
ProbGaussian::ProbGaussian(int nDims, const double * means, const double * variances, int nSamples)
: ContProbDistribution(nDims, nSamples)
{
	m_aMeans = new double[m_iNDims];
	m_aVariances = new double[m_iNDims];
	for (int i=0; i<m_iNDims; i++)
	{
		if (means!=NULL)
			m_aMeans[i] = means[i];
		else
			m_aMeans[i] = 0;
		
		if (variances!=NULL)
			m_aVariances[i] = variances[i];
		else
			m_aVariances[i] = 0;
	}
}

ProbGaussian::~ProbGaussian()
{
	delete [] m_aMeans;
	delete [] m_aVariances;
}

ContProbDistribution * ProbGaussian::CreateCopy() const
{
	return new ProbGaussian(m_iNDims, m_aMeans, m_aVariances, m_iNSamples);
}

void ProbGaussian::WriteToFile(FILE * file) const
{
	// write meta data: class name, number of states
	VKString::WriteString("ProbGaussian", file);
	fwrite(&m_iNDims, sizeof(int), 1, file);
	
	// write probabilites
	fwrite(m_aMeans, sizeof(double), m_iNDims, file);
	fwrite(m_aVariances, sizeof(double), m_iNDims, file);
	fwrite(&m_iNSamples, sizeof(int), 1, file);

	for (int i=0; i<m_iNDims; i++)
		assert(m_aVariances[i]>=0);

}

void ProbGaussian::ReadFromFile(FILE * file)
{
	// read probabilites
	fread(m_aMeans, sizeof(double), m_iNDims, file);
	fread(m_aVariances, sizeof(double), m_iNDims, file);
	fread(&m_iNSamples, sizeof(int), 1, file);
	for (int i=0; i<m_iNDims; i++)
	{
		if (m_aVariances[i]<0)
		{
			std::cout<<"[WARNING] Negative variance in Gaussian: "<<m_aVariances[i]<<std::endl;
			m_aVariances[i] = 0;
		}
		assert(m_aVariances[i]>=0);
	}
}

ProbGaussian * ProbGaussian::SafeCastToGaussian()
{
	return (ProbGaussian*)this;
}

const double * ProbGaussian::GetMostProbable() const 
{
	return m_aMeans;
}

const double * ProbGaussian::GetVariance() const
{
	for (int i=0; i<m_iNDims; i++)
		if (m_aVariances[i]<10e-10 && m_aVariances[i]>0)
			m_aVariances[i] = 0;
	return m_aVariances;
}

const double* ProbGaussian::GenRandVar() const
{
	for (int i=0; i<m_iNDims; i++)
	{
		if (m_aVariances[i]==0)
			m_aTempRVHolder[i] = m_aMeans[i];
		else
			assert(false);
	}
	return m_aTempRVHolder;
}

double ProbGaussian::LogProb(const double * randVar) const
{
	return LogProb(m_aVariances, m_aMeans, m_iNDims, m_iNSamples, randVar);
}

double ProbGaussian::LogProb(const double * vars, const double * mean, int ndims, int nsamp, const double * rv)
{
	double logProb = 0;
	for (int i=0; i<ndims; i++)
	{
		double var = vars[i];
//		if (nsamp>1)
//			var = (var*nsamp) / (nsamp-1);	// as in MATLAB: unbiased normalizer?
		
		if (var==0 && rv[i]!=mean[i])
			return -NUM_DBL_MAX;
		
		if (rv[i] != mean[i])
			logProb += pow(rv[i] - mean[i], 2.) / (2 * var);
	}
	
	return -logProb;
}

void ProbGaussian::UpdateDistribution(const double * sample, bool include)
{
	if (include)
	{
		for (int i=0; i<m_iNDims; i++)
		{
			int N = m_iNSamples;
			double oldMean = m_aMeans[i];
			double newMean = (oldMean * N + sample[i]) / (N+1);
			double oldVariance = m_aVariances[i];
			
			m_aVariances[i] = (N*oldVariance + N*pow(oldMean, 2.) - 2 * N * oldMean * newMean + N * pow(newMean, 2.));
			m_aVariances[i] = (m_aVariances[i] + pow(sample[i] - newMean, 2.)) / (N+1);
			m_aMeans[i] = newMean;
			
			if (m_aVariances[i]<0)
				m_aVariances[i] = 0;
			
//			if (m_aVariances[i]<0)
//			{
//				std::cout<<"[ERROR] Invalid variance: "<<m_aVariances[i]<<". OldVar: "<<oldVariance<<std::endl;
//				std::cout<<" OldMean: "<<oldMean<<" NewMean: "<<newMean<<" N: "<<N<<std::endl;
//				std::cout<<"  Old-New:"<<(oldMean-newMean)<<" sample="<<sample[i]<<std::endl;
//				std::cout<<"  EquatioN:"<<(N*oldVariance)<<"+"<<(N*pow(oldMean, 2.))<<"-"<<(2 * N * oldMean*newMean)<<"+"<<(N * pow(newMean, 2.))<<std::endl;
//				m_aVariances[i] = (N*oldVariance + N*pow(oldMean, 2.) - 2 * N * oldMean * newMean + N * pow(newMean, 2.));
//				std::cout<<" -- var1="<<m_aVariances[i]<<std::endl;
//				m_aVariances[i] = (m_aVariances[i] + pow(sample[i] - newMean, 2.)) / (N+1);
//				std::cout<<"Eqn=["<<pow(sample[i] - newMean, 2.)<<std::endl;
//				std::cout<<" -- var2="<<m_aVariances[i]<<std::endl;
//			}
			assert(m_aVariances[i]>=0);
		}
				
		m_iNSamples++;
	}
	else
	{
		assert(m_iNSamples>0);
		for (int i=0; i<m_iNDims; i++)
		{
			int N = m_iNSamples;
			double oldMean = m_aMeans[i];
			double newMean = (oldMean * N - sample[i]) / (N-1);
			double oldVariance = m_aVariances[i];
			
			m_aVariances[i] = (N*oldVariance + N*pow(oldMean, 2.) - 2 * N * oldMean * newMean + N * pow(newMean, 2.));
			m_aVariances[i] = (m_aVariances[i] - pow(sample[i] - newMean, 2.)) / (N-1);
			m_aMeans[i] = newMean;
			if (m_aVariances[i]<0)
				m_aVariances[i] = 0;
			assert(m_aVariances[i]>=0);
		}
		
		m_iNSamples--;
	}
}

void ProbGaussian::Reset()
{
	m_iNSamples = 0;
}

void ProbGaussian::Rescale(int dim, double shift, double scale)
{
	assert(dim>=0 && dim < m_iNDims);
	m_aMeans[dim] = (m_aMeans[dim] + shift) * scale;
	m_aVariances[dim] *= pow(scale, .2);
}

VKString ProbGaussian::Print() const
{
	VKString str = "   Gaussian:\n       Mean=[ ";
	str += "";
	for (int i=0; i<NDims(); i++)
		str += VKString::number(m_aMeans[i])+" ";
	str += "]\n       Variance=[ ";
	for (int i=0; i<NDims(); i++)
		str += VKString::number(m_aVariances[i])+" ";
	str += VKString("]\n       NSamples=")+VKString::number(m_iNSamples)+"\n";
	return str;
}

//////// DiscreteProbDistribution /////////
DiscreteProbDistribution::DiscreteProbDistribution(int nStates)
{
	m_iNStates = nStates;
	m_iNSamples = 0;
}

int DiscreteProbDistribution::MaxNonzeroState() const
{
	return m_iNStates;
}

DiscreteProbDistribution * DiscreteProbDistribution::CreateFromFile(FILE * file)
{
	VKString className = VKString::ReadString(file);
	int nstates = 0;
	fread(&nstates, sizeof(int), 1, file);
	
	DiscreteProbDistribution * newProb = NULL;
	if (className=="ProbDiscreteStats")
		newProb = new ProbDiscreteStats(nstates);
	else
		std::cout<<"[ERROR] Unknown stats type="<<className.c_str()<<std::endl;
	
	assert(newProb!=NULL);
	newProb->ReadFromFile(file);
	return newProb;
}

int DiscreteProbDistribution::NSamples() const
{
	return m_iNSamples;
}


//////// ProbDiscreteStats /////////
ProbDiscreteStats::ProbDiscreteStats(int nStates)
: DiscreteProbDistribution(nStates)
{
}

ProbDiscreteStats::~ProbDiscreteStats()
{
}

DiscreteProbDistribution * ProbDiscreteStats::CreateCopy() const
{
	ProbDiscreteStats * copy = new ProbDiscreteStats(0);
	copy->m_aStateProb = m_aStateProb;
	copy->m_iNSamples = m_iNSamples;
	copy->m_iNStates = m_iNStates;
	return copy;
}

void ProbDiscreteStats::WriteToFile(FILE * file) const
{
	// write meta data: class name, number of states
	VKString::WriteString("ProbDiscreteStats", file);
	fwrite(&m_iNStates, sizeof(int), 1, file);
	
	// write probabilites
	if ((int)m_aStateProb.size()!=m_iNStates)
		std::cout<<"[ERROR] Invalid number of states: "<<m_aStateProb.size()<<" != "<<m_iNStates<<std::endl;
	assert((int)m_aStateProb.size()==m_iNStates);
	for (int i=0; i<m_iNStates; i++)
		fwrite(&(m_aStateProb[i]), sizeof(double), 1, file);
	fwrite(&m_iNSamples, sizeof(int), 1, file);
}

void ProbDiscreteStats::ReadFromFile(FILE * file)
{
	// read probabilities
	assert(m_aStateProb.size()==0);
	for (int i=0; i<m_iNStates; i++)
		m_aStateProb.push_back(0);
	for (int i=0; i<m_iNStates; i++)
		fread(&(m_aStateProb[i]), sizeof(double), 1, file);
	fread(&m_iNSamples, sizeof(int), 1, file);
}

int ProbDiscreteStats::GetMostProbable() const
{
	int state = -1;
	int maxProb = 0;
	for (int i=0; i<(int)m_aStateProb.size(); i++)
	{
		if (m_aStateProb[i] > maxProb)
		{
			state = i;
			maxProb = m_aStateProb[state];
		}
	}
	return state;
}

int ProbDiscreteStats::GenRandVar() const
{
	assert(false);
}

double ProbDiscreteStats::Prob(int state) const
{
	if (state>=0 && state < (int)m_aStateProb.size() && m_iNSamples>0)
		return m_aStateProb[state] / m_iNSamples;
	
	return 0;
}

void ProbDiscreteStats::UpdateDistribution(int state, bool include)
{
	while (state >= (int)m_aStateProb.size())
		m_aStateProb.push_back(0);
		
	if (include)
	{
		m_iNSamples++;
		m_aStateProb[state]+=1.;
	}
	else
	{
		if (m_iNSamples>0 && m_aStateProb[state]>0)
		{
			m_iNSamples--;
			m_aStateProb[state]-=1.;
		}
		else
		{
			std::cout<<"[WARNING] Cannot exclude. NSamp: "<<m_iNSamples<<" state: "<<m_aStateProb[state]<<std::endl;
		}
		
	}
	m_iNStates = (int)m_aStateProb.size();
}

void ProbDiscreteStats::Reset()
{
	for (int i=0; i<(int)m_aStateProb.size(); i++)
		m_aStateProb[i] = 0;
	m_iNSamples = 0;
}

