/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __ProbDistribution_H
#define __ProbDistribution_H

#include <assert.h>
#include <math.h>
#include <vector>
#include <iostream>
#include "VKString.h"

class ProbGaussian;

/** 
 * Helper class (quite general math)
 *
 * Probabilitiy distribution interface class
 */
class ContProbDistribution
{
	public:
		ContProbDistribution(int nDims, int nSamples);
		virtual ~ContProbDistribution();
		virtual ProbGaussian * SafeCastToGaussian();
		virtual const ProbGaussian * SafeCastToGaussian() const;
	
		virtual ContProbDistribution * CreateCopy() const = 0;
		virtual void WriteToFile(FILE * file) const = 0;
		virtual void ReadFromFile(FILE * file) = 0;
		static ContProbDistribution * CreateFromFile(FILE * file);

		virtual int NDims() const;
		virtual int NSamples() const;
		virtual const double * GetMostProbable() const = 0;
		virtual const double* GenRandVar() const = 0;
		virtual double LogProb(const double * randVar) const = 0;
	
		virtual void UpdateDistribution(const double * sample, bool include) = 0;
		virtual void Reset() = 0;
		virtual void Rescale(int dim, double shift, double scale) = 0;
	
		virtual VKString Print() const = 0;
	
			
	protected:
		int m_iNDims;
		double * m_aTempRVHolder;
		int m_iNSamples;
};

/** 
 * Gaussian distribution for probabilities
 */
class ProbGaussian : public ContProbDistribution
{
	public:
		ProbGaussian(int nDims, const double * means = NULL, const double * variances = NULL, int nSamples = 0);
		virtual ~ProbGaussian();
		virtual ProbGaussian * SafeCastToGaussian();
	
		virtual ContProbDistribution * CreateCopy() const;
		virtual void WriteToFile(FILE * file) const;
		virtual void ReadFromFile(FILE * file);

		virtual const double * GetMostProbable() const;
		virtual const double * GetVariance() const;
		virtual const double* GenRandVar() const;
		virtual double LogProb(const double * randVar) const;

		static double LogProb(const double * var, const double * mean, int ndims, int nsamp, const double * rv);
	
		virtual void UpdateDistribution(const double * sample, bool include);
		virtual void Reset();
		virtual void Rescale(int dim, double shift, double scale);
	
		virtual VKString Print() const;
	
	
	//protected:
		double * m_aMeans;
		double * m_aVariances;
		
};

///**
// * Uniform distribution for probabilities
// */
//class ProbUniform : public ContProbDistribution
//{
//	public:
//		ProbUniform(int nDims, int nSamples = 0);
//		virtual ~ProbUniform();
//		virtual ProbGaussian * SafeCastToGaussian();
//		
//		virtual void WriteToFile(FILE * file) const;
//		virtual void ReadFromFile(FILE * file);
//		
//		virtual const double * GetMostProbable() const;
//		virtual const double * GetVariance() const;
//		virtual int NSamples() const;
//		virtual const double* GenRandVar() const;
//		virtual double LogProb(const double * randVar, double optTol = 0) const;
//		virtual double LogProb(const double * randVar, double * optTol) const;
//		virtual void UpdateDistribution(const double * sample, bool include);
//		virtual void Reset();
//		virtual void Rescale(int dim, double shift, double scale);
//		
//		virtual VKString Print() const;
//		
//		
//	protected:
//		double * m_aMeans;
//		double * m_aVariances;
//		int m_iNSamples;
//};

/**
 * Statistics for discrete probabilities
 */
class DiscreteProbDistribution
{
	public:
		DiscreteProbDistribution(int nStates = 0);
		virtual DiscreteProbDistribution * CreateCopy() const = 0;
	
		virtual void WriteToFile(FILE * file) const = 0;
		virtual void ReadFromFile(FILE * file) = 0;
		static DiscreteProbDistribution * CreateFromFile(FILE * file);
		virtual int NSamples() const;
		virtual int MaxNonzeroState() const;
		virtual int GetMostProbable() const = 0;
		virtual int GenRandVar() const = 0;
		virtual double Prob(int state) const = 0;
		virtual void UpdateDistribution(int sampleState, bool include) = 0;
		virtual void Reset() = 0;
	
	protected:
		int m_iNStates;
		int m_iNSamples;
};

/**
 * Probability discrete stats
 */
class ProbDiscreteStats : public DiscreteProbDistribution
{
	public:
		ProbDiscreteStats(int nStates = 0);
		virtual ~ProbDiscreteStats();
		virtual DiscreteProbDistribution * CreateCopy() const;
	
		virtual void WriteToFile(FILE * file) const;
		virtual void ReadFromFile(FILE * file);
	
		virtual int GetMostProbable() const;
		virtual int GenRandVar() const;
		virtual double Prob(int state) const;
		virtual void UpdateDistribution(int sampleState, bool include);
		virtual void Reset();
	
	protected:
		std::vector<double> m_aStateProb;
		
};


#endif

