/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "FileManagement.h"

#include "STemplate.h"
#include "TMatch.h"
#include "ProbDistribution.h"
#include "PartOfEdge.h"
#include "STemplateNode.h"

#include "TMatchAlg.h"
#include "TMatchEnergy.h"



STemplate * STemplate::CreateTemplate(ShapeLabels * labels, bool trustLabels, bool deformTemplate, double fUseMeshThresh,
									  ShapeLabels * meshLabels)
{
	std::cout<<"----- Creating Template ("<<fUseMeshThresh<<") ------"<<std::endl;
	GeomShape * shape = labels->GetShape();
	assert(shape!=NULL);
	ShapePointCloud * pnts = shape->SafeCastToPointCloud();
	assert(pnts!=NULL);
	//assert(mesh!=NULL || pnts!=NULL);
	
	STemplate * st = new STemplate(labels);
	st->UseMeshInParts(fUseMeshThresh);
	st->m_CreatedFromModels.push_back(labels->GetShape()->GetModel()->ModelName());
	
	// create all nodes
	std::map<VKString, int> labelToTemplateNode;
	std::vector<int> templateNodeCounter;
	std::vector<int> sceneNodeToTemplateNode;
	std::vector<int> templateNodeParentIDs;

	// count parts
	std::map<VKString, int> label2nparts;
	for (int sID=0; sID<labels->NNodes(); sID++)
	{
		VKString label = labels->NodeLabel(sID);
		std::map<VKString, int>::iterator iter = label2nparts.find(label);
		if (iter==label2nparts.end())
			label2nparts[label] = 1;
		else if (trustLabels)
			iter->second++;
	}
	
	// nodes
	for (int sID=0; sID<labels->NNodes(); sID++)
	{
		// get (or create) node for a segment
		VKString label = labels->NodeLabel(sID);
		std::map<VKString, int>::iterator iter = labelToTemplateNode.find(label);
		STemplateNode * tNode = NULL;
		assert(label2nparts.find(label)!=label2nparts.end());
		int nParts = label2nparts[label];
		if (iter == labelToTemplateNode.end() || !trustLabels)
		{
			// create node
			tNode = new STemplateNode(label, (int)st->m_AllNodes.size(), st);
			labelToTemplateNode[label] = (int)st->m_AllNodes.size();
			st->m_AllNodes.push_back(tNode);
			templateNodeCounter.push_back(0);
			
			// add parent
			int sParentID = labels->NodeParent(sID);
			if (sParentID>=0)
			{
				// parent should have been added (also necessary for transformations)
				assert(sParentID < (int)sceneNodeToTemplateNode.size());
				int tParentID = sceneNodeToTemplateNode[sParentID];
				assert(tParentID>=0 && tParentID < (int)st->m_AllNodes.size()-1);
				st->m_AllNodes[tParentID]->AddPartOfChild(tNode, st, 0, false);
				templateNodeParentIDs.push_back(tParentID);
			}
			else
				templateNodeParentIDs.push_back(-1);
		}
		else
		{
			// verify that nodes have the same parent
			int sParentID = labels->NodeParent(sID);
			assert(sParentID < (int)sceneNodeToTemplateNode.size());
			int tParentID = sParentID>=0 ? sceneNodeToTemplateNode[sParentID] : -1;
			tNode = st->m_AllNodes[iter->second];
			assert(tParentID == templateNodeParentIDs[tNode->NodeID()]);
		}
		assert(tNode!=NULL);
		sceneNodeToTemplateNode.push_back(tNode->NodeID());

		// keep track of part count (if multiple parts for a label)
		assert(tNode->NodeID()>=0 && tNode->NodeID() < (int)templateNodeCounter.size());
		int partID = templateNodeCounter[tNode->NodeID()];
		templateNodeCounter[tNode->NodeID()]++;
		
		// add geometry to a node
		PartBox ellipsoid(st->m_pAllParts);
		bool addNode = false;
		if (pnts!=NULL)
		{
			std::vector<int> pIDs;
			labels->NodeAllPoints(sID, pIDs);
			std::vector<int> faceSubset;
			if (meshLabels!=NULL)
				meshLabels->NodeAllFaces(sID, faceSubset);
			
			STemplateNode * parent = (sID==0 ? NULL : st->m_AllNodes[0]);
			if (partID==0 && tNode->PartOfParent()!=NULL)
			{
				tNode->PartOfParent()->UpdateNParts(0, nParts, true);
				tNode->PartOfParent()->UpdateNParts(0, 0, true);	// set to 50%
			}
			tNode->InitShapeAndXForm(0, parent, nParts, partID, pnts, pIDs, fUseMeshThresh, faceSubset);
			VKString nodeName = tNode->NodeName();
			while (nodeName.length() < 15)
				nodeName += " ";
			VKString fitErr = VKString::number(tNode->GetPart(0)->GetFittingErr());
			while (fitErr.length() < 15)
				fitErr += " ";
		}
		else
			assert(false);		 
	}
	
	// add null node
	STemplateNode * nullNode = new STemplateNode("NULLNode", (int)st->m_AllNodes.size(), st);
	nullNode->InitNullGeometry();
	st->m_AllNodes.push_back(nullNode);
	st->m_AllNodes[0]->AddPartOfChild(nullNode, st, 0, true);
	//nullNode->SetNPartsAlwaysOne();
	//st->m_pMesh2BoxAreaRatio = pts2area;
	
	st->m_iNTopologyChanges = 0;
//	
//	// deform template to get best matching to labels
//	if (deformTemplate)	// NOTE: not working at the moment, optimized in part initialization
//	{
//		assert(false);
//		assert(pnts!=NULL);
//		TMatchAlgParams matchParams;
//		matchParams.m_InitAlignments = TMatchAlgParams::ALIGN_IdentityOnly;
//		matchParams.m_iOptimFlags = (TMatchAlgParams::OPT_GlobalScale | TMatchAlgParams::OPT_GlobalRotate 
//									 | TMatchAlgParams::OPT_GlobalTrans 
//									 | TMatchAlgParams::OPT_NodeXForms | TMatchAlgParams::OPT_NodeShapes);
//		TMatch * match = st->CreateSeedConfig(0, pnts, std::vector<int>(), R3identity_affine);
//		TMatchEnergy * energy = new TMatchEnergy(match);
//		
//		assert(!trustLabels);
//		assert(match->NShapePrimitives() == pnts->NumPoints());
//		for (int i=0; i<pnts->NumPoints(); i++)
//		{
//			int segID = (int)labels->Value(pnts->GetPoint(i), 0);
//			int nodeID = segID - 1;
//			match->SetLabel(i, nodeID);
//		}
//		match->UpdateLabels(false);
//		
//		int mStepIterations;
//		int pntsEnergyType = 0;
//		pntsEnergyType |= (energy->GetTerms() & TMatchEnergy::ET_PntsDistance);
//		pntsEnergyType |= (energy->GetTerms() & TMatchEnergy::ET_PntsOrientation);
//		energy->SetTerms(TMatchEnergy::ET_Pnts | TMatchEnergy::ET_NodeFit2Pnts | pntsEnergyType);
//		TMatchAlg::MStep_DeformTemplate(energy, matchParams, mStepIterations);
//		
//		st->SetDeformAsRestPose(match);
//		st->m_iNDescribedInstances = 1;
//	}
	std::cout<<"-----------------------------"<<std::endl;
	return st;
}

STemplate::STemplate(ShapeLabels * labels)
{
	m_pRootMatch = NULL;
	m_pGeomLabels = labels;
	//m_pMesh2BoxAreaRatio = NULL;
	m_iNTopologyChanges = -1;
	m_pAllParts = NULL;
	m_fUseMeshThresh = -1;	// never use mesh just boxes
	m_iInitCoreSeeds = 0;
}

STemplate::~STemplate()
{
	for (int i=0; i<(int)m_AllNodes.size(); i++)
		delete m_AllNodes[i];
}

void STemplate::MergeTemplate(const STemplate * st)
{
	// if no seeds - create from 0
	if (st->m_SeedDeformations.size()==0)
	{
		int seedID = st->NSeedConfigs();
		std::vector<int> emptySet2;
		TMatch * newSeed = st->CreateSeedConfig(0, NULL, emptySet2, R3identity_affine);
		newSeed->SetSeedID(seedID);
		
		VKString newSeedName = VKString("NewSeed_")+VKString::number(seedID);
		if (st->m_CreatedFromModels.count()>0)
			newSeedName = VKString("Seed_")+st->m_CreatedFromModels[0];
		((STemplate*)st)->AddOrReplaceSeed(seedID, newSeedName, newSeed, 0);
	}
	
	// Copy part geometries
	if (m_pAllParts==NULL)
		m_pAllParts = new PartsCollection();
	assert(m_pAllParts!=NULL && st->m_pAllParts!=NULL);		// assume there are parts
	int partsoffset = m_pAllParts->NParts();	// all ellipsoids need to be updated
	for (int i=0; i<st->m_pAllParts->NParts(); i++)
		m_pAllParts->AddPart(st->m_pAllParts->GetPart(i)->CreateCopy());
 
	// Copy part nodes (update geometry IDs)
	int nodes1 = (int)m_AllNodes.size();
	int configoffset  = (int)m_SeedDeformations.size();
	std::map<int, int> old2newnode;
	std::map<int, int> new2oldnode;
	old2newnode[0] = 0;
	int nullID = -1;
	for (int i=0; i<m_AllNodes.size() && nullID==-1; i++)
		if (m_AllNodes[i]->IsNull())
			nullID = i;
	assert(nullID>=0);
	
	for (int i=1; i<(int)st->m_AllNodes.size(); i++)
	{
		if (!st->m_AllNodes[i]->IsNull())	// no need for two null nodes
		{
			int newNodeID = (int)m_AllNodes.size();
			old2newnode[i] = newNodeID;
			new2oldnode[newNodeID] = i;
			m_AllNodes.push_back(st->m_AllNodes[i]->CreateCopyNoEdges(this, partsoffset, configoffset, newNodeID));
		}
		else
			old2newnode[i] = nullID;
	}
	
	// Copy part of edges
	for (int i=nodes1; i<(int)m_AllNodes.size(); i++)
		m_AllNodes[i]->CopyEdges(partsoffset, configoffset, old2newnode, new2oldnode, (STemplate*)st);
	
	// Copy meta data
	for (int i=0; i<st->m_CreatedFromModels.count(); i++)
		if (!m_CreatedFromModels.contains(st->m_CreatedFromModels[i]))
			m_CreatedFromModels.push_back(st->m_CreatedFromModels[i]);

	// Copy seeds and classifiers (NOTE: m_iCoreSeeds - is wrong, but hard to fix, also irrelevant)
	for (int i=1; i<(int)st->m_apLocalClassifiers.size(); i++)
		m_apLocalClassifiers.push_back(st->m_apLocalClassifiers[i]->CreateCopy());
	
	// add new seeds
	assert(st->m_sSeedNames.count()==(int)st->m_SeedDeformations.size());
	std::vector<TemplateNodeDeform *> emptySet;
	for (int i=0; i<(int)st->m_SeedDeformations.size(); i++)
	{
		m_sSeedNames.push_back(st->m_sSeedNames[i]); 
		TMatch * newSeed = st->m_SeedDeformations[i]->CreateCopy(emptySet, this, &old2newnode, configoffset);
		m_SeedDeformations.push_back(newSeed);
	}

	// configuration hierarchy
	for (int i=1; i<(int)st->m_ConfigParents.size(); i++)
	{
		if (st->m_ConfigParents[i]==0)
			m_ConfigParents.push_back(0);
		else
			m_ConfigParents.push_back(configoffset + st->m_ConfigParents[i]);
	}
}

void STemplate::Normalize(const R3Point & centroid, double aveRadius )
{
	// shift by centroid and scale by ave radius
	if (R3Distance(centroid, R3Point(0,0,0))>10E-10 || aveRadius!=1)
	{
		for (int i=0; i<(int)m_AllNodes.size(); i++)
			m_AllNodes[i]->Normalize(centroid, aveRadius);
	}

	for (int i=0; i<(int)m_SeedDeformations.size(); i++)
	{
		assert(m_SeedDeformations[i]!=NULL);
		for (int j=0; j<m_SeedDeformations[i]->NTemplateNodes(); j++)
		{
			TemplateNodeDeform & deform = (TemplateNodeDeform &)m_SeedDeformations[i]->Deformation(j);
			//const double * s = GetNode(deform.NodeID())->ScaleDistribution()->GetMostProbable();
			int flags, ndims;
			int seedID = m_SeedDeformations[i]->SeedID();
			const PartOfEdge * edge = GetNode(deform.NodeID())->PartOfParent();
			const double * xform = edge->GetMeanTransform(seedID, deform.NParts(), deform.PartID(), flags, ndims);
			
			const R3Vector & ms=GetNode(deform.NodeID())->GetMeanScale(seedID);
			
			deform.SetScale(GetNode(deform.NodeID())->GetMeanScale(seedID));
			deform.SetTransform(xform);
		}
		m_SeedDeformations[i]->SetDirtyFlags();
		m_SeedDeformations[i]->UpdateEllipsoids();
	}
}

/*
void STemplate::Normalize()
{
	R3Point centroid(0,0,0);
	TMatch * everything = CreateSeedConfig(0, NULL, std::vector<int>(), R3identity_affine);
	everything->CreateTemplateSamples(10000);
	std::vector<R3Point> pnts;
	for (int i=0; i<everything->NTemplateNodes(); i++)
	{
//		if (everything->Deformation(i).IsIncluded())
		{
			const PartBox & box = everything->NodeEllipsoid(i);
			R3Point pos;
			double eigs[3];
			
			for (int j=0; j<box.NTemplatePointsOnNode(); j++)
			{
				box.TemplatePoint(j, pos, eigs);
				centroid += pos;
				pnts.push_back(pos);
			}
		}
	}
	
	centroid /= (int)pnts.size();
	
	//double aveRadius = 1;
	
	double aveRadius = 0;
	for (int i=0; i<(int)pnts.size(); i++)
		aveRadius += R3Distance(centroid, pnts[i]);
	aveRadius = aveRadius/(int)pnts.size();
	
	centroid = -centroid;
	
	delete everything;
	
	std::cout<<"Normalizing template:"<<std::endl;;
	std::cout<<"      centroid=["<<centroid.X()<<" "<<centroid.Y()<<" "<<centroid.Z()<<"] "<<" radius="<<aveRadius<<std::endl;
	Normalize(centroid, aveRadius);
}*/

void STemplate::UseMeshInParts(double thresh)
{
	m_fUseMeshThresh = thresh;
	if (m_pAllParts==NULL && m_fUseMeshThresh >= 0)
		m_pAllParts = new PartsCollection();
	
}

const ShapeLabels * STemplate::TrainingLabels() const
{
	return m_pGeomLabels;
}

int STemplate::NNodes() const
{
	return (int)m_AllNodes.size();
}

const STemplateNode * STemplate::GetNode(int nodeID) const
{
	if (! (nodeID>=0 && nodeID < (int)m_AllNodes.size()))
	{
		std::cout<<"[ERROR] Cannot get node: "<<nodeID<<" / "<<m_AllNodes.size()<<std::endl;
		assert(false);
	}
	return m_AllNodes[nodeID];
}

void STemplate::InitAndAddNewNode(int parentID, const VKString & label, const PartBox & geometry, int configID,
								  std::vector<TemplateNodeDeform*> * newDeforms)
{
	STemplateNode * node = new STemplateNode(label, (int)m_AllNodes.size(), this);
	m_AllNodes.push_back(node);

	assert(parentID < (int)m_AllNodes.size());
	if (parentID>=0)
		m_AllNodes[parentID]->AddPartOfChild(node, this, configID, geometry.IsType(PartBox::BOXTYPE_NullBox));
	
	if (node->PartOfParent()!=NULL)
		node->PartOfParent()->UpdateNParts(configID, 1, true);

	node->InitShapeAndXForm(configID, parentID>=0 ? GetNode(parentID) : NULL, 1, 0, geometry);
	
	if (newDeforms!=NULL)
	{ 
		assert(node->PartOfParent()!=NULL);
		int partID = 0;
		int nparts = 1;
		int flags, ndims;
		//const PartTransforms * pxform = node->PartOfParent()->GetPartTransforms(nparts);
		//const ContProbDistribution * xform = pxform->TransformsForPart(partID, xformflags);
		const double * xform = node->PartOfParent()->GetMeanTransform(configID, nparts, partID, flags, ndims);
		TemplateNodeDeform * nodeDeform = NULL;
		nodeDeform = new TemplateNodeDeform(node->NodeID(), nparts, partID,
											xform, ndims, flags, node->GetMeanScale(configID));
		newDeforms->push_back(nodeDeform);
	}
}

void STemplate::RemoveNode(int nodeID)
{
	assert(nodeID>=0 && nodeID<(int)m_AllNodes.size());
	for (int i=0; i<(int)m_SeedDeformations.size(); i++)
	for (int n=0; n<m_SeedDeformations[i]->NTemplateNodes(); n++)
	{
		if (m_SeedDeformations[i]->Deformation(n).NodeID()==nodeID)
		{
			std::cout<<"[WARNING] Cannot remove node: "<<GetNode(nodeID)->NodeName().c_str();
			std::cout<<" since it's still used in seed: "<<(i+1)<<std::endl;
			return;
		}
	}

	PartOfEdge * edge = m_AllNodes[nodeID]->PartOfParent();
	if (edge!=NULL)
	{
		edge->ParentTemplate()->RemoveChild(m_AllNodes[nodeID]);
		delete edge;
	}

	delete m_AllNodes[nodeID];
	m_AllNodes.erase(m_AllNodes.begin()+nodeID);
	for (int n=nodeID; n < (int)m_AllNodes.size(); n++)
		m_AllNodes[n]->SetNodeID(n);
	
	for (int i=0; i<(int)m_SeedDeformations.size(); i++)
	for (int n=0; n<m_SeedDeformations[i]->NTemplateNodes(); n++)
	{
		TemplateNodeDeform & deform = (TemplateNodeDeform&)m_SeedDeformations[i]->Deformation(n);
		if (deform.NodeID()>nodeID)
			deform.SetNodeID(deform.NodeID()-1);
	} 
}

//const ContProbDistribution * STemplate::GetPointsToAreaRatio() const
//{
//	return m_pMesh2BoxAreaRatio;
//}

// template configuration
void STemplate::DeformEllipsoid(const TemplateNodeDeform & deform, PartBox & deformMe) const
{
	//if (deform.m_bXFormChanged)	// update deformation
	{
		assert(deform.m_iNodeID >= 0 && deform.m_iNodeID < (int)m_AllNodes.size());
		
		R3Affine orig2deform = R3identity_affine;
		
		R3Affine xform = PartTransforms::CreateAffineTransform(deform.m_fTransform, deform.m_TransformType);
		R3Affine scale = R3identity_affine;
		const R3Vector & undeformed = deformMe.UndeformedScale();
		const R3Vector & deformed = deform.m_PCAScale;

		R3Affine reOrient(deformMe.UndeformedOrientation().Matrix());
		scale.Transform(reOrient);
		for (int axis=0; axis<3; axis++)
			if (undeformed[axis]>0)
				scale.Scale(axis, deformed[axis]/undeformed[axis]);
		scale.InverseTransform(reOrient);
		
		orig2deform.Transform(xform);
		orig2deform.Transform(scale);
		
		((TemplateNodeDeform&)deform).m_bXFormChanged = false;
		((TemplateNodeDeform&)deform).m_AffineXForm = orig2deform;
	}
	deformMe.ApplyDeform(&deform);
}

PartsCollection * STemplate::GetAllParts()
{
	return m_pAllParts;
}

// seed (high-probability) template configurations
int STemplate::NSeedConfigs() const
{
	return 1 + m_sSeedNames.count();
}

/**
 * Create a seed deformation
 */
TMatch * STemplate::CreateSeedConfig(int configID, GeomShape * shape,
									 const std::vector<int> & subset, const R3Affine & shape2root) const
{
	if (configID>0)
	{
		
		// need to have more seeds
		configID--;
		if (!(configID>=0 && configID<(int)m_SeedDeformations.size()))
			std::cout<<"[ERROR] Cannot create seed config: "<<configID<<std::endl;
		assert(configID>=0 && configID<(int)m_SeedDeformations.size());
		
		std::vector<TemplateNodeDeform*> tConfig;
		for (int i=0; i<m_SeedDeformations[configID]->NTemplateNodes(); i++)
			tConfig.push_back(m_SeedDeformations[configID]->Deformation(i).CreateCopy());
		
		return new TMatch(this, configID+1, shape, subset, shape2root, tConfig);
	}
	
	assert(configID==0);	// most probable config
	
	std::vector<TemplateNodeDeform*> tConfig;
	
	// copy shape template tree to scene graph tree
	std::vector<int> nParts;
	std::vector<int> partIDs;
	
	std::vector<const STemplateNode*> explore;
	assert(m_AllNodes.size()>=1);
	explore.push_back(m_AllNodes[0]);
	nParts.push_back(1);
	partIDs.push_back(0);
	
	while (explore.size()>0)
	{
		assert(explore.size()==nParts.size() && explore.size()==partIDs.size());
		int lID = 0;
		const STemplateNode * cst = explore[lID];
		//std::cout<<"Exploring node: "<<cst->NodeID()<<" - "<<cst->NodeName().c_str()<<std::endl;

		int np = nParts[lID];
		int pID = partIDs[lID];
		nParts.erase(nParts.begin()+lID);
		partIDs.erase(partIDs.begin()+lID);
		explore.erase(explore.begin()+lID);
		
		// add as a final deformed node
		if (cst->NPartOfChildren()==0)
		{
			// get most probable transform
			const PartOfEdge * edge = cst->PartOfParent();
			assert(edge!=NULL);
			int flags, ndims; 
			const double * xform = edge->GetMeanTransform(configID, np, pID, flags, ndims);
//			if (configID==3)
//			{
//				std::cout<<"Creating: "<<edge->ChildTemplate()->NodeName().c_str()<<": ";
//				std::cout<<"["<<xform[0]<<" "<<xform[1]<<" "<<xform[2]<<"]"<<std::endl;
//			}
			
			// get most probable scale
			R3Vector ellScale(0,0,0);
			if (!cst->GetPart(configID)->IsType(PartBox::BOXTYPE_NullBox)
				&& !cst->GetPart(configID)->IsType(PartBox::BOXTYPE_EmptyBox))
				ellScale = cst->GetMeanScale(configID);
			tConfig.push_back(new TemplateNodeDeform(cst->NodeID(), np, pID, xform, ndims, flags, ellScale));
		}
		
		// add all configurations (e.g. based on number of parts)
		for (int c=0; c<cst->NPartOfChildren(); c++)
		{
			const PartOfEdge * edge = cst->PartOfChild(c);
			
			std::map<int, bool> nparts;
			for (int configID = 0; configID < edge->NConfigs(); configID++)
			{
				const DiscreteProbDistribution * npartsStats = edge->GetProbNParts(configID);
				for (int i=1; i<npartsStats->MaxNonzeroState(); i++)
				{
					if (npartsStats->Prob(i) > 0)
						nparts[i] = true;
				}
			}
			
			for (std::map<int, bool>::iterator npartsIter = nparts.begin(); npartsIter != nparts.end(); npartsIter++)
			for (int pi=0; pi<npartsIter->first; pi++)
			{
				nParts.push_back(npartsIter->first);
				partIDs.push_back(pi);
				explore.push_back(edge->ChildTemplate());
			}
		}
	}
	
	return new TMatch(this, configID, shape, subset, shape2root, tConfig);
}

R3Point STemplate::TemplatePointInOtherConfig(int configID, const TMatch * match, int tmpltPntID) const
{
	if (configID==0)
		return TemplatePointInCanonicalSpace(match, tmpltPntID);
	
	// create appropriate match
	STemplate * st = (STemplate*)this;
	while (configID >= (int)m_CanonicalMatches.size())
		st->m_CanonicalMatches.push_back(NULL);
	if (m_CanonicalMatches[configID]==NULL)
		st->m_CanonicalMatches[configID] = CreateSeedConfig(configID, NULL, std::vector<int>(), R3identity_affine);
	
	// assume deformation ids are the same (never matchdifferent configs)
	int deformID, localPntID;
	match->LocalPntID(tmpltPntID, deformID, localPntID);
	const TemplateNodeDeform & inDeform = match->Deformation(deformID);
	const PartBox & inBox = match->NodeEllipsoid(deformID);
	int nodeID = match->Deformation(deformID).NodeID();
	
	assert(match->Deformation(deformID).PartID()==0 && match->Deformation(deformID).NParts()==1);
	const TemplateNodeDeform & outDeform = st->m_CanonicalMatches[configID]->Deformation(deformID);
	const PartBox & outBox = st->m_CanonicalMatches[configID]->NodeEllipsoid(deformID);
	
	// find corresponding point in canonical space (mesh point or box)
	if (inBox.IsType(PartBox::BOXTYPE_BoxSurface))	// box
	{
		assert(inBox.m_pSamplingOfABox!=NULL);
		assert(localPntID>=0 && localPntID < (int)inBox.m_pSamplingOfABox->samples.size());
		const PointOnPartBox & pntOnBox = inBox.m_pSamplingOfABox->samples[localPntID];
		
		R3Point undeformedPnt = outBox.TemplatePointPos(pntOnBox);
		return outDeform.Orig2Deformed(undeformedPnt);
	}
	else if (inBox.IsType(PartBox::BOXTYPE_MeshPoints)) // mesh
	{
		R3Point undeformedPnt = outBox.GetPart()->PointPosition(localPntID);
		return outDeform.Orig2Deformed(undeformedPnt);
	}
	else
		assert(false);
	
	//return deform.Orig2Deformed(match->UndeformedTemplateSample(tmpltPntID));
}

R3Point STemplate::TemplatePointInCanonicalSpace(const TMatch * match, int tmpltPntID) const
{
	// create / find root match (i.e. all parts co-aligned)
	if (m_pRootMatch==NULL)
	{
		STemplate * st = (STemplate*)this;
		st->m_pRootMatch = CreateSeedConfig(0, NULL, std::vector<int>(), R3identity_affine);
		for (int i=0; i<NNodes(); i++)
			st->m_NodeID2RootMatchDeform.push_back(-1);
		for (int i=0; i<m_pRootMatch->NTemplateNodes(); i++)
		{
			int nodeID = m_pRootMatch->Deformation(i).NodeID();
			int partID = m_pRootMatch->Deformation(i).PartID();
			int nparts = m_pRootMatch->Deformation(i).NParts();
			assert(partID==0 && nparts==1 && nodeID>=0 && nodeID < NNodes());
			st->m_NodeID2RootMatchDeform[nodeID] = i;
		}
		m_pRootMatch->UpdateEllipsoids();
	} 

	// find corresponding node (in match)
	int deformID, localPntID;
	match->LocalPntID(tmpltPntID, deformID, localPntID);
	const TemplateNodeDeform & inDeform = match->Deformation(deformID);
	const PartBox & inBox = match->NodeEllipsoid(deformID);
	int nodeID = match->Deformation(deformID).NodeID();
	assert(match->Deformation(deformID).PartID()==0 && match->Deformation(deformID).NParts()==1);
	
	// map match node id to root match node id. I.e. find canonical deformation
	assert(nodeID>=0 && nodeID < (int)m_NodeID2RootMatchDeform.size() && m_NodeID2RootMatchDeform[nodeID]>=0);
	const TemplateNodeDeform & outDeform = m_pRootMatch->Deformation(m_NodeID2RootMatchDeform[nodeID]);
	const PartBox & outBox = m_pRootMatch->NodeEllipsoid(m_NodeID2RootMatchDeform[nodeID]);
	
	// find corresponding point in canonical space (mesh point or box)
	if (inBox.IsType(PartBox::BOXTYPE_BoxSurface))	// box
	{
		assert(inBox.m_pSamplingOfABox!=NULL);
		assert(localPntID>=0 && localPntID < (int)inBox.m_pSamplingOfABox->samples.size());
		const PointOnPartBox & pntOnBox = inBox.m_pSamplingOfABox->samples[localPntID];
		
		R3Point undeformedPnt = outBox.TemplatePointPos(pntOnBox);
		return outDeform.Orig2Deformed(undeformedPnt);
	}
	else if (inBox.IsType(PartBox::BOXTYPE_MeshPoints)) // mesh
	{ 
		R3Point undeformedPnt = outBox.GetPart()->PointPosition(localPntID);
		return outDeform.Orig2Deformed(undeformedPnt);
	}
	else
		assert(false);
	return R3Point(0,0,0);
	
//	// canonical point: deform using root's deformation
//	if (tmpltPntID < m_pRootMatch->NTemplatePoints())
//	{
//		m_pRootMatch->LocalPntID(tmpltPntID, deformID, localPntID);
//		const TemplateNodeDeform & tmpdeform = m_pRootMatch->Deformation(deformID);
//		//return tmpdeform.Orig2Deformed(m_pRootMatch->UndeformedTemplateSample(tmpltPntID));
//		return m_pRootMatch->UndeformedTemplateSample(tmpltPntID);
//	}
//	else
//		return R3Point(0,0,0);
//	return deform.Orig2Deformed(match->UndeformedTemplateSample(tmpltPntID));
	//return inDeform.Orig2Deformed(match->UndeformedTemplateSample(tmpltPntID));
	//return match->UndeformedTemplateSample(tmpltPntID);
}

const VKString & STemplate::ConfigName(int configID) const
{
	assert(configID>=1 && configID < NSeedConfigs());
	return m_sSeedNames[configID-1];
}

int STemplate::MaxDescribedInstances(int configID) const
{
	int nMax = 0;
	for (int i=0; i<(int)m_AllNodes.size(); i++)
		nMax = vkMax(nMax, m_AllNodes[i]->NDescribedInstances(configID));
	return nMax;
}

// improve stats
void STemplate::UpdateAllStats(const TMatch * match, bool include)
{
	const VKString & mName = match->ModelName();
	if (mName!="none")
	{
		if (include)
		{
			if (m_CreatedFromModels.contains(mName))
				std::cout<<"[WARNING] STemplate was updated from "<<mName.c_str()<<" before"<<std::endl;
			else
			{
				m_CreatedFromModels.push_back(mName);
				//m_iNDescribedInstances++;
			}
		}
		else	// exclude
		{
			assert(m_CreatedFromModels.contains(mName));
			//m_iNDescribedInstances--;
			m_CreatedFromModels.remove(mName);
		}
	}
			
	PartBox rootEllipsoid(m_pAllParts);
	R3Triad xyz(R3Vector(1, 0, 0), R3Vector(0, 1, 0), R3Vector(0, 0, 1));
	rootEllipsoid.InitBox(R3Point(0,0,0), xyz, R3Vector(1, 1, 1), 0);
	
	for (int i=0; i<match->NTemplateNodes(); i++)
	{
		const TemplateNodeDeform & deform = match->Deformation(i);
		
		assert (deform.NParts()==1);	// assume for now
		if (m_AllNodes[deform.NodeID()]->PartOfParent()!=NULL)
		{
			int nparts = deform.IsIncluded() ? deform.NParts() : 0;
			m_AllNodes[deform.NodeID()]->PartOfParent()->UpdateNParts(match->SeedID(), nparts, include);
		}
				
		m_AllNodes[deform.NodeID()]->UpdateShapeAndXForm(match->SeedID(), deform, include);
	}
	
	// update seed deformations
	Normalize(R3Point(0,0,0), 1);
	
	// update classifier (from real geometry)
	if (match->GetMatchedGeomAsPnts()!=NULL)
	{
		int configID = match->SeedID();
		if (configID>=(int)m_apLocalClassifiers.size())
			TrainClassifiers();
		if (m_apLocalClassifiers[configID]==NULL)
			TrainClassifier(configID);
		
		if (configID>=(int)m_apLocalClassifiers.size() || m_apLocalClassifiers[configID]==NULL)
		{
			std::cout<<"[ERROR] Invalid Config: "<<configID<<" / "<<m_apLocalClassifiers.size();
			std::cout<<" seeds="<<m_SeedDeformations.size()<<std::endl;
		}
			
		assert(configID < (int)m_apLocalClassifiers.size() && m_apLocalClassifiers[configID]!=NULL);
		std::vector<const PartLocalGeometry *> shapes;
		LocalGeometryTrainingSet * trainingSet = CreateTrainingSet(match, shapes);
		m_apLocalClassifiers[configID]->UpdateClassifier(*trainingSet);
//		delete trainingSet;
//		for (int i=0; i<(int)shapes.size(); i++)
//			delete shapes[i];	// vk: memory leak
	}
}

LocalGeometryTrainingSet * STemplate::CreateTrainingSet(const TMatch * match, std::vector<const PartLocalGeometry *> & shapes)
{
	ShapePointCloud * reSortedPoints = new ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD, "allt", "none", NULL);
	std::vector<int> nodeIDs;
	
	for (int i=0; i<match->NTemplateNodes(); i++)
	{
		int offset = reSortedPoints->NumPoints();
		int npoints = match->NPointsLabeledAsNode(i);
		if (match->NodeEllipsoid(i).IsType(PartBox::BOXTYPE_NullBox))
			npoints = 0;
		
		for (int j=0; j<npoints; j++)
		{
			// NOTE: if node changes orientation need to change this...
			R3Point p = match->AlignedPointPosition(match->GetPointLabeledAsNode(i, j));
			reSortedPoints->AddPoint(ShapePoint(p, reSortedPoints, reSortedPoints->NumPoints()));
		}
		
		if (npoints>0)
		{
			shapes.push_back(new PartLocalGeometry(reSortedPoints, offset, npoints, true));
			nodeIDs.push_back(i);
		}
	}
	LocalGeometryTrainingSet * ptr = new LocalGeometryTrainingSet(shapes, nodeIDs, match->NTemplateNodes());
	return ptr;
}

void STemplate::AddOrReplaceSeed(int myseedid, const VKString & name, TMatch * match, int configParent)
{
	int actualSeed = myseedid;
	int localSeed = myseedid-1;
	assert(m_sSeedNames.count()==(int)m_SeedDeformations.size());
	if (localSeed == m_sSeedNames.count())
	{
		m_sSeedNames.push_back(name);
		m_SeedDeformations.push_back(NULL);
	}
	assert(localSeed>=0 && localSeed < (int)m_SeedDeformations.size());
	if (configParent==0)
		m_iInitCoreSeeds = actualSeed;
	
	while (actualSeed>=(int)m_ConfigParents.size())
		m_ConfigParents.push_back(-1);
	
	m_ConfigParents[actualSeed] = configParent;
	
	if (m_SeedDeformations[localSeed]!=match)
	{
		if (m_SeedDeformations[localSeed]!=NULL)
			delete m_SeedDeformations[localSeed];
		
		for (int i=0; i<match->NTemplateNodes(); i++)
		{
			int nodeID = match->Deformation(i).NodeID();
			assert(match->Deformation(i).PartID()==0);
			
			PartOfEdge * edge = m_AllNodes[nodeID]->PartOfParent();
			if (actualSeed < edge->NConfigs())
				((DiscreteProbDistribution*)edge->GetProbNParts(actualSeed))->Reset();
			if (match->Deformation(i).IsIncluded())
				edge->UpdateNParts(actualSeed, match->Deformation(i).NParts(), true);
			else
				edge->UpdateNParts(actualSeed, 0, true);
		}
	}
	m_sSeedNames[localSeed] = name;
	m_SeedDeformations[localSeed] = match;
	UpdateAllStats(match, true);
}

void STemplate::RemoveSeed(int seedID)
{
	seedID--;
	assert(m_sSeedNames.count()==(int)m_SeedDeformations.size());
	assert(seedID>=0 && seedID < m_sSeedNames.count());
	m_sSeedNames.remove(seedID);
	delete m_SeedDeformations[seedID];
	m_SeedDeformations.erase(m_SeedDeformations.begin()+seedID);
	for (int i=seedID; i<(int)m_SeedDeformations.size(); i++)
		m_SeedDeformations[i]->SetSeedID(i+1);
	m_ConfigParents.erase(m_ConfigParents.begin() + (seedID+1));
}

void STemplate::RemoveExcludedDeformsFromSeeds()
{
	for (int i=0; i<(int)m_SeedDeformations.size(); i++)
	{
		for (int j=0; j<(int)m_SeedDeformations[i]->NTemplateNodes(); j++)
		{
			if (!m_SeedDeformations[i]->Deformation(j).IsIncluded())
			{
				m_SeedDeformations[i]->RemoveDeform(j);
				j--;
			}
		}
	}
}

const VKStringList & STemplate::GetCreatedFromModelsList() const
{
	return m_CreatedFromModels;
}

// write template
void STemplate::WriteTemplate(const VKString & filename) const
{
	FileManagement::CreateDirsOnPathToFile(filename.c_str());
	FILE * file = fopen(filename.c_str(), "wb");
	if (file==NULL)
		std::cout<<"[ERROR] Cannot open "<<filename.c_str()<<std::endl;
	assert(file!=NULL);
	
	int version = STMPLT_V0_4;
	fwrite(&version, sizeof(STemplateVersions), 1, file);
	
	fwrite(&m_fUseMeshThresh, sizeof(double), 1, file);
	bool exists = (m_pAllParts!=NULL);
	fwrite(&exists, sizeof(bool), 1, file);
	if (exists)
		m_pAllParts->WriteParts(file);
	 
	int nNodes = (int)m_AllNodes.size();
	fwrite(&nNodes, sizeof(int), 1, file);
		
	// save all node data
	for (int i=0; i<nNodes; i++)
		m_AllNodes[i]->WriteNode(file);
		
	// part of relationships
	for (int i=0; i<nNodes; i++)
		m_AllNodes[i]->WriteSubparts(file);
	
	// connector edge (symmetry, other constraints)
	VKStringList::WriteList(file, m_CreatedFromModels);
	fwrite(&m_iNTopologyChanges, sizeof(int), 1, file);
	
	assert(m_sSeedNames.count()==(int)m_SeedDeformations.size());
	int nseeds = m_sSeedNames.count();
	fwrite(&nseeds, sizeof(int), 1, file);
	fwrite(&m_iInitCoreSeeds, sizeof(int), 1, file);
	
	if (nseeds+1 != (int)m_ConfigParents.size())
	{
		if (m_ConfigParents.size()==0 && m_iInitCoreSeeds==(int)nseeds)	// all seeds are core
		{
			for (int i=0; i<=nseeds; i++)
				((STemplate*)this)->m_ConfigParents.push_back(0);
		}
		else
		{
			std::cout<<"[ERROR] Invalid number of seeds: "<<nseeds<<" / "<<m_ConfigParents.size()<<std::endl;
			std::cout<<"   NCORR="<<m_iInitCoreSeeds<<std::endl;
			assert(false);
		}
	}
	
	assert(nseeds+1 == (int)m_ConfigParents.size());
	for (int i=0; i<(int)m_ConfigParents.size(); i++)
		fwrite(&(m_ConfigParents[i]), sizeof(int), 1, file);

	VKStringList::WriteList(file, m_sSeedNames);
	for (int i=0; i<nseeds; i++)
	{
		int ndeforms = m_SeedDeformations[i]->NTemplateNodes();
		fwrite(&ndeforms, sizeof(int), 1, file);
		for (int j=0; j<ndeforms; j++)
		{
//			int nid = m_SeedDeformations[i]->Deformation(j).NodeID();
//			std::cout<<"seed: "<<i<<" nodeid: "<<nid<<" name: "<<GetNode(nid)->NodeName().c_str()<<std::endl;
			m_SeedDeformations[i]->Deformation(j).WriteNodeDeform(file);
		}
	}
	
	// write classifier
	int nclassifiers = (int)m_apLocalClassifiers.size();
	fwrite(&nclassifiers, sizeof(int), 1, file);
	for (int i=0; i<nclassifiers; i++)
	{
		LocalGeometryClassifier * locCla = m_apLocalClassifiers[i];
		bool claExists = (locCla!=NULL);
		fwrite(&claExists, sizeof(bool), 1, file);
		if (claExists)
			locCla->WriteClassifier(file);
	}
	
	
	fclose(file);
}

STemplate * STemplate::ReadTemplate(const VKString & filename, ModelDatabase * db)
{
	FILE * file = fopen(filename.c_str(), "rb");
	if (file==NULL)
		std::cout<<"[ERROR] Cannot open "<<filename.c_str()<<std::endl;
	assert(file!=NULL);
	
	STemplateVersions version;
	fread(&version, sizeof(STemplateVersions), 1, file);
	assert(version<=(int)STMPLT_V0_4);

	STemplate * st = new STemplate(NULL);	// it's ok to forget the initial geometry
	// part library
	fread(&(st->m_fUseMeshThresh), sizeof(double), 1, file);
	bool exists;
	fread(&exists, sizeof(bool), 1, file);
	if (exists)
		st->m_pAllParts = PartsCollection::ReadParts(file);
	
	int nNodes = 0;
	fread(&nNodes, sizeof(int), 1, file);
	
	// read all node data
	for (int i=0; i<nNodes; i++)
		st->m_AllNodes.push_back(STemplateNode::ReadNode(file, st, version));
	
	// part of relationships
	for (int i=0; i<nNodes; i++)
		st->m_AllNodes[i]->ReadSubparts(file, st, version);

	//	// todo: connector edge (symmetry, other constraints)
	VKStringList::ReadList(file, st->m_CreatedFromModels);
	fread(&(st->m_iNTopologyChanges), sizeof(int), 1, file);
	
	int nseeds;
	fread(&nseeds, sizeof(int), 1, file);
	if (feof(file))
	{
		fclose(file);
		return st;	// no default configs
	}
	
	if (version==STMPLT_V0_1)
		st->m_iInitCoreSeeds = nseeds;
	else if ((int)version>=(int)STMPLT_V0_2)
		fread(&(st->m_iInitCoreSeeds), sizeof(int), 1, file);

	if ((int)version >= (int)STMPLT_V0_3)
	{
		// hierarchy
		for (int i=0; i<nseeds+1; i++)
		{
			st->m_ConfigParents.push_back(0);
			fread(&(st->m_ConfigParents[i]), sizeof(int), 1, file);
			if (i > 0 && st->m_ConfigParents[i] < 0)
				st->m_ConfigParents[i] = 0;
		}
	}
	
	VKStringList::ReadList(file, st->m_sSeedNames);
	
	assert(nseeds==st->m_sSeedNames.count());
	
	if (version<STMPLT_V0_3)
	{
		for (int i=0; i<(int)st->m_AllNodes.size(); i++)
			st->m_AllNodes[i]->DuplicateConfigs(0);
		
		for (int i=0; i<(int)st->m_AllNodes.size(); i++)
			if (st->m_AllNodes[i]->PartOfParent()!=NULL)
				st->m_AllNodes[i]->PartOfParent()->DuplicateConfigs(0);
	}
	
	
	for (int i=0; i<nseeds; i++)
	{
		int ndeforms;
		fread(&ndeforms, sizeof(int), 1, file);
		std::vector<TemplateNodeDeform*> tConfig;
		for (int j=0; j<ndeforms; j++)
		{
			tConfig.push_back(TemplateNodeDeform::ReadNodeDeform(file));
//			// TODO: fix a bug in template creation app... removeme
//			if (i==2 && tConfig[j]->IsIncluded())
//			{
//				VKString nname = st->GetNode(tConfig[j]->NodeID())->NodeName();
//				std::cout<<"   "<<nname.c_str()<<": "<<j<<std::endl;
//				if (nname == "seat")
//					tConfig[j]->m_fTransform[2] += 1.2;
//				else if (nname=="leg")
//					tConfig[j]->m_fTransform[2] += .5;
//			}
		}
//		std::cout<<"Reading Seed: "<<i<<std::endl;
		TMatch * match = new TMatch(st, i+1, NULL, std::vector<int>(), R3identity_affine, tConfig);
		st->m_SeedDeformations.push_back(match);
//		if (i==2)
//		{
//			int myPartID = 1; 
//			STemplateNode * n = (STemplateNode*)st->GetNode(tConfig[myPartID]->NodeID());
//			int flags, ndims;
//			double * mt = ((double*)(n->m_pParentPart->GetMeanTransform(i+1, 1, 0, flags, ndims)));
//			std::cout<<"N: "<<n->NodeName().c_str()<<": ["<<mt[0]<<" "<<mt[1]<<" "<<mt[2]<<"] "<<std::endl;
//			std::cout<<"D: "<<n->NodeName().c_str()<<": ["<<tConfig[myPartID]->m_fTransform[0]<<" "<<tConfig[myPartID]->m_fTransform[1]<<" "<<tConfig[myPartID]->m_fTransform[2]<<"] "<<std::endl;
//			for (int k=0; k<2; k++)
//				mt[k] = tConfig[1]->m_fTransform[k];
//			for (int k=0; k<2; k++)
//				((ProbGaussian*)n->m_pParentPart->m_apPartTransforms[i+1][1]->m_apXForms[0])->m_aMeans[k] = tConfig[myPartID]->m_fTransform[k];
//			((ProbGaussian*)n->m_apPartScales[i+1])->m_aMeans[2] = tConfig[1]->m_PCAScale[2];
//		}
		
//		for (int n=0; n<match->NTemplateNodes(); n++)
//		{
//			((PartBox&)match->NodeEllipsoid(n)).SetScale(match->Deformation(n).GetScale());
//		}
//		if (i==2)
//			st->SetDeformAsRestPose(st->m_SeedDeformations[i]);
	}
	assert(st->m_sSeedNames.count()==(int)st->m_SeedDeformations.size());
	//std::cout<<"Template: v="<<version<<std::endl;
	if (version==STMPLT_V0_4)
	{
		// read classifiers
		int nclassifiers;
		fread(&nclassifiers, sizeof(int), 1, file);
		for (int i=0; i<nclassifiers; i++)
		{
			bool claExists;
			fread(&claExists, sizeof(bool), 1, file);
			if (claExists)
				st->m_apLocalClassifiers.push_back(LocalGeometryClassifier::ReadClassifier(file));
			else
				st->m_apLocalClassifiers.push_back(NULL);
		}
		//std::cout<<"Read classifiers: "<<nclassifiers<<std::endl;
	}
		
	
	fclose(file);
	return st;
}

void STemplate::SetDeformAsRestPose(TMatch * match)
{
	for (int i=0; i<(int)m_AllNodes.size(); i++)
		m_AllNodes[i]->Reset();
	
	for (int i=0; i<match->NTemplateNodes(); i++)
	{
		const TemplateNodeDeform & deform = match->Deformation(i);
//		if (m_AllNodes[deform.NodeID()]->PartOfParent()!=NULL)
//			m_AllNodes[deform.NodeID()]->PartOfParent()->UpdateNParts(match->SeedID(), deform.NParts(), true);
		m_AllNodes[deform.NodeID()]->UpdateShapeAndXForm(match->SeedID(), deform, true);
	}
}

int STemplate::NSamples(int configID) const
{
	int nsamp = 0;
	for (int i=0; i<(int)m_AllNodes[0]->m_SubpartTypes.size(); i++)
	{
		PartOfEdge * edge = m_AllNodes[0]->m_SubpartTypes[i];
		if (configID<edge->NConfigs())
			nsamp = vkMax(nsamp, edge->GetProbNParts(configID)->NSamples());
	}
	return nsamp;
}

int STemplate::ConfigParent(int configID) const
{
	if (m_ConfigParents.size()==0 )	// all seeds are core
	{
		for (int i=0; i<=m_iInitCoreSeeds; i++)
			((STemplate*)this)->m_ConfigParents.push_back(0);
	}
		
	if (!(configID>=0 && configID < (int)m_ConfigParents.size()))
	{
		std::cout<<"[ERROR] Invalid config: "<<configID<<" / "<<m_ConfigParents.size()<<std::endl;
		assert(false);
	}
	return m_ConfigParents[configID];
}

const std::vector<int> & STemplate::GetConfigHierarchy() const
{
	return m_ConfigParents;
}

void STemplate::TrainClassifier(int configID)
{
	//assert(configID > 0);
	TMatch * match = CreateSeedConfig(configID, NULL, std::vector<int>(), R3identity_affine);

	std::vector<const PartLocalGeometry *> shapes;
	std::vector<int> nodeIDs;
	
	for (int nid=0; nid<match->NTemplateNodes(); nid++)
	{
		int globNodeID = match->Deformation(nid).NodeID();
		const PartLocalGeometry * lg = match->NodeEllipsoid(nid).GetLocalGeometry();
		if (lg!=NULL && match->Deformation(nid).IsIncluded())
		{
			shapes.push_back(lg);
			nodeIDs.push_back(nid);
		}
	}
	assert(shapes.size()>0);
	
	while (configID >= (int)m_apLocalClassifiers.size())
		m_apLocalClassifiers.push_back(NULL);
	assert(m_apLocalClassifiers[configID] == NULL);
	m_apLocalClassifiers[configID] = new LocalGeometryClassifier(LocalGeometryTrainingSet(shapes, nodeIDs, match->NTemplateNodes()));
	//delete match;
}

void STemplate::TrainClassifiers()
{
	TimeProfiler profiler;
	profiler.startedProcess("Training");
	std::cout<<"Classifier Training ("<<NSeedConfigs()<<") "<<std::flush;
	while ((int)m_apLocalClassifiers.size() < NSeedConfigs())
		m_apLocalClassifiers.push_back(NULL);

	for (int i=0; i<NSeedConfigs(); i++)
	{
		if (m_apLocalClassifiers[i] == NULL)
			TrainClassifier(i);
		std::cout<<i<<" "<<std::flush;
		
	}
	std::cout<<"\n     ";
	profiler.finishedProcess("Training", true);
}

const double * STemplate::LogProbsLocalDescriptors(int pntID, const TMatch * match) const
{
//	std::cout<<"Getting local descriptor - 0"<<std::endl;
	int nfeat;
	const double * feat = match->GetLocalPntsFeatures(pntID, -1, nfeat);
	
	const LocalGeometryClassifier * classifier = GetLocalClassifier(match->SeedID());
//	std::cout<<"Getting local descriptor - 1 claPtr="<<classifier<<std::endl;
	const double * val = classifier->LogProbs(nfeat, feat);
//	std::cout<<"Getting local descriptor - 2"<<std::endl;
	return val;
}

const LocalGeometryClassifier * STemplate::GetLocalClassifier(int configID) const
{
	if (m_apLocalClassifiers.size()==0)
		((STemplate*)this)->TrainClassifiers();
	
	if (configID >= (int)m_apLocalClassifiers.size())
		((STemplate*)this)->TrainClassifier(configID);
	
	assert(configID < (int)m_apLocalClassifiers.size());
	assert(m_apLocalClassifiers[configID]!=NULL);
	return m_apLocalClassifiers[configID];
}



