/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "TMatch.h"

#include "VkFunctions.h"
#include "LinAlg.h"

#include "CorrsAffineAlign.h"

#include "PartTransforms.h"
#include "STemplate.h"
#include "STemplateNode.h"
#include "PartOfEdge.h"
#include "PartConnectorEdge.h"

#include "TMatchEnergy.h"
#include "ConsistentPartsAlg.h"


TMatch::TMatch(const STemplate * st, int seedID, GeomShape * shape, const std::vector<int> & subset, 
			   const R3Affine & pnts2nodes, const std::vector<TemplateNodeDeform *> & nodeDeformations)
{
	m_pTemplateSamples = NULL;
	m_pPntsLocalGeometry = NULL;
	m_bAlignedNullNodes = false;
	m_pSTemplate = st;
	m_iSeedID = seedID;
	m_SeedAlignment = pnts2nodes;

	m_pPnts = NULL;
	m_pPntLabels = NULL;
	m_AllPnts2Subset = NULL;
	m_Subset2AllPnts = NULL;
	m_iSubsetSize = -1;
	
	m_AllPntsRadius = -1;
	m_AllPntsCentroid = R3Point(0,0,0);
	
	m_sModelName = "none";
	m_sGeomName = "none";
	m_iNPoints = 0;
	
	// geometry is optional
	if (shape!=NULL)
	{
		if (shape->GetModel()==NULL)
			m_sModelName = "UnknownModel";
		else
		{
			assert(shape->GetModel()!=NULL);
			m_sModelName = shape->GetModel()->ModelName();
		}
		
		m_sGeomName = shape->GeometryName();
		
		
		m_pPnts = shape->SafeCastToPointCloud();
		assert(m_pPnts!=NULL);
		m_iNPoints = m_pPnts->NumPoints();
		
		m_AllPntsCentroid = m_pPnts->Centroid();
		m_AllPntsRadius = m_pPnts->AverageRadius(m_AllPntsCentroid);
		
		// create point subset
		m_iSubsetSize = (int)subset.size();
		if (m_iSubsetSize==0)
			m_iSubsetSize = m_pPnts->NumPoints();

		m_AllPnts2Subset = new int[m_pPnts->NumPoints()];
		m_Subset2AllPnts = new int[m_iSubsetSize];
		for (int i=0; i<m_pPnts->NumPoints(); i++)
			m_AllPnts2Subset[i] = -1;
				
		for (int i=0; i<m_iSubsetSize; i++)
		{
			int sID = ((int)subset.size()==0) ? i : subset[i];
			m_Subset2AllPnts[i] = sID;
			m_AllPnts2Subset[sID] = i;
		}
	}
		
	m_Pnts2Nodes = pnts2nodes;
	m_fGlobalTranslate = R3Vector(0,0,0);
	m_fGlobalRotate = R3Vector(0,0,0);
	m_fGlobalScale = R3Vector(1., 1., 1.);

	// include all nodes
	m_TemplateMap = nodeDeformations;
	for (int i=0; i<(int)m_TemplateMap.size(); i++)
	{
		m_FinalEllipsoids.push_back(PartBox(((STemplate*)st)->GetAllParts()));
		m_FinalEllipsoids[i].InitCopy(*(st->GetNode(m_TemplateMap[i]->NodeID())->GetPart(seedID)));
		m_FinalEllipsoids[i].SetOrigin(R3Point(0,0,0));
	}
	
	UpdateEllipsoids();
	UpdateLabels(false);
	CreateTemplateSamples();
}

TMatch::~TMatch()
{
	if (m_Subset2AllPnts!=NULL)
	{
		delete [] m_Subset2AllPnts;
		delete [] m_AllPnts2Subset;
	}
	
	for (int i=0; i<(int)m_TemplateMap.size(); i++)
		delete m_TemplateMap[i];
	
	if (m_pPntsLocalGeometry!=NULL)
		delete m_pPntsLocalGeometry;
	
	if (m_pTemplateSamples!=NULL)
		delete m_pTemplateSamples;
}

TMatch * TMatch::CreateCopy(const std::vector<TemplateNodeDeform *> & addNodes,
							const STemplate * newTemplate, std::map<int, int> * old2newNode, int configoffset) const
{
	std::vector<TemplateNodeDeform * > ndeforms;
	
	for (int i=0; i<(int)m_TemplateMap.size(); i++)
		ndeforms.push_back(m_TemplateMap[i]->CreateCopy(old2newNode));
	for (int i=0; i<(int)addNodes.size(); i++)
		ndeforms.push_back(addNodes[i]->CreateCopy(old2newNode));
	 
	std::vector<int> subset;
	if (m_iSubsetSize!=-1 && (m_Subset2AllPnts!=NULL))
	{
		for (int i=0; i<m_iSubsetSize; i++)
			subset.push_back(m_Subset2AllPnts[i]);
	} 
	
	if (newTemplate==NULL)
		newTemplate = m_pSTemplate;
	TMatch * newMatch = new TMatch(newTemplate, configoffset + m_iSeedID, m_pPnts, subset, m_SeedAlignment, ndeforms);
	
	// check things got transfered
	newMatch->m_iNPoints = m_iNPoints;
	newMatch->m_sModelName = m_sModelName;
	newMatch->m_sGeomName = m_sGeomName;
	
	assert(newMatch->m_iSubsetSize == m_iSubsetSize);
	for (int i=0; i<m_iNPoints; i++)
		assert(newMatch->m_AllPnts2Subset[i] == m_AllPnts2Subset[i]);
	for (int i=0; i<m_iSubsetSize; i++)
		assert(newMatch->m_Subset2AllPnts[i] == m_Subset2AllPnts[i]);
	
	// copy global alignment
	newMatch->m_fGlobalTranslate = m_fGlobalTranslate;
	newMatch->m_fGlobalRotate = m_fGlobalRotate;
	newMatch->m_fGlobalScale = m_fGlobalScale;
	newMatch->m_Pnts2Nodes = m_Pnts2Nodes;
		
	// copy mapping
	for (int i=0; i<m_aPnts2Segments.size(); i++)
		newMatch->m_aPnts2Segments[i] =m_aPnts2Segments[i];

	newMatch->UpdateLabels(false);
	newMatch->CreateTemplateSamples();
	return newMatch;
}


// matched geometry and template
const ShapePointCloud * TMatch::GetMatchedGeomAsPnts() const
{
	return m_pPnts;
}

const VKString & TMatch::ModelName() const
{
	return m_sModelName;
}

const VKString & TMatch::GeomName() const
{
	return m_sGeomName;
}

int TMatch::NPointsLabeledAsNode(int nodeID) const
{
	if (!(nodeID>=0 && nodeID < (int)m_NodeToGlobPntIDs.size()))
		std::cout<<"[ERROR] NodeID = "<<nodeID<<" Cached: "<<m_NodeToGlobPntIDs.size()<<" thisptr="<<this<<std::endl;
	assert(nodeID>=0 && nodeID < (int)m_NodeToGlobPntIDs.size());
	return (int) m_NodeToGlobPntIDs[nodeID].size();
}

int TMatch::GetPointLabeledAsNode(int nodeID, int pntid) const
{
	assert(nodeID>=0 && nodeID < (int)m_NodeToGlobPntIDs.size());
	assert(pntid>=0 && pntid < (int)m_NodeToGlobPntIDs[nodeID].size());
	return (int)m_NodeToGlobPntIDs[nodeID][pntid];
}

int TMatch::NearestPointWithLabel(int nodeID, const R3Point & posNormalizedPntsSpace) const
{
	assert(m_NodeToGlobPntIDs[nodeID].size()>0);
	TMatch * ncThis = (TMatch*)this;
	if (NTemplateNodes() >= m_apKdTreesForLabeledPnts.size())
	{
		ncThis->FreeKdTrees();
		while (NTemplateNodes() >= m_apKdTreesForLabeledPnts.size())
			ncThis->m_apKdTreesForLabeledPnts.push_back(NULL);
	}

	assert(nodeID>=0 && nodeID < (int)m_apKdTreesForLabeledPnts.size());
	R3Kdtree<const PointInSet *> * kdTree = ncThis->m_apKdTreesForLabeledPnts[nodeID];

	if (kdTree==NULL)
	{
		assert(nodeID < (int)m_NodeToGlobPntIDs.size());
		kdTree = GetMatchedGeomAsPnts()->CreateSubsetKdTree(m_NodeToGlobPntIDs[nodeID]);
		ncThis->m_apKdTreesForLabeledPnts[nodeID] = kdTree;
	}

	assert(kdTree!=NULL);
	
	R3Point origPnt = m_AllPntsRadius * posNormalizedPntsSpace + m_AllPntsCentroid;
	const PointInSet * closest = kdTree->FindClosest(origPnt, 0, RN_INFINITY*10000);
	assert(closest!=NULL);
	return closest->id;
}

int TMatch::NearestTemplatePointGlobalID(int nodeID, const R3Point & posDeformedTemplateSpace) const
{
	assert(nodeID>=0 && nodeID < (int)m_FinalEllipsoids.size());
	
	if (m_FinalEllipsoids[nodeID].IsType(PartBox::BOXTYPE_NullBox))
	{
		std::cout<<"[WARNING] Cannot map NULL points. Doing expensive nearest neighbor (reimplement)"<<std::endl;
		int gid = -1;
		double bestDist = -1;
		for (int p=0; p<NTemplatePoints(); p++)
		{
			double dist = R3Distance(DeformedTemplateSample(p), posDeformedTemplateSpace);
			if (gid < 0 || dist<bestDist)
			{
				gid = p;
				bestDist = dist;
			}
		}
		assert(gid>=0 && bestDist>=0);
		return gid;
	}
	return GlobalTmpltPnt(nodeID, m_FinalEllipsoids[nodeID].NearestTemplatePointLocalID(posDeformedTemplateSpace));
}

int TMatch::GlobalTmpltPnt(int nodeID, int pntID) const
{
	assert(nodeID>=0 && nodeID < (int)m_PerNodeTemplateSamplesOffset.size());
	return m_PerNodeTemplateSamplesOffset[nodeID] + pntID;
}

void TMatch::LocalPntID(int globalID, int & nodeID, int & localID) const
{
	assert(globalID >= 0 && globalID < (int)m_GlobalTemplateIDToNodeID.size());
	nodeID = m_GlobalTemplateIDToNodeID[globalID];
	assert(nodeID>=0 && nodeID < (int)m_PerNodeTemplateSamplesOffset.size());
	localID = globalID - m_PerNodeTemplateSamplesOffset[nodeID];
	assert(localID >= 0);
}

R3Point TMatch::UndeformedTemplateSample(int globalTmpltPnt) const
{
	int nodeID, locID;
	LocalPntID(globalTmpltPnt, nodeID,locID);
	return m_FinalEllipsoids[nodeID].UndeformedTemplatePointPosition(locID);
}

R3Point TMatch::DeformedTemplateSample(int globalTmpltPnt) const
{
	int nodeID, locID;
	LocalPntID(globalTmpltPnt, nodeID, locID);
	return m_FinalEllipsoids[nodeID].DeformedTemplatePointPosition(locID);
}

R3Point TMatch::NormalizedPosition(int gpntid) const
{
	R3Point pos = m_pPnts->PointPosition(gpntid);
	R3Vector ppv = (pos-m_AllPntsCentroid) / m_AllPntsRadius;
	return R3Point(ppv.X(), ppv.Y(), ppv.Z());
}

R3Point TMatch::AlignedPointPosition(int gPntID) const
{
	R3Point pos = NormalizedPosition(gPntID);
	m_Pnts2Nodes.Apply(pos);
	return pos;
}

/*
R3Triad TMatch::AlignedPoint(int globalPntID, const ShapeLocalPCA * pcaAxes, R3Point & pos, double * pntEigs) const
{
	assert(m_pPnts!=NULL && m_AllPntsRadius>0);
	 
	const ShapePoint & pnt = m_pPnts->GetPoint(globalPntID);
	R3Triad axes;
	if (pcaAxes!=NULL)
		axes = pcaAxes->GetPCA(pnt, pntEigs);
		
	//R3Vector ppv = (pnt.GetPosition()-m_AllPntsCentroid) / m_AllPntsRadius;
	//pos = R3Point(ppv.X(), ppv.Y(), ppv.Z());
	pos = NormalizedPosition(globalPntID);
	m_Pnts2Nodes.Apply(pos);
	
	R3Vector a1 = axes[0];
	R3Vector a2 = axes[1];
	R3Vector a3 = axes[2];
	m_Pnts2Nodes.Apply(a1);
	m_Pnts2Nodes.Apply(a2);
	m_Pnts2Nodes.Apply(a3);
	
	R3Triad pntOrient(a1,a2,a3);
	pntOrient.Normalize();
	return pntOrient;
}
*/

R3Point TMatch::AlignedToNodePointPosition(int nodeID, int gPntID, int configID) const
{
	assert(nodeID>=0);
	assert(m_pPnts!=NULL && m_AllPntsRadius>0);
	R3Point pos = m_pPnts->PointPosition(gPntID);
	R3Vector ppv = (pos-m_AllPntsCentroid) / m_AllPntsRadius;
	pos = R3Point(ppv.X(), ppv.Y(), ppv.Z());
	m_Pnts2Nodes.Apply(pos);
	if (configID<0)
		return pos;
	
	//return NodeEllipsoid(nodeID).Deformation()->Deformed2Orig(pos);

	// special case: NULL nodes
	int alignToNode = nodeID;
	const STemplateNode * node = m_pSTemplate->GetNode(Deformation(nodeID).NodeID());
	if (node->GetPart(configID)->IsType(PartBox::BOXTYPE_NullBox))
	{
		((TMatch*)this)->AlignNullNodes();
		std::map<int, int>::iterator iter = ((TMatch*)this)->m_NullToAlignment.find(gPntID);
		assert(iter != m_NullToAlignment.end());
		alignToNode = iter->second;
	}
	
	// a bit inefficient (also non-robust) - should be stored in 'UndeformedCentroid()'
	node = m_pSTemplate->GetNode(Deformation(alignToNode).NodeID());
	
	int flags, ndims;
	int nparts = Deformation(alignToNode).NParts();
	int partID = Deformation(alignToNode).PartID();
	const double * xform = node->PartOfParent()->GetMeanTransform(configID, nparts, partID, flags, ndims);
	R3Affine aff = PartTransforms::CreateAffineTransform(xform, flags);
	
	R3Point nc0 = R3Point(0,0,0);
	aff.Apply(nc0);
	R3Triad no0 = NodeEllipsoid(alignToNode).UndeformedOrientation();
	R3Vector ns0 = NodeEllipsoid(alignToNode).UndeformedScale();

	R3Point nc1 = Deformation(alignToNode).Orig2Deformed(R3Point(0,0,0));
	R3Triad no1 = Deformation(alignToNode).Orig2Deformed(no0);
	no1.Normalize();
	R3Vector ns1 = Deformation(alignToNode).Orig2Deformed(ns0);

	R3Vector p1 = pos - nc1;
	p1 = R3Vector(p1.Dot(no1[0]) / ns1[0], p1.Dot(no1[1]) / ns1[1], p1.Dot(no1[2]) / ns1[2]);
	
	pos = nc0 + no0[0] * p1[0] * ns0[0] + no0[1] * p1[1] * ns0[1] + no0[2] * p1[2] * ns0[2];
		
	return pos;
}

int TMatch::NTemplatePoints() const
{
	return m_iNTemplateSamples;
}

int TMatch::NPointsOnTemplateNode(int nodeid) const
{
	assert(nodeid>=0 && nodeid < m_FinalEllipsoids.size());
	return m_FinalEllipsoids[nodeid].NTemplatePointsOnNode(); 
}

R3Point TMatch::Template2PntCloud(const R3Point & tpnt) const
{
	R3Point ppnt = tpnt;
	m_Pnts2Nodes.ApplyInverse(ppnt);
	ppnt = m_AllPntsRadius * ppnt + m_AllPntsCentroid;
	return ppnt;
}

const STemplate * TMatch::GetSTemplate() const
{
	return m_pSTemplate;
}

void TMatch::SetSeedID(int seedID)
{
	m_iSeedID = seedID;
}

int TMatch::SeedID() const
{
	return m_iSeedID;
}

const R3Affine & TMatch::SeedAlignment() const
{
	return m_SeedAlignment;
}

// per point labels
void TMatch::RemoveDeform(int deformID)
{
	assert(m_TemplateMap.size()==m_FinalEllipsoids.size());
	assert(deformID>=0 && deformID < (int)m_FinalEllipsoids.size());
	
	if (m_TemplateMap[deformID]->IsIncluded())
	{
		std::cout<<"[WARNING] Cannot remove deformation: "<<deformID<<" it's included"<<std::endl;
		return;
	}
	delete m_TemplateMap[deformID];
	m_TemplateMap.erase(m_TemplateMap.begin()+deformID);
	m_FinalEllipsoids.erase(m_FinalEllipsoids.begin()+deformID);
}


int TMatch::NShapePrimitives() const
{
	return m_iSubsetSize;
}

int TMatch::Loc2GlobPid(int localPntID) const
{
	assert(m_Subset2AllPnts!=NULL);
	return m_Subset2AllPnts[localPntID];
}

int TMatch::LocalNodeID(int globalPntID) const
{
//	assert(m_pPntLabels!=NULL);
//	const ShapePoint & pnt = m_pPnts->GetPoint(globalPntID);
//	return (int)m_pPntLabels->Value(pnt, 0)-1;
	assert(globalPntID < m_aPnts2Segments.size());
	return m_aPnts2Segments[globalPntID]-1;
}

int TMatch::Glob2LocPid(int globalPntID) const
{
	return m_AllPnts2Subset[globalPntID];
}

const ShapeLabels * TMatch::GetPerPointLabels() const
{
	return m_pPntLabels;
}


// template configuration
int TMatch::NTemplateNodes() const
{
	return (int)m_TemplateMap.size();
}

const R3Affine & TMatch::GlobalXForm() const
{
	return m_Pnts2Nodes;
}

const R3Vector & TMatch::GetRotation() const
{
	return m_fGlobalRotate;
}
 
const R3Vector & TMatch::GetTranslation() const
{
	return m_fGlobalTranslate;
}

const R3Vector & TMatch::GetScale() const
{
	return m_fGlobalScale;
}

double TMatch::GetAllPointsRadius() const
{
	return m_AllPntsRadius;
}
const R3Point & TMatch::GetAllPointsCentroid() const
{
	return m_AllPntsCentroid;
}


const TemplateNodeDeform & TMatch::Deformation(int localNodeID) const
{
	assert(localNodeID>=0 && localNodeID < (int)m_TemplateMap.size());
	return *(m_TemplateMap[localNodeID]);
}

const PartBox & TMatch::NodeEllipsoid(int localNodeID) const
{
	assert(localNodeID>=0 && localNodeID < (int)m_FinalEllipsoids.size());
	return m_FinalEllipsoids[localNodeID];
}

// set labels
bool TMatch::SetLabel(int localPntID, int localNodeID)
{
	localNodeID++;
	assert(Loc2GlobPid(localPntID) < (int)m_aPnts2Segments.size());
	bool diff = (m_aPnts2Segments[Loc2GlobPid(localPntID)]!=localNodeID);
	m_aPnts2Segments[Loc2GlobPid(localPntID)] = localNodeID;
	return diff;
}

void TMatch::UpdateLabels(bool excludeNodes)
{
	if (m_pPnts==NULL)
		return;
	
	if (m_pPntLabels==NULL)
	{
		// create labels object
		VKString labelsName = "PntsSegmentation_1";
		int counter = 2;
		while (m_pPnts->GetShapeVal(labelsName)!=NULL)
		{
			labelsName = VKString("PntsSegmentation_")+VKString::number(counter);
			counter++;
		}
		m_pPntLabels = new ShapeLabels(m_pPnts, labelsName);
		if (m_pPnts->GetShapeVal(labelsName)==NULL)
			m_pPnts->AddShapeVal(labelsName, m_pPntLabels);

		// create tree hierarchy (root -> segments)
		std::vector<SceneGraphLabel*> labelsTree;
		
		SceneGraphLabel * root = new SceneGraphLabel();
		root->m_sLabel = "OtherPoints";
		root->m_iID = (int)labelsTree.size();
		root->m_pParent = NULL;
		labelsTree.push_back(root);

		for (int i=0; i<(int)m_TemplateMap.size(); i++)
		{
			const STemplateNode * tNode = m_pSTemplate->GetNode(m_TemplateMap[i]->NodeID());
			
			SceneGraphLabel * sgNode = new SceneGraphLabel();
			sgNode->m_sLabel = tNode->NodeName();
			sgNode->m_iID = (int)labelsTree.size();
			sgNode->m_pParent = root;
			labelsTree.push_back(sgNode);
			
			root->m_apChildren.push_back(sgNode);
		}
		m_pPntLabels->SetAllLabels(labelsTree);
		
		// provide initial labeling (just assign everything to a root)
		assert(m_pPnts!=NULL);
		if (m_aPnts2Segments.size()==0)
		{
			for (int i=0; i<m_pPnts->NumPoints(); i++)
				m_aPnts2Segments.push_back(0);	// everything is at a root
		}
		else
			assert(m_pPnts->NumPoints()==(int)m_aPnts2Segments.size());
	}
	
	assert(m_pPntLabels!=NULL);
	m_pPntLabels->SetPntsSegments(m_aPnts2Segments);
	
	// set excluded nodes
	//excludeNodes = true;	// do not exclude
	if (excludeNodes)
	{
		std::vector<int> npoints;
		for (int i=0; i<NTemplateNodes(); i++)
			npoints.push_back(0);
		for (int i=0; i<NShapePrimitives(); i++)
		{
			int segID = LocalNodeID(Loc2GlobPid(i));
			assert(segID>=0 && segID < (int)npoints.size());
			npoints[segID]++;
		}
			
		VKString exclStr = "";
		for (int i=0; i<(int)m_TemplateMap.size(); i++)
		{
			if (npoints[i]==0 && !m_FinalEllipsoids[i].IsType(PartBox::BOXTYPE_NullBox))
			{
				m_TemplateMap[i]->SetIncluded(false);
				if (!m_FinalEllipsoids[i].IsType(PartBox::BOXTYPE_EmptyBox))
				{
					assert(m_FinalEllipsoids[i].Deformation()==m_TemplateMap[i]);
					const STemplateNode * cNode = m_pSTemplate->GetNode(m_TemplateMap[i]->NodeID());
					int nparts = m_TemplateMap[i]->NParts();
					UpdateEllipsoid(i);
					exclStr += cNode->NodeName()+" ";
				}
			}
			else
				m_TemplateMap[i]->SetIncluded(true);
		}
		if (exclStr!="")
		{
			std::cout<<"  Excluding: "<<exclStr.c_str()<<std::endl;
			std::cout<<"   I=---";
//			assert(m_TemplateMap[i]->IsIncluded()==false);
		}
		
	}
	
	// cache node clusters
	// initialize
	int nNodes = NTemplateNodes();
	if ((int)m_NodeToGlobPntIDs.size()!=nNodes)
	{
		m_NodeToGlobPntIDs.clear();
		for (int i=0; i<nNodes; i++)
			m_NodeToGlobPntIDs.push_back(std::vector<int>());
	}
	else
	{
		for (int i=0; i<nNodes; i++)
			m_NodeToGlobPntIDs[i].clear();
	}
	
	// cache
	int npnts = NShapePrimitives();
	for (int i=0; i<npnts; i++)
	{
		int nodeID = LocalNodeID(Loc2GlobPid(i));
		if (nodeID==-1)
			continue;
		assert(nodeID>=0 && nodeID < nNodes);
		m_NodeToGlobPntIDs[nodeID].push_back(Loc2GlobPid(i));
	}
	
	// cached kdTrees
	FreeKdTrees();

	// update energy
	for (int i=0; i<(int)m_EnergyUpdates.size(); i++)
		m_EnergyUpdates[i]->UpdateLabeling();	
}

void TMatch::UpdateEllipsoids()
{
	// find final ellipsoids in space
	assert(m_TemplateMap.size() == m_FinalEllipsoids.size());
	for (int i=0; i<(int)m_FinalEllipsoids.size(); i++)
		UpdateEllipsoid(i);
}

void TMatch::UpdateEllipsoid(int localNodeID)
{
	assert(m_TemplateMap.size() == m_FinalEllipsoids.size());
	assert(localNodeID>=0 && localNodeID < (int)m_TemplateMap.size());
	m_pSTemplate->DeformEllipsoid(*(m_TemplateMap[localNodeID]), m_FinalEllipsoids[localNodeID]);
	// update energy
	for (int i=0; i<(int)m_EnergyUpdates.size(); i++)
	{
		m_EnergyUpdates[i]->UpdateNodeXForm(localNodeID);
		m_EnergyUpdates[i]->UpdateNodeShape(localNodeID);
	}
}

void TMatch::SetTransform(int localNodeID, const double * transform)
{
	assert(localNodeID>=0 && localNodeID < (int)m_TemplateMap.size());
//	for (int i=0; i<m_TemplateMap[localNodeID].m_iXFormNDim; i++)
//		m_TemplateMap[localNodeID].m_fTransform[i] = transform[i];
	m_TemplateMap[localNodeID]->SetTransform(transform);
	
	// update energy
	for (int i=0; i<(int)m_EnergyUpdates.size(); i++)
		m_EnergyUpdates[i]->UpdateNodeXForm(localNodeID);	
}

void TMatch::SetPCAScale(int localNodeID, const R3Vector & scale)
{
	assert(localNodeID>=0 && localNodeID < (int)m_TemplateMap.size());
	//m_TemplateMap[localNodeID].m_PCAScale = scale;
	m_TemplateMap[localNodeID]->SetScale(scale);
	
	// update energy
	for (int i=0; i<(int)m_EnergyUpdates.size(); i++)
		m_EnergyUpdates[i]->UpdateNodeShape(localNodeID);	
}

void TMatch::SetDirtyFlags()
{
	for (int i=0; i<(int)m_TemplateMap.size(); i++)
		m_TemplateMap[i]->SetDirtyFlag();
}

void TMatch::SetGlobalTranslate(const R3Vector & translate)
{
	m_fGlobalTranslate = translate;
	RecomputeGlobalAlignment();
}

void TMatch::SetGlobalRotate(const R3Vector & rotateRadians)
{
	//assert(rotateRadians[0]==0 && rotateRadians[1]==0);	// assume rotation around z-axis only
	m_fGlobalRotate = rotateRadians;
	RecomputeGlobalAlignment();
}

void TMatch::SetGlobalScale(const R3Vector & scale)
{
	assert(scale[0]==scale[1] && scale[1]==scale[2] && scale[0]>0);	// assume isometric scaling
	m_fGlobalScale = scale;
	RecomputeGlobalAlignment();
}

void TMatch::SetGlobalXForm(const R3Affine & xform)
{
	m_Pnts2Nodes = xform;
	for (int i=0; i<(int)m_EnergyUpdates.size(); i++)
		m_EnergyUpdates[i]->UpdateGlobalXForm();
	//FreeKdTrees();
}

void TMatch::RecomputeGlobalAlignment()
{
	m_Pnts2Nodes = m_SeedAlignment;
	m_Pnts2Nodes.Rotate(m_fGlobalRotate);
	//m_Pnts2Nodes.Scale(m_fGlobalScale);	
	//m_Pnts2Nodes.Translate(m_fGlobalTranslate/m_fGlobalScale);
	m_Pnts2Nodes.Translate(m_fGlobalTranslate);
	
	// update energy
	for (int i=0; i<(int)m_EnergyUpdates.size(); i++)
		m_EnergyUpdates[i]->UpdateGlobalXForm();	
}

void TMatch::WriteMatch(FILE * file) const
{
	TMatchVersion ver = TMatchVer1;
	fwrite(&ver, sizeof(TMatchVersion), 1, file);

	// seed: id and aligning transformation
	fwrite(&m_iSeedID, sizeof(int), 1, file);
	bool mirr = m_SeedAlignment.IsMirrored();
	double m[16];
	
	for (int i=0; i<4; i++)
	for (int j=0; j<4; j++)	
		m[i*4 + j] = m_SeedAlignment.Matrix()[i][j];
	
	fwrite(&mirr, sizeof(bool), 1, file);
	fwrite(m, sizeof(double), 16, file);

	// geometry: model and geometry name, subset		
	VKString::WriteString(m_sModelName, file);
	VKString::WriteString(m_sGeomName, file);
	fwrite(&m_iNPoints, sizeof(int), 1, file);
	
	if (m_iNPoints>0)
	{
		fwrite(&m_iSubsetSize, sizeof(int), 1, file);
		if (m_iSubsetSize!=m_iNPoints)
			fwrite(m_AllPnts2Subset, sizeof(int), m_iSubsetSize, file);
	}
	
	// write template configuration parameters
		// global alignment
//	double v[9];
//	v[0] = m_fGlobalRotate[0];		v[1] = m_fGlobalRotate[1];		v[2] = m_fGlobalRotate[2];
//	v[3] = m_fGlobalScale[0];		v[4] = m_fGlobalScale[1];		v[5] = m_fGlobalScale[2];
//	v[6] = m_fGlobalTranslate[0];	v[7] = m_fGlobalTranslate[1];	v[8] = m_fGlobalTranslate[2];
//	fwrite(v, sizeof(double), 9, file);

	mirr = m_Pnts2Nodes.IsMirrored();
	for (int i=0; i<4; i++)
	for (int j=0; j<4; j++)
		m[i*4 + j] = m_Pnts2Nodes.Matrix()[i][j];
	fwrite(&mirr, sizeof(bool), 1, file);
	fwrite(m, sizeof(double), 16, file);

		// per node deformation
	int ndeforms = (int)m_TemplateMap.size();
	fwrite(&ndeforms, sizeof(int), 1, file);	
	for (int i=0; i<ndeforms; i++)
		m_TemplateMap[i]->WriteNodeDeform(file);
	
	if (m_iNPoints>0)
	{
		// write labeling
		assert((int)m_aPnts2Segments.size() == m_iNPoints);
		for (int i=0; i<m_iNPoints; i++)
			fwrite(&(m_aPnts2Segments[i]), sizeof(int), 1, file);
	}
}

TMatch * TMatch::ReadMatch(FILE * file, const STemplate * st, ModelDatabase * db)
{
	// seed: id and aligning transformation
	TMatchVersion ver;
	fread(&ver, sizeof(TMatchVersion), 1, file);
	if (ver!=TMatchVer1)
		std::cout<<"[ERROR] invalid version ID: "<<ver<<" vs curr: "<<TMatchVer1<<std::endl;
	assert(ver == TMatchVer1);
	
	int seedID = 0;
	fread(&seedID, sizeof(int), 1, file);
	bool mirr;
	double m[16];
	fread(&mirr, sizeof(bool), 1, file);
	fread(m, sizeof(double), 16, file);
	R3Affine seedAlignment(R4Matrix(m), mirr);
	
	// geometry: model and geometry name, subset
	int nPnts = 0;	
	VKString mName = VKString::ReadString(file);
	VKString gName = VKString::ReadString(file);

	fread(&nPnts, sizeof(int), 1, file);
		
	int subsetSize = 0;
	std::vector<int> subset;
	if (nPnts>0)
	{
		fread(&subsetSize, sizeof(int), 1, file);
		if (subsetSize!=nPnts)
		{
			double * subsetArray = new double[subsetSize];
			fread(subsetArray, sizeof(int), subsetSize, file);
			for (int i=0; i<subsetSize; i++)
				subset.push_back(i);
			delete [] subsetArray;
		}
	}
	
	// read template configuration parameters
	// global alignment
//	R3Vector globRotate(0,0,0);
//	R3Vector globScale(0,0,0);
//	R3Vector globTranslate(0,0,0);
	R3Affine globXForm;
//	if (version < STemplate::STMPLT_V0_4)
//	{
//		double v[9];
//		fread(v, sizeof(double), 9, file);
//		globRotate = R3Vector(v[0], v[1], v[2]);
//		globScale = R3Vector(v[3], v[4], v[5]);
//		globTranslate = R3Vector(v[6], v[7], v[8]);
//	}
//	else
//	{
		fread(&mirr, sizeof(bool), 1, file);
		fread(m, sizeof(double), 16, file);
		globXForm = R3Affine(R4Matrix(m), mirr);
//	}
	
	// per node deformation
	int ndeforms;
	fread(&ndeforms, sizeof(int), 1, file);
	double scale[3];	
	std::vector<TemplateNodeDeform *> nodeDeformations;
	for (int i=0; i<ndeforms; i++)
		nodeDeformations.push_back(TemplateNodeDeform::ReadNodeDeform(file));

	TMatch * newMatch = new TMatch(st, seedID, NULL, subset, seedAlignment, nodeDeformations);
	
	newMatch->m_sModelName = mName;
	newMatch->m_sGeomName = gName;
	newMatch->m_iNPoints = nPnts;
//	if (version < STemplate::STMPLT_V0_4)
//	{
//		newMatch->SetGlobalTranslate(globTranslate);
//		newMatch->SetGlobalRotate(globRotate);
//		newMatch->SetGlobalScale(globScale);
//		newMatch->RecomputeGlobalAlignment();
//	}
//	else
//	{
	newMatch->SetGlobalXForm(globXForm);
//	}
	if (subset.size()==0)
	{
		newMatch->m_iSubsetSize = nPnts;
		newMatch->m_Subset2AllPnts = new int[nPnts];
		newMatch->m_AllPnts2Subset = new int[nPnts];
		for (int i=0; i<nPnts; i++)
		{
			newMatch->m_Subset2AllPnts[i] = i;
			newMatch->m_AllPnts2Subset[i] = i;
		}
	}
	
	if (mName!="none" && db!=NULL)	// can be configuration only without geometry
		newMatch->ReadGeometry(db);
	
	// read labeling
	int * pntsSegs = NULL;
	if (nPnts>0)
	{
		pntsSegs = new int[nPnts];
		fread(pntsSegs, sizeof(int), nPnts, file);
	}
	
	if (newMatch->m_pPnts!=NULL)
	{
		assert((int)newMatch->m_aPnts2Segments.size()==nPnts && pntsSegs!=NULL);
		for (int i=0; i<nPnts; i++)
			newMatch->m_aPnts2Segments[i] = pntsSegs[i];
	}
	else if (nPnts>0)
	{
		assert(newMatch->m_aPnts2Segments.size()==0);
		for (int i=0; i<nPnts; i++)
			newMatch->m_aPnts2Segments.push_back(pntsSegs[i]);
	}		
	
	if (pntsSegs!=NULL)
		delete [] pntsSegs;
		
	newMatch->UpdateEllipsoids();
	newMatch->UpdateLabels(false);
	return newMatch;
}

void TMatch::ReadGeometry(ShapePointCloud * pnts)
{
	m_pPnts = pnts;
	m_AllPntsCentroid = m_pPnts->Centroid();
	m_AllPntsRadius = m_pPnts->AverageRadius(m_AllPntsCentroid);
	
	UpdateLabels(false);
}

void TMatch::ReadGeometry(ModelDatabase * db)
{
	assert(db!=NULL);
	Model3D * model = db->GetModel(m_sModelName);
	if (model==NULL)
		std::cout<<"[ERROR] Cannot find model: "<<m_sModelName.c_str()<<std::endl;
	assert(model!=NULL);
	GeomShape * shape = model->GetGeomShape(m_sGeomName);
	if (shape==NULL)
		std::cout<<"[ERROR] Cannot find geometry "<<m_sGeomName.c_str()<<" in model "<<m_sModelName.c_str()<<std::endl;
	assert(shape!=NULL);
	m_pPnts = shape->SafeCastToPointCloud();
	assert(m_pPnts!=NULL);

	m_AllPntsCentroid = m_pPnts->Centroid();
	m_AllPntsRadius = m_pPnts->AverageRadius(m_AllPntsCentroid);

	UpdateLabels(false);
}

void TMatch::AddEnergyUpdates(TMatchEnergy * energy)
{
	m_EnergyUpdates.push_back(energy);
}

void TMatch::RemoveEnergyUpdates(TMatchEnergy * energy)
{
	for (int i=0; i<(int)m_EnergyUpdates.size(); i++)
		if (m_EnergyUpdates[i]==energy)
		{
			m_EnergyUpdates.erase(m_EnergyUpdates.begin() + i);
			i--;
		}
}

void TMatch::CreateTemplateSamples(int nsamples)
{
	int nExpectedSamples = nsamples;
	m_iNTemplateSamples = 0;
	
	if (nExpectedSamples < 0 && m_pPnts==NULL)
		nExpectedSamples = 1000;	// vk: const
	else if (nExpectedSamples < 0 && m_pPnts!=NULL)
		nExpectedSamples = m_pPnts->NumPoints();
	else
		assert(nExpectedSamples>0);

	double totalArea = 0;
	for (int i=0; i<(int)m_FinalEllipsoids.size(); i++)
		totalArea += m_FinalEllipsoids[i].Area();

	// sample ellipsoids
	for (int i=0; i<(int)m_FinalEllipsoids.size(); i++)
	{
		int nToCreate = (int)(m_FinalEllipsoids[i].Area() * nExpectedSamples / totalArea);
		m_FinalEllipsoids[i].CreateTemplateSamples(nToCreate);
	}
	m_PerNodeTemplateSamplesOffset.clear();
	m_GlobalTemplateIDToNodeID.clear();
	
	m_pTemplateSamples = new ShapePointCloud(GeomShape::SHAPE_TYPE_POINT_CLOUD, "allt", "none", NULL);
	
	// enumerate points and create point set
	for (int i=0; i<(int)m_FinalEllipsoids.size(); i++)
	{
		m_PerNodeTemplateSamplesOffset.push_back(m_iNTemplateSamples);
		for (int j=0; j<m_FinalEllipsoids[i].NTemplatePointsOnNode(); j++)
		{
			//R3Point  pos = m_FinalEllipsoids[i].UndeformedTemplatePointPosition(j);
			m_GlobalTemplateIDToNodeID.push_back(i);
			R3Point pos = DeformedTemplateSample(m_pTemplateSamples->NumPoints());
			m_pTemplateSamples->AddPoint(ShapePoint(pos, m_pTemplateSamples, m_pTemplateSamples->NumPoints()));
		}		
		m_iNTemplateSamples += m_FinalEllipsoids[i].NTemplatePointsOnNode();
	}
	
	// local shape descriptors
	for (int i=0; i<(int)m_FinalEllipsoids.size(); i++)
		m_FinalEllipsoids[i].ComputeLocalShapeDescriptors(m_pTemplateSamples, m_PerNodeTemplateSamplesOffset[i],
														  m_FinalEllipsoids[i].NTemplatePointsOnNode());
}

const double * TMatch::GetLocalPntsFeatures(int pntID, int matchNodeID, int & nfeat) const
{
	assert(matchNodeID==-1);
	// TODO: note that affine xform is more than pnts2nodes if node's orientation changes
	nfeat = GetPntsLocalGeometry()->NLocalFeatures();
	return GetPntsLocalGeometry()->LocalOrientAsFeatures(pntID, m_Pnts2Nodes);
}

const PartLocalGeometry * TMatch::GetPntsLocalGeometry() const
{
	if (m_pPntsLocalGeometry==NULL)
		((TMatch*)this)->m_pPntsLocalGeometry = new PartLocalGeometry(m_pPnts, 0, m_pPnts->NumPoints());
	return m_pPntsLocalGeometry;
}


void TMatch::AlignNullNodes()
{
	if (m_bAlignedNullNodes)
		return;	// compute only once

	m_bAlignedNullNodes = true;
	m_NullToAlignment.clear();
	m_Alignment.clear();

	// find null node id
	int nullNodeID = -1;
	for (int i=0; i<(int)m_FinalEllipsoids.size(); i++)
		if (m_FinalEllipsoids[i].IsType(PartBox::BOXTYPE_NullBox))
			nullNodeID = Deformation(i).NodeID();
	assert(nullNodeID>=0);
	
	// assign to nearest
	std::vector<TMatchEnergy*> energy;
	energy.push_back(new TMatchEnergy(this));
	
	int methods = ConsistentPartsAlg::CPM_InitGreedy_AssignToNonNull | ConsistentPartsAlg::CPM_ClusterTakeAll;
	methods |= ConsistentPartsAlg::CPM_AveClusterTakeAny;
	
	ConsistentPartsAlg * cPartsAlg = new ConsistentPartsAlg((STemplate*)m_pSTemplate, energy, methods, false);
	std::vector<PointsCluster> parts;
	cPartsAlg->ExtractConsistentParts(nullNodeID, 0, 1, parts);
	
	// take affine transformation from the nearest part...
	for (int i=0; i<(int)parts.size(); i++)
	{
		VKStringList splitLabel = parts[i].m_sLabel.split("_");
		assert(splitLabel.count()==6);
		assert(splitLabel[0] == "AttachedTo");
		bool ok;
		int localID = splitLabel[1].toInt(&ok);
		assert(ok);
		for (int j=0; j<(int)parts[i].m_aSubset.size(); j++)
			m_NullToAlignment[parts[i].m_aSubset[j]] = localID;
	}
	
	delete energy[0];
	delete cPartsAlg;
}

void TMatch::FreeKdTrees()
{
	for (int i=0; i<(int)m_apKdTreesForLabeledPnts.size(); i++)
	{
		if (m_apKdTreesForLabeledPnts[i]!=NULL)
			delete m_apKdTreesForLabeledPnts[i];
		m_apKdTreesForLabeledPnts[i] = NULL;
	}
}



