/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __TMatch_H
#define __TMatch_H

#include "Model3D.h"
#include "ShapeMesh.h"
#include "ShapeLabels.h"
#include "ShapePointCloud.h"
#include "ShapeLocalPCA.h"

#include "PartBox.h"
#include "TemplateNodeDeform.h"

class PartTransforms;
class STemplate;
class TMatchEnergy;


/**
 * This class describes two parameters: labeling of points and template configuration
 */
class TMatch
{
	public:
		enum TMatchVersion
		{
			TMatchVer1 = 2000
		};
	
		TMatch(const STemplate * st, int seedID, GeomShape * shape, const std::vector<int> & subset, 
			   const R3Affine & pnt2nodes, const std::vector<TemplateNodeDeform *> & nodeDeformations);
		virtual ~TMatch();
	
		virtual TMatch * CreateCopy(const std::vector<TemplateNodeDeform *> & additionalNodes
									= std::vector<TemplateNodeDeform *>(), const STemplate * newTemplate = NULL,
									std::map<int, int> * old2newNode = NULL, int configoffset = 0) const;
	// matched geometry (also get aligned point)
		virtual const ShapePointCloud * GetMatchedGeomAsPnts() const;
		virtual const VKString & ModelName() const;
		virtual const VKString & GeomName() const;
	
		virtual int NPointsLabeledAsNode(int nodeid) const;
		virtual int GetPointLabeledAsNode(int nodeid, int pntid) const;
		virtual R3Point NormalizedPosition(int gpntid) const;
		//virtual R3Triad AlignedPoint(int gPntID, R3Point & pos, double pntEigs[3]) const;

		virtual R3Point AlignedPointPosition(int gPntID) const;
		// NOTE: aligns to DEFORMED template
		//virtual R3Triad AlignedToNodeLocalOrientation(int nodeID, int gPntID, R3Point & pos, double pntEigs[3]) const;
		virtual R3Point AlignedToNodePointPosition(int nodeID, int gPntID, int configID) const;

	// get sampling of a template
		virtual int NTemplatePoints() const;
		virtual int NPointsOnTemplateNode(int nodeID) const;
		virtual int GlobalTmpltPnt(int nodeID, int pntID) const;
		virtual void LocalPntID(int globalID, int & nodeID, int & localID) const;
		virtual R3Point DeformedTemplateSample(int globalTmpltPnt) const;
			// NOTE: function below aligns part TO THE ORIGIN!
			// use STemplate->TemplatePointInCanonicalSpace if you want to co-align templates
		virtual R3Point UndeformedTemplateSample(int globalTmpltPnt) const;
	
		virtual R3Point Template2PntCloud(const R3Point & tpnt) const; 
	
	// template
		virtual const STemplate * GetSTemplate() const;
		virtual void SetSeedID(int seedID);	// NOTE: should only be used by classes that create new seeds
		virtual int SeedID() const;
		virtual const R3Affine & SeedAlignment() const;
		virtual void RemoveDeform(int deformID);// NOTE: should only be used by classes that create new seeds
			
	// per point labels
		virtual int NShapePrimitives() const;
		virtual int Loc2GlobPid(int localPntID) const;
		virtual int LocalNodeID(int globalPntID) const;
		virtual int Glob2LocPid(int globalPntID) const;
		virtual const ShapeLabels * GetPerPointLabels() const;
		virtual int NearestPointWithLabel(int nodeID, const R3Point & posNormalizedPntsSpace) const;
	
	// template configuration
		// global 
		virtual const R3Affine & GlobalXForm() const;
		virtual const R3Vector & GetRotation() const;
		virtual const R3Vector & GetTranslation() const;
		virtual const R3Vector & GetScale() const;
		virtual double GetAllPointsRadius() const;
		virtual const R3Point & GetAllPointsCentroid() const;
	
		// per node
		virtual int NTemplateNodes() const;
		virtual const TemplateNodeDeform & Deformation(int localNodeID) const;
		virtual const PartBox & NodeEllipsoid(int localNodeID) const;	// pre-computed given current transform
		virtual int NearestTemplatePointGlobalID(int nodeID, const R3Point & posDeformedTemplateSpace) const;

	// for matching algorithm (optimization) and for initialiation 
		// set labels
		virtual bool SetLabel(int localPntID, int localNodeID);	// returns true if labels changed
		virtual void UpdateLabels(bool excludeNodes);
	
		// set template deformation
		virtual void UpdateEllipsoids();
		virtual void UpdateEllipsoid(int localNodeID);
		//virtual void AddPotentialNewNode(const PotentialNewNode & newNode);
	
		// explicitly setting deformation (should only be used in TMatchAlg)
		virtual void SetTransform(int localNodeID, const double * transform);
		virtual void SetPCAScale(int localNodeID, const R3Vector & scale);
		virtual void SetDirtyFlags();
		// global alignment
		virtual void SetGlobalXForm(const R3Affine & xform);
		virtual void SetGlobalTranslate(const R3Vector & translate);
		virtual void SetGlobalRotate(const R3Vector & rotateRadians);
		virtual void SetGlobalScale(const R3Vector & scale);
		
		// read / write match: all deformation parameters + labeling
		virtual void WriteMatch(FILE * file) const;
		static TMatch * ReadMatch(FILE * file, const STemplate * st, ModelDatabase * db);
		virtual void ReadGeometry(ModelDatabase * db);
		virtual void ReadGeometry(ShapePointCloud * pnts);
		
		// corresponding energy term (for updates ONLY)
		virtual void AddEnergyUpdates(TMatchEnergy * energy);
		virtual void RemoveEnergyUpdates(TMatchEnergy * energy);
	
		// sample template deformation
		virtual void CreateTemplateSamples(int nsamples = -1);

		// get local geometries
		virtual const PartLocalGeometry * GetPntsLocalGeometry() const;
		virtual const double * GetLocalPntsFeatures(int pntID, int matchNodeID, int & nfeat) const;
	
	protected:
	// original template
		const STemplate * m_pSTemplate;
		int m_iSeedID;
		R3Affine m_SeedAlignment;

	// matched geometry (subset of points)
		ShapePointCloud * m_pPnts;
		PartLocalGeometry * m_pPntsLocalGeometry;
		double m_AllPntsRadius;
		R3Point m_AllPntsCentroid;
		VKString m_sModelName;
		VKString m_sGeomName;
		int m_iNPoints;
	
		int * m_AllPnts2Subset;
		int * m_Subset2AllPnts;
		int m_iSubsetSize;

	// deformation parameters:
		// global alignment
		R3Vector m_fGlobalTranslate;
		R3Vector m_fGlobalRotate;
		R3Vector m_fGlobalScale;
		virtual void RecomputeGlobalAlignment();
		R3Affine m_Pnts2Nodes;
	
		// list of existing nodes (mapped to some global node in a template)
		std::vector<TemplateNodeDeform * > m_TemplateMap;
		std::vector<PartBox> m_FinalEllipsoids;
		ShapePointCloud * m_pTemplateSamples;	// TODO: cleanup when deleting data
	
		// new nodes
		//std::vector<PotentialNewNode> m_PotentialNewNodes;
	
	// labeling parameters:
		std::vector<int> m_aPnts2Segments;
		ShapeLabels * m_pPntLabels;
		std::vector<std::vector<int> > m_NodeToGlobPntIDs;
		virtual void FreeKdTrees();
		std::vector<R3Kdtree<const PointInSet *> * > m_apKdTreesForLabeledPnts;
		//std::vector<std::vector<int> > m_PerNodeTemplateSamplesNearestGPnt;

	
	// corresponding energy terms for updates
		std::vector<TMatchEnergy*> m_EnergyUpdates;
	
	// sampling of template functions
		int m_iNTemplateSamples;	// set to be same as point samples
		std::vector<int> m_PerNodeTemplateSamplesOffset;
		std::vector<int> m_GlobalTemplateIDToNodeID;
	
	// EXPERIMENTAL
		void AlignNullNodes();
		bool m_bAlignedNullNodes;
		std::map<int, int> m_NullToAlignment;
		std::vector<R3Affine> m_Alignment;	
};

#endif
