/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "TMatchEnergy.h"
#include "TMatchAlg.h"

#include "VkFunctions.h"

#include "STemplate.h"
#include "STemplateNode.h"
#include "PartOfEdge.h"
#include "ShapeManifoldMesh.h"
#include "DistGeodesicDijkstra.h"

std::map<const ShapeMesh *, VKGraph * > TMatchEnergy::m_Mesh2Graph;
int GlobalConstants::s_iMinNSamplesForVariance = 0;
int GlobalConstants::s_iMinNSamplesForMean = 0;

bool TMatchEnergy::m_bPrintDebug = false;

//int TMatchEnergy::GetNhdSizeConstant()
//{
//	assert(false);
//	return 50;		// vk: const
//}

void TMatchEnergy::Compatible(TMatch * m1, TMatch * m2, bool & sameParts, bool & sameIncluded)
{
	sameParts = true;
	sameIncluded = true;
	if (m1->NTemplateNodes() != m2->NTemplateNodes())
	{
		sameParts = false;
		sameIncluded = false;
		return;
	}

	// NOTE: should only learn if same parts are included.
	// otherwise updates might be uneven (e.g. arm-rests are OFF for chairs vs benches due to uneven averaging...
	for (int d1=0; d1 < m1->NTemplateNodes(); d1++)
	{
		const TemplateNodeDeform & deform1 = m1->Deformation(d1);
		const TemplateNodeDeform & deform2 = m2->Deformation(d1);

		if (deform1.IsIncluded() != deform2.IsIncluded())
			sameIncluded = false;
		if (deform1.NodeID() != deform2.NodeID() || deform1.PartID() != deform2.PartID()
			|| deform1.NParts() != deform2.NParts())
		{
			sameParts = false;
			sameIncluded = false;
			break;
		}
	}
}

double TMatchEnergy::DeformationDistance(TMatch * m1, TMatch * m2, bool &sameParts, bool &sameIncluded)
{
	// old deformation penalty
//	Compatible(m1, m2, sameParts, sameIncluded);
//	if (!sameParts)
//		return RN_INFINITY;
//	
//	double penalty = 0;
//
//	assert(m1->NTemplateNodes()==m2->NTemplateNodes());
//	for (int d1=0; d1 < m1->NTemplateNodes(); d1++)
//	{
//		int d2 = d1;
//		const TemplateNodeDeform & deform1 = m1->Deformation(d1);
//		const TemplateNodeDeform & deform2 = m2->Deformation(d2);
//		
//		if (deform1.IsIncluded()!=deform2.IsIncluded())
//			continue;
//		
//		//const R3Vector & GetScale() const;
//		int dims1, flags1;
//		const double * t1 = deform1.GetTransform(dims1, flags1);
//		int dims2, flags2;
//		const double * t2 = deform2.GetTransform(dims2, flags2);
//		assert(dims1==dims2 && flags1==flags2);
//		assert(flags1==PartTransforms::PARTXFORM_Translation && dims1==3);
//		
//		const R3Vector & s1 = deform1.GetScale();
//		const R3Vector & s2 = deform2.GetScale();
//		
//		double xformPenalty =0;
//		for (int i=0; i<dims1; i++)
//			xformPenalty += pow(t1[i]-t2[i], 2);
//		
//		double scalePenalty = 0;
//		for (int i=0; i<3; i++)
//			scalePenalty += pow(s1[i]-s2[i], 2);
//		
//		penalty = vkMax(penalty, sqrt(scalePenalty)+sqrt(xformPenalty));
//	}
//	return penalty;
	
 	// new deformation distance
	Compatible(m1, m2, sameParts, sameIncluded);
	if (!sameParts)
		return RN_INFINITY;

	// points distance
	double distPenalty = 0;
	assert(m1->NTemplateNodes()==m2->NTemplateNodes());
	distPenalty += PntsDistances(m1, m2);
	distPenalty += PntsDistances(m2, m1);
	
	// descriptor distance
	double featPenalty = 0;
	featPenalty += LocalDescriptorDistance(m1, m2);
	featPenalty += LocalDescriptorDistance(m2, m1);
	
	return 1000 * distPenalty/2 + 2. * featPenalty;
}

double TMatchEnergy::PntsDistances(TMatch * m1, TMatch * m2)
{
	double nullPenalty = 1.; // vk: const
	
	int norm = 0;
	double pntDist = 0;

	for (int d1=0; d1 < m1->NTemplateNodes(); d1++)
	for (int p=0; p < m1->NPointsOnTemplateNode(d1); p++)
	{
		R3Point pos = m1->DeformedTemplateSample(m1->GlobalTmpltPnt(d1, p));
		int m2p = m2->NearestTemplatePointGlobalID(d1, pos);
		if (m2p==-1)
			pntDist += nullPenalty;
		else
			pntDist += R3SquaredDistance(pos, m2->DeformedTemplateSample(m2p));
		norm++;
	}
	return pntDist / norm;
}

double TMatchEnergy::LocalDescriptorDistance(TMatch * m1, TMatch * m2)
{
	int npnts = 0;
	double penalty = 0;
	double scale = 1.;
	
	// check if one is template
	TMatch * tmpltMatch = NULL;
	TMatch * pntsMatch = NULL;
	if (m1->GetMatchedGeomAsPnts()==NULL)
	{
		tmpltMatch = m1;
		pntsMatch = m2;
	}
	else if (m2->GetMatchedGeomAsPnts()==NULL)
		return 0;
	
	if (m1->GetMatchedGeomAsPnts()!=NULL && m2->GetMatchedGeomAsPnts()!=NULL)
		scale = .5;
	
	// compute energies
	if (tmpltMatch!=NULL)	// measure energy w.r.t. to the template
	{
		assert(pntsMatch->GetMatchedGeomAsPnts()!=NULL);
		const STemplate * st = tmpltMatch->GetSTemplate();
		for (int i=0; i<pntsMatch->NShapePrimitives(); i++)
		{
			const double * logProb = st->LogProbsLocalDescriptors(pntsMatch->Loc2GlobPid(i), pntsMatch);
			penalty += logProb[pntsMatch->LocalNodeID(i)];
			npnts++;
		}
	}
	else	// measure energy w.r.t. to the average of two models
	{
		assert(m1->GetMatchedGeomAsPnts()!=NULL && m2->GetMatchedGeomAsPnts()!=NULL);

		// suppose m1 serves as the template - train classifier
		std::vector<const PartLocalGeometry *> shapes;
		LocalGeometryTrainingSet * trainingSet = STemplate::CreateTrainingSet(m1, shapes);
		LocalGeometryClassifier classifier(*trainingSet);

		// for each point - check its quality w.r.t. to the classifier;
		for (int i=0; i<m2->NShapePrimitives(); i++)
		{
			int nfeat;
			const double * feat = m2->GetLocalPntsFeatures(m2->Loc2GlobPid(i), -1, nfeat);
			const double * logProb = classifier.LogProbs(nfeat, feat);
			penalty += logProb[m2->LocalNodeID(i)];
			npnts++;
		}
		
		// cleanup. 	// vk: memory leak
//		delete trainingSet;
//		for (int i=0; i<(int)shapes.size(); i++)
//			delete shapes[i];
	}
	// check shape descriptor distance
	assert(npnts>0);
	return scale * penalty / npnts;
}

double TMatchEnergy::DeformationDistanceThreshold()
{
	// NOTE: setting to lower value: better results, more templates, i.e. slower ( ≈ 50 )
	// NOTE: setting to higher value: less templates, i.e. faster ( ≈ 75 )
	return 75;	// vk: const
	//return 50;	// vk: const
}

int TMatchEnergy::DefaultMatchQualityEnergy()
{
	int testFlags =ET_Pnts | ET_PntEdges | ET_NodeExists | ET_NodeXForm | ET_NodeShape | ET_NodeFit2Pnts;
	testFlags = testFlags | ET_PntsDistance | ET_PntsLocalClass | ET_DistsModelSpace ;
	//testFlags = testFlags | ET_SameWeightParts;	// new stuff - not sure...
	return testFlags;
}

TMatchEnergy::TMatchEnergy(TMatch * match)
{
	// ET_Pnts | ET_PntEdges | ET_NodeExists | ET_NodeXForm | ET_NodeShape | ET_NodeFit2Pnts;
	// THIS IS WHAT SHOULD BE HERE:
	m_EnergyFlags = (ET_Pnts | ET_PntEdges | ET_NodeExists | ET_NodeXForm | ET_NodeShape | ET_NodeFit2Pnts);
	m_EnergyFlags |= (ET_PntsDistance | ET_PntsLocalClass | ET_DistsModelSpace);
	
	// TESTING:
	//m_EnergyFlags = ET_Pnts | ET_NodeExists | ET_NodeFit2Pnts | ET_NodeXForm | ET_NodeShape;
	//m_EnergyFlags = ET_Pnts | ET_NodeFit2Pnts;
	//m_EnergyFlags |= ET_PntsDistance | ET_PntsLocalClass | ET_DistsModelSpace;
	//m_EnergyFlags |= ET_PntsDistance | ET_DistsModelSpace;
	
	
	// IRRELEVANT
	s_fPntInEllipsoidVar = .1;			
	s_fPointEnergyWeight = 10.;
	
	// Global scaling of energy (just to make it easier to read)
	m_fGlobalEnergyScale = 1000.;

	// SCALING OF PARTS
		// scaling of parts: normal
	//s_fEllScaleMinVarScale = 10.;
	//s_fEllScaleMinVarScale = 1.;			// non-conservative
	//s_fEllScaleMinVarOffset = .5;

		// scaling of parts: no scaling
	//s_fEllScaleMinVarScale = 0;
	//s_fEllScaleMinVarOffset = 2.;
	//s_fEnoughSamples = 2;
		
	// TRANSLATION OF PARTS
	//s_fSubpartXFormMinVar = 2.;			// translating center
	//s_fSubpartXFormMinVar = 2.;				// non-conservative
	//s_fSubpartXFormMinVar = .05;			// conservative
	//s_fSubpartXFormMinVar = 0;		// not part translation
	
	// PER POINT ENERGY
	//s_fPnt2NullPenalty = .1;	// with distance to box
	//s_fPnt2NullPenalty = .15;	// with distance to box
	//s_fPnt2NullPenalty = .3;	// with distance to box
	s_fPnt2NullPenalty = 1.;	// vk: const
	//s_fPnt2NullPenalty = 20.;
	//s_fPnt2NullPenalty = 10.;
	
	// PER EDGE (pnt nhd) ENERGY
	s_fEdgeEnergyWeight = 2.;	// vk: const
	//s_fEdgeEnergyWeight = 0;

	// LOCAL ORIENTATIONS
	//s_fPntLocNhdDist = .2;	// NOTE: should be same as in TMatch::CreateTemplate...
	//s_iPntLocNhdSize = GetNhdSizeConstant();
	
	s_fPointOrientWeight = 2.;	// vk: const
	//s_fPointOrientWeight = 1.;
	//s_fPointOrientWeight = 0;
	
	// NODE PROBABILITIES
	s_ZeroNodeProbPenalty = 1000.;
	
	// NODE DEFORMS
	s_fXFormWeight = 5.;
	s_fScaleWeight = 5.;
	
	// parameters
	m_pMatch = match;	
	
	// to be pre-computed stuff
	m_pNNGraph = NULL;
	m_aPerPointEnergy = NULL;
	m_aPerLabelEnergy = NULL;
	m_afPerPointLocShapeLogProb = NULL;
	m_abRecomputeLocalShapeProb = NULL;
	
	
	// energy terms caching
	m_CachedGlobalEnergyFlags = 0;
	for (int i=0; i<m_pMatch->NTemplateNodes(); i++)
	{
		m_CachedPerNodeTerms.push_back(std::map<EnergyTerms, double>());
		m_CachePerNodeFlags.push_back(0);
	}
	for (int i=0; i<m_pMatch->NTemplateNodes(); i++)
	{
		UpdateNodeShape(i);
		UpdateNodeXForm(i);
	}

	UpdateLabeling();
	
	m_pMatch->AddEnergyUpdates(this);
}

TMatchEnergy::~TMatchEnergy()
{
	if (m_pNNGraph!=NULL)
		delete m_pNNGraph;
	if (m_aPerPointEnergy!=NULL)
		delete [] m_aPerPointEnergy;
	if (m_aPerLabelEnergy!=NULL)
		delete [] m_aPerLabelEnergy;
}

TMatch * TMatchEnergy::GetMatch()
{
	return m_pMatch;
}

//// Energy ////
double TMatchEnergy::Energy() const
{
	TMatchAlg::s_sProfiler.startedProcess("2.1_EnergyComputation");
	double e = Energy(m_EnergyFlags);
	TMatchAlg::s_sProfiler.finishedProcess("2.1_EnergyComputation");
	return e;
}

double TMatchEnergy::Energy(int terms) const
{
	double energy = 0;
//	TMatchAlg::s_sProfiler.startedProcess("2.1.6_EnergyEdges");
	if (terms & ET_PntEdges)
		energy += EEdges();
//	TMatchAlg::s_sProfiler.finishedProcess("2.1.6_EnergyEdges");
	for (int nodeID=0; nodeID<m_pMatch->NTemplateNodes(); nodeID++)
	{
		if (terms & ET_Pnts)
			energy += EPnts(nodeID);
		if (terms & ET_NodeExists)
			energy += ENodeExists(nodeID);
		if (terms & ET_NodeXForm)
			energy += ENodeXForm(nodeID);
		if (terms & ET_NodeShape)
			energy += ENodeShape(nodeID);
		if (terms & ET_NodeFit2Pnts)
			energy += ENodeFit2Pnts(nodeID);
	}
	return energy;
}

double TMatchEnergy::NodeEnergy(int nodeID) const
{
	double energy = 0;
	if (m_EnergyFlags & ET_Pnts)
		energy += EPnts(nodeID);
	if (m_EnergyFlags & ET_NodeExists)
		energy += ENodeExists(nodeID);
	if (m_EnergyFlags & ET_NodeXForm)
		energy += ENodeXForm(nodeID);
	if (m_EnergyFlags & ET_NodeShape)
		energy += ENodeShape(nodeID);
	if (m_EnergyFlags & ET_NodeFit2Pnts)
		energy += ENodeFit2Pnts(nodeID);
	return energy;
}

void TMatchEnergy::SetTerms(int terms)
{
	m_EnergyFlags = terms;
}

int TMatchEnergy::GetTerms() const
{
	return m_EnergyFlags;
}

double TMatchEnergy::MissingPartPenalty() const
{
	return s_ZeroNodeProbPenalty;
}

void TMatchEnergy::ClearCache()
{
	UpdateLabeling();
	UpdateGlobalXForm();
	for (int i=0; i<m_pMatch->NTemplateNodes(); i++)
	{
		UpdateNodeShape(i);
		UpdateNodeXForm(i);
	}
}

void TMatchEnergy::UpdateLabeling()
{
	if (m_aPerPointEnergy!=NULL)
		delete [] m_aPerPointEnergy;
	m_aPerPointEnergy = NULL;
	
	// global terms
	m_CachedGlobalEnergyFlags &= ~ET_PntEdges;
	
	// local terms
	for (int i=0; i<(int)m_CachePerNodeFlags.size(); i++)
	{
		m_CachePerNodeFlags[i] &= ~ET_Pnts;
		m_CachePerNodeFlags[i] &= ~ET_NodeFit2Pnts;
		m_CachePerNodeFlags[i] &= ~ET_NodeExists;
	}	
}

void TMatchEnergy::UpdateNodeShape(int nodeID)
{
	if (m_aPerPointEnergy!=NULL)
		delete [] m_aPerPointEnergy;
	m_aPerPointEnergy = NULL;

	assert(nodeID>=0 && nodeID < (int)m_CachePerNodeFlags.size());
	m_CachePerNodeFlags[nodeID] &= ~ET_Pnts;
	m_CachePerNodeFlags[nodeID] &= ~ET_NodeShape;
	m_CachePerNodeFlags[nodeID] &= ~ET_NodeFit2Pnts;
}

void TMatchEnergy::UpdateNodeXForm(int nodeID)
{
	if (m_aPerPointEnergy!=NULL)
		delete [] m_aPerPointEnergy;
	m_aPerPointEnergy = NULL;

	assert(nodeID>=0 && nodeID < (int)m_CachePerNodeFlags.size());
	m_CachePerNodeFlags[nodeID] &= ~ET_Pnts;
	m_CachePerNodeFlags[nodeID] &= ~ET_NodeXForm;
	m_CachePerNodeFlags[nodeID] &= ~ET_NodeFit2Pnts;
	
	// TODO: if transform includes orientation, set m_abRecomputeLocalShapeProb[i] = false...
}

void TMatchEnergy::UpdateGlobalXForm()
{
	if (m_aPerPointEnergy!=NULL)
		delete [] m_aPerPointEnergy;
	m_aPerPointEnergy = NULL;

	for (int nodeID=0; nodeID<m_pMatch->NTemplateNodes(); nodeID++)
	{
		assert(nodeID>=0 && nodeID < (int)m_CachePerNodeFlags.size());
		
		m_CachePerNodeFlags[nodeID] &= ~ET_Pnts;
		m_CachePerNodeFlags[nodeID] &= ~ET_NodeFit2Pnts;
		
		if (m_abRecomputeLocalShapeProb!=NULL)
			m_abRecomputeLocalShapeProb[nodeID] = true;
	}
	
}

VKString TMatchEnergy::EnergyStr() const
{
	VKString eStr("E=");
	eStr += VKString::number(Energy())+"  :\n";
	eStr += VKString("   Epnts    = ")+VKString::number(Energy(ET_Pnts))+"\n";
	eStr += VKString("   Eedges   = ")+VKString::number(Energy(ET_PntEdges))+"\n";
	eStr += VKString("   ENExists = ")+VKString::number(Energy(ET_NodeExists))+"\n";
	eStr += VKString("   ENXForm  = ")+VKString::number(Energy(ET_NodeXForm))+"\n";
	eStr += VKString("   ENShape  = ")+VKString::number(Energy(ET_NodeShape))+"\n";
	eStr += VKString("   ENodeFit = ")+VKString::number(Energy(ET_NodeFit2Pnts))+"\n";
	return eStr;
}

bool TMatchEnergy::GetCachedValue(EnergyTerms term, int nodeID, double & value) const
{
	assert(nodeID < (int)m_CachePerNodeFlags.size() && nodeID < (int)m_CachedPerNodeTerms.size()); 
	bool valid = ((m_CachePerNodeFlags[nodeID] & term) != 0);
	if (valid)
		value = ((TMatchEnergy*)this)->m_CachedPerNodeTerms[nodeID][term];
	return valid;
}

void TMatchEnergy::SetCacheValue(EnergyTerms term, int nodeID, double value) const
{
	assert(nodeID < (int)m_CachePerNodeFlags.size() && nodeID < (int)m_CachedPerNodeTerms.size()); 
	((TMatchEnergy*)this)->m_CachePerNodeFlags[nodeID] |= term;
	((TMatchEnergy*)this)->m_CachedPerNodeTerms[nodeID][term] = value;
}

bool TMatchEnergy::GetCachedGlobalValue(EnergyTerms term, double & value) const
{
	value = ((TMatchEnergy*)this)->m_CachedGlobalEnergyTerms[term];
	return ((m_CachedGlobalEnergyFlags & term) !=0);
}

void TMatchEnergy::SetCacheGlobalValue(EnergyTerms term, double value) const
{
	((TMatchEnergy*)this)->m_CachedGlobalEnergyFlags |= term;
	((TMatchEnergy*)this)->m_CachedGlobalEnergyTerms[term] = value;		
}

double TMatchEnergy::EPnts(int nodeID) const
{
	if ((m_EnergyFlags & ET_Pnts)==0)
		return 0;

	double epnts = 0;
	if (GetCachedValue(ET_Pnts, nodeID, epnts))
		return epnts;
	epnts = 0;
	for (int i=0; i<m_pMatch->NPointsLabeledAsNode(nodeID); i++)
	{
		int gpID = m_pMatch->GetPointLabeledAsNode(nodeID, i);
		epnts += ComputeEGlobalPnt(gpID, m_pMatch->LocalNodeID(gpID));
	}
	
	SetCacheValue(ET_Pnts, nodeID, epnts);
	return epnts;
}

double TMatchEnergy::EEdges() const
{
	double edgeEnergy = 0;
	if (GetCachedGlobalValue(ET_PntEdges, edgeEnergy))
		return edgeEnergy;
	
	edgeEnergy = 0;
	
	const VKGraph * nnGraph = GetEdgeEnergy();
	assert(nnGraph!=NULL);
	// go over neighborhood
	for (int e=0; e<nnGraph->NumEdges(); e++)
	{
		const VKGraphEdge * edge = ((VKGraph*)nnGraph)->Edge(e);
		int gp1 = m_pMatch->Loc2GlobPid(edge->n1->id);
		int gp2 = m_pMatch->Loc2GlobPid(edge->n2->id);
		
		if (m_pMatch->LocalNodeID(gp1) != m_pMatch->LocalNodeID(gp2))
			edgeEnergy += edge->weight;
	}

	SetCacheGlobalValue(ET_PntEdges, edgeEnergy);
	
	return edgeEnergy;
}

double TMatchEnergy::ENodeExists(int nodeID) const
{
	if ((m_EnergyFlags & ET_NodeExists)==0)
		return 0;
	
	double nExistsEnergy = 0;
	if (GetCachedValue(ET_NodeExists, nodeID, nExistsEnergy))
		return nExistsEnergy;
	
	nExistsEnergy = 0;
	
	const STemplateNode * stn = m_pMatch->GetSTemplate()->GetNode(m_pMatch->Deformation(nodeID).NodeID());
	const PartBox * part = stn->GetPart(m_pMatch->SeedID());
	if (part->IsType(PartBox::BOXTYPE_EmptyBox) || part->IsType(PartBox::BOXTYPE_NullBox))
		return 0;
 
	bool included = m_pMatch->Deformation(nodeID).IsIncluded();
	included = included && m_pMatch->NPointsLabeledAsNode(nodeID)>0;
	assert((m_pMatch->NPointsLabeledAsNode(nodeID)==0) == (!included));
	
//	if (!included)
//	{
//		std::cout<<"Excluded "<<stn->NodeName().c_str();
//		std::cout<<". Prob(0) = "<<<<std::endl;
//	}
	
	assert(part->IsType(PartBox::BOXTYPE_BoxSurface) || part->IsType(PartBox::BOXTYPE_MeshPoints));
	
	if (!included)
	{
		double prob0 = stn->PartOfParent()->GetProbNParts(m_pMatch->SeedID())->Prob(0);
		if (prob0==0)
			nExistsEnergy += s_ZeroNodeProbPenalty;
	}
	
	SetCacheValue(ET_NodeExists, nodeID, nExistsEnergy);
	
	return nExistsEnergy;	// no penalty for now
}

double TMatchEnergy::ENodeXForm(int nodeID) const
{
	if ((m_EnergyFlags & ET_NodeXForm)==0)
		return 0;
	
	double nXFormEnergy = 0;
	if (GetCachedValue(ET_NodeXForm, nodeID, nXFormEnergy))
		return nXFormEnergy;
	
	nXFormEnergy = 0;
	
	const STemplate * st = m_pMatch->GetSTemplate();
	
	const TemplateNodeDeform & deform = m_pMatch->Deformation(nodeID);
	
	if (!deform.IsIncluded())
		return 0;
	
	const STemplateNode * stn = st->GetNode(deform.NodeID());
	const PartOfEdge * pe = stn->PartOfParent();
	assert(pe!=NULL);
	
	int deformXFormType, deformXFormDims;
	const double * sample = deform.GetTransform(deformXFormDims, deformXFormType);

	double s = GetNodeXFormWeight();
	nXFormEnergy = s * -pe->LogProb(m_pMatch->SeedID(), deform.NParts(), deform.PartID(), sample, deformXFormType);
	SetCacheValue(ET_NodeXForm, nodeID, nXFormEnergy);
	return nXFormEnergy;
}

double TMatchEnergy::ENodeShape(int nodeID) const
{
	if ((m_EnergyFlags & ET_NodeShape)==0)
		return 0;
	
	double nShapeEnergy = 0;
	if (GetCachedValue(ET_NodeShape, nodeID, nShapeEnergy))
		return nShapeEnergy;
	
	nShapeEnergy = 0;

	const STemplate * st = m_pMatch->GetSTemplate();
	const TemplateNodeDeform & deform = m_pMatch->Deformation(nodeID);
	const PartBox & ep = m_pMatch->NodeEllipsoid(nodeID);
	const STemplateNode * stn = st->GetNode(deform.NodeID());
	if (ep.IsType(PartBox::BOXTYPE_NullBox) || ep.IsType(PartBox::BOXTYPE_EmptyBox) || !deform.IsIncluded())
		return 0;
	double s = GetNodeShapeWeight();
	nShapeEnergy = s * -stn->LogProb(m_pMatch->SeedID(), deform.GetScale());
	SetCacheValue(ET_NodeShape, nodeID, nShapeEnergy);
		
	return nShapeEnergy;
}

double TMatchEnergy::GetNodeShapeWeight() const
{
	return s_fScaleWeight / m_pMatch->NTemplateNodes();
}

double TMatchEnergy::GetNodeXFormWeight() const
{
	return s_fXFormWeight / m_pMatch->NTemplateNodes();
}

double TMatchEnergy::GetModelPntsDistWeight(int nodeID) const
{
	if (m_EnergyFlags & ET_SameWeightParts)
	{
		int nparts = m_pMatch->NTemplateNodes();
		int npartpnts = m_pMatch->NPointsLabeledAsNode(nodeID);
		//int ntotalpnts = m_pMatch->NShapePrimitives();
		return m_fGlobalEnergyScale / (nparts * npartpnts);
	}
	else	// does not depend on nodeID
	{
		return m_fGlobalEnergyScale / m_pMatch->NShapePrimitives();
	}
}

double TMatchEnergy::GetTmpltPntsDistWeight(int nodeID) const
{
	if (m_EnergyFlags & ET_SameWeightParts)
	{
		int nparts = m_pMatch->NTemplateNodes(); 
		int npartpnts = m_pMatch->NPointsOnTemplateNode(nodeID);
		//int ntotalpnts = m_pMatch->NShapePrimitives();
		return m_fGlobalEnergyScale / (nparts * npartpnts);
	}
	else	// does not depend on nodeID
	{
		// changed for ICP procedure (easy to include, but weights should be same)
		//return m_fGlobalEnergyScale / m_pMatch->NTemplatePoints();
		return m_fGlobalEnergyScale / m_pMatch->NShapePrimitives();		
	}
}

double TMatchEnergy::ENodeFit2Pnts(int nodeID) const
{
	if ((m_EnergyFlags & ET_NodeFit2Pnts)==0)
		return 0;
	
	double nFitEnergy = 0;
	if (GetCachedValue(ET_NodeFit2Pnts, nodeID, nFitEnergy))
		return nFitEnergy;

	nFitEnergy = 0;
	double s = GetTmpltPntsDistWeight(nodeID);
	R3Point ellPos;
//	double ellEigs[3];
	R3Point pntPos;
//	double pntEigs[3];
	
	const PartBox & node = m_pMatch->NodeEllipsoid(nodeID);
	TMatch * match = ((TMatchEnergy*)this)->GetMatch();
	VKString nName = match->GetSTemplate()->GetNode(match->Deformation(nodeID).NodeID())->NodeName();
//	if (nName=="base")
//		std::cout<<"\n"<<nName.c_str()<<" - Included="<<node.IsIncluded()<<std::endl;	
	
	if (!node.IsIncluded() || node.IsType(PartBox::BOXTYPE_EmptyBox) || node.IsType(PartBox::BOXTYPE_NullBox))
		nFitEnergy = 0;
	else
	{
//		if (nName=="base")
//		{
//			std::cout<<"   - Npoints: "<<node.NTemplatePointsOnNode()<<std::endl;
//			std::cout<<"   - Labeled as me: "<<m_pMatch->NPointsLabeledAsNode(nodeID)<<std::endl;
//		}
		for (int i=0; i<node.NTemplatePointsOnNode(); i++)
		{
			//R3Triad ellOrient = node.TemplatePoint(i, ellPos, ellEigs);
			ellPos = node.DeformedTemplatePointPosition(i);
			int gPntID = FindNearestPoint(nodeID, i, ellPos);
			
//			if (nName=="base")
//			{
//				
//				std::cout<<" - gPntID = "<<gPntID<<std::endl;
//			}
			if (gPntID==-1)
			{
				nFitEnergy += s * GetNullPenalty();
			}
			else
			{
				assert(gPntID!=-1);
				int ncount = 0;
				double objSpaceE = 0;
				if (m_EnergyFlags & ET_DistsModelSpace) // OLD: (object space)
				{
					//R3Triad pntOrient = m_pMatch->AlignedPoint(gPntID, GetLocalPCAVals(), pntPos, pntEigs);
					//objSpaceE = ComputeEPnt(pntPos, pntOrient, pntEigs, ellPos, ellOrient, ellEigs);
					if (m_EnergyFlags & ET_PntsDistance)
					{
						pntPos = m_pMatch->AlignedPointPosition(gPntID);
						objSpaceE += R3SquaredDistance(pntPos, ellPos);
					}
					ncount ++;
				}
				
				double tmpltSpaceE = 0;
				if (m_EnergyFlags & ET_DistsTmpltSpace) // NEW: (template space)
				{
					assert(false);
//					R3Triad ellOrient = node.UndeformedTemplatePoint(i, ellPos, ellEigs);
					if (m_EnergyFlags & ET_PntsDistance)
					{
						ellPos = node.UndeformedTemplatePointPosition(i); 
						//pntPos = node.Deformation()->Deformed2Orig(m_pMatch->AlignedPointPosition(gPntID));
						pntPos = m_pMatch->AlignedToNodePointPosition(nodeID, gPntID, m_pMatch->SeedID());
						tmpltSpaceE += R3SquaredDistance(pntPos, ellPos);
					}
					//R3Triad pntOrient = node.Deformation()->Deformed2Orig(pntOrient);
//					tmpltSpaceE = ComputeEPnt(pntPos, pntOrient, pntEigs, ellPos, ellOrient, ellEigs);
					ncount ++;
				}
				if (ncount>0)
					nFitEnergy += s * (objSpaceE + tmpltSpaceE) / ncount;
				//std::cout<<"nodeFit["<<nodeID<<"] = "<<nFitEnergy<<std::endl;
			}
		}
	}
	
	SetCacheValue(ET_NodeFit2Pnts, nodeID, nFitEnergy);
	//std::cout<<"NodeFit["<<nodeID<<"] = "<<nFitEnergy<<std::endl;
	return nFitEnergy;
}

double TMatchEnergy::ComputeEGlobalPnt(int globalPntID, int localNodeID) const
{
	if ((m_EnergyFlags & ET_Pnts)==0)
		return 0;

	// get node data
	const PartBox & ellipsoid = m_pMatch->NodeEllipsoid(localNodeID);
	double s = GetModelPntsDistWeight(localNodeID);
	if (ellipsoid.IsType(PartBox::BOXTYPE_NullBox))
		return s * GetNullPenalty();
	
	if (ellipsoid.IsType(PartBox::BOXTYPE_EmptyBox) || !ellipsoid.IsIncluded())
		return s * (GetNullPenalty()+5.);
	
	// get global alignment ;
	//double pntEigs[3];
	//R3Triad pntOrient = m_pMatch->AlignedPoint(globalPntID, GetLocalPCAVals(), pntPos, pntEigs);
	R3Point pntPos = m_pMatch->AlignedPointPosition(globalPntID);

	// compute energy
//	double tPntEigs[3];

	// NOTE: nearest template point function changed
	//R3Triad tPntOrient = ellipsoid.NearestTemplatePoint(pntPos, tPntPos, tPntEigs);
	int globTmpltID =  m_pMatch->NearestTemplatePointGlobalID(localNodeID, pntPos);

	int ncount = 0;
	double objSpaceE = 0;
	if (m_EnergyFlags & ET_DistsModelSpace) // OLD: (object space)
	{
		R3Point tPnt = m_pMatch->DeformedTemplateSample(globTmpltID);
		objSpaceE += R3SquaredDistance(tPnt, pntPos);
		ncount++;
		//R3Point objSpaceTPnt = ellipsoid.Deformation()->Orig2Deformed(tPntPos);
		//R3Triad objSpaceTOrient = ellipsoid.Deformation()->Orig2Deformed(tPntOrient);
		//objSpaceE = ComputeEPnt(pntPos, pntOrient, pntEigs, objSpaceTPnt, objSpaceTOrient, tPntEigs);
	}
	
	double tmpltSpaceE = 0;
	if (m_EnergyFlags & ET_DistsTmpltSpace) // NEW: (template space)
	{
		assert(false);	// for now assume it's not on. simplifies ICP 
		pntPos = m_pMatch->AlignedToNodePointPosition(localNodeID, globTmpltID, m_pMatch->SeedID());
		R3Point tPnt = m_pMatch->UndeformedTemplateSample(globTmpltID);
		tmpltSpaceE += R3SquaredDistance(tPnt, pntPos);
		ncount++;
		//R3Point tmpltSpacePoint = ellipsoid.Deformation()->Deformed2Orig(pntPos);
		//R3Triad tmpltSpaceOrient = ellipsoid.Deformation()->Deformed2Orig(pntOrient);
		//tmpltSpaceE = ComputeEPnt(tmpltSpacePoint, tmpltSpaceOrient, pntEigs, tPntPos, tPntOrient, tPntEigs);
	}
	
	// local shape descriptors
	double locOrient = 0;
	if (m_EnergyFlags & ET_PntsLocalClass)
		locOrient += LogProbLocalShape(globalPntID, localNodeID);
	//std::cout<<"locOrient="<<locOrient<<std::endl;
	
	//std::cout<<"Global Pnt Value: "<<(s * (objSpaceE + tmpltSpaceE) / ncount + locOrient)<<std::endl;
	double finalE = 0;
	if (ncount>0)
		finalE += (objSpaceE + tmpltSpaceE) / ncount;
	finalE += locOrient;
//	std::cout<<"TMAtchEnergy["<<localNodeID<<"].pnts = "<<finalE<<" * "<<s<<" = "<<(finalE*s)<<std::endl;
	return s * finalE;
}

double TMatchEnergy::LogProbLocalShape(int globalPntID, int localNodeID) const
{
	int nlabels = m_pMatch->NTemplateNodes();
	int npoints = m_pMatch->GetMatchedGeomAsPnts()->NumPoints();
	TMatchEnergy * ncThis = (TMatchEnergy*)this;
	if (m_afPerPointLocShapeLogProb==NULL)
	{
		assert(m_abRecomputeLocalShapeProb==NULL);
		ncThis->m_afPerPointLocShapeLogProb = new double[npoints * nlabels];
		ncThis->m_abRecomputeLocalShapeProb = new bool[npoints];
		for (int i=0; i<npoints; i++)
			ncThis->m_abRecomputeLocalShapeProb[i] = true;
	}
	if (m_abRecomputeLocalShapeProb[globalPntID])
	{
		const double * logProbs = m_pMatch->GetSTemplate()->LogProbsLocalDescriptors(globalPntID, m_pMatch);
		for (int i=0; i<nlabels; i++)
		{
			double logProb = 0;
			if (logProbs[i]>=0 && m_pMatch->Deformation(i).IsIncluded())
				logProb = logProbs[i];
			ncThis->m_afPerPointLocShapeLogProb[globalPntID * nlabels + i] = logProb / npoints * s_fPointOrientWeight;
		}
		ncThis->m_abRecomputeLocalShapeProb[globalPntID] = false;
	}
	return m_afPerPointLocShapeLogProb[globalPntID * nlabels + localNodeID];
}

/*
double TMatchEnergy::ComputeEPnt(const R3Point & pntPos, const R3Triad & pntOrient, const double * pntEigs,
								 const R3Point & nodePos, const R3Triad & nodeOrient, const double * nodeEigs) const
{
	double pntEnergy = 0;
	if (m_EnergyFlags & ET_PntsDistance)
	{
		double distE =pow(R3Distance(pntPos, nodePos), 2.);
		double w = 1.;
		if (m_EnergyFlags & ET_PntsOrientation)
			w = 1. - OrientationSimilarity(pntOrient, pntEigs, nodeOrient, nodeEigs);
		pntEnergy += distE + s_fPointOrientWeight * w;
	}
	return pntEnergy;
}*/



const double * TMatchEnergy::GetPntsEnergy() const
{
	if (m_aPerPointEnergy==NULL)
	{
		TMatchEnergy * ncThis = (TMatchEnergy*)this;
		int npoints = m_pMatch->NShapePrimitives();
		int nlabels = m_pMatch->NTemplateNodes();
		ncThis->m_aPerPointEnergy = new double[npoints * nlabels];
	
		for (int l=0; l<nlabels; l++)
		for (int p=0; p<npoints; p++)
			ncThis->m_aPerPointEnergy[p*nlabels + l] = ComputeEGlobalPnt(m_pMatch->Loc2GlobPid(p), l);
	}
	
	return m_aPerPointEnergy;
}

const double * TMatchEnergy::GetLabelEnergy() const
{
	if (m_aPerLabelEnergy==NULL)
	{
		int nlabels = m_pMatch->NTemplateNodes();
		((TMatchEnergy*)this)->m_aPerLabelEnergy = new double[nlabels*nlabels];
		for (int i=0; i<nlabels; i++)
		for (int j=0; j<nlabels; j++)
		{
			if (i!=j)
				((TMatchEnergy*)this)->m_aPerLabelEnergy[i*nlabels + j] = 1;
			else
				((TMatchEnergy*)this)->m_aPerLabelEnergy[i*nlabels + j] = 0;
		}
	}
	return m_aPerLabelEnergy;
}

void TMatchEnergy::SetEdgeEnergy(const VKGraph * graph)
{
	m_pNNGraph = (VKGraph*)graph;
}

const VKGraph * TMatchEnergy::GetEdgeEnergy() const
{
	// NOTE: invariant to all parameters
	if (m_pNNGraph==NULL)
	{
		// manifold vs pnt cloud
		const ShapePointCloud * pnts = m_pMatch->GetMatchedGeomAsPnts();
		const ShapeMesh * mesh = pnts->GetMesh();
		
		std::map<const ShapeMesh *, VKGraph * >::iterator iter = m_Mesh2Graph.find(mesh);
		if (iter != m_Mesh2Graph.end())
		{
			((TMatchEnergy*)this)->m_pNNGraph = iter->second;
			return m_pNNGraph;
		}
		
//		TimeProfiler tempTimer;
//		tempTimer.startedProcess("GetEdgeEnergy");

		double pi = 3.14159265359;
		int nhdSize = 30;//vk: const
		double epsilon = 0.5;
		assert(pnts!=NULL);
		double distScale = pnts->AverageRadius(pnts->Centroid());
		VKGraph * nnGraph = new VKGraph(pnts->NumPoints());
		m_Mesh2Graph[mesh] = nnGraph;
		
		int nEdges = 0;

		if (mesh!=NULL && mesh->SafeCastToManifold()!=NULL)		 // MANIFOLD
		{
			ShapeManifoldMesh * origMesh = (ShapeManifoldMesh*)mesh->SafeCastToManifold();
			assert(origMesh!=NULL);
			R3Mesh * r3origmesh = origMesh->GetMesh();
			assert(r3origmesh!=NULL);

			const R3Kdtree<const PointInSet *> * kdTree = pnts->GetKdTree();
			double maxDist = PartLocalGeometry::GetNhdDist()*distScale; 
			int f1id, f2id;
			double b1, b2;
			int npoints = m_pMatch->NShapePrimitives();
			for (int p1=0; p1<npoints; p1++)
			{
				// find geodesic neighborhoods (and also path to neighbors via face centers)
				pnts->GetPoint(p1).GetBarycentricCoords(&f1id, &b1, &b2, &mesh);

				// find potential neighbors
				std::vector<int> nhd;
				// vk: const
				ShapeLocalPCA::GetNhd(pnts->PointPosition(p1), kdTree, maxDist, nhdSize, nhd);

				for (int n=0; n<(int)nhd.size(); n++)
				{
					// compute maximal dihedral angle: theta
					int p2 = nhd[n];
					if (p2 == p1)
						continue;

					double euclDist = R3Distance(pnts->PointPosition(p1), pnts->PointPosition(p2));
					pnts->GetPoint(p2).GetBarycentricCoords(&f2id, &b1, &b2, &mesh);
					assert(f2id>=0 && f2id < r3origmesh->NFaces());
					R3Vector n1 = r3origmesh->FaceNormal(r3origmesh->Face(f1id));
					R3Vector n2 = r3origmesh->FaceNormal(r3origmesh->Face(f2id));
					if (!nnGraph->Adjacent(p1, p2))
					{
						nEdges++;
						// vk: const
						//double value = pow(vkAbs(n1.Dot(n2)), 10.) * exp(-pow(euclDist / (.5*maxDist), 2.));
						double theta = acos(n1.Dot(n2));
						
						double orientSim = -log(vkMin(1., epsilon + 1-theta/pi));
						double distSim = exp(-pow(euclDist / (.5*maxDist), 2.));

						if (orientSim*distSim>0)
							nnGraph->AddEdge(p1, p2, orientSim * distSim);
					}
				}
			}
		}
		else	// POINT CLOUD
		{
			// local orientation
			//const ShapeLocalPCA * pcaAxes = GetLocalPCAVals();
			const PartLocalGeometry * locGeom = m_pMatch->GetPntsLocalGeometry();
			int npoints = m_pMatch->NShapePrimitives();
			bool subsetExists = (npoints!=pnts->NumPoints());
			const R3Kdtree<const PointInSet *> * kdTree = NULL;
			if (!subsetExists)
				kdTree = pnts->GetKdTree();
			else
			{
				std::vector<int> pntSubset;
				for (int i=0; i<m_pMatch->NShapePrimitives(); i++)
					pntSubset.push_back(m_pMatch->Loc2GlobPid(i));
				kdTree = pnts->CreateSubsetKdTree(pntSubset);
			}
			
			
			// to normalize energy by distance
			double maxDist = PartLocalGeometry::GetNhdDist() * distScale;
			double e1[3], e2[3];
			
			// get connected components (if points are on a mesh);
			const ShapeMesh * mesh = pnts->GetMesh();
			const ShapeLabels * cc = (mesh!=NULL ? mesh->GetConnectedComponents() : NULL);
			std::vector<bool> bigEnoughCC;
			
			if (cc!=NULL)
			{
				double ccAreaThresh = mesh->Area() * .05;	// 5% of total mesh area, vk: const
				int ccSizeThresh = 1;
				for (int i=0; i<cc->NNodes(); i++)
				{
					std::vector<int> faceIDs;
					cc->NodeAllFaces(i, faceIDs);
					double a = 0;
					for (int j=0; j<(int)faceIDs.size(); j++)
						a += mesh->FaceArea(faceIDs[j]);
					bigEnoughCC.push_back(a > ccAreaThresh && faceIDs.size() > ccSizeThresh);
				}
			}
			
			if ((m_EnergyFlags & ET_PntEdges)!=0)
			for (int n=0; n<npoints; n++)
			{
				int gid = m_pMatch->Loc2GlobPid(n);
				const ShapePoint & pnt = pnts->GetPoint(m_pMatch->Loc2GlobPid(n));
				// pca orientation
				R3Triad a1 = locGeom->LocalOrient(gid, e1, R3identity_affine);
				//pcaAxes->GetPCA(pnt, e1);
				R3Vector e1vec(e1[0], e1[1], e1[2]);
				e1vec.Normalize();
				R3Point pos1 = pnts->PointPosition(gid);
				
				// neighborhood
				std::vector<int> ids;
				assert(maxDist>0);
				ShapeLocalPCA::GetNhd(pos1, kdTree, maxDist, -1, ids);
				
				// compute number of nn stats
				for (int nn=0; nn<(int)ids.size(); nn++)
				{
					int subsetID = m_pMatch->Glob2LocPid(ids[nn]);
					if (n==subsetID)	// ignore self
						continue;
					if (!nnGraph->Adjacent(n, subsetID))
					{
						// if different connected components - no penalty
						// NOTE: need to modify this: detect meaningful CC, etc...
						
						bool ccTest = true;
						if (cc!=NULL)	// additional cc test
						{
							int ccn1 = cc->Value(pnt, 0);
							int ccn2 = cc->Value(pnts->GetPoint(ids[nn]), 0);
							assert(ccn1 >= 0 && ccn1 < (int)bigEnoughCC.size());
							assert(ccn2 >= 0 && ccn2 < (int)bigEnoughCC.size());
							if (bigEnoughCC[ccn1] && bigEnoughCC[ccn2])
								ccTest = (ccn1 == ccn2);
						}
						if (ccTest)
						{
							R3Point pos2 = pnts->PointPosition(ids[nn]);
							//R3Triad a2 = pcaAxes->GetPCA(pnts->GetPoint(ids[nn]), e2);
							R3Triad a2 = locGeom->LocalOrient(ids[nn], e2, R3identity_affine);
							
							nEdges++;
							
							//double orientSim = OrientationSimilarity(a1, e1, a2, e2);
							double theta = acos(a1[2].Dot(a2[2]));	// TODO - not very robust...
							double orientSim = -log(vkMin(1., epsilon + 1-theta/pi));
							double distSim = exp(-pow(R3Distance(pos1, pos2) / (.5*maxDist), 2.));
							if (orientSim * distSim > 0)
								nnGraph->AddEdge(n, subsetID, orientSim * distSim);	// vk: const
						}
					}
				}
			}
					
			if (subsetExists)
				delete (R3Kdtree<const PointInSet *> *)kdTree;
//			tempTimer.finishedProcess("GetEdgeEnergy", true);
		}
		((TMatchEnergy*)this)->m_pNNGraph = nnGraph;
		for (int i=0; i<nnGraph->NumEdges(); i++)
			nnGraph->Edge(i)->weight *= m_fGlobalEnergyScale*s_fEdgeEnergyWeight / nEdges;
			//nnGraph->Edge(i)->weight = 0;
	}
	
	
	return m_pNNGraph;
}

double TMatchEnergy::GetNullPenalty() const
{
	return s_fPnt2NullPenalty;
}

void TMatchEnergy::SetNullPenalty(double penalty)
{
	s_fPnt2NullPenalty = penalty;
	
	if (m_aPerPointEnergy!=NULL)
		delete [] m_aPerPointEnergy;
	m_aPerPointEnergy = NULL;
	
	for (int nodeID=0; nodeID<m_pMatch->NTemplateNodes(); nodeID++)
	{
		assert(nodeID>=0 && nodeID < (int)m_CachePerNodeFlags.size());		
		m_CachePerNodeFlags[nodeID] &= ~ET_Pnts;
		m_CachePerNodeFlags[nodeID] &= ~ET_NodeFit2Pnts;
	}
}


//// HELPER FUNCTIONS ////
/*
void TMatchEnergy::GetEVectorWeights(const double e[3], double w[3], std::vector<int> * outSort) const
{
	w[0]=0;w[1]=0;w[2]=0;
	if (e[0]<0)
		return;

	std::vector<Sortable> srtEvals;
	for (int i=0; i<3; i++)
		srtEvals.push_back(Sortable(-e[i], NULL, i));
	std::sort(srtEvals.begin(), srtEvals.end());
	
	if (e[srtEvals[2].id] / e[srtEvals[1].id]< .5)	// plane
		w[srtEvals[2].id] = 1.;
	if (e[srtEvals[1].id] / e[srtEvals[0].id] < .5) // cylinder
		w[srtEvals[0].id] = 1.;
	
	if (outSort!=NULL)
	{
		(*outSort).push_back(srtEvals[0].id);
		(*outSort).push_back(srtEvals[1].id);
		(*outSort).push_back(srtEvals[2].id);
	}

	return;

}

double TMatchEnergy::OrientationSimilarity(const R3Triad & a1, const double e1[3],
										   const R3Triad & a2, const double e2[3]) const
{
	double sim = 0;
	// if orientation is similar - lessen weight
//	const R3Triad & a2 = pntOrient;
//	const double * e2 = pntEigs;
//	const R3Triad & a1 = nodeOrient;
//	const double * e1 = nodeEigs;
	if (e1[0]>0 && e2[1]>0)
	{
		double w1[3];
		double w2[3];
		std::vector<int> srt1, srt2;
		GetEVectorWeights(e1, w1, &srt1);
		GetEVectorWeights(e2, w2, &srt2);
		
		// difference in eigenvalue structure
		R3Vector ev0(e1[srt1[0]], e1[srt1[1]], e1[srt1[2]]);
		R3Vector ev1(e2[srt2[0]], e2[srt2[1]], e2[srt2[2]]);
		ev0.Normalize();
		ev1.Normalize();
		double epsilon = 10e-5;
		
		//close to 1 => similar
		sim = pow(vkAbs(ev0.Dot(ev1)), 10.);	// vk: const
		double e2Plane = e2[srt2[2]] / e2[srt2[1]];
		double e2Cylinder = e2[srt2[1]] / e2[srt2[0]];
		
		bool e1Plane = w1[srt1[2]]>0;
		bool e1Cylinder = w1[srt1[0]]>0;
		
		int axesToCompare = -1;
		if (e1Plane && e1Cylinder )		// plane or cylinder
		{
			if (e2Plane < e2Cylinder)
				axesToCompare = 2;
			else
				axesToCompare = 0;
		}
		else if (e1Plane)	// plane
			axesToCompare = 2;
		else if (e1Cylinder)
			axesToCompare = 0;
		
		if (axesToCompare>=0)
			sim *= vkAbs(a1[srt1[axesToCompare]].Dot(a2[srt2[axesToCompare]]));
		else
			sim = 0;
	}
	return vkMin(sim, 1.);
}
 */

void TMatchEnergy::WriteEnergyAndMatch(FILE * file) const
{
	// write energy terms
	int nentries = 7;
	double e[7];
	e[1] = Energy(ET_Pnts);
	e[2] = Energy(ET_PntEdges);
	e[3] = Energy(ET_NodeExists);
	e[4] = Energy(ET_NodeXForm);
	e[5] = Energy(ET_NodeShape);
	e[6] = Energy(ET_NodeFit2Pnts);
	
	fwrite(&nentries, sizeof(int), 1, file);
	fwrite(e, sizeof(double), nentries, file);
	
	m_pMatch->WriteMatch(file);
}

TMatchEnergy * TMatchEnergy::ReadEnergyAndMatch(FILE * file, const STemplate * st, ModelDatabase * db, double *energy)
{
	assert(file!=NULL);
	int nentries;
	fread(&nentries, sizeof(int), 1, file);
	double * e = new double[nentries];
	fread(e, sizeof(double), nentries, file);
	if (energy!=NULL)
	{
		*energy = 0;
		for (int i=0; i<nentries; i++)
			*energy += e[i];
	}
	TMatch * match = TMatch::ReadMatch(file, st, db);
	delete [] e;
	return new TMatchEnergy(match);
}


/**
 * Given a point position (in template space) tSpacePos, find a nearest point 
 * in the point cloud assigned to nodeID template node.
 */ 
int TMatchEnergy::FindNearestPoint(int nodeID, int pntID, const R3Point & tSpacePos) const
{
	if (m_pMatch->NPointsLabeledAsNode(nodeID)==0)
		return -1;

	R3Point pntsPos = tSpacePos;
	m_pMatch->GlobalXForm().ApplyInverse(pntsPos);
	return m_pMatch->NearestPointWithLabel(nodeID, pntsPos);
}



