/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "TemplateNodeDeform.h"

//// TemplateElement (template matching datastructures) //////
TemplateNodeDeform::TemplateNodeDeform(int nodeID, int nParts, int partID,
									   const double * transform, int ndim, int type, const R3Vector & scale)
{	
	m_iNodeID = nodeID;
	m_iNParts = nParts;
	m_iPartID = partID;
	
	m_iXFormNDim = ndim;
	m_TransformType = type;
	
	if (m_iXFormNDim>0 && transform!=NULL)
	{
		m_fTransform = new double[m_iXFormNDim];
		for (int i=0; i<ndim; i++)
			m_fTransform[i] = transform[i];
	}
	else
		m_fTransform = NULL;
	
	m_PCAScale = scale;
	
	m_bIncluded = true;
	m_bXFormChanged = false;
	m_AffineXForm = R3identity_affine;
}

TemplateNodeDeform::TemplateNodeDeform(const TemplateNodeDeform & deform, std::map<int, int> * old2newNode)
{
	m_iNodeID = deform.m_iNodeID;
	if (old2newNode!=NULL)
	{
		assert(old2newNode->find(deform.m_iNodeID)!=old2newNode->end());
		m_iNodeID = (*old2newNode)[deform.m_iNodeID];
	}
	m_iNParts = deform.m_iNParts;
	m_iPartID = deform.m_iPartID;
	
	m_iXFormNDim = deform.m_iXFormNDim;
	m_TransformType = deform.m_TransformType;
	if (m_iXFormNDim>0 && deform.m_fTransform!=NULL)
	{
		m_fTransform = new double[m_iXFormNDim];
		for (int i=0; i<deform.m_iXFormNDim; i++)
			m_fTransform[i] = deform.m_fTransform[i];
	}
	else
		m_fTransform = NULL;
	m_PCAScale = deform.m_PCAScale;
	
	m_bXFormChanged = deform.m_bXFormChanged;
	m_AffineXForm = deform.m_AffineXForm;
	m_bIncluded = deform.m_bIncluded;
}

TemplateNodeDeform::~TemplateNodeDeform()
{
	if (m_fTransform!=NULL)
		delete [] m_fTransform;
}

TemplateNodeDeform * TemplateNodeDeform::CreateCopy(std::map<int, int> * old2newNode) const
{ 
	return new TemplateNodeDeform(*this, old2newNode);
}

int TemplateNodeDeform::NodeID() const
{
	return m_iNodeID;
}

int TemplateNodeDeform::PartID() const
{
	return m_iPartID;
}

int TemplateNodeDeform::NParts() const
{
	return m_iNParts;
}

void TemplateNodeDeform::SetNodeID(int nodeID)
{
	m_iNodeID = nodeID;
}

const double * TemplateNodeDeform::GetTransform(int & xFormDims, int & xFormFlags) const
{
	xFormDims = m_iXFormNDim;
	xFormFlags = m_TransformType;
	return m_fTransform;
}

void TemplateNodeDeform::SetTransform(const double * newTransform)
{
	m_bXFormChanged = true;
	for (int i=0; i<m_iXFormNDim; i++)
		m_fTransform[i] = newTransform[i];
}

bool TemplateNodeDeform::IsIncluded() const
{
	return m_bIncluded;
}

void TemplateNodeDeform::SetIncluded(bool included)
{
	//m_bXFormChanged = true;
	m_bIncluded = included;
}

const R3Vector & TemplateNodeDeform::GetScale() const
{
	return m_PCAScale;
}

void TemplateNodeDeform::SetScale(const R3Vector & scale)
{
	m_bXFormChanged = true;
	m_PCAScale = scale;
}

void TemplateNodeDeform::SetDirtyFlag()
{
	m_bXFormChanged = true;
}

void TemplateNodeDeform::WriteNodeDeform(FILE * file) const
{
	fwrite(&m_iNodeID, sizeof(int), 1, file);
	fwrite(&m_iNParts, sizeof(int), 1, file);
	fwrite(&m_iPartID, sizeof(int), 1, file);
	
	fwrite(&m_iXFormNDim, sizeof(int), 1, file);
	fwrite(&m_TransformType, sizeof(int), 1, file);
	assert(m_iXFormNDim>0 && m_fTransform!=NULL);
	fwrite(m_fTransform, sizeof(double), m_iXFormNDim, file);
	
	double scale[3];
	for (int d=0; d<3; d++)
		scale[d] = m_PCAScale[d];
	
	fwrite(scale, sizeof(double), 3, file);
	
//	if (m_iNodeID==10 || m_iNodeID==11)
//		((TemplateNodeDeform*)this)->m_bIncluded = false;
	fwrite(&m_bIncluded, sizeof(bool), 1, file);
}

TemplateNodeDeform * TemplateNodeDeform::ReadNodeDeform(FILE * file)
{
	int nodeID, nParts, partID;
	double * xform = NULL;
	int ndim, type;
	
	fread(&nodeID, sizeof(int), 1, file);
	fread(&nParts, sizeof(int), 1, file);
	fread(&partID, sizeof(int), 1, file);
	
	fread(&ndim, sizeof(int), 1, file);
	fread(&type, sizeof(int), 1, file);
	assert(ndim>0);
	xform = new double[ndim];
	fread(xform, sizeof(double), ndim, file);
	
	double s[3];
	fread(s, sizeof(double), 3, file);
	
	TemplateNodeDeform * deform = NULL;
	deform = new TemplateNodeDeform(nodeID, nParts, partID, xform, ndim, type, R3Vector(s[0], s[1], s[2]));
	fread(&(deform->m_bIncluded), sizeof(bool), 1, file);
	delete [] xform;
	return deform;
}

R3Point TemplateNodeDeform::Orig2Deformed(const R3Point & pnt) const
{
	assert(!m_bXFormChanged);
    R3Point p = pnt;
    m_AffineXForm.Apply(p);
    return p;
}

R3Point TemplateNodeDeform::Deformed2Orig(const R3Point & pnt) const
{
	assert(!m_bXFormChanged);
    R3Point p = pnt;
    m_AffineXForm.ApplyInverse(p);
    return p;
}

R3Vector TemplateNodeDeform::Orig2Deformed(const R3Vector & vec) const
{
	assert(!m_bXFormChanged);
    R3Vector v = vec;
    m_AffineXForm.Apply(v);
    return v;
}

R3Vector TemplateNodeDeform::Deformed2Orig(const R3Vector & vec) const
{
	assert(!m_bXFormChanged);
    R3Vector v = vec;
    m_AffineXForm.ApplyInverse(v);
    return v;
}

R3Triad TemplateNodeDeform::Orig2Deformed(const R3Triad & axes) const
{
    return R3Triad(Orig2Deformed(axes[0]), Orig2Deformed(axes[1]), Orig2Deformed(axes[2]));
}

R3Triad TemplateNodeDeform::Deformed2Orig(const R3Triad & axes) const
{
    return R3Triad(Deformed2Orig(axes[0]), Deformed2Orig(axes[1]), Deformed2Orig(axes[2]));
}

///////// CREATE IDENTITY TRANSFORMATION //////////
IdentityTemplateDeform IdentityTemplateDeform::s_IdentityDeform;

IdentityTemplateDeform::IdentityTemplateDeform()
: TemplateNodeDeform(-1, -1, -1, NULL, 0, 0, R3Vector(1,1,1))
{
	m_bIncluded = true;
	m_bXFormChanged = false;
	m_AffineXForm = R3identity_affine;
}

void IdentityTemplateDeform::SetTransform(double * newTransform)
{
	assert(false);	// non-deformable
}

void IdentityTemplateDeform::SetIncluded(bool included)
{
	assert(false);	// non-deformable
}

void IdentityTemplateDeform::SetScale(const R3Vector & scale)
{
	assert(false);	// non-deformable
}

void IdentityTemplateDeform::WriteNodeDeform(FILE * file) const
{
	assert(false);	// non-writeable
}

