/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#include "VisShapeTemplate.h"
#include "PartOfEdge.h"
#include "TMatch.h"
#include "TMatchAlg.h"
#include "STemplateNode.h"

void VisShapeTemplate::DrawTemplateConfig(const TMatch * tConfig, const DatabaseVisWindow * window,
										  int flags, std::vector<RenderableFace> & faces, const R3Affine & xform)
{
	window->LoadStandardSceneProjectionCamera();
	for (int i=0; i<tConfig->NTemplateNodes(); i++)
	{
		int nodeID = i+1;
		const PartBox & ellipsoid = tConfig->NodeEllipsoid(i);
		const TemplateNodeDeform & deform = tConfig->Deformation(i);
		
		if ((flags & VISTMPLT_DrawTemplateAsBoxes) && ellipsoid.IsType(PartBox::BOXTYPE_BoxSurface))
			DrawTemplateAsBoxes(ellipsoid, nodeID, window, faces, xform, false, -1);
		else if ((flags & VISTMPLT_DrawTemplateAsBoxes) && ellipsoid.IsType(PartBox::BOXTYPE_MeshPoints))
			DrawTemplateMesh(ellipsoid, nodeID, window, faces, xform, false);

		if (flags & VISTMPLT_DrawTemplateAsAxes)
		{
			if (flags & VISTMPLT_DrawDistribution)
			{
				int cid = tConfig->SeedID();
				const STemplateNode * node = tConfig->GetSTemplate()->GetNode(deform.NodeID());
				R3Vector svar(0,0,0);
				int anc1, anc2;
				if (node!=NULL)
					svar = node->ScaleVariance(cid, anc1);
				const PartOfEdge * edge = node->PartOfParent();
				
				double prob = edge->GetProbNParts(cid)->Prob(deform.NParts());
				R3Vector pvar = edge->PosVariance(cid, deform.NParts(), deform.PartID(), anc2);
				DrawTemplateAsAxes(ellipsoid, nodeID, window, xform, &svar, &pvar, -1);
			}
			else
				DrawTemplateAsAxes(ellipsoid, nodeID, window, xform, NULL, NULL, -1);
			
		}
		
		if ((flags & VISTMPLT_DrawTemplateSampling) && tConfig->NTemplatePoints()>0 && deform.IsIncluded())
		{
			glBegin(GL_POINTS);
			glPointSize(2.);
			for (int p=0; p<ellipsoid.NTemplatePointsOnNode(); p++)
			{
				double r, g, b;
				window->MapIntToRadicalColor(nodeID, r, g, b);
				glColor3d(r,g,b);
				R3Point pos = ellipsoid.DeformedTemplatePointPosition(p);
				xform.Apply(pos);
				glVertex3d(pos.X(), pos.Y(), pos.Z());
			}
			glEnd();
		}
	}
}

void VisShapeTemplate::DrawTemplateMesh(const PartBox & ellipsoid, int nodeID, const DatabaseVisWindow * window,
										std::vector<RenderableFace> & faces, const R3Affine & xform, bool outline)
{
	if (ellipsoid.IsType(PartBox::BOXTYPE_NullBox)
		|| ellipsoid.IsType(PartBox::BOXTYPE_EmptyBox) || !ellipsoid.IsIncluded())
		return;
	
	double r, g, b;
	window->MapIntToRadicalColor(nodeID, r, g, b);

	const PartGeometry * partGeometry = ellipsoid.GetPart();
	assert(partGeometry!=NULL);
	const ShapeMesh * mesh = partGeometry->GetNormalizedMesh();
	if (mesh==NULL)
		return;

	for (int fid=0; fid<mesh->NFaces(); fid++)
	{
		int f = (int)faces.size();
		faces.push_back(RenderableFace());
		faces[f].r = r;		faces[f].g = g;		faces[f].b = b;		faces[f].a = .2;
		
		for (int v=0; v<3; v++)
		{
			R3Point pos = mesh->VertexPosition(mesh->VertexOnFace(fid, v));
			pos = ellipsoid.Deformation()->Orig2Deformed(pos);
			xform.Apply(pos);
			faces[f].vtx[v] = pos;
		}
	}
	return;
}

void VisShapeTemplate::DrawTemplateAsBoxes(const PartBox & ellipsoid, int nodeID, const DatabaseVisWindow * window,
										   std::vector<RenderableFace> & faces, const R3Affine & xform,
										   bool outline, int highlightFace)
{
	if (ellipsoid.IsType(PartBox::BOXTYPE_NullBox)
		|| ellipsoid.IsType(PartBox::BOXTYPE_EmptyBox) || !ellipsoid.IsIncluded())
		return;
	
	R3Point pnt = ellipsoid.UndeformedCentroid();
    pnt = ellipsoid.Deformation()->Orig2Deformed(pnt);
	xform.Apply(pnt);
	
	double r, g, b;
	window->MapIntToRadicalColor(nodeID, r, g, b);
	
	//const double * scale = ellipsoid.Scale()->GetMostProbable();
	const R3Vector & scale = ellipsoid.UndeformedScale();
    const R3Triad & orient = ellipsoid.UndeformedOrientation();
	
	if (outline)
		glLineWidth(4.);
	
	for (int p=0; p<6; p++)
	{
		int v0ID = p / 2;
		int v1ID = (v0ID!=0 ? 0 : 2);
		int v2ID = (v0ID!=1 ? 1 : 2);
		int s0 = pow(-1, p%2);
		R3Vector v0 = orient[v0ID]*scale[v0ID];
		R3Vector v1 = orient[v1ID]*scale[v1ID];
		R3Vector v2 = orient[v2ID]*scale[v2ID];
        v0 = ellipsoid.Deformation()->Orig2Deformed(v0);
        v1 = ellipsoid.Deformation()->Orig2Deformed(v1);
        v2 = ellipsoid.Deformation()->Orig2Deformed(v2);
		xform.Apply(v0);
		xform.Apply(v1);
		xform.Apply(v2);
		
		int f = (int)faces.size();
		faces.push_back(RenderableFace());
		faces[f].r = r;		faces[f].g = g;		faces[f].b = b;		faces[f].a = .3;
		faces[f].vtx[0] = pnt + s0*v0 - v1 + v2;
		faces[f].vtx[1] = pnt + s0*v0 + v1 - v2;
		faces[f].vtx[2] = pnt + s0*v0 - v1 - v2;
		
		f = (int)faces.size();
		faces.push_back(RenderableFace());
		faces[f].r = r;		faces[f].g = g;		faces[f].b = b;		faces[f].a = .3;
		faces[f].vtx[0] = pnt + s0*v0 - v1 + v2;
		faces[f].vtx[1] = pnt + s0*v0 + v1 - v2;
		faces[f].vtx[2] = pnt + s0*v0 + v1 + v2;
		
		if (outline)
		{
			glBegin(GL_LINE_STRIP);
			glColor3d(1., 1., 0.);
			R3Point p = pnt + s0*v0 + v1 + v2;
			glVertex3d(p.X(), p.Y(), p.Z());
			p = pnt + s0*v0 + v1 - v2;
			glVertex3d(p.X(), p.Y(), p.Z());
			p = pnt + s0*v0 - v1 - v2;
			glVertex3d(p.X(), p.Y(), p.Z());
			p = pnt + s0*v0 - v1 + v2;
			glVertex3d(p.X(), p.Y(), p.Z());
			p = pnt + s0*v0 + v1 + v2;
			glVertex3d(p.X(), p.Y(), p.Z());
			glEnd();
		}
		
		if (highlightFace>=0 && (highlightFace/2)==(p/2))
		{
			glBegin(GL_TRIANGLES);
			glColor3d(1., 1., 0.);
			int f0 = (int)faces.size()-1;
			int f1 = f0-1;
			glVertex3d(faces[f0].vtx[0][0], faces[f0].vtx[0][1], faces[f0].vtx[0][2]);
			glVertex3d(faces[f0].vtx[1][0], faces[f0].vtx[1][1], faces[f0].vtx[1][2]);
			glVertex3d(faces[f0].vtx[2][0], faces[f0].vtx[2][1], faces[f0].vtx[2][2]);
			glVertex3d(faces[f1].vtx[0][0], faces[f1].vtx[0][1], faces[f1].vtx[0][2]);
			glVertex3d(faces[f1].vtx[1][0], faces[f1].vtx[1][1], faces[f1].vtx[1][2]);
			glVertex3d(faces[f1].vtx[2][0], faces[f1].vtx[2][1], faces[f1].vtx[2][2]);
			glEnd();
		}
	}
	
	glColor3d(1., 1., 1.);
	
	return;
}

void VisShapeTemplate::DrawTemplateAsAxes(const PartBox & ellipsoid, int nodeID,
										  const DatabaseVisWindow * window, const R3Affine & xform,
										  const R3Vector * scaleVar, const R3Vector * posVar, double prob)
{
	if (ellipsoid.IsType(PartBox::BOXTYPE_NullBox)
		|| ellipsoid.IsType(PartBox::BOXTYPE_EmptyBox) || !ellipsoid.IsIncluded())
		return;
	
	R3Point pnt = ellipsoid.Deformation()->Orig2Deformed(ellipsoid.UndeformedCentroid());
	//const double * scale = ellipsoid.Scale()->GetMostProbable();
	const R3Vector & scale = ellipsoid.UndeformedScale();
    const R3Triad & orient = ellipsoid.UndeformedOrientation();
	xform.Apply(pnt);
	
	glDisable(GL_LIGHTING);
	glLineWidth(2.);
	double r, g, b;
	window->MapIntToRadicalColor(nodeID, r, g, b);
	glColor4d(r, g, b, 1.);
	
	for (int i=0; i<3; i++)
	{
		R3Vector v = orient[i]*scale[i];
        v = ellipsoid.Deformation()->Orig2Deformed(v);
		xform.Apply(v);
		glBegin(GL_LINES);
		glVertex3d(-v.X()+pnt.X(), -v.Y()+pnt.Y(), -v.Z()+pnt.Z());
		glVertex3d(v.X()+pnt.X(), v.Y()+pnt.Y(), v.Z()+pnt.Z());
		glEnd();
		
		if (scaleVar!=NULL)
		{
			glEnable (GL_LINE_STIPPLE);
			glLineStipple (1, 0xF00F);
			glLineWidth(12.);
			R3Vector vOff = 2. * sqrt((*scaleVar)[i]) * orient[i];
			vOff = ellipsoid.Deformation()->Orig2Deformed(vOff);
			xform.Apply(vOff);
			R3Point p0 = pnt + v;
			
			double us = 1.;
			glBegin(GL_LINES);
			glVertex3d(p0.X(), p0.Y(), p0.Z());
			p0 += us*vOff;
			glVertex3d(p0.X(), p0.Y(), p0.Z());
			p0 = pnt - v;
			glVertex3d(p0.X(), p0.Y(), p0.Z());
			p0 -= us*vOff;
			glVertex3d(p0.X(), p0.Y(), p0.Z());
			glEnd();
			glDisable (GL_LINE_STIPPLE);
		}
	}
	
	if (posVar!=NULL)
	{
		glColor3d(.2, .2, .2);
		double spVar = 1.;
		//glTranslated(pnt.X(), pnt.Y(), pnt.Z());
		R3Vector posVarVec = spVar * R3Vector(sqrt((*posVar)[0]), sqrt((*posVar)[1]), sqrt((*posVar)[2]));
		R3Vector vec(1,1,1);
		vec.Normalize();
		xform.Apply(vec);
		posVarVec *= vec.Length();
		//posVarVec = ellipsoid.Orig2Deformed(posVarVec);
		//xform.Apply(posVarVec);
		R3CoordSystem coords = R3xyz_coordinate_system;
		coords.SetOrigin(ellipsoid.Deformation()->Orig2Deformed(ellipsoid.UndeformedCentroid()));
		coords.Transform(xform);
		R3Ellipsoid(coords, posVarVec).Draw();
		//glTranslated(-pnt.X(), -pnt.Y(), -pnt.Z());
	}

	VKString probStr = VKString::number((int)(prob*100.))+"%";
	if (prob>=0)
		window->GetRenderer()->RenderText(pnt, probStr, .8, .8, .8, VisWindowRenderer::TEXT_SIZE_LARGE);
}



void VisShapeTemplate::DrawDeformedGeometry(DatabaseVisWindow * window, const R3Affine & camera, TMatch * currMatch,
											std::vector<TMatchEnergy*> & allMatches, CoAlignedParts * coAligned,
											int clusteringID, int clusterID, bool drawCurrent, bool drawAll,
											bool drawConfidence, int selectedPart, bool onlyRigid,
											BufferID * isectBuffer, std::vector<std::vector<bool> > * matchPnts)
{
#ifndef MESA_RENDERING
	if (isectBuffer!=NULL)
	{
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, isectBuffer->m_iFrameBufferObject);
		glViewport( 0, 0, isectBuffer->m_iWidth, isectBuffer->m_iHeight);
		glEnable(GL_DEPTH_TEST);
		glDisable(GL_LIGHTING);
		glDisable(GL_POLYGON_SMOOTH);
		glShadeModel(GL_FLAT);
		glDisable(GL_BLEND);
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_CULL_FACE);
		glDisable(GL_DITHER);
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		glClearColor(0, 0, 0, 0);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT
				| GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	}
#endif
	
	for (int i=0; i<(int)allMatches.size(); i++)
	{
		TMatch * match = allMatches[i]->GetMatch();
		
		if (drawAll && clusteringID>=0 && clusterID>=0)
		{
			if (coAligned==NULL)
				continue;
			bool partThere = false;
			
			for (int j=0; j<coAligned->NPartsInCluster(clusteringID, clusterID) && !partThere; j++)
				partThere = coAligned->GetPart(clusteringID, clusterID, j).m_iMatchID == i;
			
			if (!partThere)
				continue;
		}
		
		// rendering single model
		if (drawCurrent && match != currMatch)
			continue;
		
		// rendering all models
		if (drawAll && match->SeedID() != currMatch->SeedID())
			continue;
		
		window->LoadStandardSceneProjectionCamera();
		
		for (int p=0; p<match->NShapePrimitives(); p++)
		{
			int gp = match->Loc2GlobPid(p);
			int lnid = match->LocalNodeID(gp);
			//int globNodeID = match->Deformation(lnid).NodeID();

			if (matchPnts!=NULL && i <(int)matchPnts->size() && gp < (int)(*matchPnts)[i].size() && (*matchPnts)[i][gp])
				glPointSize(4.);
			else
				glPointSize(2.);
			
			if (selectedPart==-1 || lnid==selectedPart)
			{
				R3Point pos(0,0,0);
				if (onlyRigid)
					pos = match->AlignedToNodePointPosition(lnid, gp, -1);
				else
					pos = match->AlignedToNodePointPosition(lnid, gp, match->SeedID());
				camera.Apply(pos);
				double r, g, b;
				if (drawConfidence && coAligned!=NULL && clusteringID>=0)
				{
					int cid = 0, pid = 0;
					double mnv = 0, mxv = 1;
					bool valid = true;
					coAligned->GetClusterForPoint(clusteringID, i, p, &cid, &pid, valid);
					double cv = valid ? coAligned->GetConfidenceValue(clusteringID, cid, pid, i, p, valid) : -1;
					//double cv = 1.;
					if (!valid)
					{
						//std::cout<<"[WARNING] Computing confidence"<<std::endl;;
						clusteringID = coAligned->GetClusteringFromMatchesID(match->SeedID());
						const TemplateNodeDeform & deform = match->Deformation(lnid);
						int method = CoAlignedParts::CCF_Dist2All;
						int clusterID, partInClusterID;
						bool valid;
						coAligned->GetClusterWithPart(clusteringID, i, lnid, &clusterID, &partInClusterID, valid);
						assert(valid);
						coAligned->ComputeConfidences(clusteringID, clusterID, method);
						cv = coAligned->GetConfidenceValue(clusteringID, cid, pid, i, p, valid);
						assert(valid);
					}
					
					coAligned->GetConfidenceValBounds(clusteringID, cid, &mnv, &mxv);
					window->MapScalarToColor(cv, mnv, mxv, r, g, b);
				}
				else if (isectBuffer==NULL)
					window->MapIntToColor(i, r, g, b);
				else
				{
					int mid = i+1;
					unsigned char r1 = (char)mid;
					unsigned char g1 = (char)(mid>>8);
					unsigned char b1 = (char)(p);
					unsigned char a1 = (char)(p>>8);
					glColor4ub(r1, g1, b1, a1);
				}
				
				if (isectBuffer==NULL)
					glColor3d(r, g, b);
				
				glBegin(GL_POINTS);
				glVertex3d(pos.X(), pos.Y(), pos.Z());
				glEnd();
			}
		}

	}
	if (isectBuffer!=NULL)
		window->LoadCurrentBuffer();
}

void VisShapeTemplate::DrawTemplateMesh(bool drawMesh, bool drawLabels, bool drawMislabeled, bool drawGTMesh,
										DatabaseVisWindow * window, int subwindow, 
										const R3Affine & camera, const ShapeLabels * labels)
{
	if (labels==NULL)
		return;
	
	const GeomShape * shape = labels->GetShape();
	const ShapeLabels * gtLabels = NULL;
	if (drawMislabeled || drawGTMesh)
	{
		const ShapeValue * gtLabelsVals = shape->GetShapeVal(ShapeValue::m_sDefaultSceneGraphLabels);
		if (gtLabelsVals==NULL)
		{
			const GeomShape * mainShape = ((Model3D*)shape->GetModel())->GetGeomShape(Model3D::m_sMainGeomName);
			if (mainShape!=NULL);
			gtLabelsVals = mainShape->GetShapeVal(ShapeValue::m_sDefaultSceneGraphLabels);
		}
		
		if (gtLabelsVals!=NULL)
			gtLabels = gtLabelsVals->SafeCastToLabels();
			
		if (gtLabels==NULL)
			std::cout<<"[WARNING] Could not find gt labels."<<std::endl;
	}

//	std::vector<VisModel3D *> & allVis = window->GetDrawableModels();
//	
//	int mid = shape->GetModel()->ModelID();
//	assert(mid>=0 && mid < (int)allVis.size());
//	int camID = allVis[mid]->SubwindowCameraID(subwindow, window);
//	R3Affine camera = allVis[mid]->GetCameraTransform(camID);
	window->LoadStandardSceneProjectionCamera();
	
	if ((drawMesh || drawLabels) && !drawGTMesh)
	{
		glEnable(GL_COLOR);
		const ShapePointCloud * pnts = shape->SafeCastToPointCloud();
		const ShapeMesh * mesh = shape->SafeCastToMesh();
		
		if (pnts!=NULL)
		{
			//glPointSize(8.);
			glPointSize(4.);
			if (drawMesh)
				glColor3d(.2, .2, .2);
			glDisable(GL_LIGHTING);
			glBegin(GL_POINTS);
			for (int i=0; i<pnts->NumPoints(); i++)
			{
				int mnid = (int)labels->Value(pnts->GetPoint(i), 0);
				if (gtLabels!=NULL)
				{
					int gnid = (int)gtLabels->Value(pnts->GetPoint(i), 0);
					if (gtLabels->NodeLabel(gnid)==labels->NodeLabel(mnid))
						continue;
				}
				R3Point p = pnts->PointPosition(i);
				camera.Apply(p);
				double r, g, b;
				if (drawLabels)
				{
					window->MapIntToRadicalColor(mnid, r, g, b);
					glColor3d(r,g,b);
				}
					
				glVertex3d(p.X(), p.Y(), p.Z());
			}
			glEnd();
		}
		else if (mesh!=NULL)
			VisModel3D::DrawMeshFaces(mesh, false, window, camera);
	}
	if (drawGTMesh && gtLabels!=NULL)
	{
		glShadeModel(GL_SMOOTH);
		glEnable(GL_DEPTH_TEST);
		glEnable(GL_BLEND);
		
		// TODO: do it properly w/ gfx card (i.e. upload geometry)
		glEnable(GL_LIGHTING);
		glDisable(GL_CULL_FACE);
		glEnable(GL_NORMALIZE);
		glEnable(GL_RESCALE_NORMAL);
		glDisable(GL_TEXTURE_2D);
		glDisable(GL_COLOR);

		PointCloudAllFaces * allFaces = NULL;
		Model3D * mdl = (Model3D*)shape->GetModel();
		allFaces = ShapePointCloud::CreateAllFacesSet("allfaces", Model3D::m_sMainGeomName, mdl);

		GeomShape * shape = mdl->GetGeomShape(Model3D::m_sMainGeomName);
		assert(shape!=NULL);
		ShapeMesh * mesh = shape->SafeCastToMesh();
		assert(mesh!=NULL);
		static GLfloat surfaceMaterial[4];
		glBegin(GL_TRIANGLES);
		for (int i=0; i<mesh->NFaces(); i++)
		{
			int gnid = (int)gtLabels->Value(allFaces->GetPoint(i), 0);
			window->MapIntToRadicalColor(gnid, surfaceMaterial);
			glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, surfaceMaterial); 
			R3Point p1 = mesh->FacePoint(i, 0, 0);
			R3Point p2 = mesh->FacePoint(i, 1., 0);
			R3Point p3 = mesh->FacePoint(i, 0, 1.);
			camera.Apply(p1);
			camera.Apply(p2);
			camera.Apply(p3);
			glVertex3d(p1.X(), p1.Y(), p1.Z());
			glVertex3d(p2.X(), p2.Y(), p2.Z());
			glVertex3d(p3.X(), p3.Y(), p3.Z());
		}
		glEnd();
	}
}

void VisShapeTemplate::DrawDebugNearestSurface(bool drawMislabeled, bool surf1, bool surf2, const TMatch * matched,
											   int partID, const DatabaseVisWindow * window, const R3Affine & model2cam)
{
	const ShapeLabels * gtLabels = NULL;
	if (drawMislabeled)
	{
		const GeomShape * shape = matched->GetMatchedGeomAsPnts();
		const ShapeValue * gtLabelsVals = NULL;
		if (shape!=NULL)
			gtLabelsVals = shape->GetShapeVal(ShapeValue::m_sDefaultSceneGraphLabels);
		if (gtLabelsVals==NULL)
		{
			const GeomShape * mainShape = ((Model3D*)shape->GetModel())->GetGeomShape(Model3D::m_sMainGeomName);
			if (mainShape!=NULL);
			gtLabelsVals = mainShape->GetShapeVal(ShapeValue::m_sDefaultSceneGraphLabels);
		}
		
		if (gtLabelsVals!=NULL)
			gtLabels = gtLabelsVals->SafeCastToLabels();
		
		if (gtLabels==NULL)
			std::cout<<"[WARNING] Could not find gt labels."<<std::endl;
	}
	
	window->LoadStandardSceneProjectionCamera();
	const ShapeLabels * labels = matched->GetPerPointLabels();
	const ShapePointCloud * pnts = labels->GetShape()->SafeCastToPointCloud();
	if (pnts==NULL)
	{
		std::cout<<"[WARNING] Cannot draw nearest surface: no points"<<std::endl;
		return;
	}
	glPointSize(4.);
	
	R3Point centroid = pnts->Centroid();
	double aveRad = pnts->AverageRadius(centroid);
	std::vector<R3Point> edges;
	glBegin(GL_POINTS);
	if (surf1)
	{
		for (int nodeID = 0; nodeID < matched->NTemplateNodes(); nodeID++)
		for (int p=0; p<matched->NPointsLabeledAsNode(nodeID) && (partID == nodeID || partID==-1); p++)
		{
			// tpnt - point cloud deformed to template
			int gid = matched->GetPointLabeledAsNode(nodeID, p);
			R3Point pPos = pnts->PointPosition(gid);
			//R3Point tpnt = matched->AlignedToNodePointPosition(nodeID, gid, matched->SeedID());
			double tempeigs[3];
			R3Point tpnt = matched->AlignedPointPosition(gid);

			// find nearest point on nodeID in template
			int tid = matched->NearestTemplatePointGlobalID(nodeID, tpnt);

			R3Point tPos = matched->DeformedTemplateSample(tid);
			matched->GlobalXForm().ApplyInverse(tPos);	// to (normalized) point space
			tPos *= aveRad;
			tPos += centroid;
			model2cam.Apply(pPos);
			model2cam.Apply(tPos);
			
			glColor3d(.2, .2, .2);
			glVertex3d(pPos.X(), pPos.Y(), pPos.Z());
			glColor3d(.2, 1., .2);
			glVertex3d(tPos.X(), tPos.Y(), tPos.Z());
			edges.push_back(tPos);
			edges.push_back(pPos);
		}
	}
	if (surf2)
	{
		
		for (int nodeID = 0; nodeID < matched->NTemplateNodes(); nodeID++)
		for (int p=0; p<matched->NPointsOnTemplateNode(nodeID) && (partID == nodeID || partID==-1); p++)
		{
			if (!matched->Deformation(nodeID).IsIncluded())
				continue;
			// pointsT[corrid] - find point in deformed template space
			int tid = matched->GlobalTmpltPnt(nodeID, p);
			R3Point tpnt = matched->DeformedTemplateSample(tid);
			matched->GlobalXForm().ApplyInverse(tpnt);	// to (normalized) point space
			int gid = matched->NearestPointWithLabel(nodeID, tpnt);

			tpnt *= aveRad;
			tpnt += centroid;
			R3Point p = pnts->PointPosition(gid);
			
			model2cam.Apply(tpnt);
			model2cam.Apply(p);
			
			glColor3d(.2, .2, .2);
			glVertex3d(p.X(), p.Y(), p.Z());
			glColor3d(.2, 1., .2);
			glVertex3d(tpnt.X(), tpnt.Y(), tpnt.Z());
			edges.push_back(tpnt);
			edges.push_back(p);
		}
			
	}
	glEnd();
	
	glBegin(GL_LINES);
	glLineWidth(1.);
	glColor3d(.2, 1., .2);
	for (int i=0; i<(int)edges.size(); i+=2)
	{
		glVertex3d(edges[i].X(), edges[i].Y(), edges[i].Z());
		glVertex3d(edges[i+1].X(), edges[i+1].Y(), edges[i+1].Z());
	}
	glEnd();
}

void VisShapeTemplate::DrawDebugEnergy(TMatchEnergy * energy, DatabaseVisWindow * window, TMatch * seedConfig,
									   const R3Affine & xform, int selectedPart, int selectedFace, int testValue)
{
	TMatch * match = ((TMatchEnergy*)energy)->GetMatch();
	const ShapePointCloud * pnts = match->GetMatchedGeomAsPnts();
	assert(pnts!=NULL);
	VisModel3D * mvis = NULL;

	int id = pnts->GetModel()->ModelID();
	double pntsRad = pnts->AverageRadius(pnts->Centroid());
	
	if (id<0 || id >= (int)window->GetDrawableModels().size())
	{
		std::cout<<"[WARNING] Cannot find selected point. NDraw="<<window->GetDrawableModels().size()<<std::endl;
		return;
	}
	mvis = window->GetDrawableModels()[id];
		
	DrawPntsIntersectionBuffer(match, window, xform);
	
	bool isPntSel = !mvis->SelectedPoint().Invalid();
	if (mvis!=NULL)
	{
		if (isPntSel && mvis->SelectedPoint().PointID()==-1)	// point not in set
		{
			double mindist = 10000000;
			int bestp2 = -1;
			for (int p2=0; p2<pnts->NumPoints(); p2++)
			{
				double dist = R3Distance(pnts->PointPosition(p2), mvis->SelectedPoint().GetPosition());
				if (dist < mindist || bestp2<0)
				{
					mindist = dist;
					bestp2 = p2;
				}
			}
			
			assert(bestp2>=0);
			mvis->SelectPoint(pnts->GetPoint(bestp2));
		}
		
		int selPartLocalID = -1;
		if (selectedPart!=-1)
		{
			int targetID = seedConfig->Deformation(selectedPart).NodeID();
			//std::cout<<"Target ID = "<<targetID<<std::endl;
			for (int pid = 0; pid < match->NTemplateNodes(); pid++)
				if (match->Deformation(pid).NodeID()==targetID)
					selPartLocalID = pid;
			if (selPartLocalID==-1)
				std::cout<<"[WARNING] Selected part does not exist in current config"<<std::endl;
		}
		if (isPntSel)
		{
			int selID = mvis->SelectedPoint().PointID();
			static int prevSelID = -1;
			static int prevSelPartID = -1;

			if (prevSelID!=selID || prevSelPartID!=selPartLocalID)
			{
				prevSelID = selID;
				prevSelPartID = selPartLocalID;
				
				int nfeat;
				const double * feat1 = energy->GetMatch()->GetLocalPntsFeatures(selID, -1, nfeat);

				std::cout<<"F = [ ";
				for (int i=0; i<nfeat; i++)
					std::cout<<feat1[i]<<" ";
				std::cout<<"]"<<std::endl;

				if (selPartLocalID!=-1)
				{
					const TemplateNodeDeform & deform = match->Deformation(selPartLocalID);
					VKString nname = match->GetSTemplate()->GetNode(deform.NodeID())->NodeName();

					std::cout<<"Label ["<<prevSelID<<"] = "<<nname.c_str();
					std::cout<<" E="<<energy->ComputeEGlobalPnt(selID, selPartLocalID)<<std::endl;
				}
				else
				{
					const TemplateNodeDeform & deform = match->Deformation(match->LocalNodeID(selID));
					VKString nname = match->GetSTemplate()->GetNode(deform.NodeID())->NodeName();

					std::cout<<"Label ["<<prevSelID<<"] = "<<nname.c_str();
					std::cout<<" E="<<energy->ComputeEGlobalPnt(selID, match->LocalNodeID(selID))<<std::endl;
				}
			}
		}
		window->LoadStandardSceneProjectionCamera();
		
		int p1 = -1;
		if (isPntSel)
			p1 = mvis->SelectedPoint().PointID();

		// initialize per point values to render
		std::vector<double> eglobalpnts;
		std::vector<bool> valid;
		for (int p2=0; p2<pnts->NumPoints(); p2++)	
		{
			eglobalpnts.push_back(0);
			valid.push_back(false);
		}

		// part not selected point not selected - nothing to render
		if (selPartLocalID<0 && !isPntSel)
			return;
		
		std::vector<double> minmax;
		if (selPartLocalID<0 && isPntSel)	// point is selected - render smoothness
		{
			if (testValue==-1)
			{
				assert(false);
			}
			else if (testValue>=0)	// difference between objects in feature
			{
				int nfeat;
				const double * feat1 = energy->GetMatch()->GetLocalPntsFeatures(p1, -1, nfeat);
				double feat1cpy[6];
				for (int i=0; i<6; i++)
					feat1cpy[i] = feat1[i];
				if (testValue <= nfeat)
				{
					for (int p2=0; p2<pnts->NumPoints(); p2++)
					{
						const double * feat2 = energy->GetMatch()->GetLocalPntsFeatures(p2, -1, nfeat);
						double eval = 0;
						if (testValue<nfeat)
							eval += pow(feat1cpy[testValue]- feat2[testValue], 2.);
						else
						{
							for (int i=0; i<6; i++)
								eval += pow(feat1cpy[i]- feat2[i], 2.);
						}
						minmax.push_back(eval);
						eglobalpnts[p2] = eval;
						valid[p2] = true;
					}
				}
//				std::cout<<"MinMax=["<<minval<<" "<<maxval<<"]"<<std::endl;
			}
		}
		else if (selPartLocalID>=0)	// point and parts are selected - render point energy
		{
			if (testValue==-1)
			{
				for (int p2=0; p2<pnts->NumPoints(); p2++)
				{
					double eval = -energy->ComputeEGlobalPnt(p2, selPartLocalID);
					minmax.push_back(eval);
					eglobalpnts[p2] = eval;
					valid[p2] = true;
				}
			}
			else if (testValue>=0)	// draw difference in feature (to closest point)
			{
				for (int p2=0; p2<pnts->NumPoints(); p2++)
				{
					int nfeat;
					const double * feat1 = energy->GetMatch()->GetLocalPntsFeatures(p2, -1, nfeat);
					double feat1cpy[6];
					for (int i=0; i<6; i++)
						feat1cpy[i] = feat1[i];

					double bestVal = 1000000;
					const PartLocalGeometry * locGeom = energy->GetMatch()->NodeEllipsoid(selPartLocalID).GetLocalGeometry();
					for (int np=0; np<locGeom->NPnts(); np++)
					{
						const double * feat2 = locGeom->LocalOrientAsFeatures(np, R3identity_affine);
						double eval = 0;
						if (testValue<nfeat)
							eval += pow(feat1cpy[testValue]- feat2[testValue], 2.);
						else
						{
							for (int i=0; i<6; i++)
								eval += pow(feat1cpy[i]- feat2[i], 2.);
						}
						bestVal = vkMin(eval, bestVal);
					}
					minmax.push_back(bestVal);
					eglobalpnts[p2] = bestVal;
					valid[p2] = true;
				}
			}
		}
		else
			assert(false);
		
		if (testValue==-2 || testValue==-3) // temp debugging: render template points
		{
			if (selPartLocalID >= 0 || testValue == -3)
			{
				eglobalpnts.clear();
				const LocalGeometryClassifier * classifier = match->GetSTemplate()->GetLocalClassifier(match->SeedID());
				std::vector<int> bestNodeIDs;
//				std::ofstream testDebug("atrender.txt");
				for (int n=0; n < match->NTemplateNodes(); n++)
				{
					const PartLocalGeometry * part = match->NodeEllipsoid(n).GetLocalGeometry();
//					std::cout<<"Part["<<n<<"] : "<<part<<" NPnts="<<(part==NULL ? 0 : part->NPnts())<<std::endl;
					for (int i=0; i<match->NPointsOnTemplateNode(n); i++)
					{
						assert(part!=NULL);
						const double * feat = part->LocalOrientAsFeatures(i, R3identity_affine);
						const double * logProbs = classifier->LogProbs(part->NLocalFeatures(), feat);
						int bestNode = -1;
						for (int pid = 0; pid < match->NTemplateNodes(); pid++)
							if (bestNode==-1 || logProbs[pid] < logProbs[bestNode] && logProbs[pid] >= 0)
								bestNode = pid;
						int nid1 = match->Deformation(n).NodeID();
						int nid2 = match->Deformation(bestNode).NodeID();
						const STemplate * st =match->GetSTemplate();
//						testDebug<<st->GetNode(nid1)->NodeName().c_str()<<" "<<st->GetNode(nid2)->NodeName().c_str()<<"\n";
//						testDebug<<feat[0]<<" "<<feat[1]<<" "<<feat[2]<<" "<<feat[3]<<" "<<feat[4]<<" "<<feat[5]<<"\n";
//						
//						R3Point pos = part->m_pPnts->PointPosition(i);
//						testDebug<<pos[0]<<" "<<pos[1]<<" "<<pos[2]<<"\n";

						//bestNodeIDs.push_back(match->Deformation(bestNode).NodeID());
						//bestNodeIDs.push_back(n+1);
						bestNodeIDs.push_back(bestNode+1);
						double value = selPartLocalID >= 0 ? logProbs[selPartLocalID] : 0;
						minmax.push_back(value);
						eglobalpnts.push_back(value);
					}
				}
//				testDebug.close();
//				exit(0);
				R3Point centroid = pnts->Centroid();
				double aveRad = pnts->AverageRadius(centroid);
				std::sort(minmax.begin(), minmax.end());
				double minval = 0;
				double maxval = 1;
				//int minid = (int) minmax.size()*.2;
				//int maxid = (int) minmax.size()*.8;
				int minid = 0;
				int maxid =(int) minmax.size() - 1;
				if (minmax.size()>0)
				{
					minval = minmax[minid];
					maxval = minmax[maxid];
				}
				for (int i=0; i<match->NTemplatePoints(); i++)
				{
					assert(i < (int)eglobalpnts.size());
					double value = eglobalpnts[i];
					R3Point tPos = (match->DeformedTemplateSample(i)*aveRad+centroid);
					xform.Apply(tPos);
					double r = 1., g = 0., b = 0.;
					if (testValue==-2)
						window->MapScalarToColor(value, minval, maxval, r, g, b);
					else if (testValue==-3)
						window->MapIntToRadicalColor(bestNodeIDs[i], r, g, b);
					glColor3d(r, g, b);
					R3Sphere(tPos, .01).Draw();
				}
			}
		}
		else   // normally render points to be labeled
		{
			std::sort(minmax.begin(), minmax.end());
			double minval = 0;
			double maxval = 1;
			int minid = (int) minmax.size()*.1;
			int maxid = (int) minmax.size()*.9;
			if (minmax.size()>0)
			{
				minval = minmax[minid];
				maxval = minmax[maxid];
			}

			for (int p2=0; p2<pnts->NumPoints(); p2++)
			{
				assert(p2 < (int)eglobalpnts.size());
				double value = eglobalpnts[p2];
				R3Point pos2 = pnts->PointPosition(p2);
				xform.Apply(pos2);
				double r = 1., g = 0., b = 0.;
				if (valid[p2])
					window->MapScalarToColor(value, minval, maxval, r, g, b);
				glColor3d(r, g, b);
				R3Sphere(pos2, p2 == p1 ? .015 : .01).Draw();
			}
		}
	}
}

void VisShapeTemplate::DrawPart(TMatchEnergy * energy, std::vector<bool> & selectedPnt,
								DatabaseVisWindow * window, const R3Affine & camera)
{
	window->LoadStandardSceneProjectionCamera();
	const ShapePointCloud * pnts = energy->GetMatch()->GetMatchedGeomAsPnts();
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		int lnid = energy->GetMatch()->LocalNodeID(i);
		glPointSize(2.);
		int nodeID = energy->GetMatch()->Deformation(lnid).NodeID();
		R3Point pos = pnts->PointPosition(i);
		camera.Apply(pos);
		double r, g, b;
		//if (selectedPart.find(lnid)!=selectedPart.end())
		if (i < (int)selectedPnt.size() && selectedPnt[i])
		{
			glPointSize(2.);
			window->MapIntToColor(nodeID, r, g, b);
			glColor3d(r,g,b);
		}
		else
		{
			glColor4d(.5, .5, .5, .5);
			glPointSize(1.);
		}
		glBegin(GL_POINTS);
		glVertex3d(pos.X(), pos.Y(), pos.Z());
		glEnd();
	}
}


bool VisShapeTemplate::Draw(DatabaseVisWindow * window)
{
	if (m_bDisable)
		return false;

	std::vector<RenderableFace> trFaces;
	std::vector<VisModel3D *> & allVis = window->GetDrawableModels();
	
	// draw mesh
	double br, bg, bb;
	window->GetBackgroundColor(&br, &bg, &bb);
	glClearColor(br, bg, bb, 1.0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	
	if (m_iCurrMatched<0)
		return true;
	
	// DRAW ORIGINAL TEMPLATE
	int templateSubwindow = 0;

		// original template camera
	R3Affine cam0 = window->GetEmptySubwindowCamera(templateSubwindow);
	window->LoadStandardSceneProjectionCamera();
	if (m_CoAlignedPntSeeds.size()>0)
	{
		std::cout<<"Drawing co-aligned :"<<m_CoAlignedPoints.size()<<std::endl;
		assert(m_CoAlignedPoints.size()==m_CoAlignedPntSeeds.size());
		glPointSize(1.);
		glBegin(GL_POINTS);
		for (int i=0; i<(int)m_CoAlignedPoints.size(); i++)
		{
			//if (m_CoAlignedPntSeeds[i]==m_iCurrMatched)
			{
				//std::cout<<"   Di="<<i<<std::endl;
				double r, g, b;
				window->MapIntToColor(i, r, g, b);
				glColor3d(r, g, b);
				for (int p=0; p < (int)m_CoAlignedPoints[i].size(); p++)
				{
					R3Point pos = m_CoAlignedPoints[i][p];
					cam0.Apply(pos);
					//std::cout<<"pos="<<pos.X()<<" "<<pos.Y()<<" "<<pos.Z()<<std::endl;
					glVertex3d(pos.X(), pos.Y(), pos.Z());
				}
			}
		}
		glEnd();
	}
	
		// camera for training model
	const ShapeLabels * trainLabels = m_apSeedConfigs[m_iCurrMatched]->GetSTemplate()->TrainingLabels();
	if (trainLabels!=NULL)
	{
		R3Affine train2cam = cam0;
		R3Affine I2train = IdentityToModelXForm(window, templateSubwindow, trainLabels->GetShape());
		R3Affine train2template = m_apSeedConfigs[m_iCurrMatched]->SeedAlignment();
		train2template.Apply(train2cam);
		I2train.ApplyInverse(train2cam);

		// draw model on which template was trained (first model)		
		DrawTemplateMesh(IsFlagSet(VISTMPLT_TemplatePnts), IsFlagSet(VISTMPLT_TemplateSegPnts),
						 false, IsFlagSet(VISTMPLT_TemplateGTSegMesh),
						 window, templateSubwindow, train2cam, trainLabels);
	}
	
		// draw template
	if (IsFlagSet(VISTMPLT_SeedTemplate))
		DrawTemplateConfig(m_apSeedConfigs[m_iCurrMatched], window, m_iVisFlags, trFaces, cam0);

	DrawIntersectionBuffers(m_apSeedConfigs[m_iCurrMatched], window, cam0);
	m_TemplateCamera = cam0;
	
	// DRAW EDITABLE TEMPLATE
	if (m_iInteractionFlags & INTERACT_CreateTempalte)
	{
		if (m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apSeedConfigs.size())
		{
			TMatch * match = m_apSeedConfigs[m_iCurrMatched];
			if (m_iSelectedPart>=0 && m_iSelectedPart < match->NTemplateNodes())
			{
				const PartBox & ellipsoid = match->NodeEllipsoid(m_iSelectedPart);
				const TemplateNodeDeform & deform = match->Deformation(m_iSelectedPart);
				
				std::vector<RenderableFace> tmpFaces;
				if ((m_iVisFlags & VISTMPLT_DrawTemplateAsBoxes) && ellipsoid.IsType(PartBox::BOXTYPE_BoxSurface))
					DrawTemplateAsBoxes(ellipsoid, m_iSelectedPart+1, window, tmpFaces, cam0, true, m_iSelectedFace);
			}
		}
	}
		
	// DRAW DEFROMED TEMPLATE
	// draw reference template
	if (IsFlagSet(VISTMPLT_DrawReferenceModel) && window->WindowToModel().size()==1)
	{
		assert(m_apMatchedConfigs.size()>=1);
		TMatch * matched = m_apMatchedConfigs[0]->GetMatch();
		R3Affine model2cam = window->GetEmptySubwindowCamera(templateSubwindow);
		R3Affine I2model = IdentityToModelXForm(window, templateSubwindow, matched->GetMatchedGeomAsPnts());
		matched->GlobalXForm().Apply(model2cam);
		I2model.ApplyInverse(model2cam);
		DrawTemplateMesh(true, false, false, false, window, templateSubwindow, model2cam, matched->GetPerPointLabels());
	}
	
	TMatch * matched = NULL;
	if (m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apMatchedConfigs.size())
		matched = m_apMatchedConfigs[m_iCurrMatched]->GetMatch();
//	std::cout<<"Configs: "<<m_iCurrMatched<<" / "<<m_apMatchedConfigs.size()<<std::endl;
	
	int matchedSubwindow = window->NumMainWindows()>1 ? 1 : 0;
	if (matched!=NULL)
	{
		// template camera
		R3Affine cam1 = window->GetEmptySubwindowCamera(matchedSubwindow);
		
		// point cloud camera
		R3Affine model2cam = cam1;
		R3Affine I2model = IdentityToModelXForm(window, matchedSubwindow, matched->GetMatchedGeomAsPnts());
		R3Affine pnts2template = matched->GlobalXForm();
		pnts2template.Apply(model2cam);
		I2model.ApplyInverse(model2cam);
		
		// draw textured
		if (IsFlagSet(VISTMPLT_DrawMeshTextures))
		{
			const ShapeMesh * mesh = matched->GetMatchedGeomAsPnts()->GetMesh();
			assert(mesh!=NULL);
			assert(window->GetDrawableModels().size()==1);
			VisModel3D * mvis = window->GetDrawableModels()[0];
			GeometryTextures * textures = mvis->GetTextures(mesh->GeometryName());
			assert(textures!=NULL);
			VisModel3D::DrawMeshFaces(mesh, true, window, model2cam,
									  textures->m_mMaterialImageToTextureID, textures->m_iNTextures);
		}
		
		// draw segmented
		if (IsFlagSet(VISTMPLT_DrawSegmentedMesh))
		{
			const ShapeMesh * mesh = matched->GetMatchedGeomAsPnts()->GetMesh();
			assert(mesh!=NULL);
			if (mesh->GetShapeVal("FaceSegments")==NULL)
				std::cout<<"[WARNING] Could not draw faces see IsFlagSet(VISTMPLT_DrawSegmentedMesh)"<<std::endl;
			else
			{
				const ShapeLabels * labels = mesh->GetShapeVal("FaceSegments")->SafeCastToLabels();
				assert(labels!=NULL);
				VisModel3D::DrawGeomWithLabels(mesh, labels, window, model2cam, -1);
			}
		}
		
//		// TEMP: Debug triangle smoothness (projecting points to trinalges)
//		VisModel3D * mvis = window->GetDrawableModels()[0];
//		ShapePoint p = mvis->SelectedPoint();
//		if (!p.Invalid())
//		{
//			const ShapeMesh * mesh = matched->GetMatchedGeomAsPnts()->GetMesh();
//			
//			
//			int fid;
//			double b1, b2;
//			p.GetBarycentricCoords(&fid, &b1, &b2, &mesh);
//			std::vector<int> fids;
//			mesh->AdjacentFaces(fid, fids);
//			glBegin(GL_TRIANGLES);
//			double pi = 3.14159265359;
//			R3Vector n1 = mesh->FaceNormal(fid);
//			fids.push_back(fid);
//			double eps = 0.000001;
//			for (int i=0; i<(int)fids.size(); i++)
//			{
//				R3Vector n2 = mesh->FaceNormal(fids[i]);
//				double dot = vkAbs(n1.Dot(n2));
//				if (dot >1.)
//					dot = 1.;
//				double angle = acos(dot)/pi;
//				std::cout<<"Penalty: "<<-log(angle+eps)<<" Angle["<<i<<"] = "<<angle<<" Dot: "<<n1.Dot(n2)<<std::endl;
//				double r, g, b;
//				window->MapScalarToColor(angle, 0, 1, r, g, b);
//				glColor3d(r, g, b);
//				for (int v=0; v<3; v++)
//				{
//					R3Point p=mesh->VertexPosition(mesh->VertexOnFace(fids[i], v));
//					model2cam.Apply(p);
//					glVertex3d(p.X(), p.Y(), p.Z());
//				}
//			}
//			glEnd();
//		}
//		// END TEMP DEBUG
		
		
		
		// draw labeled point cloud
		//std::cout<<"Draw labeled points: "<<IsFlagSet(VISTMPLT_MatchedPnts)<<" "<<IsFlagSet(VISTMPLT_MatchedSegPnts)<<std::endl;
		DrawTemplateMesh(IsFlagSet(VISTMPLT_MatchedPnts), IsFlagSet(VISTMPLT_MatchedSegPnts),
						 IsFlagSet(VISTMPLT_DrawMislabeledOnly), IsFlagSet(VISTMPLT_MatchedGTSegMesh), 
						 window, matchedSubwindow, model2cam, matched->GetPerPointLabels());
		
		
		// draw undeformed template over matched model
		if (IsFlagSet(VISTMPLT_SeedOverMatched))
		{
			R3Affine seed2I = cam1;
			matched->SeedAlignment().ApplyInverse(seed2I);		// seed -> pnts
			matched->GlobalXForm().Apply(seed2I);				// pnts -> I 
			DrawTemplateConfig(m_apSeedConfigs[m_iCurrMatched], window, m_iVisFlags, trFaces, seed2I);
		}
		
		// draw deformed template over matched model
		if (IsFlagSet(VISTMPLT_DeformedOverMatched))
			DrawTemplateConfig(matched, window, m_iVisFlags, trFaces, cam1);
		
		if (IsFlagSet(VISTMPLT_DeformedPoints) || IsFlagSet(VISTMPLT_AllDeformedPoints))
		{
			assert(m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apMatchedConfigs.size());
			TMatch * currMatch = m_apMatchedConfigs[m_iCurrMatched]->GetMatch();

			std::vector<int> * currCluster = NULL;

			DrawDeformedGeometry(window, window->GetEmptySubwindowCamera(matchedSubwindow), currMatch,
								 m_apMatchedConfigs, m_pCoAlignedParts, m_iSelectedClustering, m_iSelectedCluster,
								 IsFlagSet(VISTMPLT_DeformedPoints), IsFlagSet(VISTMPLT_AllDeformedPoints),
								 IsFlagSet(VISTMPLT_DrawPointConfidence), m_iSelectedPart, m_bOnlyRigidAlignment, NULL,
								 &m_SelectedPointsFromDeformed);
			
			if (m_bSelectingDeformedPnts)
			{
				if (!m_bDeformedPntsBufferInitialized)
				{
					m_bDeformedPntsBufferInitialized = true;
					m_DeformedPntsBuffer = window->CreateBuffer();
				}
				DrawDeformedGeometry(window, window->GetEmptySubwindowCamera(matchedSubwindow), currMatch,
									 m_apMatchedConfigs, m_pCoAlignedParts, m_iSelectedClustering, m_iSelectedCluster,
									 IsFlagSet(VISTMPLT_DeformedPoints), IsFlagSet(VISTMPLT_AllDeformedPoints),
									 IsFlagSet(VISTMPLT_DrawPointConfidence), m_iSelectedPart, m_bOnlyRigidAlignment,
									 &m_DeformedPntsBuffer);
			}
			if (m_iSelDeformedX>=0)
			{
				glMatrixMode(GL_MODELVIEW);
				glLoadIdentity();
				
				glMatrixMode(GL_PROJECTION);
				glLoadIdentity();
				glOrtho(0.0, 1.0, 0.0, 1.0, 0., 10.);
				
				double minx = (double)vkMin(m_iSelDeformedX, m_iSelDeformedNewX) / (double)window->Width();
				double miny = (double)vkMin(m_iSelDeformedY, m_iSelDeformedNewY) / (double)window->Height();
				double w =(double)vkAbs(m_iSelDeformedX-m_iSelDeformedNewX) / (double)window->Width();
				double h =(double)vkAbs(m_iSelDeformedY-m_iSelDeformedNewY) / (double)window->Height();
				window->DrawSquare(minx, miny, w, h, 1., 1., 1., .5, 1., .5, .5, .5);
			}
		}
		
		// debug nearest surface
		if (IsFlagSet(VISTMPLT_DebugNearestSurface1) || IsFlagSet(VISTMPLT_DebugNearestSurface2))
			DrawDebugNearestSurface(IsFlagSet(VISTMPLT_DrawMislabeledOnly),
									IsFlagSet(VISTMPLT_DebugNearestSurface1),
									IsFlagSet(VISTMPLT_DebugNearestSurface2),
									matched, m_iSelectedPart, window, model2cam);
		
		// debug smoothness term
		if (IsFlagSet(VISTMPLT_DebugSmoothness))
		{
			if (m_iCurrMatched>=0 && m_iCurrMatched < (int)m_apMatchedConfigs.size())
				DrawDebugEnergy(m_apMatchedConfigs[m_iCurrMatched], window, m_apSeedConfigs[0],
								model2cam, m_iSelectedPart, m_iSelectedFace, m_iTestingSomething);
		}
	}
		
	window->LoadStandardSceneProjectionCamera();
	//std::cout<<"TrFaces: "<<trFaces.size()<<std::endl;
	window->GetRenderer()->DrawTransparentFaces(trFaces);
	
	DrawSelectedParts(window);
	
	if (window->IsDesignFlagOn(DatabaseVisWindow::RENDER_MODEL_LIST))
		window->DrawModelListSubwindow();
	return true;
}

void VisShapeTemplate::DrawSelectedParts(DatabaseVisWindow * window)
{
	int nAdditional = window->NumAdditionalWindows();
	int currSubwindow = window->NumMainWindows();
	int nMax = nAdditional + currSubwindow;
	for (int i=0; i<(int)m_SelectedPointsFromDeformed.size(); i++)
	{
		if (m_SelectedPointsFromDeformed[i].size()>0)
		{
			assert(m_pDB!=NULL);
			int mid = m_pDB->GetModel(m_apMatchedConfigs[i]->GetMatch()->ModelName())->ModelID();
			VisModel3D * m = window->GetDrawableModels()[mid];
			int camID = m->SubwindowCameraID(currSubwindow, window);
			R3Affine camera = m->GetCameraTransform(camID);
			
			assert(currSubwindow < (int)window->WindowToModel().size());
			window->WindowToModel()[currSubwindow] = mid;
			
			DrawPart(m_apMatchedConfigs[i], m_SelectedPointsFromDeformed[i], window, camera);
			currSubwindow++;
			if (currSubwindow >= nMax)
				break;
		}
	}
}

void VisShapeTemplate::DrawIntersectionBuffers(const TMatch * tConfig, DatabaseVisWindow * window, const R3Affine & cam)
{
#ifndef MESA_RENDERING
	if (!m_bIBufferInitialized)
	{
		m_bIBufferInitialized = true;
		m_IntersectionBuffer = window->CreateBuffer();
	}
	
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_IntersectionBuffer.m_iFrameBufferObject);
    glViewport( 0, 0, m_IntersectionBuffer.m_iWidth, m_IntersectionBuffer.m_iHeight);

	glEnable(GL_DEPTH_TEST);
    glDisable(GL_LIGHTING);
    glDisable(GL_POLYGON_SMOOTH);
    glShadeModel(GL_FLAT);
    glDisable(GL_BLEND);
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_CULL_FACE);
    glDisable(GL_DITHER);
    glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
	
	glClearColor(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	glBegin(GL_TRIANGLES);
	for (int i=0; i<tConfig->NTemplateNodes(); i++)
	{
		int nodeID = i+1;
		const PartBox & ellipsoid = tConfig->NodeEllipsoid(i);
		const TemplateNodeDeform & deform = tConfig->Deformation(i);
		
		std::vector<RenderableFace> boxFaces;
		if (ellipsoid.IsType(PartBox::BOXTYPE_BoxSurface))
		{
			DrawTemplateAsBoxes(ellipsoid, nodeID, window, boxFaces, cam, false, -1);
			if (boxFaces.size()==0)
				continue;
			assert(boxFaces.size()==12);
			for (int f=0; f<(int)boxFaces.size(); f++)
			{
				// rgba: nodeID (3 bytes), face(1 byte)
				unsigned char r = (char)nodeID;
				unsigned char g = (char)(nodeID>>8);
				unsigned char b = (char)(nodeID>>16);
				unsigned char a = (char)(f / 2 + 1);
				glColor4ub(r, g, b, a);
				for (int fid=0; fid<3; fid++)
					glVertex3d(boxFaces[f].vtx[fid].X(), boxFaces[f].vtx[fid].Y(), boxFaces[f].vtx[fid].Z());
			}
		}
	}
	glEnd();
	
	window->LoadCurrentBuffer();
	
#else
	std::cout<<"[WARNING] Cannot do interaction w/ MESA renderer"<<std::endl;
#endif
}

void VisShapeTemplate::DrawPntsIntersectionBuffer(const TMatch * tConfig, DatabaseVisWindow * window, const R3Affine & cam)
{
#ifndef MESA_RENDERING
	const ShapePointCloud * pnts = tConfig->GetMatchedGeomAsPnts();
	if (pnts==NULL)
		return;
	
	if (!m_bPntsIBufferInitialized)
	{
		m_bPntsIBufferInitialized = true;
		m_PntsBuffer = window->CreateBuffer();
	}
	
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, m_PntsBuffer.m_iFrameBufferObject);
	glViewport( 0, 0, m_PntsBuffer.m_iWidth, m_PntsBuffer.m_iHeight);
	
	window->LoadStandardSceneProjectionCamera();
	glClearColor(0, 0, 0, 0);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_ACCUM_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
	
	glPointSize(4.);
	
	assert(pnts!=NULL);
//	static const ShapePointCloud * precompPnts = NULL;
//	static R3Point centroid(0,0,0);
//	static double aveRad = 0;
//	if (precompPnts!=pnts)
//	{
//		centroid = pnts->Centroid();
//		aveRad = pnts->AverageRadius(centroid);
//		precompPnts = pnts;
//	}
	
	glBegin(GL_POINTS);
	for (int i=0; i<pnts->NumPoints(); i++)
	{
		int pid = i+1;
		unsigned char r = (char)pid;
		unsigned char g = (char)(pid>>8);
		unsigned char b = (char)(pid>>16);
		unsigned char a = (char)(128);
		glColor4ub(r, g, b, a);
		R3Point p = pnts->PointPosition(i);
		cam.Apply(p);
		R3Sphere(p, .01).Draw();
	}
	glEnd();
	
	window->LoadCurrentBuffer();
#endif
	
}

R3Affine VisShapeTemplate::IdentityToModelXForm(DatabaseVisWindow * window, int wndID, const GeomShape * shape) const
{
    if (shape==NULL)
        shape = IdentityToModelXForm(window, wndID);
    if (shape==NULL)
        return R3identity_affine;

    // find transform
    R3Point centroid(0,0,0);
    double aveRad = 1.;
    if (shape->SafeCastToMesh()!=NULL)
    {
        centroid = shape->SafeCastToMesh()->Centroid();
        aveRad = shape->SafeCastToMesh()->AverageRadius(centroid);
    }
    else if (shape->SafeCastToPointCloud()!=NULL)
    {
        centroid = shape->SafeCastToPointCloud()->Centroid();
        aveRad = shape->SafeCastToPointCloud()->AverageRadius(centroid);
    }
        
    // align centroids -> scale
    R3Affine affIdentityToModel = R3identity_affine;
    affIdentityToModel.Translate(R3Vector(centroid.X(), centroid.Y(), centroid.Z()));
	affIdentityToModel.Scale(aveRad);
    return affIdentityToModel;
}

GeomShape * VisShapeTemplate::IdentityToModelXForm(DatabaseVisWindow * window, int wndID) const
{
	std::vector<VisModel3D *> & allVis = window->GetDrawableModels();
	
	// find transformation
	assert(wndID < (int)window->WindowToModel().size());
	int m0ID = window->WindowToModel()[wndID]; assert(m0ID < (int)allVis.size());
	if (m0ID==-1)
		return NULL;
	
	// get mesh
	static bool noGeomError = false;
	GeomShape * geom = allVis[m0ID]->GetModel()->GetGeomShape(Model3D::m_sMainGeomName);
	if (geom==NULL)
	{	
		if (!noGeomError)
		{
			std::cout<<"[WARNING] No main geom. mID="<<m0ID<<" mName=";
			std::cout<<allVis[m0ID]->GetModel()->ModelName().c_str()<<std::endl;
		}
		noGeomError = true;
		return NULL;
	}
	assert(geom!=NULL);
	
	return geom;
}






