/***************************************************************************
 *   Copyright (C) 2008 by Vladimir Kim									   *
 *   Author: Vladimir Kim (vk@princeton.edu)							   *
 ***************************************************************************/

#ifndef __VisWindow_H
#define __VisWindow_H

#include <vector>
#include <iostream>
#include <assert.h>

#include "ParamParser.h"
#include "VkFunctions.h"
#include "OpenGLHeaders.h"

class VisWindow;

// the off-screen buffers
struct BufferID
{
	GLuint m_iFrameBufferObject;
	GLuint m_iDepthBuffer;
	GLuint m_iColorBuffer;
	GLuint m_iTextureBuffer;
	unsigned char * m_aColorBuffer;
	int m_iWidth;
	int m_iHeight;
};

struct RenderableFace
{
	public:
		R3Point vtx[3];
		R3Vector norm;
		double r, g, b, a;
};


/**
 * Window renderer is a class that contains all implementation-specific details
 * for text rendering currently only Glut version is implemented, Qt is another option.
 * There is mutual dependence with VisWindow class - VisWindowRenderer makes calls 
 * for user input events, and VisWindow might make calls to issue special rendering commands.
 */
class VisWindowRenderer
{
	public:
		VisWindowRenderer(VisWindow * window);
	
		enum TextSize
		{
			TEXT_SIZE_SMALL,
			TEXT_SIZE_MEDIUM,
			TEXT_SIZE_LARGE,
		};

		virtual void RenderText(R3Point pnt, const VKString & text, 
								double r, double g, double b, TextSize size) const;
		virtual void RenderText(R2Point pnt, const VKString & text, 
								double r, double g, double b, TextSize size) const;

		virtual void TerminateWindow();
		virtual void UpdateScreen();
	
	// buffer
		virtual void CreateBuffer(BufferID & bufferID);
		virtual void LoadBuffer(BufferID & bufferID);
		virtual void FillDrawBuffer(BufferID & bufferID);
	
	// render transparent
		virtual void DrawTransparentFaces(const std::vector<RenderableFace> & faces, bool lighting = false) const;

	
	protected:
		VisWindow * m_pWindow;
};


/**
 * Vis window is a general window class. It has abstract functions for user input events
 * and handles some abstract rendering (subwindows / etc...)
 */
class VisWindow
{
	public:
		enum SpecialKeys
		{
			SPECIAL_NONE  = 0,
			SPECIAL_ALT   = 2,
			SPECIAL_CTRL  = 4,
			SPECIAL_SHIFT = 8,
		};
		enum MouseButton
		{
			MOUSE_LEFT,
			MOUSE_RIGHT,
			MOUSE_MIDDLE
		};
		enum MouseButtonState
		{
			MOUSE_BT_UP,
			MOUSE_BT_DOWN
		};
	
		enum WndPosition
		{
			WND_LEFT	= 1, 
			WND_VCENTER = 2,
			WND_RIGHT	= 4,
			WND_TOP		= 8,
			WND_HCENTER = 16,
			WND_BOTTOM  = 32
		};
		VisWindow();
		
		void SetRenderer(VisWindowRenderer * renderer);
		VisWindowRenderer * GetRenderer();
		const VisWindowRenderer * GetRenderer() const;
		virtual VKString GetClassName() const = 0;
		virtual void LoadStandardSceneProjectionCamera(const R3Affine & mview = R3identity_affine ) const;
		
		virtual void KeyPressed(const char key, int mouseX, int mouseY);
		virtual void KeyReleased(const char key, int mouseX, int mouseY);
		virtual void MouseButtonEvent(int x, int y, MouseButton mouseButtonID, 
									  MouseButtonState newState);
		virtual void MouseMotionEvent(int x, int y, int dx, int dy);
	
		virtual void UpdateMouseButtons(MouseButton mouseButtonID, MouseButtonState newState);
		virtual bool IsMouseButtonDown(MouseButton bt);
	
		virtual void SetSpecialKey(int specialKeyFlags);
		virtual bool Shift();
		virtual bool Ctrl();
		virtual bool Alt();
		
		virtual void Draw() = 0;
		virtual void InitializeGL();
		virtual void Resize(int w, int h);
		virtual int Width() const;
		virtual int Height() const;
	
		virtual void MapScalarToColor(double value, double minValue, double maxValue, 
									  double & r, double & g, double & b) const;
		virtual void MapIntToColor(int i, double & r, double & g, double & b) const;
		virtual void StandardSurfaceColor(double & r, double & g, double & b, double & a) const;
		virtual double GetAlpha() const;
		virtual void SetAlpha(double alpha);

		virtual void MapIntToColor(int i, GLfloat material[4]) const;
		virtual void MapScalarToColor(double value, double minValue, double maxValue,
									  GLfloat material[4]) const;
		virtual void StandardSurfaceColor(GLfloat material[4]) const;
		
		virtual void MapIntToRadicalColor(int i, double & r, double & g, double & b) const;
		virtual void MapIntToRadicalColor(int i, GLfloat material[4]) const;

		virtual void StartRenderSubwindow(int top, int left, double width, double height);
		virtual bool InSubWindow(int x, int y, double * xWnd, double * yWnd,
								 double width, double height, int subwindowFlags);
	
		virtual void DrawSquare(double minX, double minY, double width, double height,
								double r, double g, double b, double a,
								double fillR, double fillG, double fillB, double fillA);
		virtual void DrawArrows(double minX, double minY, double width, double height,
								double lr, double lg, double lb, double la,
								double rr, double rg, double rb, double ra, bool fill);
	
		// rendering global values
		//virtual void ClearAllGlobalValues();
		//virtual void AddToGlobalValuePool(const VKString & valueName, ModelPerVertexValues * values);
		//virtual void NormalizePoolAndCalculateColors(const VKString & valueName, 
		//											 const VKString & interpolator="none");
			
		bool m_bGlobalColors;
		bool m_bDefaultDrawIsocontours;
		double m_fMinPercentile;
		double m_fMaxPercentile;
		
		bool InMainWndBuffer();
		void StartDrawToBuffer();
		BufferID & EndDrawToBuffer();
	
		void LoadCurrentBuffer();
		
		BufferID CreateBuffer();
	
		static const int BUFFER_COLOR_COMPONENTS;
	
	protected:
		double m_fStandardAlpha;
	
		int m_iCurrentFramebufferID;
		std::vector<BufferID> m_Buffers;

		// global values
		//std::map<VKString, std::vector<ModelPerVertexValues*> > m_GlobalValuePool;
	
		// width / height
		int m_iWidth, m_iHeight;
		VisWindowRenderer * m_pRenderer;
		MouseButtonState m_MouseStates[3];
		int m_iSpecialKeys;
		std::vector<RNRgb> m_MulticolorSequence;
};

#endif

