// Source file for GAPS scalar harmonics class



////////////////////////////////////////////////////////////////////////
// This class stores spherical harmonics for a set of shells
// at increasing radii.
////////////////////////////////////////////////////////////////////////



// Include files

#include "R3Harmonics/R3Harmonics.h"



R3SphericalHarmonics::
R3SphericalHarmonics(int bandwidth, int nshells)
  : weights(NULL),
    shells(NULL),
    nshells(nshells)
{
  // Create arrays of shells and weights
  if ((nshells > 0) && (bandwidth >= 0)) {
    // Create array of shells
    shells = new R2SphericalHarmonics * [ nshells ];
    assert(shells);
    for (int i = 0; i < nshells; i++) {
      shells[i] = new R2SphericalHarmonics(bandwidth);   
      assert(shells[i]);
    }

    // Create array of weights
    RNScalar total_weight = 0;
    weights = new RNScalar [ nshells ];
    assert(weights);
    for (int i = 0; i < nshells; i++) {
      RNScalar radius = (RNScalar) i / (RNScalar) nshells;
      weights[i] = radius * radius;
      total_weight += weights[i];
    }
    for (int i = 0; i < nshells; i++) {
      weights[i] /= total_weight;
    }
  }
}



R3SphericalHarmonics::
R3SphericalHarmonics(const R3SphericalHarmonics& harmonics)
  : weights(NULL),
    shells(NULL),
    nshells(harmonics.nshells)
{
  // Create array of shells
  if (nshells > 0) {
    shells = new R2SphericalHarmonics * [ nshells ];
    assert(shells);
    for (int i = 0; i < nshells; i++) {
      shells[i] = new R2SphericalHarmonics(*(harmonics.shells[i]));   
      assert(shells[i]);
    }

    // Create array of weights
    RNScalar total_weight = 0;
    weights = new RNScalar [ nshells ];
    assert(weights);
    for (int i = 0; i < nshells; i++) {
      weights[i] = (RNScalar) i / (RNScalar) nshells;
      total_weight += weights[i];
    }
    for (int i = 0; i < nshells; i++) {
      weights[i] /= total_weight;
    }
  }
}



R3SphericalHarmonics::
~R3SphericalHarmonics(void)
{
  // Delete arrays of shells and weights
  if (nshells > 0) {
    // Delete shells
    if (shells) {
      for (int i = 0; i < nshells; i++) 
        if (shells[i]) delete shells[i];
      delete [] shells;
    }

    // Delete weights
    if (weights) delete [] weights;
  }
}



RNScalar R3SphericalHarmonics::
L2Norm(void) const
{
  // This is not quite right, because weighting is linear
  // But sum terms are quadratic

  // Return sum of squared coefficients
  RNScalar mean = 0.0;
  RNScalar max_radius = RadiusResolution()-1;
  RNScalar max_radius_cubed = max_radius * max_radius * max_radius;
  RNScalar last_radius_cubed = 0;
  RNScalar total_scale_squared = 0;
  for (int i = 0; i < nshells; i++) {
    RNScalar radius = i; // Radius(i);
    RNScalar radius_cubed = radius * radius * radius;
    RNScalar radius_scale = (radius_cubed - last_radius_cubed) / max_radius_cubed;
    last_radius_cubed = radius_cubed;
    RNScalar scale_squared = radius_scale * radius_scale;
    mean += scale_squared * shells[i]->L2NormSquared();
    total_scale_squared += scale_squared;
  }

  // Return mean times volume of sphere (4/3 PI R^3)
  RNScalar volume = (4.0/3.0) * RN_PI * max_radius_cubed;
  return volume * sqrt(mean / total_scale_squared);
}



RNScalar R3SphericalHarmonics::
L2NormSquared(void) const
{
  RNScalar l2norm = L2Norm();
  return l2norm * l2norm;
}



RNScalar R3SphericalHarmonics::
L2NormSquared(int S) const
{
  // Return sum of squared coefficients
  assert((0 <= S) && (S < nshells));
  return shells[S]->L2NormSquared();
}



RNScalar R3SphericalHarmonics::
L2NormSquared(int S, int L) const
{
  // Return sum of squared coefficients
  assert((0 <= S) && (S < nshells));
  return shells[S]->L2NormSquared(L);
}



RNScalar R3SphericalHarmonics::
L2NormSquared(int S, int L, int M) const
{
  // Return sum of squared coefficients
  assert((0 <= S) && (S < nshells));
  return shells[S]->L2NormSquared(L, M);
}



void R3SphericalHarmonics::
Clear(RNScalar value)
{
  // Set all coefficients to zero
  for (int i = 0; i < nshells; i++) 
     shells[i]->Clear(value);
}



RNScalar R3SphericalHarmonics::
Dot(const R3SphericalHarmonics& harmonics) const
{
  // Return the dot product of the vectors
  RNScalar sum = 0.0;
  for (int i = 0; i < nshells; i++) 
    sum += shells[i]->Dot(harmonics.Shell(i));
  return sum;
}



RNScalar R3SphericalHarmonics::
L1Distance(const R3SphericalHarmonics& harmonics) const
{
  // Return the sum of distances between coefficients
  RNScalar sum = 0.0;
  for (int i = 0; i < nshells; i++) 
    sum += shells[i]->L1Distance(harmonics.Shell(i));
  return sum;
}



RNScalar R3SphericalHarmonics::
L2DistanceSquared(const R3SphericalHarmonics& harmonics) const
{
  // Return the sum of squared distances between coefficients
  RNScalar sum = 0.0;
  for (int i = 0; i < nshells; i++) 
    sum += shells[i]->L2DistanceSquared(harmonics.Shell(i));
  return sum;
}



int R3SphericalHarmonics::
ReadFile(const char *filename)
{
  // Open file
  FILE *fp = fopen(filename, "rb");
  if (!fp) {
    RNFail("Unable to open voxel file %s", filename);
    return 0;
  }

  // Read 
  int status = Read(fp);

  // Close file
  fclose(fp);

  // Return status
  return status;
}



int R3SphericalHarmonics::
WriteFile(const char *filename) const
{
  // Open file
  FILE *fp = fopen(filename, "wb");
  if (!fp) {
    RNFail("Unable to open voxel file %s", filename);
    return 0;
  }

  // Write
  int status = Write(fp);

  // Close file
  fclose(fp);

  // Return status
  return status;
}



int R3SphericalHarmonics::
Read(FILE *fp)
{
  // Check file
  if (!fp) fp = stdin;

  // Read number of shells from file
  if (fread(&nshells, sizeof(int), 1, fp) != 1) {
    RNFail("Unable to read nshells from spherical harmonics file");
    return 0;
  }

  // Read shells
  if (nshells > 0) {
    // Create array of shells
    shells = new R2SphericalHarmonics * [ nshells ];
    assert(shells);

    // Read shells
    for (int i = 0; i < nshells; i++) {
      shells[i] = new R2SphericalHarmonics();   
      assert(shells[i]);
      shells[i]->Read(fp);
    }

    // Create array of weights
    RNScalar total_weight = 0;
    weights = new RNScalar [ nshells ];
    assert(weights);
    for (int i = 0; i < nshells; i++) {
      RNScalar radius = (RNScalar) i / (RNScalar) nshells;
      weights[i] = radius * radius;
      total_weight += weights[i];
    }
    for (int i = 0; i < nshells; i++) {
      weights[i] /= total_weight;
    }
  }

  // Return number of shells
  return nshells;
}



int R3SphericalHarmonics::
Write(FILE *fp) const
{
  // Check file
  if (!fp) fp = stdout;

  // Write number of shells to file
  if (fwrite(&nshells, sizeof(int), 1, fp) != 1) {
    RNFail("Unable to write nshells to spherical harmonics file");
    return 0;
  }

  // Write shells
  for (int i = 0; i < nshells; i++) 
    shells[i]->Write(fp);

  // Return number of shells written
  return nshells;
}



int R3SphericalHarmonics::
Print(FILE *fp) const
{
  // Check file
  if (!fp) fp = stdout;

  // Print coefficients
  for (int S = 0; S < nshells; S++) {
    for (int L = 0; L < Bandwidth(); L++) {
      for (int M = -L ; M <= L; M++) {
        const RNScalar *c = shells[S]->Coefficients(L, M);
        fprintf(fp, "( %g %g )", c[0], c[1]);
      }
      fprintf(fp, "\n");
    }
    fprintf(fp, "\n");
  }

  // Return number of shells written
  return nshells;
}



int 
ForwardSphericalHarmonicTransform(const R3SphericalGrid& grid, R3SphericalHarmonics& harmonics)
{
  // Check resolutions
  int size = grid.ThetaResolution();
  assert(size == grid.PhiResolution());
  assert(size == 2*harmonics.Bandwidth());
  assert(RadiusResolution() == harmonics.RadiusResolution());

  // Perform transform
  int count = 0;
  const RNScalar *grid_valuesp = grid.GridValues();
  for (int i = 0; i < grid.RadiusResolution(); i++) {
    RNScalar *coefficients = (RNScalar *) harmonics.Shell(i).Coefficients();
    count += R2ForwardSphericalHarmonicTransform(grid_valuesp, coefficients, size); 
    grid_valuesp += size * size;
  }

  // Return status
  return count;
}



int
InverseSphericalHarmonicTransform(const R3SphericalHarmonics& harmonics, R3SphericalGrid& grid)
{
  // Check resolutions
  int size = grid.ThetaResolution();
  assert(size == grid.PhiResolution());
  assert(size == 2*harmonics.Bandwidth());
  assert(harmonics.RadiusResolution() == grid.RadiusResolution());

  // Perform transform
  int count = 0;
  RNScalar *grid_valuesp = (RNScalar *) grid.GridValues();
  for (int i = 0; i < harmonics.nshells; i++) {
    const RNScalar *coefficients = harmonics.shells[i]->Coefficients();
    count += R2InverseSphericalHarmonicTransform(coefficients, grid_valuesp, size); 
    grid_valuesp += size * size;
  }

  // Return status
  return count;
}



