// Header file for GAPS spherical scalar grid class



// Class definition

class R2SphericalGrid {
public:
  // Constructors
  R2SphericalGrid(int phi_resolution = 0, int theta_resolution = 0);
  R2SphericalGrid(const R2SphericalGrid& grid);

  // Grid property functions
  int NEntries() const;
  int PhiResolution(void) const;
  int ThetaResolution(void) const;
  int Resolution(int dim) const;
  RNScalar Sum(void) const;
  RNScalar Mean(void) const;
  RNScalar Maximum(void) const;
  RNScalar Minimum(void) const;
  RNInterval Range(void) const;
  RNScalar L1Norm(void) const;
  RNScalar L2Norm(void) const;
  RNScalar Area(void) const;
  RNScalar Integral(RNScalar (*f)(RNScalar)) const;
  RNScalar Integral(const R2SphericalGrid& g, RNScalar (*f)(RNScalar, RNScalar)) const;
  int Cardinality(void) const;

  // Grid value access functions
  RNScalar GridValue(int i, int j) const;
  RNScalar GridValue(RNCoord x, RNCoord y) const;
  RNScalar GridValue(const R2Point& grid_point) const;
  RNScalar& operator()(int i, int j);

  // Grid manipulation functions
  void Abs(void);
  void Sqrt(void);
  void Square(void);
  void Negate(void);
  void Invert(void);
  void Normalize(void);
  void Clear(RNScalar value = 0);
  void Add(RNScalar value);
  void Add(const R2SphericalGrid& grid);
  void Subtract(RNScalar value);
  void Subtract(const R2SphericalGrid& grid);
  void Multiply(RNScalar value);
  void Multiply(const R2SphericalGrid& grid);
  void Divide(RNScalar value);
  void Divide(const R2SphericalGrid& grid);
  void Threshold(RNScalar threshold, RNScalar low, RNScalar high);
  void SetGridValue(int i, int j, RNScalar value);
  void AddGridValue(int i, int j, RNScalar value);

  // Arithmetic operators
  R2SphericalGrid& operator=(const R2SphericalGrid& grid);
  R2SphericalGrid& operator+=(RNScalar scale);
  R2SphericalGrid& operator+=(const R2SphericalGrid& grid);
  R2SphericalGrid& operator-=(RNScalar scale);
  R2SphericalGrid& operator-=(const R2SphericalGrid& grid);
  R2SphericalGrid& operator*=(RNScalar scale);
  R2SphericalGrid& operator*=(const R2SphericalGrid& grid);
  R2SphericalGrid& operator/=(RNScalar scale);
  R2SphericalGrid& operator/=(const R2SphericalGrid& grid);

  // Relationship functions
  RNScalar Dot(const R2SphericalGrid& grid) const;
  RNScalar L1Distance(const R2SphericalGrid& grid) const;
  RNScalar L2Distance(const R2SphericalGrid& grid) const;
  RNScalar L2DistanceSquared(const R2SphericalGrid& grid) const;

  // Transformation utility functions
  R3Point WorldPosition(const R2Point& grid_point) const;
  R3Vector WorldVector(const R2Vector& grid_vector) const;
  R2Point GridPosition(const R3Point& world_point) const;
  R2Point GridPosition(const R3Vector& world_vector) const;
  R3Point WorldPosition(RNCoord x, RNCoord y) const;
  R2Point GridPosition(RNCoord x, RNCoord y, RNCoord z) const;

  // I/O functions
  int ReadFile(const char *filename);
  int WriteFile(const char *filename) const;
  // int Write(FILE *fp = NULL) const;
  // int Read(FILE *fp = NULL);
  int Print(FILE *fp = NULL) const;

  // Spherical harmonics transform
  int ForwardSphericalHarmonicTransform(R2SphericalHarmonics& harmonics) const;

  // Debugging functions
  const RNScalar *GridValues(void) const;

private:
  // Utility functions
  static RNScalar L1NormFunction(RNScalar value);
  static RNScalar L2NormFunction(RNScalar value);
  static RNScalar AreaFunction(RNScalar value);
  static RNScalar L1DistanceFunction(RNScalar value1, RNScalar value2);
  static RNScalar L2DistanceFunction(RNScalar value1, RNScalar value2);
  static RNScalar DotFunction(RNScalar value1, RNScalar value2);

private:
  // Utility functions
  RNAngle Phi(RNScalar y) const;
  RNAngle Theta(RNScalar x) const;
  RNCoord X(RNAngle phi) const;
  RNCoord Y(RNAngle theta) const;

private:
  R2Grid grid;
};



// Inline functions

inline int R2SphericalGrid::
NEntries(void) const
{
  // Return total number of grid entries
  return grid.NEntries();
}



inline int R2SphericalGrid::
PhiResolution(void) const
{
  // Return resolution in phi dimension
  return grid.XResolution();
}




inline int R2SphericalGrid::
ThetaResolution(void) const
{
  // Return resolution in theta dimension
  return grid.YResolution();
}




inline int R2SphericalGrid::
Resolution(int dim) const
{
  // Return resolution in dimension
  return grid.Resolution(dim);
}




inline RNScalar R2SphericalGrid::
L1Norm(void) const
{
  // Return L1 norm of grid
  return Integral(L1NormFunction);
}



inline RNScalar R2SphericalGrid::
L2Norm(void) const
{
  // Return L2 norm of grid
  return sqrt(Integral(L2NormFunction));
}



inline RNScalar R2SphericalGrid::
Area(void) const
{
  // Return L2 norm of grid
  return sqrt(Integral(AreaFunction));
}



inline RNScalar R2SphericalGrid::
Sum(void) const
{
  // Return sum of all grid values
  return L1Norm();
}



inline RNScalar R2SphericalGrid::
Mean(void) const
{
  // Find average value
  if (grid.NEntries() == 0) return 0.0;
  return Sum() / grid.NEntries();
}



inline RNScalar R2SphericalGrid::
Minimum(void) const
{
  // Return smallest value
  return grid.Minimum();
}



inline RNScalar R2SphericalGrid::
Maximum(void) const
{
  // Return largest value
  return grid.Maximum();
}



inline RNInterval R2SphericalGrid::
Range(void) const
{
  // Return smallest and largest values
  return grid.Range();
}



inline int R2SphericalGrid::
Cardinality(void) const
{
  // Return number of non-zero grid values
  return grid.Cardinality();
}



inline RNScalar& R2SphericalGrid::
operator()(int i, int j) 
{
  // Return value at grid point
  return grid(i, j);
}



inline const RNScalar *R2SphericalGrid::
GridValues(void) const
{
  // Return pointer to grid values
  return grid.GridValues();
}



inline RNScalar R2SphericalGrid::
GridValue(int i, int j) const
{
  // Return value at grid point
  return grid.GridValue(i, j);
}



inline RNScalar R2SphericalGrid::
GridValue(const R2Point& point) const
{
  // Return value at grid point
  return grid.GridValue(point);
}



inline void R2SphericalGrid::
Abs(void) 
{
  // Take absolute value of every grid value
  grid.Abs();
}



inline void R2SphericalGrid::
Sqrt(void) 
{
  // Take sqrt of every grid value
  grid.Sqrt();
}



inline void R2SphericalGrid::
Square(void) 
{
  // Square every grid value
  grid.Square();
}



inline void R2SphericalGrid::
Negate(void) 
{
  // Negate every grid value
  grid.Negate();
}



inline void R2SphericalGrid::
Invert(void) 
{
  // Negate every grid value
  grid.Invert();
}



inline void R2SphericalGrid::
Normalize(void) 
{
  // Scale so that length of "vector" is one
  RNScalar length = L2Norm();
  if (RNIsNotZero(length)) Divide(length);
}



inline void R2SphericalGrid::
Clear(RNScalar value) 
{
  // Set all grid values to value
  grid.Clear(value);
}



inline void R2SphericalGrid::
Add(RNScalar value) 
{
  // Add value to every grid 
  grid.Add(value);
}



inline void R2SphericalGrid::
Add(const R2SphericalGrid& g) 
{
  // Add passed grid values to corresponding entries of this grid
  grid.Add(g.grid);
}



inline void R2SphericalGrid::
Subtract(RNScalar value) 
{
  // Add the opposite
  grid.Subtract(value);
}



inline void R2SphericalGrid::
Subtract(const R2SphericalGrid& g) 
{
  // Subtract passed grid values from corresponding entries of this grid
  grid.Subtract(g.grid);
}



inline void R2SphericalGrid::
Multiply(RNScalar value) 
{
  // Multiply grid values by value
  grid.Multiply(value);
}



inline void R2SphericalGrid::
Multiply(const R2SphericalGrid& g) 
{
  // Multiply passed grid values by corresponding entries of this grid
  grid.Multiply(g.grid);
}



inline void R2SphericalGrid::
Divide(RNScalar value) 
{
  // Multiply by recipricol
  grid.Divide(value);
}



inline void R2SphericalGrid::
Divide(const R2SphericalGrid& g) 
{
  // Divide passed grid values by corresponding entries of this grid
  grid.Divide(g.grid);
}



inline void R2SphericalGrid::
Threshold(RNScalar threshold, RNScalar low, RNScalar high) 
{
  // Set grid value to low (high) if less/equal (greater) than threshold
  grid.Threshold(threshold, low, high);
}



inline void R2SphericalGrid::
SetGridValue(int i, int j, RNScalar value)
{
  // Set value at grid point
  grid.SetGridValue(i, j, value);
}



inline void R2SphericalGrid::
AddGridValue(int i, int j, RNScalar value)
{
  // Add value at grid point
  grid.AddGridValue(i, j, value);
}



inline R2SphericalGrid& R2SphericalGrid::
operator=(const R2SphericalGrid& g) 
{
  // Return this
  grid = g.grid;
  return *this;
}



inline R2SphericalGrid& R2SphericalGrid::
operator+=(RNScalar value) 
{
  // Add value to all grid values 
  grid.Add(value);
  return *this;
}



inline R2SphericalGrid& R2SphericalGrid::
operator+=(const R2SphericalGrid& g) 
{
  // Add passed grid values to corresponding entries of this grid
  grid.Add(g.grid);
  return *this;
}



inline R2SphericalGrid& R2SphericalGrid::
operator-=(RNScalar value) 
{
  // Subtract value from all grid values 
  grid.Subtract(value);
  return *this;
}



inline R2SphericalGrid& R2SphericalGrid::
operator-=(const R2SphericalGrid& g) 
{
  // Subtract passed grid values from corresponding entries of this grid
  grid.Subtract(g.grid);
  return *this;
}



inline R2SphericalGrid& R2SphericalGrid::
operator*=(RNScalar value) 
{
  // Multiply grid values by value
  grid.Multiply(value);
  return *this;
}



inline R2SphericalGrid& R2SphericalGrid::
operator*=(const R2SphericalGrid& g) 
{
  // Multiply passed grid values by corresponding entries of this grid
  grid.Multiply(g.grid);
  return *this;
}



inline R2SphericalGrid& R2SphericalGrid::
operator/=(RNScalar value) 
{
  // Divide grid values by value
  grid.Divide(value);
  return *this;
}



inline R2SphericalGrid& R2SphericalGrid::
operator/=(const R2SphericalGrid& g) 
{
  // Divide passed grid values by corresponding entries of this grid
  grid.Divide(g.grid);
  return *this;
}



inline RNScalar R2SphericalGrid::
Dot(const R2SphericalGrid& g) const
{
  // Compute dot product between this and grid
  return Integral(g, DotFunction);
}



inline RNScalar R2SphericalGrid::
L1Distance(const R2SphericalGrid& g) const
{
  // Return L1 norm of grid
  return Integral(g, L1DistanceFunction);
}




inline RNScalar R2SphericalGrid::
L2DistanceSquared(const R2SphericalGrid& g) const
{
  // Return L2 norm of grid
  return Integral(g, L2DistanceFunction);
}



inline RNScalar R2SphericalGrid::
L2Distance(const R2SphericalGrid& g) const
{
  // Return L2 norm of grid
  return sqrt(L2DistanceSquared(g));
}



inline R3Point R2SphericalGrid::
WorldPosition(const R2Point& grid_point) const
{
  // Return 3D point represented by position in grid
  return WorldPosition(grid_point[0], grid_point[1]);
}



inline R3Vector R2SphericalGrid::
WorldVector(const R2Vector& grid_vector) const
{
  // Return 3D vector represented by vector in grid
  return WorldPosition(grid_vector[0], grid_vector[1]).Vector();
}



inline R2Point R2SphericalGrid::
GridPosition(const R3Point& world_point) const
{
  // Return grid position along vector towards world_point (assuming sphere is at origin)
  return GridPosition(world_point[0], world_point[1], world_point[2]);
}



inline R2Point R2SphericalGrid::
GridPosition(const R3Vector& world_vector) const
{
  // Return grid position along world_vector 
  return GridPosition(world_vector[0], world_vector[1], world_vector[2]);
}



inline int R2SphericalGrid::
ReadFile(const char *filename)
{
  // Read grid from file
  return grid.Read(filename);
}



inline int R2SphericalGrid::
WriteFile(const char *filename) const
{
  // Write grid to file
  return grid.Write(filename);
}



/*

inline int R2SphericalGrid::
Read(FILE *fp)
{
  // Read grid from file
  return grid.Read(fp);
}



inline int R2SphericalGrid::
Write(FILE *fp) const
{
  // Write grid to file
  return grid.Write(fp);
}

*/


inline int R2SphericalGrid::
Print(FILE *fp) const
{
  // Print grid values
  return grid.Print(fp);
}



inline RNAngle R2SphericalGrid::
Phi(RNScalar x) const
{
  // Return phi, the angle counterclockwise around the equator (x-z plane) from the x-axis
  return RN_TWO_PI * x / PhiResolution();
}



inline RNAngle R2SphericalGrid::
Theta(RNScalar y) const
{
  // Return theta, the angle from the up axis (y in world coords) to the vector at row j in the grid
  return RN_PI * (2.0*y + 1)/(2.0 * ThetaResolution());
}



inline RNCoord R2SphericalGrid::
X(RNAngle phi) const
{
  // Return the X grid coordinate mapping to phi
  RNScalar x = phi * PhiResolution() / (2.0 * RN_PI);
  while (x >= PhiResolution()) x -= PhiResolution();
  while (x < 0) x += PhiResolution();
  return x;
}



inline RNCoord R2SphericalGrid::
Y(RNAngle theta) const
{
  // Return the Y grid coordinate mapping to theta
  RNScalar y = (2.0 * ThetaResolution() * theta - RN_PI) / (2.0 * RN_PI);
  if (y < 0) y = -y;
  while (y >= ThetaResolution()) y = 2 * ThetaResolution() - y;
  return y;
}



