// Source file for GAPS scalar harmonics class



////////////////////////////////////////////////////////////////////////
// This class stores spherical harmonic coefficients.
// The coefficients are arranged in one array with both
// real and imaginary coefficients interleaved.  For each
// f(l,m) there are two entries, the real coefficient followed
// by the imaginary coefficient.  So, the entire array has 
// 2*bandwidth*bandwidth entries.  The coefficients f(m,l) 
// are ordered as follows (m is index within order, l is order):
//   f(0,0) f(0,1) f(0,2) ... f(0,bw-1)
//          f(1,1) f(1,2) ... f(1,bw-1)
//          etc.
//                 f(bw-2,bw-2), f(bw-2,bw-1)
//		               f(bw-1,bw-1)
//			       f(-(bw-1),bw-1)
//		 f(-(bw-2),bw-2) f(-(bw-2),bw-1)
//	  etc.
//	          f(-2,2) ... f(-2,bw-1)
//	  f(-1,1) f(-1,2) ... f(-1,bw-1)
////////////////////////////////////////////////////////////////////////



// Include files

#include "R3Harmonics/R3Harmonics.h"



R2SphericalHarmonics::
R2SphericalHarmonics(int bandwidth)
  : coefficients(NULL),
    ncoefficients(2*bandwidth*bandwidth),
    bandwidth(bandwidth)
{
  // Allocate array for coefficients
  if (bandwidth >= 0) {
    coefficients = new RNScalar [ ncoefficients ];
    assert(coefficients);
  }
}



R2SphericalHarmonics::
R2SphericalHarmonics(const R2SphericalHarmonics& harmonics)
  : coefficients(NULL),
    ncoefficients(harmonics.ncoefficients),
    bandwidth(harmonics.bandwidth)
{
  // Allocate/copy array for coefficients
  if (bandwidth >= 0) {
    // Allocate coefficients
    coefficients = new RNScalar [ ncoefficients ];
    assert(coefficients);

    // Copy coefficients
    for (int i = 0; i < ncoefficients; i++) 
      coefficients[i] = harmonics.coefficients[i];
  }
}



R2SphericalHarmonics::
~R2SphericalHarmonics(void)
{
  // Delete coefficients
  if (coefficients) delete [] coefficients;
}



RNScalar R2SphericalHarmonics::
L1Norm(void) const
{
  // Return sum of absolute values of coefficients
  RNScalar sum = 0.0;
  for (int i = 0; i < ncoefficients; i++)
    sum += fabs(coefficients[i]);
  return sum;
}



RNScalar R2SphericalHarmonics::
L1Norm(int L) const
{
  // Return sum of squared coefficients with frequency L
  RNScalar sum = 0.0;
  const RNScalar *p = Coefficients(L, 0);
  for (int M = -L; M <= L; M++) {
    sum += fabs(*(p++)); 
    sum += fabs(*(p++));
  }
  return sum;
}



RNScalar R2SphericalHarmonics::
L1Norm(int L, int M) const
{
  // Return sum of squared coefficients for both real and imaginary terms
  const RNScalar *p = Coefficients(L, M);
  return (fabs(p[0]) + fabs(p[1]));
}



RNScalar R2SphericalHarmonics::
L2NormSquared(void) const
{
  // Return sum of squared coefficients
  RNScalar sum = 0.0;
  for (int i = 0; i < ncoefficients; i++)
    sum += coefficients[i] * coefficients[i];
  return sum;
}



RNScalar R2SphericalHarmonics::
L2NormSquared(int L) const
{
  // Return sum of squared coefficients with frequency L
  RNScalar sum = 0.0;
  for (int M = -L; M <= L; M++) 
    sum += L2NormSquared(L, M);
  return sum;
}



RNScalar R2SphericalHarmonics::
L2NormSquared(int L, int M) const
{
  // Return sum of squared coefficients for both real and imaginary terms
  const RNScalar *p = Coefficients(L, M);
  return (p[0]*p[0] + p[1]*p[1]);
}



int R2SphericalHarmonics::
Index(int L, int M) const
{
  // Return index of coefficients for F,L]
  assert((L >= 0) && (L < bandwidth));
  assert((M >= -L) && (M <= L));
  return 2 * seanindex(M, L, bandwidth);
}



void R2SphericalHarmonics::
Clear(RNScalar value)
{
  // Set all coefficients to zero
  for (int i = 0; i < ncoefficients; i++) 
    coefficients[i] = value;
}



void R2SphericalHarmonics::
SetRealCoefficient(int L, int M, RNScalar value)
{
  // Set coefficient
  coefficients[Index(L, M)] = value;
}



void R2SphericalHarmonics::
SetImaginaryCoefficient(int L, int M, RNScalar value)
{
  // Set coefficient
  coefficients[Index(L, M)+1] = value;
}



RNScalar R2SphericalHarmonics::
Dot(const R2SphericalHarmonics& harmonics) const
{
  // Return the dot product of the vectors
  RNScalar sum = 0.0;
  for (int i = 0; i < ncoefficients; i++) 
    sum += coefficients[i] * harmonics.coefficients[i];
  return sum;
}



RNScalar R2SphericalHarmonics::
L1Distance(const R2SphericalHarmonics& harmonics) const
{
  // Return the sum of distances between coefficients
  RNScalar sum = 0.0;
  for (int i = 0; i < ncoefficients; i++)
    sum += fabs(coefficients[i] - harmonics.coefficients[i]);
  return sum;
}



RNScalar R2SphericalHarmonics::
L2DistanceSquared(const R2SphericalHarmonics& harmonics) const
{
  // Return the sum of squared distances between coefficients
  RNScalar sum = 0.0;
  for (int i = 0; i < ncoefficients; i++) {
    RNScalar distance = coefficients[i] - harmonics.coefficients[i];
    sum += distance * distance;
  }
  return sum;
}



int R2SphericalHarmonics::
ReadFile(const char *filename)
{
  // Open file
  FILE *fp = fopen(filename, "rb");
  if (!fp) {
    RNFail("Unable to open file %s", filename);
    return 0;
  }

  // Read 
  int status = Read(fp);

  // Close file
  fclose(fp);

  // Return status
  return status;
}



int R2SphericalHarmonics::
WriteFile(const char *filename) const
{
  // Open file
  FILE *fp = fopen(filename, "wb");
  if (!fp) {
    RNFail("Unable to open file %s", filename);
    return 0;
  }

  // Read 
  int status = Write(fp);

  // Close file
  fclose(fp);

  // Return status
  return status;
}



int R2SphericalHarmonics::
Read(FILE *fp)
{
  // Check file
  if (!fp) fp = stdin;

  // Read bandwidth from file
  if (fread(&bandwidth, sizeof(int), 1, fp) != 1) {
    RNFail("Unable to read bandwidth from spherical harmonics file");
    return 0;
  }

  // Compute number of coefficients
  ncoefficients = 2 * bandwidth * bandwidth;

  // Allocate coefficients
  coefficients = new RNScalar [ ncoefficients ];
  assert(coefficients);

  // Read coefficients
  if (fread(coefficients, sizeof(RNScalar), ncoefficients, fp) != (unsigned int) ncoefficients) {
    RNFail("Unable to read %d coefficients from spherical harmonics file", ncoefficients);
    return 0;
  }

  // Return number of coefficients read
  return ncoefficients;
}



int R2SphericalHarmonics::
Write(FILE *fp) const
{
  // Check file
  if (!fp) fp = stdout;

  // Write bandwidth to file
  if (fwrite(&bandwidth, sizeof(int), 1, fp) != 1) {
    RNFail("Unable to write bandwidth to spherical harmonics file");
    return 0;
  }

  // Write coefficients
  if (fwrite(coefficients, sizeof(RNScalar), ncoefficients, fp) != (unsigned int) ncoefficients) {
    RNFail("Unable to write coefficients to spherical harmonics file");
    return 0;
  }

  // Return number of coefficients written
  return ncoefficients;
}



int R2SphericalHarmonics::
Print(FILE *fp) const
{
  // Check file
  if (!fp) fp = stdout;

  // Print coefficients
  for (int L = 0; L < bandwidth; L++) {
    for (int M = -L ; M <= L; M++) {
      fprintf(fp, "( %g %g )", RealCoefficient(L, M), ImaginaryCoefficient(L, M));
    }
    fprintf(fp, "\n");
  }

  // Return number of coefficients written
  return ncoefficients;
}



int R2SphericalHarmonics::
InverseSphericalHarmonicTransform(R2SphericalGrid& grid) const
{
  // Check resolutions
  int size = grid.ThetaResolution();
  assert(size == grid.PhiResolution());
  assert(size == 2*Bandwidth());

  // Perform transform
  return R2InverseSphericalHarmonicTransform((RNScalar *) Coefficients(), (RNScalar *) grid.GridValues(), size); 
}



int
R2ForwardSphericalHarmonicTransform(const RNScalar *grid, RNScalar *harmonics, int size)
{
  // Get convenient variables
  int bw = size/2;
  int cutoff = bw;
  assert((size % 2) == 0);
  assert((bw % 2) == 0);

  // Allocate temporary variables
  double *workspace1=new double[8*bw*bw+32*bw];
  double *workspace2=new double[Spharmonic_TableSize(bw)];
  double **table=Spharmonic_Pml_Table(bw, workspace2, workspace1);
  double *realIn=new double[size * size];
  double *imagIn=new double[size * size];
  double *realOut=new double[bw * bw];
  double *imagOut=new double[bw * bw];
  assert(table && workspace1 && workspace2 && realIn && imagIn && realOut && realOut);

  // Fill input variables
  double *realInp = realIn;
  double *imagInp = imagIn;
  const RNScalar *gridp = grid;
  for (int i = 0; i < size*size; i++) {
    *(realInp++) = *(gridp++);
    *(imagInp++) = 0.0;
  }

  // Perform forward harmonic transform
  FST_semi_memo(realIn, imagIn, realOut, imagOut, size, table, workspace1, 1, cutoff);

  // Copy output variables
  double *realOutp = realOut;
  double *imagOutp = imagOut;
  RNScalar *harmonicsp = harmonics;
  for (int i = 0; i < bw*bw; i++) {
    *(harmonicsp++) = *(realOutp++);
    *(harmonicsp++) = *(imagOutp++);
  }

  // Delete temporary variables
  delete [] workspace1;
  delete [] workspace2;
  delete [] realIn;
  delete [] imagIn;
  delete [] realOut;
  delete [] imagOut;
  free(table);

  // Return success
  return size;
}



int
R2InverseSphericalHarmonicTransform(const RNScalar *harmonics, RNScalar *grid, int size)
{
  // Get convenient variables
  assert((size % 2) == 0);
  int bw = size/2;
  int cutoff = bw;

  // Allocate temporary variables
  double *workspace1=new double[8*bw*bw+32*bw];
  double *workspace2=new double[Spharmonic_TableSize(bw)];
  double *workspace3=new double[Spharmonic_TableSize(bw)];
  double **table=Spharmonic_Pml_Table(bw, workspace2, workspace1);
  double **transpose_table = Transpose_Spharmonic_Pml_Table(table, bw, workspace3,workspace1);  
  double *realIn=new double[size * size];
  double *imagIn=new double[size * size];
  double *realOut=new double[bw * bw];
  double *imagOut=new double[bw * bw];
  assert(table && workspace1 && workspace2 && realIn && imagIn && realOut && realOut);

  // Fill input variables
  double *realOutp = realOut;
  double *imagOutp = imagOut;
  const RNScalar *harmonicsp = harmonics;
  for (int i = 0; i < bw*bw; i++) {
    *(realOutp++) = *(harmonicsp++);
    *(imagOutp++) = *(harmonicsp++);
  }

  // Perform inverse harmonic transform
  InvFST_semi_memo(realOut, imagOut, realIn, imagIn, size, transpose_table, workspace1, 0, cutoff);

  // Parse output variables
  double *realInp = realIn;
  RNScalar *gridp = grid;
  for (int i = 0; i < size*size; i++) 
    *(gridp++) = *(realInp++);

  // Delete temporary variables
  delete [] workspace1;
  delete [] workspace2;
  delete [] workspace3;
  delete [] realIn;
  delete [] imagIn;
  delete [] realOut;
  delete [] imagOut;
  free(table);
  free(transpose_table);

  // Return success
  return size;
}



