/***************************************************************************
  **************************************************************************
  
                SOFT: SO(3) Fourier transform code
  
   Peter Kostelec, Dan Rockmore
   {geelong,rockmore}@cs.dartmouth.edu
  
   Contact: Peter Kostelec
            geelong@cs.dartmouth.edu
  
  
   Copyright 2003 Peter Kostelec, Dan Rockmore
  
  
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
  
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
  
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
  
  
   Commercial use is absolutely prohibited.
  
   See the accompanying LICENSE file for details.
  
  ************************************************************************
  ************************************************************************/



/*

  This program will test the accuracy and speed of the routines
  (as they stand now) by doing the following

  1) generate random wigner coefficients
  2) do a Wigner synthesis to get sample values
  3) do a Wigner analysis to get new coefficients
  4) compare the new coefficients with the old


  input: orders m1, m2
         bandwidth bw
	 number of loops
	 output file (optional) - to write the differences
         between the original coeffs and new ones

  example: test_Wigner_Naive m1 m2 bw loops output_file

  example: test_Wigner_Naive 3 4 16 coeffs.dat splat.dat


*/

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "utils.h"
#include "makeWigner.h"
#include "wignerTransforms.h"
#include "csecond.h"

void srand48(int s){srand(s);}
double drand48(void){return ((double)rand())/RAND_MAX;}

int main ( int argc ,
	   char **argv )
{
  int i, j, m1, m2, bw, n ;
  int loops, m ;
  long int seed ;
  double *coeffs, *signal, *newcoeffs;
  double *wigners, *wignersTrans ;
  double *workspace, *scratch ;
  double *sinPts, *cosPts ;
  double *sinPts2, *cosPts2 ;
  double tmp_error, max_error, sum_error;
  double tmp_relerror, max_relerror, sum_relerror;
  double tstart, tstop, runtime ;

  FILE *fp ;

  
  if (argc < 5)
    {
      fprintf(stdout,"Usage: test_Wigner_Naive m1 m2 bw loops [output_file]\n");
      exit(0);
    }

  m1 = atoi( argv[1] );
  m2 = atoi( argv[2] );
  bw = atoi( argv[3] );
  loops = atoi( argv[4] ) ;
  m = MAX( ABS( m1 ) , ABS( m2 ) ) ;
  n = 2 * bw ;
  runtime = 0.0 ;

  coeffs = ( double * ) malloc(sizeof( double ) * (bw - m) ) ;
  newcoeffs = ( double * ) malloc(sizeof( double ) * (bw - m) ) ;
  signal = ( double * ) malloc(sizeof( double ) * n ) ;
  wigners = ( double * ) malloc( sizeof( double ) * ( bw - m ) * n ) ;
  wignersTrans = ( double * ) malloc( sizeof( double ) * ( bw - m ) * n ) ;
  workspace = (double *) malloc(sizeof( double ) * (4 + 8) * n ) ;
  sinPts = workspace ;
  cosPts = sinPts + n ;
  sinPts2 = cosPts + n ;
  cosPts2 = sinPts2 + n ;
  scratch = cosPts2 + n ; /* scratch needs to be of size n */

  /* note that the definition of wigSpec requires that instead of
     evaluating at beta, I need to evaluate at beta/2; ergo I call
     SinEvalPts2 instead of SinEvalPts, etc etc
  */


  /* generate seed for random number generator */
  time ( &seed ) ;
  srand48( seed ) ;

  /* precompute sines and cosines appropriate for making the
     wigners */

  SinEvalPts( n, sinPts ) ;
  CosEvalPts( n, cosPts ) ;
  SinEvalPts2( n, sinPts2 ) ;
  CosEvalPts2( n, cosPts2 ) ;

  /* make the wigners */
  genWig_L2( m1, m2, bw,
		  sinPts, cosPts,
		  sinPts2, cosPts2,
		  wigners, scratch ) ;

  /* now make the wigners - transpose version! */
  genWigTrans_L2( m1, m2, bw,
		  sinPts, cosPts,
		  sinPts2, cosPts2,
		  wignersTrans, scratch ) ;


  max_error = 0.0 ;
  max_relerror = 0.0 ;
  sum_error = 0.0 ;
  sum_relerror = 0.0 ;
  for ( i = 0 ; i < loops ; i ++ )
    {
      /* generate random coeffs */
      for( j = 0 ; j < (bw - m) ; j++ )
	coeffs[ j ] = drand48() ;
      
      /* turn on stop watch */
      tstart = csecond () ;

      /* now synthesize */
      wigNaiveSynthesis( m1, m2, bw, coeffs,
			 wignersTrans, signal,
			 scratch ) ;

      /* now analyze */
      wigNaiveAnalysis( m1, m2, bw, signal,
			wigners, newcoeffs,
			scratch ) ;

      /* turn off stop watch */
      tstop = csecond () ;

      runtime += (tstop - tstart);

      /* now figure out errors */
      for( j = 0 ; j < bw - m ; j ++ )
	{
	  tmp_error = fabs( coeffs[j] - newcoeffs[j] );
	  tmp_relerror = tmp_error / ( fabs( coeffs[j] ) +
				       pow( 10.0, -50.0 ) );
	  max_error = MAX( max_error, tmp_error );
	  max_relerror = MAX( max_relerror, tmp_relerror );
	}
      sum_error += max_error ;
      sum_relerror += max_relerror ;
    }

  fprintf(stderr,"error r-o:\t\t %.5e\n",sum_error/((double) loops));
  fprintf(stderr,"rel error (r-o)/o:\t %.5e\n",sum_relerror/((double) loops));
  fprintf(stderr,"max error:\t\t %.5e\n", max_error );
  fprintf(stderr,"max rel error:\t\t %.5e\n", max_relerror );
  fprintf(stderr,"\naverage runtime: %.5f seconds per loop\n",
	  runtime/((double) loops));

  if ( argc == 6 )
    {
      fp = fopen(argv[5], "w");
      for ( i = 0 ; i < bw - m ; i ++ )
	fprintf(fp,"%.16f\n", coeffs[i] - newcoeffs[i]);
    }


  free( workspace ) ;
  free( wignersTrans ) ;
  free( wigners ) ;
  free( signal ) ;
  free( newcoeffs ) ;
  free( coeffs ) ;

  return 0 ;
}
