//
// $Id: R3Io.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include <iostream.h>
#include "R3Nodes/R3Nodes.h"
#include "R3Io.h"
#include "Files.h"
#include "VrmlInfo.h"
#include "parse_nodes.h"
#include "Runtime.h"
#if (RN_OS == RN_WINDOWSNT)
#include "MathConstants.h"
#endif

VrmlFile *cur_vrml_file;
VrmlInfo *cur_vrml_info = new VrmlInfo;

static RNBoolean y_up = TRUE;
static RNBoolean ring_only = FALSE;
static RNBoolean info_output = FALSE;

static int recursion_level = 0;




/*################## Parsing functions #############################*/

void R3Vrml_set_info_output(RNBoolean value)
{
  info_output = value;
}  // R3Vrml_set_info_output



RNBoolean R3Vrml_get_info_output()
{
  return info_output;
}  // R3Vrml_get_info_output



void R3Vrml_set_y_up(RNBoolean value)
{
  y_up = value;
}  // R3Vrml_set_y_up



RNBoolean R3Vrml_get_y_up()
{
  return y_up;
}  // R3Vrml_get_y_up



void R3Vrml_set_ring_only(RNBoolean value)
{
  ring_only = value;
}  // R3Vrml_set_ring_only



RNBoolean R3Vrml_get_ring_only()
{
  return ring_only;
}  // R3Vrml_get_ring_only



R3Model *
R3ReadVrml2File(const char *filename)
{
    // Create model
    R3Model *model = new R3Model();
    if (!R3LoadVrml2File(model, model, filename)) {
	delete model;
	return NULL;
    }
    else {
	return model;
    }
}



static RNArray<VrmlNode *> top_nodes;

void delete_vrml_data()
{
  int nr_entries = top_nodes.NEntries();
  for(int i=0; i<nr_entries; i++) {
    VrmlNode *vrmlnode_ptr = top_nodes.Kth(i);
    if (vrmlnode_ptr) {
      delete vrmlnode_ptr;
#ifdef __DEBUG_DESTRUCTORS__
      cout << i << ": about to delete a " 
	   << vrmlnode_ptr->get_node_name() << endl;
#endif
    }
  }  // for
  top_nodes.Empty();
}  // delete_vrml_data



R3GroupNode *
R3LoadVrml2File(R3Model *model, R3GroupNode *group, const char *filename)
{
  //  RNSetErrorLevel(1);  // disable warnings...

  cur_vrml_file = new VrmlFile;

  if (!cur_vrml_file->open((char *) filename)) {
    // RNFail("Unable to open file %s", filename);
    cerr << "Error: could not open " << filename << endl;
    return NULL;
  }

  // Load initial group into group stack
  group_stack.Push(group);

  recursion_level++;

  R3TransformationNode *transform_node;
  R3Affine *affine;

  if (y_up && (recursion_level == 1)) {
    affine = new R3Affine(R4identity_matrix);
    transform_node = new R3TransformationNode(affine);
    group_stack.Head()->InsertNode(transform_node);
    group_stack.Push(transform_node);
  }

  // Parse Vrml Nodes
  char token[MAX_TOKEN_LENGTH];
  int done = 0;

  while (!done) {
    cur_vrml_file->get_token(token);
    if (strcmp(token, "") != 0) {
      top_nodes.Insert(parse_node(token));
    }
    else done = 1;
  }  // while

  if (recursion_level == 1) {
    if (ring_only) {
      delete_vrml_data();
    }
    if (y_up) {
      affine->Rotate(R3Vector(1, 0, 0), M_PI/2);
      //      affine->Rotate(R3Vector(0, 1, 0), M_PI/2);
      transform_node->SetTransformation(affine);
      group_stack.RemoveHead();
    }
  }
  recursion_level--;
  group_stack.RemoveHead();

  if (info_output) {
    cout << "Read VRML file " << filename << ", "
	 << cur_vrml_file->get_line_nr() - 1 << " lines" << endl;
  }
  cur_vrml_info->add_nr_lines(cur_vrml_file->get_line_nr() - 1);

  delete cur_vrml_file;

  // Return success
  return group;

}  // R3LoadVrml2File




/*################## Writing functions #############################*/

static int
R3WriteVrml2Node(FILE *fp, R3Node *node, const char *indent)
{
    // Construct next indent string
    char next_indent[256];
    strcpy(next_indent, indent);
    strcat(next_indent, "    ");

    // Write nodes by type
    if (node->ClassID() == R3SwitchNode::CLASS_ID()) {
        // Write switch node
        R3SwitchNode *switch_node = (R3SwitchNode *) node;
        fprintf(fp, "%sSwitch {\n", indent);
        fprintf(fp, "%s  whichChoice %d\n", indent, switch_node->Index());
        fprintf(fp, "%s  choice [\n", indent);
	for (int i = 0; i < node->NChildren(); i++) 
	    R3WriteVrml2Node(fp, node->ChildNode(i), next_indent);
        fprintf(fp, "%s  ]\n", indent);
        fprintf(fp, "%s}\n", indent);
    }
    else if (node->ClassID() == R3LodNode::CLASS_ID()) {
        // Write LOD node
        R3Point c = node->BShape().Centroid();
        fprintf(fp, "%sLOD {\n", indent);
        fprintf(fp, "%s  center %g %g %g\n", indent, c.X(), c.Y(), c.Z());
        fprintf(fp, "%s  level [\n", indent);
	for (int i = 0; i < node->NChildren(); i++) 
	    R3WriteVrml2Node(fp, node->ChildNode(i), next_indent);
        fprintf(fp, "%s  ]\n", indent);
        fprintf(fp, "%s}\n", indent);
    }
    else if (node->ClassID() == R3TransformationNode::CLASS_ID()) {
        // Write transformation node
        fprintf(fp, "%sTransform {\n", indent);
        fprintf(fp, "%s  children [\n", indent);
	for (int i = 0; i < node->NChildren(); i++) 
	    R3WriteVrml2Node(fp, node->ChildNode(i), next_indent);
        fprintf(fp, "%s  ]\n", indent);
        fprintf(fp, "%s}\n", indent);
    }
    else if (node->ClassID() == R3LightNode::CLASS_ID()) {
        // Write header of enclosing group
        fprintf(fp, "%sGroup {\n", indent);
	fprintf(fp, "%s  children [\n", indent);

	// Write light node
        R3LightNode *light_node = (R3LightNode *) node;
	const R3Light *light = light_node->Light();
	if (light->ClassID() == R3DirectionalLight::CLASS_ID()) {
	    // Write directional light node
	    const R3DirectionalLight *directional_light = (const R3DirectionalLight *) light;
	    const RNRgb& c = directional_light->Color();
	    const R3Vector& v = directional_light->Direction();
	    fprintf(fp, "%s    DirectionalLight {\n", indent);
	    fprintf(fp, "%s      color %g %g %g\n", indent, c.R(), c.G(), c.B());
	    fprintf(fp, "%s      direction %g %g %g\n", indent, v.X(), v.Y(), v.Z());
	    fprintf(fp, "%s      intensity %g\n", indent, directional_light->Intensity());
	    fprintf(fp, "%s      on %s\n", indent, (directional_light->IsActive()) ? "TRUE" : "FALSE");
	    fprintf(fp, "%s    }\n", indent);
	}
	else if (light->ClassID() == R3PointLight::CLASS_ID()) {
	    const R3PointLight *point_light = (const R3PointLight *) light;
	    const RNRgb& c = point_light->Color();
	    const R3Point& p = point_light->Position();
	    fprintf(fp, "%s    PointLight {\n", indent);
	    fprintf(fp, "%s      color %g %g %g\n", indent, c.R(), c.G(), c.B());
	    fprintf(fp, "%s      location %g %g %g\n", indent, p.X(), p.Y(), p.Z());
	    fprintf(fp, "%s      intensity %g\n", indent, point_light->Intensity());
	    fprintf(fp, "%s      on %s\n", indent, (point_light->IsActive()) ? "TRUE" : "FALSE");
	    fprintf(fp, "%s    }\n", indent);
	}
	else if (light->ClassID() == R3SpotLight::CLASS_ID()) {
	    const R3SpotLight *spot_light = (const R3SpotLight *) light;
	    const RNRgb& c = spot_light->Color();
	    const R3Point& p = spot_light->Position();
	    const R3Vector& v = spot_light->Direction();
	    fprintf(fp, "%s    SpotLight {\n", indent);
	    fprintf(fp, "%s      color %g %g %g\n", indent, c.R(), c.G(), c.B());
	    fprintf(fp, "%s      location %g %g %g\n", indent, p.X(), p.Y(), p.Z());
	    fprintf(fp, "%s      direction %g %g %g\n", indent, v.X(), v.Y(), v.Z());
	    fprintf(fp, "%s      intensity %g\n", indent, spot_light->Intensity());
	    fprintf(fp, "%s      cutOffAngle %g\n", indent, spot_light->CutOffAngle());
	    fprintf(fp, "%s      on %s\n", indent, (spot_light->IsActive()) ? "TRUE" : "FALSE");
	    fprintf(fp, "%s    }\n", indent);
	}

	// Write rest of enclosing group node
	for (int i = 0; i < node->NChildren(); i++) 
	    R3WriteVrml2Node(fp, node->ChildNode(i), next_indent);
	fprintf(fp, "%s  ]\n", indent);
	fprintf(fp, "%s}\n", indent);
    }
    else if (node->ClassID() == R3ShapeNode::CLASS_ID()) {
    }
    else {
        // Write other interior nodes
        if (node->NChildren() > 0) {
	    if (node->NChildren() == 1) {
	        R3WriteVrml2Node(fp, node->ChildNode(0), indent);
	    }
	    else {
	        fprintf(fp, "%sGroup {\n", indent);
		fprintf(fp, "%s  children [\n", indent);
		for (int i = 0; i < node->NChildren(); i++) 
		    R3WriteVrml2Node(fp, node->ChildNode(i), next_indent);
		fprintf(fp, "%s  ]\n", indent);
		fprintf(fp, "%s}\n", indent);
	    }
	}
    }

    // Return success
    return TRUE;
}



R3Model *
R3WriteVrml2File(R3Model *model, const char *filename)
{
    FILE *fp;

    // Open file
    if (!(fp = fopen(filename, "w"))) {
	RNFail("Unable to open file %s", filename);
	return NULL;
    }

    // Write VRML 2.0 header
    fprintf(fp, "#VRML V2.0 utf8\n");

    // Write model
    R3WriteVrml2Node(fp, model, "");

    // Close file
    fclose(fp);

    // Return success
    return model;
}



