//
// $Id: VsExpression.C,v 1.1 2000/01/17 18:47:40 min Exp min $
//

#include "VsExprNodes.h"
#include "VsExpression.h"
#include "VsSymbolTable.h"


static const char binary_ops[][5] = {"=", "+=", "-=", "*=", "/=", "%=", "&=",
				     "|=", "^=", "<<=", ">>=", ">>>=", "==",
				     "!=", "<", "<=", ">=", ">", "+", "-",
				     "*", "/", "%", "&&", "||", "&", "|",
				     "^", "<<", ">>", ">>>", ""};
//
// precedence:  operator:
// 1            comma
// 2            assignment (e.g. =, +=)
// 3            conditional (?:)
// 4            ||
// 5            &&
// 6            |
// 7            ^
// 8            &
// 9            == !=
// 10           < <= > >=
// 11           << >> >>>
// 12           + -
// 13           * / %
// 14           unary operators (! ~ - ++ --)
// 15           () [] .
//
static const int binary_ops_precedence[] = { 2, 2, 2, 2, 2, 2, 2,
					     2, 2, 2, 2, 2, 9,
					     9, 10, 10, 10, 10, 12, 12,
					     13, 13, 13, 5, 4, 8, 6,
					     7, 11, 11, 11 };





void
VsExpression::parse_entry(Tokenfile *tokenfile_ptr)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "VsExpression::parse_entry" << endl;
#endif
  parse_tree = parse(tokenfile_ptr, 0);
  if (parse_tree == NULL) {
    tokenfile_ptr->print_error_line();
    cerr << "Expected an expression" << endl;
    tokenfile_ptr->read_until(SEMI_COLON);
  }
}  // VsExpression::parse_entry



VsExprNode *
VsExpression::parse(Tokenfile *tokenfile_ptr, int precedence)
{
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "VsExpression::parse" << endl;
#endif

  VsExprNode *lhs = parse_primary(tokenfile_ptr);
  if (!lhs) return NULL;

  int end_of_expression = 0;

  char token[MAX_TOKEN_LENGTH];
  tokenfile_ptr->lookahead_token(token);
  int op_index;

  while (!end_of_expression) {
    if ((op_index = is_binary_op(token)) != -1) {
      int my_precedence = binary_ops_precedence[op_index];

      if (my_precedence > precedence) {
	tokenfile_ptr->get_token(token);

	VsExprNode *rhs = parse(tokenfile_ptr, my_precedence);

	lhs = new VsBinaryExprNode(lhs, token, rhs);

	tokenfile_ptr->lookahead_token(token);
      }
      else
	end_of_expression = 1;
    }
    else if ((strcmp(token, "++") == 0) || (strcmp(token, "--") == 0)) {
      tokenfile_ptr->get_token(token);
      // unlikely, but possible: i++ or i-- of unknown variable...
      if (((VsVariableNode *) lhs)->get_field_ptr() == NULL) {
	((VsVariableNode *) lhs)->set_field_ptr(new SFFloat(0));
      }
      // also possible: ++ or -- of not a variable (e.g. 3++)
      if (lhs->get_node_type() != VsVariableNodeType) {
	tokenfile_ptr->print_error_line();
	cerr << "Operator [" << token << "] only valid for variables" << endl;
      }
      else
	lhs = new VsBinaryExprNode(lhs, token, NULL);
      tokenfile_ptr->lookahead_token(token);
    }  // else, postfix ++ or --
    else end_of_expression = 1;
  }  // while

  // else if token is ? (tertiary operator)
  // test for ? and put it in the binary ops array?

  return lhs;

}  // VsExpression::parse



int
VsExpression::is_binary_op(char *token)
{
  for(int i=0; binary_ops[i][0] != '\0'; i++) {
    if (strcmp(token, binary_ops[i]) == 0) return i;
  }  // for
  return -1;
}  // VsExpression::is_binary_op



VsExprNode *
VsExpression::parse_primary(Tokenfile *tokenfile_ptr)
{
  char token[MAX_TOKEN_LENGTH];

  tokenfile_ptr->lookahead_token(token);
  if (!token[0]) return NULL;

#ifdef __DEBUG_PARSE_VRMLSCRIPT__
  cout << "VsExpression::parse_primary, got token [" << token << "]" << endl;
#endif

  //  
  // new operator first, otherwise it is treated as an identifier
  //
  if (strcmp(token, "new") == 0) {
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
    cout << "VsExpression::parse_primary, new operator" << endl;
#endif
    tokenfile_ptr->get_token(token);
    VsConstructorNode *constr_ptr = new VsConstructorNode;
    constr_ptr->parse(tokenfile_ptr);
    return constr_ptr;
  }
  //
  // then Math standard object
  //
  else if (strcmp(token, "Math") == 0) {
    VsMathCallNode *math_call_ptr = new VsMathCallNode;
    math_call_ptr->parse(tokenfile_ptr);
    return math_call_ptr;
  }
  //
  // Browser standard object
  //
  else if (strcmp(token, "Browser") == 0) {
    VsBrowserCallNode *browser_call_ptr = new VsBrowserCallNode;
    browser_call_ptr->parse(tokenfile_ptr);
    return browser_call_ptr;
  }
  //
  // some more standard stuff: parseInt, parseFloat, VrmlMatrix...
  //
  else if ((strcmp(token, "parseInt") == 0) || 
	   (strcmp(token, "parseFloat") == 0) ||
	   (strcmp(token, "VrmlMatrix") == 0)) {
    tokenfile_ptr->print_error_line();
    tokenfile_ptr->read_until(SEMI_COLON);
    cerr << "Support for [" << token << "] not implemented yet" << endl;
    return NULL;
  }
  //
  // now check for constant (also before identifier, because of TRUE/FALSE
  //
  else if (is_constant(token)) {
    tokenfile_ptr->get_token(token);
    // note that we still may have problems with numbers like 1e-7 ... :-(
    // now lookahead for a dot, if there, get it, and get the rest
    char k = tokenfile_ptr->skip_whitespace();
    if (k == '.') {
      char rest[MAX_TOKEN_LENGTH];
      tokenfile_ptr->get_token(rest);
      //      cout << "Got the dot, current char is: " 
      //	   << (char) tokenfile_ptr->get_current_char() << endl;
      tokenfile_ptr->get_token(rest);
      strcat(token, ".");
      strcat(token, rest);
    }
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
    cout << "VsExpression::parse_primary, creating new VsConstantNode [" 
	 << token << "]" << endl;
#endif
    return new VsConstantNode(token);
  }
  //
  // check for identifier
  //
  else if (is_identifier(tokenfile_ptr, token)) {
    tokenfile_ptr->get_token(token);
    //
    // todo: check for object method call
    // 
    VsSymbolNode *symbol_ptr = cur_symbol_table->find_symbol(token);
    if (symbol_ptr) {
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
      cout << "  found [" << token << "] in cur_var_table, adding a pointer"
	<< endl;
#endif
      //
      // if it is a function, parse the arguments
      //
      if (symbol_ptr->get_node_type() == VsFunctionNodeType) {
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
	cout << "  it's a function call, adding a call node" << endl;
#endif
	VsFunctionCallNode *function_call = new VsFunctionCallNode;
	function_call->set_function_ptr((VsFunctionNode *) symbol_ptr);
	function_call->parse(tokenfile_ptr);
	return function_call;
      }
      //
      // lookahead, if there's a [, then it's an array reference
      // create a new VsArrayNode for it
      //
      else {
	int k = tokenfile_ptr->skip_whitespace();
	if (k == OPEN_BRACKET) {
	  VsArrayNode *array_node_ptr =
	    new VsArrayNode(*((VsVariableNode *) symbol_ptr));
	  array_node_ptr->parse(tokenfile_ptr);
	  return array_node_ptr;
	}
	//
	// todo: if k == . then we have a "properties" or method selection
	//
	else if (k == '.') {
	  tokenfile_ptr->print_error_line();
	  tokenfile_ptr->read_until(SEMI_COLON);
	  cerr << "Object method selector . not implemented yet" << endl;
	  return symbol_ptr;
	}
	//
	// if not, just return the variable reference
	//
	else
	  return symbol_ptr;
      }
    }
    //
    // found unknown identifier (legal in VrmlScript/JavaScript)
    //
    else {
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
      cout << "VsExpression::parse_primary, creating new VsVariableNode [" 
	   << token << "]" << endl;
#endif
      //
      // except when it's a function call, lookahead for ( here
      //
      int k = tokenfile_ptr->skip_whitespace();
      if (k == OPEN_PAREN) {
	tokenfile_ptr->print_error_line();
	cerr << "Function [" << token
	     << "] should be defined before it is called" << endl;
	// skip the arguments part, treat it as an uninitialized variable
	tokenfile_ptr->read_until(SEMI_COLON);
      }
      symbol_ptr = new VsVariableNode(token);
      cur_symbol_table->add(symbol_ptr);
      return symbol_ptr;
    }  // else, unknown identifier
  }  // if identifier
  //
  // check for unary operators
  //
  else if ((strcmp(token, "--") == 0) || (strcmp(token, "++") == 0) ||
	   (strcmp(token, "-") == 0) || (strcmp(token, "!") == 0) ||
	   (strcmp(token, "~") == 0)) {
    tokenfile_ptr->get_token(token);
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
    cout << "VsExpression::parse_primary, creating new VsBinaryExprNode" << endl;
#endif
    return new VsBinaryExprNode(new VsConstantNode("0"), // just for the -
			      token, parse_primary(tokenfile_ptr));
  }
  //
  // ignore unary +
  //
  else if (token[0] == '+') {
    tokenfile_ptr->get_token(token);
    return parse_primary(tokenfile_ptr);
  }
  //
  // check for (, if there: parse expression, accept )
  //
  else if (token[0] == OPEN_PAREN) {
    tokenfile_ptr->get_token(token);
    VsExprNode *expr_ptr = parse(tokenfile_ptr, 0);
    tokenfile_ptr->accept_token(CLOSE_PAREN);
    return expr_ptr;
  }
  //
  // check for single quote = string
  //
  else if (strcmp(token, "'") == 0) {
    SFString *string_ptr = new SFString;
    string_ptr->parse(SINGLE_QUOTE);
    VsStringConstantNode *string_node_ptr = 
      new VsStringConstantNode(string_ptr);
#ifdef __DEBUG_PARSE_VRMLSCRIPT__
    cout << "VsExpression::parse_primary, parsing string [" 
	 << string_ptr->get_string() << "]" << endl;
#endif
    return string_node_ptr;
  }
  else if (token[0] == DOUBLE_QUOTE) {
    tokenfile_ptr->print_error_line();
    cerr << "Use single quotes for strings in VrmlScript/JavaScript" << endl;
    tokenfile_ptr->read_until(SEMI_COLON);
    return NULL;
  }
  //
  // this was it, this isn't a primary expression...
  //
  else
    return NULL;

}  // VsExpression::parse_primary



int
VsExpression::is_identifier(Tokenfile *tokenfile_ptr, char *token)
{
  if (!tokenfile_ptr->is_id_firstchar(token[0])) return 0;
  int length = strlen(token);
  for(int i=1; i<length; i++) {
    if (!tokenfile_ptr->is_id_char(token[i])) return 0;
  }  // for
  return 1;
}  // VsExpression::is_identifier



int
VsExpression::is_constant(char *token)
{
  if ((strcmp(token, "TRUE") == 0) || (strcmp(token, "true") == 0) ||
      (strcmp(token, "FALSE") == 0) || (strcmp(token, "false") == 0)) return 1;

  int length = strlen(token);
  for(int i=0; i<length; i++) {
    char digit = token[i];
    if ((digit < '0') || (digit > '9')) {
      if (!((i == 1) && ((digit == 'x') || (digit == 'X'))) &&
	  !(digit == '.')) return 0;
    }
  }  // for
  return 1;

}  // VsExpression::is_constant



Field *
VsExpression::evaluate()
{
#ifdef __DEBUG_VRMLSCRIPT__
  cout << "Evaluating expression" << endl;
#endif

  if (!parse_tree) return NULL;

  return parse_tree->evaluate();

}  // VsExpression::evaluate




