%% This implementation follows paper on ImageWebs:
% Image Webs: Computing and Exploiting Connectivity in Image Collections
%     Kyle Heath, Natasha Gelfand, Maks Ovsjanikov, 
%     Mridul Aanjaneya, Leonidas J. Guibas
% http://graphics.stanford.edu/projects/lgl/...
%         /papers/hgoag-iwcecic-10/hgoag-iwcecic-10.pdf
function L = edgeRank(dataPrefix)
    tic
    try
        fid_edges = fopen([dataPrefix '.graph.edges.bin'], 'rb');
        nEdges = fread(fid_edges, 1, 'int32');
        edges = fread(fid_edges, nEdges, 'int32');
        fclose(fid_edges);

        fid_cand  = fopen([dataPrefix '.candidate.edges.bin'], 'rb');
        nCand = fread(fid_cand, 1, 'int32');
        cand = fread(fid_cand, nCand, 'int32');
        fclose(fid_cand);

        candR = cand(1:2:end)+1;
        candC = cand(2:2:end)+1;
        nCand = length(candR);

        % construct sparse laplacian
        rows = edges(1:2:end)+1;
        cols = edges(2:2:end)+1;
        vals = -ones(1, length(rows));
        N = max(max(edges))+1;

        L = sparse(rows, cols,  vals, N, N);
        L = L + L';

        rowNorms = sum(L, 2);
        L = L + diag(abs(rowNorms));

        % find  fiedler (2nd smallest) eigenvector
        ordR=[];
        ordC=[];
        
        %if abs(condest(L)) > 10E-8
        %    fprintf('[WARNING] Matrix is near-singular');
        %    ME = MException('AcctError:Incomplete', 'Mtx near-singular');
        %    throw(ME);
        %end
        [V E] = eigs(L, 2, 'SM');

        fiedler = V(:, 1)';
        MAX_ITER = 1000;
        TOL = 10E-5;

        % sort edges
        fprintf('Ranking Edges: %d\n', nCand);

        % speedup
        nMax=1000;
        appendR=[];
        appendC=[];
        if (nCand>nMax)
            fprintf('[WARNING] Will only rank first %d\n', nMax);
            nRand = randperm(nCand);
            candR = candR(nRand);
            candC = candC(nRand);
            appendR=candR((nMax+1):end);
            appendC=candC((nMax+1):end);
            candR=candR(1:nMax);
            candC=candC(1:nMax);
            nCand=nMax;
        end
        
        %fprintf('Convergence: ');
        for oID=1:(nCand-1)
            if mod(oID, 20)==0
                fprintf('%d ', oID);
            end
            % pick best edge
            [val best] = max(abs(fiedler(candR)-fiedler(candC)));
            br = candR(best);
            bc = candC(best);
            ordR = [ordR br];
            ordC = [ordC bc];

            % update laplacian
            L(br,bc)=-1;
            L(bc,br)=-1;
            L(bc,bc)=L(bc,bc)+1;
            L(br,br)=L(br,br)+1;

            % update fiedler vector
            fiedler0 = fiedler;
            for i=1:MAX_ITER
                fiedler = (2*N*eye(N,N)-L) * fiedler';
                fiedler = fiedler ./ norm(fiedler,2);
                fiedler = fiedler - sum(fiedler) / N;
                fiedler = fiedler';
                if norm(fiedler0-fiedler) < TOL
                    break;
                end
                fiedler0 = fiedler;
            end
            %fprintf('%d ', i);

            % remove candidate edges
            candR(best)=[];
            candC(best)=[];
        end
        fprintf('\n');

        % final best edges
        ordR = [ordR candR(1)];
        ordC = [ordC candC(1)];
        ordR = [ordR appendR'];
        ordC = [ordC appendC'];
    catch exception
        exception
        exception.stack
        fprintf('[WARNING] Matlab err: switch to simple randomization\n');
        ordR = cand(1:2:end)+1;
        ordC = cand(2:2:end)+1;
        rPerm = randperm(nCand);
        ordR = ordR(rPerm);
        ordC = ordC(rPerm);
    end
    
    % write output
    fid_order  = fopen([dataPrefix '.ordered.edges.bin'], 'wb');
    jLen=length(ordR)+length(ordC);
    fwrite(fid_order, jLen, 'int32');
    outVec(1:2:jLen) = ordR;
    outVec(2:2:jLen) = ordC;
    fwrite(fid_order, outVec-1, 'int32');
    fclose(fid_order);
    
    t = toc;
    fid_time = fopen([dataPrefix '.time.edgerank.bin'], 'wb');
    fwrite(fid_time, t, 'int32');
    fclose(fid_time);

    fprintf('Terminated edgerank after %fs\n', t);
end

